-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- EyeBeast
-- Created by Tony Park, July 2014
--
require('strict')
require('utils')        -- for nearby()
class = require('middleclass')
require('graphics/glyph_sets')
require('game_world_server/server_sound_man')

local Mob = require("controllers/Mob")

local EyeBeast = class('EyeBeast', Mob)


--
-- EyeBeast Constants
--

-- directions
local UP, LEFT, DOWN, RIGHT, NONE, HALF_LEFT, HALF_RIGHT, HALF_UP, HALF_DOWN, AHEAD, DOWNLEFT, DOWNRIGHT, UPLEFT, UPRIGHT, ROTATE_RIGHT, ROTATE_LEFT = 
       1,    2,    3,     4,    5,         6,          7,       8,         9,     5,       10,        11,     12,      13,           14,          15    -- none and ahead are the same!
local direction_line_modifier = { -1, 0, 1, 0 }
local direction_column_modifier = { 0, -1, 0, 1}

-- return value from off_map()
local NOT_OFF_MAP = 0

local left_animation = { AHEAD, HALF_LEFT, LEFT, HALF_LEFT }
local right_animation = { AHEAD, HALF_RIGHT, RIGHT, HALF_RIGHT }
local up_animation = { AHEAD, HALF_UP, UP, HALF_UP }
local down_animation = { AHEAD, HALF_DOWN, DOWN, HALF_DOWN }
local rotate_left_animation = {AHEAD, HALF_UP, UP, UPLEFT, LEFT, DOWNLEFT, DOWN, DOWNRIGHT, RIGHT, UPRIGHT, UP, HALF_UP }   
local rotate_right_animation = {AHEAD, HALF_DOWN, DOWN, DOWNLEFT, LEFT, UPLEFT, UP, UPRIGHT, RIGHT, DOWNRIGHT, DOWN, HALF_DOWN }    

local animation_sets = 
{ 
    [UP] = up_animation, 
    [DOWN] = down_animation,
    [LEFT] = left_animation,
    [RIGHT] = right_animation,
    [ROTATE_LEFT] = rotate_left_animation,
    [ROTATE_RIGHT] = rotate_right_animation,
}

local animation_length = 
{
    [UP] = #up_animation, 
    [DOWN] = #down_animation,
    [LEFT] = #left_animation,
    [RIGHT] = #right_animation,
    [ROTATE_LEFT] = #rotate_left_animation,
    [ROTATE_RIGHT] = #rotate_right_animation,
}


local look_time = 0.5       -- 'look' time in seconds
local rotate_look_time = 1.0
--local look_fps = #up_animation / look_time  -- assume all animations are the same length

local look_fps = 
{
    [UP] =  #up_animation / look_time,
    [DOWN] = #down_animation / look_time,
    [LEFT] = #left_animation / look_time,
    [RIGHT] = #right_animation / look_time,
    [ROTATE_LEFT] = #rotate_left_animation / rotate_look_time,
    [ROTATE_RIGHT] = #rotate_right_animation / rotate_look_time,
}

local eye_glyphs =
{
    [UP] = get_glyph("mobs",82), 
    [DOWN] = get_glyph("mobs",85),
    [LEFT] = get_glyph("mobs",50),
    [RIGHT] = get_glyph("mobs",53),
    [HALF_UP] = get_glyph("mobs",81),
    [HALF_DOWN] = get_glyph("mobs",84),
    [HALF_LEFT] = get_glyph("mobs",49),
    [HALF_RIGHT] = get_glyph("mobs",52),
    [AHEAD] = get_glyph("mobs",48),
    [DOWNLEFT] = get_glyph("mobs",54),
    [DOWNRIGHT] = get_glyph("mobs",55),
    [UPLEFT] = get_glyph("mobs",86),
    [UPRIGHT] = get_glyph("mobs",87),
}

local monster_vertical_animation = get_glyph("mobs",64)
local monster_horizontal_animation = get_glyph("mobs",96)
local monster_standing_glyph = get_glyph("mobs",8)
local tentacle_base_glyph = get_glyph("mobs",9)
local tentacle_animation_length = 2
local tentacles_fps = 5
local yumyum_base_glyph = get_glyph("mobs",104)
local yumyum_animation_length = 13
local yumyum_fps = 4

local direction_angles =
{
    [UP] =          (4/4)*math.pi, 
    [HALF_UP] =     (4/4)*math.pi, 
    [UPRIGHT] =     (3/4)*math.pi,
    [RIGHT] =       (2/4)*math.pi,
    [HALF_RIGHT] =  (2/4)*math.pi,
    [DOWNRIGHT] =   (1/4)*math.pi,
    [DOWN] =        (0)*math.pi,
    [HALF_DOWN] =   (0)*math.pi,
    [DOWNLEFT] =    (7/4)*math.pi,
    [LEFT] =        (6/4)*math.pi,
    [HALF_LEFT] =   (6/4)*math.pi,
    [UPLEFT] =      (5/4)*math.pi,
    --[AHEAD] =         { (4/4)*math.pi },
}

local pi = math.pi
local slice_of_pi = (1/4*pi)
local two_pi = (2*pi)

function are_we_looking(angle, direction)
    
    -- first check if we are looking in any direction
    if direction==AHEAD or direction == nil then return false end

    local looking = false
    
    local min = direction_angles[direction] - slice_of_pi
    local max = direction_angles[direction] + slice_of_pi
    
    -- check if we have a negative minimum or over-large maximum
    if min < 0 then
        -- if so, we check the range we're NOT looking at
        -- invert the result and swap max and min
        looking = true
        local new_max = min + (two_pi)
        min = max
        max = new_max
    elseif max >= two_pi then
        -- if so, we check the range we're NOT looking at
        -- invert the result and swap max and min
        looking = true
        local new_min = max - (two_pi)
        max = min
        min = new_min
    end
    
    -- then check if our angle is in the range we are looking in
    if angle >= min and angle <= max then
        looking = not looking
    end

    return looking
        
end

function EyeBeast:initialize(map_name)
    Mob.initialize(self,"eyebeast",map_name)
    
    self.current_search_direction = ROTATE_LEFT
    self.next_search_direction = nil
    self.last_move_direction = AHEAD
    
    self.eye_animation_frame = 1    -- set before the start of the animation, we display only on change
    self.tentacles_animation_frame = 0  -- this set runs from zero...
    self.yumyum_animation_frame = 0     -- this set runs from zero...
    self.yumyum = false
    
    local eye_glyph = eye_glyphs[animation_sets[self.current_search_direction][1]]  -- initial display is first element of animation
    
    local maze = get_maze(map_name)
    local display_element = new_hidden_mob_display_element(maze, eye_glyph, self.line, self.column, 101, get_simple_colour_by_name('BRIGHT_WHITE'))
    self.eye_display_element = Mob.register_display_element(self, display_element)
    
    --self.look_count = 0
    
    self.speed = 9
    
    self.never_seen_player = true
    
end

function EyeBeast:set_speed(speed)
    self.speed = speed
end


function EyeBeast:opposite_direction(current_direction)
    local reverse_direction = { [UP] = DOWN, [LEFT] = RIGHT, [DOWN] = UP, [RIGHT] = LEFT, [ROTATE] = DOWN }
    return reverse_direction[current_direction]
end

function EyeBeast:select_look_direction()
    
    if self.current_search_direction == ROTATE_LEFT then
        self.current_search_direction = ROTATE_RIGHT
    else
        self.current_search_direction = ROTATE_LEFT
    end
    
    --[[
    local random = math.random(0,1)
    
    if random == 0 then
        self.current_search_direction = ROTATE_LEFT
    else
        self.current_search_direction = ROTATE_RIGHT
    end
    --]]
    --[[    
    
    if self.next_search_direction ~= nil then
        self.current_search_direction = self.next_search_direction
        self.next_search_direction = nil
    else
        self.look_count = self.look_count + 1
        if self.look_count == 4 then
            self.current_search_direction = ROTATE
            self.look_count = 0
        else
            self.current_search_direction = self:opposite_direction(self.current_search_direction)
        end
    end
    --]]
end

function EyeBeast:can_i_see_any_player(direction)
    local seeing, angle
    local players = {}
    
    -- get list of players we can see
    for key,value in all_players() do
        
        seeing, angle = self:can_i_see_player(direction, value)
        
        if seeing then 
            table.insert(players, {['player']=key, ['angle']=angle })
        end
    end
    
    -- pick one at random
    -- could pick closest?
    if #players ~= 0 then 
        local player = random_entry(players)
        return true, player['player'], player['angle']
    else
        return false
    end 
end


function EyeBeast:can_i_see_player(direction, player)
    
    -- check for invisibility and being on the same map
    if player.visible_to_mobs == false then return false end
    if self.map_name ~= player.map_name then return false end
    
    
    --local player_line = location_data.line
    --local player_column = location_data.column
    
    -- calculate an angle (add pi to normalise to zero)
    local angle = math.atan2(self.column - player.column, self.line - player.line) + pi
    local looking = are_we_looking(angle, direction)
    local seeing = false
    if looking==true then
        local maze = get_maze(self.map_name)     
        seeing = maze:line_of_sight(self.line+0.5, self.column+0.5, player.line, player.column)
    end

    return seeing, angle
end

local direction_target_modifiers =
{
    [UP] = { ['line'] = -1, ['column'] = 0 }, 
    [HALF_UP] = {  ['line'] = -1, ['column'] = 0  }, 
    [UPRIGHT] = {  ['line'] = -1, ['column'] = 1  },
    [RIGHT] = {  ['line'] = 0, ['column'] = 1  },
    [HALF_RIGHT] = {  ['line'] = 0, ['column'] = 1  },
    [DOWNRIGHT] = {  ['line'] = 1, ['column'] = 1  },
    [DOWN] = {  ['line'] = 1, ['column'] = 0  },
    [HALF_DOWN] = {  ['line'] = 1, ['column'] = 0  },
    [DOWNLEFT] = {  ['line'] = 1, ['column'] = -1  },
    [LEFT] = {  ['line'] = 0, ['column'] = -1  },
    [HALF_LEFT] = {  ['line'] = 0, ['column'] = -1  },
    [UPLEFT] = {  ['line'] = -1, ['column'] = -1  },
    [AHEAD] = {  ['line'] = 0, ['column'] = 0  },
}

local direction_angle_ranges =
{
    [{ ['min'] = (15/8)*math.pi, ['max'] = (16/8)*math.pi }]=DOWN       ,
    [{ ['min'] = (13/8)*math.pi, ['max'] = (15/8)*math.pi }]=DOWNLEFT   ,
    [{ ['min'] = (11/8)*math.pi, ['max'] = (13/8)*math.pi }]=LEFT       ,
    [{ ['min'] = ( 9/8)*math.pi, ['max'] = (11/8)*math.pi }]=UPLEFT         ,
    [{ ['min'] = ( 7/8)*math.pi, ['max'] = ( 9/8)*math.pi }]=UP             , 
    [{ ['min'] = ( 5/8)*math.pi, ['max'] = ( 7/8)*math.pi }]=UPRIGHT    ,
    [{ ['min'] = ( 3/8)*math.pi, ['max'] = ( 5/8)*math.pi }]=RIGHT          ,
    [{ ['min'] = ( 1/8)*math.pi, ['max'] = ( 3/8)*math.pi }]=DOWNRIGHT      ,
    [{ ['min'] = ( 0/8)*math.pi, ['max'] = ( 1/8)*math.pi }]=DOWN       ,
}


local directions_to_try = 
{
    [UP] =      { UPRIGHT, UPLEFT }, 
    [UPRIGHT] = { UP, RIGHT  },
    [RIGHT] =   { UPRIGHT, DOWNRIGHT  },
    [DOWNRIGHT]={ DOWN, RIGHT  },
    [DOWN] =    { DOWNRIGHT, DOWNLEFT  },
    [DOWNLEFT] ={ DOWN, LEFT  },
    [LEFT] =    { UPLEFT, DOWNLEFT  },
    [UPLEFT] =  { UP, LEFT  },
}

-- a local function for adjusting lines or columns
local function adjust_location(target, current, distance)
    if target > current then
        current = current + distance
        if current > target then current = target end
    else
        current = current - distance
        if current < target then current = target end
    end
    
    return current
end

function get_direction_from_angle(angle)
    -- normalise angle
    while angle >= two_pi do
        angle = angle - two_pi
    end
    
    while angle < 0 do
        angle = angle + two_pi
    end
    
    for range, direction in pairs(direction_angle_ranges) do
        if angle >= range.min and angle < range.max then
            return direction
        end
    end
end

function direction_angle_test()
    for angle=-1,360 do
        print("get_direction_from_angle ", angle, get_direction_from_angle(math.rad(angle)))
    end
end

function EyeBeast:reset()
    --debugMessage("Eyebeast reset", self.mob_number)
    
    Mob.reset(self)
    
    self.never_seen_player = true
    
    self.tracking = nil
    self.tracking_player_id = nil
    self.tracking_direction = nil
end


function EyeBeast:update(dt)
    
    if self.held then return end
    
    -- check for player capture (all players)
    local nearby_player_ids = find_nearby_players({line=self.line+0.5, column=self.column+0.5, map_name=self.map_name}, 0.5)
    for _,id in ipairs(nearby_player_ids) do
        -- we caught the player!
        --[[
        if self.yumyum == false then
            
            -- eyebeast yumyum animation
            self.yumyum = true
            self:set_visible(false, self.eye_display_element)
            self.yumyum_frame = 0
            
            -- kill the player
            get_player_by_id(id):die()
        end
        --]]
        get_player_by_id(id):die()
    end
    
    if #nearby_player_ids > 0 then
        -- eyebeast caught something, return it to home
        
        -- do the eat sounds even if we dropped the animation
        self:sounds('eat', 0)
        
        self:reset()
        return
    end
        
    --[[
    if self.yumyum == true then
        if not self.yumyum_first_time then
            self.yumyum_first_time = true
            self:sounds('eat', 0)
        end
        
        -- we have caught the player, do the eat animation
        -- update the yumyum glyph
        self.yumyum_animation_frame = self.yumyum_animation_frame + (dt * yumyum_fps)
        if self.yumyum_animation_frame >= yumyum_animation_length then
            self.yumyum_animation_frame = self.yumyum_animation_frame - 2
        end
        
        self:set_glyph(1, yumyum_base_glyph+math.floor(self.yumyum_animation_frame))
        
        return
    end
    --]]
    
    local function track_player(direction)
        
        local spotted = false
        local id, angle
--[[        
        local print_flag = false
        if self.tracking_direction then
            print("tracking...", self.tracking_direction, self.mob_number)
            print_flag = true
        end
--]]        
        if self.tracking_player_id ~= nil then
            -- we have a target player, can we still see them?
            id = self.tracking_player_id
            spotted, angle = self:can_i_see_player(direction, get_player_by_id(id))
        end
        
        if not spotted then
            -- if we never had a target player, or we can't see them, look for other players
            spotted, id, angle = self:can_i_see_any_player(direction)
        end
                
        -- if spotted ~= true then 
        --      do nothing - leave the current target alone so we finish the move and look from there
        -- end
        
        if spotted==true then 
            -- if spotted == true, then we either saw the current location of the player we
            -- are following, or the current location of a new player
            self.tracking_player_id = id
            self.tracking_last_known_line = math.floor(get_player_by_id(id).line)
            self.tracking_last_known_column = math.floor(get_player_by_id(id).column)
            self.tracking_direction = get_direction_from_angle(angle)
            self.tracking = true
        else
            -- we didn't see a player... go to last known player location
            self.tracking_player_id = nil
            if self.tracking ~= nil then
                angle = math.atan2(self.column-self.tracking_last_known_column, self.line-self.tracking_last_known_line) + pi
                self.tracking_direction = get_direction_from_angle(angle)
                --print("tracking to last know location")
            end
        end
            
        if self.line == self.tracking_last_known_line and self.column == self.tracking_last_known_column then
            -- we have arrived at the last known player location...
            -- either we are munching them or we've lost them
            -- either way, tracking is finished
            self.tracking = nil
            self.tracking_player_id = nil
            self.tracking_direction = nil
        end
        
        if self.tracking then   
            
            local function update_tracking_step()
                self.next_tracking_line = self.line + direction_target_modifiers[self.tracking_direction]['line']
                self.next_tracking_column = self.column + direction_target_modifiers[self.tracking_direction]['column']
            end
            
            update_tracking_step()
                    
            if is_wall(self.map_name, self.next_tracking_line, self.next_tracking_column) then
                -- con't go there, it's a wall!
                -- we still have angle
                -- we have a tracking direction
                local direction_pair = directions_to_try[self.tracking_direction]
                local first_try = math.random(1,2)
                local second_try
                if first_try == 1 then second_try = 2 else second_try = 1 end
                
                self.tracking_direction = direction_pair[first_try]
                update_tracking_step()
                if is_wall(self.map_name, self.next_tracking_line, self.next_tracking_column) then
                    self.tracking_direction = direction_pair[second_try]
                    update_tracking_step()
                    if is_wall(self.map_name, self.next_tracking_line, self.next_tracking_column) then
                        -- it's all walls.. give up tracking
                        self.tracking = nil
                        self.tracking_player_id = nil
                    end
                end
            end
        end
--[[                
        if self.tracking_direction and print_flag == false then
            print("newly tracking...", self.tracking_direction, self.mob_number, spotted, self.held)
        end
--]]                
        return spotted
    end
    
    
    
    if self.tracking ~= nil then
        
        -- check if we finished a cell step, if so, relocate player
        if self.line == self.next_tracking_line and self.column == self.next_tracking_column then
            track_player(self.tracking_direction)
        end
        
        -- now do the movement and animation
        local column = self.column
        local line = self.line
        
        -- update the eye glyph with the location we are tracking in
        local eye_glyph = eye_glyphs[self.tracking_direction]
        self:set_glyph(self.eye_display_element,eye_glyph)
        
        -- update the tentacles glyph
        self.tentacles_animation_frame = self.tentacles_animation_frame + (dt * tentacles_fps)
        if self.tentacles_animation_frame >= tentacle_animation_length then
            self.tentacles_animation_frame = self.tentacles_animation_frame - tentacle_animation_length
        end
        self:set_glyph(1, tentacle_base_glyph+math.floor(self.tentacles_animation_frame))
        
        -- calculate the move distance and from that, a new location
        local distance = self.speed * dt
        
        if line == self.next_tracking_line then
            -- we are moving horizontally
            column = adjust_location(self.next_tracking_column, column, distance) 
        elseif column == self.next_tracking_column then
            -- we are moving vertically
            line = adjust_location(self.next_tracking_line, line, distance)
        else
            -- we are moving diagonally
            distance = distance * 0.707
            column = adjust_location(self.next_tracking_column, column, distance) 
            line = adjust_location(self.next_tracking_line, line, distance)
        end 
        
        self:set_location(line, column)
        
        -- do sound updates
        self:sounds('tracking',dt)
        
    else
        -- monster is searching
        -- print 'searching'
        
        local fps = look_fps[self.current_search_direction]
        local frame_count = animation_length[self.current_search_direction]
        local old_animation_frame = self.eye_animation_frame
        
        self.eye_animation_frame = self.eye_animation_frame + (dt * fps)
        
        while self.eye_animation_frame >= frame_count+1 do
            -- while loop in case we were in the debugger (or otherwise delayed a long time)
            self:select_look_direction()
            self.eye_animation_frame = self.eye_animation_frame - frame_count
            old_animation_frame = 0     -- force update below
        end
        
        local floored_animation_frame = math.floor(self.eye_animation_frame)
        local direction = animation_sets[self.current_search_direction][floored_animation_frame]
        
        -- if something changed...
        if floored_animation_frame > old_animation_frame then
            local eye_glyph = eye_glyphs[direction]
            self:set_glyph(self.eye_display_element, eye_glyph)
        end
        
        -- see if we can spot the player
        local spotted = track_player(direction)
        
        if spotted then
            -- we spotted the player, make a spotted noise
            -- print("tracking in non tracking mode")
            self:sounds('tracking', 0)
        end

    end
    
end


function EyeBeast:sounds(command, dt)
       
    local eat_sounds = { 'eb_eat1', 'eb_eat2' }
    local chase_sounds = { 'eb_come_back1', 'eb_come_back2', 'eb_gonna_eat' }
    
    if self.never_seen_player == true then
        transmit_named_located_sound('eb_first_spot', 8, self.line, self.column, self.map_name)
        self.sound_delay = (math.random(10) * 0.1) + 1  
        self.never_seen_player = false
    elseif command == 'eat' then
        transmit_named_located_sound(random_entry(eat_sounds), 8, self.line, self.column, self.map_name)
    elseif command == 'tracking' then
        self.sound_delay = self.sound_delay - dt
        if self.sound_delay < 0 then
            transmit_named_located_sound(random_entry(chase_sounds), 8, self.line, self.column, self.map_name)
            self.sound_delay = (math.random(10) * 0.1) + 1  
        end
    end
    
end

return EyeBeast
