-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Mob base class
-- Created by Tony Park, July 2014
--
require('strict')
require('utils')
class = require('middleclass')

local Mob = class('Mob')
local avatars = require("avatars/avatar_loader")
local mob_counter = 0

require('controllers/collisions')

function hold_all_mobs(map_name)
    -- NOTE! not a class method!!
    if gulp.LuaStateCodeMode == 'SERVER' then
        if check_map_exists(map_name) then 
            for _,mob in pairs(get_map(map_name).auto.mobs) do
                mob.held = true
            end
        end
    end
end

function release_all_mobs(map_name)
    -- NOTE! not a class method!!
    if gulp.LuaStateCodeMode == 'SERVER' then
        if check_map_exists(map_name) then
            for _,mob in pairs(get_map(map_name).auto.mobs) do
                mob.held = false
            end
        end
    end
end
    
-- mobs need a map
-- if you don't provide one for initialisation, you better do it atomically with the creation of the mob
function Mob:initialize(avatar, map_name)

    self.visible = false
    self.line = 0
    self.column = 0
    self.avatar = avatar
    self.avatar_data = avatars[avatar]
    self.map_name = map_name
    self.held = false
    
    mob_counter= mob_counter + 1
    self.mob_number = mob_counter
    
    -- remember the last place we printed, so we can skip the 
    -- (network) printing if we haven't moved significantly
    self.last_display_line = -1
    self.last_display_column = -1
    
    -- for remote mobs
    self.network_id = nil               -- initially not on network

    self.display_elements = {}
    local maze = get_maze(self.map_name)
    
    local base_display_element = new_hidden_mob_display_element(maze, self.avatar_data.graphic_unicode_point, self.line, self.column, 100, get_simple_colour_by_name('BRIGHT_WHITE'))
    self:register_display_element(base_display_element)

    -- register ourselves for draw and update too - and active input, if we decided to handle it
    gulp._delegate_list:add_delegate(self)
    
    -- register for collisions if necessary
    register_for_collisions("Mob"..self.mob_number, self, self.avatar_data.collision_rect)
    
    
    --if self.avatar_data.collision_rect then
    --    self.collision_ref = register_for_collisions(self.avatar_data.collision_rect)
    --end
    

end

-- a finisher method
function Mob:remove()
    gulp._delegate_list:drop_delegate(self)
    --self.gr:delete()
end

-- save the mob description as a string
function Mob:serialise()
    
    if self.network_id == nil then return end
    
    local message = {
        mt = "mob",
        id = self.network_id,
        avatar = self.avatar,
        line = self.line,
        column = self.column,
        map_name = self.map_name,
        display_elements = self.display_elements
    }
    
    return message
end

-- NOTE!!! Not a class function
-- creates a mob from a serialised message - CLIENT only
function NewMob(message)
    
    local avatar = message.avatar
    local column = message.column
    local line = message.line
    local map_name = message.map_name
    
    if not gulp.maps[map_name] then
        print("Got mob create message for "..map_name.." when it is not loaded")
        return
    end
    
    -- create the mob and set the simple data
    local mob = Mob(avatar, map_name)
    mob:local_set_display_location(line, column)
    
    mob:set_network_id(message.id)
    
    -- now set the glyphs for the display elements
    for key,element in ipairs(message.display_elements) do
            
        local glyph = element.glyph
        local layer = element.layer
        local visible = element.visible
        
        if key == 1 then
            -- first display element automatically exists, just set the glyph
            mob.display_elements[1]:set_glyph(glyph)
            mob.display_elements[1]:set_layer(layer)
        else
            -- create a display element and register it (assume it is visible for now)
            local maze = get_maze(map_name)
            local element = new_hidden_mob_display_element(maze, glyph, line, column, layer, get_simple_colour_by_name('BRIGHT_WHITE'))
            mob:register_display_element(element)
        end
        
        mob:local_set_visible(visible, key)
            
    end
    
    return mob
end

function Mob:set_map(map_name)
    self.map_name = map_name
end


function Mob:set_network_id(id)
    self.network_id = id
end

function Mob:register_display_element(element)
    table.insert(self.display_elements, element)
    return #self.display_elements       -- give a pointer into the table for this new element
end

function Mob:finish()
    
    if gulp.LuaStateCodeMode == 'SERVER' then
        if self.network_id ~= nil then 
            local message = { 
                mt = "mob_finish",
                id = self.network_id,
                map_name = self.map_name,
            }
            gulp_networking:broadcast_to_all_with_map_loaded(self.map_name, message)
        end
    end
    
    if gulp.LuaStateCodeMode == 'CLIENT' then
        for _, object in pairs(self.display_elements) do
            object:hide()
        end
    end
    
    self.display_elements = {}
    gulp._delegate_list:drop_delegate(self)
    
    deregister_for_collisions("Mob"..self.mob_number)
end

function Mob:set_glyph(element, glyph)
    
    if gulp.LuaStateCodeMode=='SERVER' then 
        
        -- SERVER
        
        -- check the glyph has actually changed before launching a network message...
        if self.display_elements['glyph'] ~= glyph then
            self:network_set_glyph(element, glyph)
        end
    
    else
    
        --CLIENT
        
        -- check the glyph has actually changed before launching a network message...
        if self.display_elements[element]:get_glyph() ~= glyph then
            self:local_set_glyph(element, glyph)
        end
    
    end
    
end

function Mob:local_set_glyph(element, glyph)
    self.display_elements[element]:set_glyph(glyph)
end

function Mob:network_set_glyph(element, glyph)
    
    if self.network_id == nil then return end
    
    local message = { 
        mt = "mob_glyph",
        element = element,
        glyph = glyph,
        id = self.network_id,
        map_name = self.map_name,
    }
    
    gulp_networking:broadcast_to_all_with_map_loaded(self.map_name, message)
end

function Mob:process_glyph_message(message)
    -- mob_glyph tags already stripped
    -- id already processed
    
    self:local_set_glyph(message.element, message.glyph)
end             


function Mob:reset()
    self:set_location(self.start_line, self.start_column)
end

function Mob:set_location(line, column, create)
    local location_delta_required = gulp.minimum_cell_movement_before_transmit_in_cells
    
    if create ~= nil then
        self.start_line = line
        self.start_column = column
    end
    
    -- if we are held, then don't move!!
    if self.held then return end
    
    -- shift the actual location
    self.line = line
    self.column = column
    
    -- check if we've moved a significant distance
    if (math.abs(line - self.last_display_line) > location_delta_required) or
        (math.abs(column - self.last_display_column) > location_delta_required) then
        
        self.last_display_line = line
        self.last_display_column = column
        
        if gulp.LuaStateCodeMode == 'CLIENT' then
            self:local_set_display_location(line, column)
        else
            self:network_set_display_location(line, column)
        end
    end
end

function Mob:local_set_display_location(line, column)
    
    self.line = line
    self.column = column
    
    for _,v in pairs(self.display_elements) do
        v:set_line(line)            -- line & column always 0 based
        v:set_column(column)
    end

end

function Mob:network_set_display_location(line, column)
    
    if self.network_id == nil then return end
    
    local message = { 
        mt = "mob_location",
        line = line,
        column = column,
        map_name = self.map_name,
        id = self.network_id 
    }
   
    gulp_networking:broadcast_to_all_with_map_loaded(self.map_name, message)
end

function Mob:process_location_message(message)
    -- mob_location tags already stripped
    -- id already processed
   
    if message.map_name == self.map_name then
        self:local_set_display_location(message.line, message.column)
    end
end             

function Mob:hide()
    self:set_visible(false)
end

function Mob:show()
    self:set_visible(true)
end

function Mob:set_visible(visible, element)
    -- element will be nil if we are operating on the whole mob
    
    -- check if there is anything to do
    if visible == self.visible and element == nil then return end
    
    if gulp.LuaStateCodeMode == 'CLIENT' then
        self:local_set_visible(visible, element)
    else
        self:network_set_visible(visible, element)
    end
    
    if element == nil then
        self.visible = visible
    end

    if gulp.LuaStateCodeMode == 'SERVER' then
        if element == nil then
            for _,element in ipairs(self.display_elements) do
                element['visible'] = visible
            end
        else
            self.display_elements[element]['visible'] = visible
        end
    end

end

function Mob:local_set_visible(visible, element)
    
    if element == nil then
    -- we are showing / hiding the whole mob
        if visible then
            for _,v in pairs(self.display_elements) do v:show() end
        else
            for _,v in pairs(self.display_elements) do v:hide() end
        end
    else
        if visible then
            self.display_elements[element]:show()
        else
            self.display_elements[element]:hide()
        end
    end
end

function Mob:network_set_visible(visible, element)
    
    if self.network_id == nil then return end
    
    local message = { 
        mt = "mob_visibility",
        visible = visible,
        element = element,
        id = self.network_id, 
        map_name = self.map_name,
    }
   
    gulp_networking:broadcast_to_all_with_map_loaded(self.map_name, message)
end

function Mob:process_visibility_message(message)
    -- mob_visibility tags already stripped
    -- id already processed
   
    self:local_set_visible(message.visible, message.element)
end             

function Mob:hold()
    self.held = true
end

function Mob:release()
    self.held = false
end


return Mob

