-- Forlorn Fox
-- Copyright © 2020 Rob Probin & Tony Park.
-- No redistribution without permission.
--
-- Caves Mob base class
--

require('strict')
class = require('middleclass')

local Mob = require("controllers/Mob")

local CavesMob = class('CavesMob', Mob)

local meter = require("controls/meter")

--[[
    stats are
    
    hp: hit points
    damage: damage per hit
    delay: time between hits
    hit chance: base chance to hit (%)
    level: level of mob
    
--]]

-- this is a general delay on opening the combat menu to allow everyone to get ready
local combat_start_delay = 2

function CavesMob:initialize(avatar, map_name, combat_data)
    
    Mob.initialize(self,avatar,map_name)
    
    self.hp = combat_data.hp
    self.damage = combat_data.damage
    self.hit_time = combat_data.hit_time
    self.hit_chance = combat_data.hit_chance
    self.display_name = combat_data.display_name
    self.level = combat_data.level

    self.players_in_combat = {}
    
    self.in_combat = false
    
    local meter_data={      -- this data should take in self.time_to_hit, and self.delay, and the random element, and the check in update
        running = false,
        initial_time = 0, 
        max_time = 6, 
    }
    self.mob_meter = meter(meter_data)

end

function CavesMob:get_level()
    return self.level
end

function CavesMob:calculate_next_hit_at()
    self.next_hit_at = self.hit_time + math.random()*(6-self.hit_time)
    if self.next_hit_at < self.mob_meter.time + 0.1 then
        self.next_hit_at = self.mob_meter.time + 0.1
    end
end

function CavesMob:start_combat(combat_id, manager_name)
   
    -- start a new combat
    --self.time_to_hit = combat_start_delay + (self.delay * math.random())
    self.manager = manager_name
    self.combat_id = combat_id
    
    self.mob_meter.running = true
    self.mob_meter.time = self.hit_time - (self.hit_time * math.random())
    self.in_combat = true
    
    -- time of the next hit will be a random time between the lowest possible (hit_time) and the top of the meter
    self:calculate_next_hit_at()
    
    self:hold()

end    

function CavesMob:take_damage(damage)
    self.hp = self.hp - damage
    self:hp_messages()
    
    if self.hp <= 0 then
        --self:death_messages(damage, player_id)
        self:die()
    end
end

function CavesMob:is_dead()
    return self.hp <= 0
end

function CavesMob:hp_messages()
    local message = { 
        mt = "game_control",
        manager = self.manager,
        command = "caves_mob_hp_update",
        mob_id = self.network_id,
        mob_hp = self.hp
    }
    local manager = find_game_manager_by_name(self.manager)
    manager:broadcast_to_all_players_in_combat(self.combat_id, message)
end

function CavesMob:die()
    -- need to end combat, but after a suitable delay
    -- caves manager should control this
    
    -- stop the mob from taking any more actions
    self.in_combat = false
    
    local manager = find_game_manager_by_name(self.manager)
    
    self:hold()
    self:hide()
    
    self:finish()
    
    manager:mob_died(self.combat_id, self.network_id)
end


function CavesMob:hit_player(player)
    
    local damage = 0
    local roll = math.random(100)
    
    local player_level = player:get_level()
    local level_diff = self.level - player_level
    
    local hit_chance = self.hit_chance
    
    if self.level < player_level then
        for i = 1,player_level-self.level do
            hit_chance = hit_chance * 0.8
        end
    elseif self.level > player_level then
        for i = 1,self.level-player_level do
            hit_chance = hit_chance * 1.2
        end
    end

    debugMessage("Mob level:"..self.level.." Player level:"..player_level.." Hit chance:"..hit_chance.." Roll:"..roll)

    if roll <= hit_chance then
        damage = self.damage
    end 
    
    return damage

end



function CavesMob:update(dt)
    
    if self.in_combat then
        
        if self.mob_meter.time > self.next_hit_at then
            
            self.mob_meter.time = self.mob_meter.time - 2
            
            -- if the last hit was very late, the next hit time might be less than the current time
            -- in which case, the mob will hit again in rapid succession
            -- in future we could adapt this algorithm on a per mob basis
            self:calculate_next_hit_at()
            
            local manager = find_game_manager_by_name(self.manager)
            local combat = manager.combats[self.combat_id]
            if combat then
                -- if we are are in combat with multiple characters, need to choose who to hit, and send messages to each appropriately
                local player_to_hit = self:choose_player_to_hit(combat)
                
                if player_to_hit then
                
                    local damage = self:hit_player(player_to_hit)
                    player_to_hit:take_damage(damage, self.manager)
                    
                    local hit_description, description_index = self:get_hit_description(damage)
                    
                    local message = { 
                        mt = "game_control",
                        manager = self.manager,
                        command = "caves_mob_attack",
                        damage = damage,
                        description = hit_description,
                        target = true,
                        meter_time = self.mob_meter.time,
                    }
                    player_to_hit.socket_ref:SendMessage(message)
                    
                    -- adjust message for other players and broadcast
                    message.description = self:get_hit_other_description(damage, player_to_hit.name, description_index)
                    message.target = false,
                    manager:broadcast_to_all_players_in_combat_except(self.combat_id, message, player_to_hit.id)                        
                    
                    -- do player death description
                    if player_to_hit:is_dead() then
                        local _message = { 
                            mt = "game_control",
                            manager = self.manager,
                            command = "caves_message",
                        }
                    
                        _message.description = self:get_kill_description(player_to_hit, damage)
                        player_to_hit.socket_ref:SendMessage(_message) 
                    
                        _message.description = self:get_other_kill_description(player_to_hit, damage)
                        manager:broadcast_to_all_players_in_combat_except(self.combat_id, _message, player_to_hit.id)                        
                    end
                    
                end
            end
        end
    end
end

return CavesMob

    
