-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
--[[

    Set of graphics helpher functions
    
--]]

local utf8 = require("utilities/gulp_utf8")
require('utils')
require('graphics/glyph_sets')
require('graphics/colours')


function load_graphics()
    local graphics = gulp.app:get_standard_graphics_context()
    for k,v in pairs(glyph_sets) do
        if v.load_from_file then
            graphics:load_textures_from_glyph_set(v);
        else
            graphics:create_texture_set(v);     
        end
    end
    
    -- write the default set
    default_font(graphics)
end


function swap_font(gr, set)
    local GTI = gr:get_GameTexInfo(set.character_base_code)
    gr:overwrite_GameTexInfo(0, GTI)
end

function default_font(gr)
    swap_font(gr, glyph_sets.spectrum2)
end    

-- defines the highest possible view layer - this is what we look at by default
top_layer = 32767

local gr
local last_colour
local adl   -- absolute draw list
--local mdl -- mobs draw list

function set_graphics_context(graphics)
    gr = graphics
end

function get_last_graphics_context()
    -- this is a hack!
    return gr
end

function set_absolute_draw_list(draw_list)
    adl = draw_list
end


function render_draw_lists(gr)
    --sdl:render_simple(gr)
    --msdl:render_simple(gr)
    adl:render_simple(gr)
end
--[[
function set_sdl_rect(left, top, right, bottom)
    sdl:set_rect(left, top, right, bottom)
end

function set_msdl_rect(left, top, right, bottom)
    msdl:set_rect(left, top, right, bottom)
end
--]]
function set_adl_rect(left, top, right, bottom)
    adl:set_rect(left, top, right, bottom)
    adl:set_draw_location_mode(adl.PIXEL_BASED)
end

function new_absolute_display_element(glyph, line, column, layer, fg_colour, bg_colour, transparent)
    local display_element = gulp_cpp.DrawListElement(adl, glyph, line, column, layer, fg_colour)
    display_element:set_bg_colour(bg_colour)
    display_element:set_bg_transparency(transparent)
    if display_element ~= nil then
        display_element:show()
    end
    
    return display_element
end

function new_hidden_absolute_display_element(glyph, line, column, layer, fg_colour, bg_colour, transparent)
    local display_element = gulp_cpp.DrawListElement(adl, glyph, line, column, layer, fg_colour)
    display_element:set_bg_colour(bg_colour)
    display_element:set_bg_transparency(transparent)
    return display_element
end

function new_mob_display_element(maze, glyph, line, column, layer, fg_colour)
    local mdl = maze:get_mobs_draw_list()
    local display_element = gulp_cpp.DrawListElement(mdl, glyph, line, column, layer, fg_colour)
    if display_element ~= nil then
        display_element:show()
    end
    
    return display_element
end

function new_hidden_mob_display_element(maze, glyph, line, column, layer, fg_colour)
    local mdl = maze:get_mobs_draw_list()
    local display_element = gulp_cpp.DrawListElement(mdl, glyph, line, column, layer, fg_colour)
    return display_element
end

function new_maze_display_element(maze, line, column, glyph, layer)
    local mdle = gulp_cpp.MazeDrawListElement(maze:get_maze_draw_list(line, column), glyph, layer, 0, 0)
    return mdle
end

function new_maze_display_animated_element(maze, line, column, glyph_list, layer)
local mdle = gulp_cpp.MazeDrawListAnimatedElement(maze:get_maze_draw_list(line, column), glyph_list, layer)
return mdle
end

function check_mdls(maze, message)
    if maze.height == nil then return end
    if maze.width == nil then return end
    print("Start check integrity", message)
    for h = 1,maze:height() do
        for w = 1,maze:width() do
            local mdl = maze:get_maze_draw_list(h, w)
            mdl:check_integrity()
        end
    end
    print("Finish check integrity", message)
end


function set_bg_colour(colour)
    if type(colour) == 'string' then
        -- descriptive colour name - look it up
        -- recurse here to allow name lookup to return either
        -- a number, a table or an SDL colour (or even another string
        -- if we wanted aliases)
        gr:set_bg_fullcolour(get_SDL_Color_by_name(colour))
    elseif type(colour) == 'table' then
        -- assume a table which is convertable to an SDL colour
        gr:set_bg_fullcolour(colour_to_SDL_Color(colour))
    elseif type(colour) == 'userdata' then
        -- assume SDL colour
        gr:set_bg_fullcolour(colour)        
    elseif type(colour) == 'number' then
        -- simple colour
        gr:set_bg_colour(colour)
    else
        print ("WARNING! set_bg_colour() did not understand colour type '" .. type(colour) .. "'")
    end
    
    last_colour = colour
end

function get_current_bg_colour()
    return last_colour
end

function set_bg_colour_transparent()
    local sdl_colour = gulp_cpp.SDL_Color()
    sdl_colour.a = 0x00 
    sdl_colour.r = 0x00 
    sdl_colour.g = 0x00 
    sdl_colour.b = 0x00
    set_bg_colour(sdl_colour)
end

function set_bg_transparent()
    gr:set_bg_transparent()
end

function set_bg_opaque()
    gr:set_bg_opaque()
end


function set_fg_colour(colour)
    if type(colour) == 'string' then
        set_fg_colour(get_simple_colour_by_name(colour))        
    elseif type(colour) == 'number' then
        gr:set_fg_colour(colour)
    else
        print ("WARNING! set_fg_colour() did not understand colour type '" .. type(colour) .. "'")
        -- could do some mapping from colour table or SDL colour to a number here is necessary
    end
end

function blank_line(line)
    gr:go_to(line, 0)
    gulp_cpp.print_string(gr, string.rep(" ", 32))
end

function blank_screen()
    for i = 0, 23 do
        blank_line(i)
    end
end

function print_string_at(line, col, str)
    gr:go_to(line, col)
    gulp_cpp.print_string(gr, str)
end

function print_glyph_at(line, col, glyph)
    gr:go_to(line, col)
    gulp_cpp.print_glyph(gr, glyph)
end

function print_large_glyph_at(line, col, glyph, cell_width, cell_height, rotation, ratio)
    gr:go_to(line, col)
    gulp_cpp.print_glyph_ex(gr, glyph, rotation, ratio, cell_width, cell_height)
end

function colour_to_pixel(colour, colour_format)
    return  bit32.band(bit32.lshift(colour.r, colour_format.Rshift), colour_format.Rmask) +
            bit32.band(bit32.lshift(colour.g, colour_format.Gshift), colour_format.Gmask) +
            bit32.band(bit32.lshift(colour.b, colour_format.Bshift), colour_format.Bmask) +
            bit32.band(bit32.lshift(colour.a, colour_format.Ashift), colour_format.Amask) 
end

function compare_colours(c1, c2)
    return (c1.r == c2.r and c1.g == c2.g and c1.b == c2.b)
end

function colour_to_SDL_Color(colour)
    local sdl_colour = gulp_cpp.SDL_Color()
    sdl_colour.a = colour.a 
    sdl_colour.r = colour.r 
    sdl_colour.g = colour.g 
    sdl_colour.b = colour.b
    return sdl_colour
end


function pixel_to_colour(pixel, colour_format)
    local colour = gulp_cpp.SDL_Color()
    colour.r = bit32.rshift(bit32.band(pixel, colour_format.Rmask), colour_format.Rshift) 
    colour.g = bit32.rshift(bit32.band(pixel, colour_format.Gmask), colour_format.Gshift)
    colour.b = bit32.rshift(bit32.band(pixel, colour_format.Bmask), colour_format.Bshift)
    colour.a = bit32.rshift(bit32.band(pixel, colour_format.Amask), colour_format.Ashift)
    
    if(compare_colours(colour, gulp.colour.hot_pink_means_transparent)) then
        return colour, 'transparent'
    end
    
    return colour, 'colour'
end


function goto_position(line, column)
    gr:go_to(line, column)
end

-- http://bluesock.org/~willg/dev/ansi.html
-- http://www.termsys.demon.co.uk/vtansi.htm
-- http://en.wikipedia.org/wiki/ANSI_escape_code
-- 
local function _next_ansi_control_code(str, index)
    index = index or 0
end

local _ANSI_reverse_flag = false
local _ANSI_fg_bright_flag = false
local _ANSI_fg_colour = gulp.colour.BLACK.simple_colour
local _ANSI_bg_colour = gulp.colour.WHITE.simple_colour
local _ANSI_cursor_save_line = 1
local _ANSI_cursor_save_column = 1

local _ANSI_bright_transform = {
    [gulp.colour.BLACK.simple_colour] = gulp.colour.BRIGHT_BLACK.simple_colour,
    [gulp.colour.RED.simple_colour] = gulp.colour.BRIGHT_RED.simple_colour,
    [gulp.colour.GREEN.simple_colour] = gulp.colour.BRIGHT_GREEN.simple_colour,
    [gulp.colour.YELLOW.simple_colour] = gulp.colour.BRIGHT_YELLOW.simple_colour,
    [gulp.colour.BLUE.simple_colour] = gulp.colour.BRIGHT_BLUE.simple_colour,
    [gulp.colour.MAGENTA.simple_colour] = gulp.colour.BRIGHT_MAGENTA.simple_colour,
    [gulp.colour.CYAN.simple_colour] = gulp.colour.BRIGHT_CYAN.simple_colour,
    [gulp.colour.WHITE.simple_colour] = gulp.colour.BRIGHT_WHITE.simple_colour,
    [gulp.colour.ORANGE.simple_colour] = gulp.colour.BRIGHT_ORANGE.simple_colour
}

function _ANSI_do_fg_stuff(c)
    local c_type = type(c)
    if c_type == "table" then
        local c2 = c['simple_colour']
        if c2 then
            c = c2
        else
            c = gulp.colour.BLACK.simple_colour -- replacement colour
        end
    elseif c_type ~= "number" then
        c = gulp.colour.RED.simple_colour
    end
    if _ANSI_fg_bright_flag then
        local c2 = _ANSI_bright_transform[c]
        if c2 then
            -- only make colours bright that we have in above table...
            c = c2
        end
    end
    return c
end

function ANSI_set_fg_colour(c)
    _ANSI_fg_colour = c
    if _ANSI_reverse_flag then
        set_bg_colour(c)
    else
        c = _ANSI_do_fg_stuff(c)
        set_fg_colour(c)
    end
end

function ANSI_set_bg_colour(c)
    _ANSI_bg_colour = c
    if not _ANSI_reverse_flag then
        set_bg_colour(c)
    else
        c = _ANSI_do_fg_stuff(c)
        set_fg_colour(c)
    end
end

local function _ANSI_set_colours()
    ANSI_set_fg_colour(_ANSI_fg_colour)
    ANSI_set_bg_colour(_ANSI_bg_colour)
end

function ANSI_reset_attributes()
    _ANSI_fg_bright_flag = false
    _ANSI_reverse_flag = false
    _ANSI_set_colours()
end


local _ANSI_colour_and_text_formatting = {
    -- There can be multiple text formatting sequence numbers separated by 
    -- a ; and ending with an m where the # is of the following values: 
    --
    --  attributes
    ['0'] = ANSI_reset_attributes, -- normal display / Reset all attributes
    ['1'] = function()
        _ANSI_fg_bright_flag = true
        _ANSI_set_colours()
        end, -- bold/bright
    --['2'] = nil, -- dim
    -- 3 italic
    -- [4] = underline (mono only)
    -- [5] = blink on
    -- 6 rapid blink
    ['7'] = function() -- [7] = reverse video on / Negative -- swap foreground and background
        _ANSI_reverse_flag = true
        _ANSI_set_colours()
    end,
    
    -- [8] = nondisplayed (invisible) / hidden
    -- foreground colors
    ['30'] = function() ANSI_set_fg_colour(gulp.colour.BLACK.simple_colour) end,
    ['31'] = function() ANSI_set_fg_colour(gulp.colour.RED.simple_colour) end,
    ['32'] = function() ANSI_set_fg_colour(gulp.colour.GREEN.simple_colour) end,
    ['33'] = function() ANSI_set_fg_colour(gulp.colour.YELLOW.simple_colour) end,
    ['34'] = function() ANSI_set_fg_colour(gulp.colour.BLUE.simple_colour) end,
    ['35'] = function() ANSI_set_fg_colour(gulp.colour.MAGENTA.simple_colour) end,
    ['36'] = function() ANSI_set_fg_colour(gulp.colour.CYAN.simple_colour) end,
    ['37'] = function() ANSI_set_fg_colour(gulp.colour.WHITE.simple_colour) end,
    
    -- 38   Reserved for extended set foreground color  typical supported next 
    -- arguments are 5;x where x is color index (0..255) or 2;r;g;b where 
    -- r,g,b are red, green and blue color channels (out of 255)

    -- background colors
    ['40'] = function() ANSI_set_bg_colour(gulp.colour.BLACK.simple_colour) end,
    ['41'] = function() ANSI_set_bg_colour(gulp.colour.RED.simple_colour) end,
    ['42'] = function() ANSI_set_bg_colour(gulp.colour.GREEN.simple_colour) end,
    ['43'] = function() ANSI_set_bg_colour(gulp.colour.YELLOW.simple_colour) end,
    ['44'] = function() ANSI_set_bg_colour(gulp.colour.BLUE.simple_colour) end,
    ['45'] = function() ANSI_set_bg_colour(gulp.colour.MAGENTA.simple_colour) end,
    ['46'] = function() ANSI_set_bg_colour(gulp.colour.CYAN.simple_colour) end,
    ['47'] = function() ANSI_set_bg_colour(gulp.colour.WHITE.simple_colour) end,
    
    -- 48   Reserved for extended set background color  typical supported next 
    -- arguments are 5;x where x is color index (0..255) or 2;r;g;b where 
    -- r,g,b are red, green and blue color channels (out of 255)
}


local _ANSI_commands = {
    m = function(params)
        for param in params:gmatch("(%d+);?") do
            local f = _ANSI_colour_and_text_formatting[param]
            if f then
                f()
            else
                print("ANSI print unrecognised parameter", param)
            end
        end
    end,
    D = function(params)
        local n = tonumber(params)
        if not n then
            n = 1
        end
        local column = gr:get_column()
        column = column - 1
        gr:go_to(gr:get_line(), column)
    end,
    s = function()
        _ANSI_cursor_save_column = gr:get_column()
        _ANSI_cursor_save_line = gr:get_line()      
    end,
    u = function()
        gr:go_to(_ANSI_cursor_save_line, _ANSI_cursor_save_column)
    end,

}

local function _process_ANSI_command(command, params)
    local func = _ANSI_commands[command]
    if func then
        func(params)
    else
        print("ANSI_print: unknown command", command, params)
    end
end

local ANSI_find_basic_pattern =       "\x1b%[%A-%a"
local ANSI_find_capture_pattern = "(.-)\x1b%[(%A-)(%a)"

function ANSI_print(str)
    local s, e, before, params, command = str:find(ANSI_find_capture_pattern)
    if not s then
        -- no ANSI control sequences - just print the line
        gulp_cpp.print_string(gr, str)
    else
        local last_e
        repeat
            last_e = e
            gulp_cpp.print_string(gr, before)
            _process_ANSI_command(command, params)
            s, e, before, params, command = str:find(ANSI_find_capture_pattern, e+1)
        until not s
        gulp_cpp.print_string(gr, str:sub(last_e+1))
    end
    
    --for a,b,c in str:gmatch "(.-)[\x1b%[(%A-)(%a)]?" do

    --[[
    local i = 1
    while i < #str do
        local s, e, param, command = str:find("\x1b%[(%A-)(%a)", i)
        if s then
            gulp_cpp.print_string(gr, str:sub(i, s-1))
        else
            gulp_cpp.print_string(gr, str:sub(i))
            break
        end
    end 
    --]]
end

function ANSI_printable_len(str)
    -- replace all Escape[....A with nothing
    -- where A is a letter
    -- We also sort out multibyte UTF8 characters
    return utf8.codepoints(str:gsub(ANSI_find_basic_pattern, ""))
end

-- various ANSI strings
ANSI_string_reset_attr = "\x1b[0m"
ANSI_string_cursor_back1 = "\x1b[D"
ANSI_string_save_cursor_position = "\x1b[s"
ANSI_string_recall_cursor_position = "\x1b[u"

function get_ANSI_string_cursor_back_x(steps)
    return string.format("\x1b[%dD", steps)
end

--
-- Triangle draw for debug purposes.
--
function DrawTriangle(gr, colour, x1, y1, x2, y2, x3, y3)
    gr:DrawLine(colour, x1, y1, x2, y2)
    gr:DrawLine(colour, x1, y1, x3, y3)
    gr:DrawLine(colour, x3, y3, x2, y2)
end

--
-- very rough circle for debug purposes.
--
function DrawCircle(gr, colour, x, y, radius)
    gr:DrawPoint(colour, x, y)
    local step = 15
    for angle = 0, 360, step do
        local x1 = x + radius * math.sin(math.rad(angle))
        local y1 = y + radius * math.cos(math.rad(angle))
        local x2 = x + radius * math.sin(math.rad(angle+step))
        local y2 = y + radius * math.cos(math.rad(angle+step))
        gr:DrawLine(colour, x1, y1, x2, y2)
    end
end

local function get_colour_for_render_debug_data(p)
    local c = gulp.colour[p]
    if c then
        return colour_to_SDL_Color(c)
    end
    return colour_to_SDL_Color(gulp.colour.MAGENTA)
end

function render_debug_data(gr)
    if gulp.server_debug_print_table then
        for _,v in pairs(gulp.server_debug_print_table) do
            local display_type = v[1]
            local line = v[2]
            local column = v[3]
            local param1 = v[4]
            local param2 = v[5]
            local param3 = v[6]
            if display_type == "text" then
                print_string_at(line, column, param1)
            elseif display_type == "rect" then
                gr:DrawRect(get_colour_for_render_debug_data(param3), line, column, param1, param2)
            elseif display_type == "mazerect" then
                gr:DrawRect(get_colour_for_render_debug_data(param3), line, column+gulp.TOP_UI_HEIGHT, param1, param2+gulp.TOP_UI_HEIGHT)
            elseif display_type == "fillrect" then
                gr:FillRect(get_colour_for_render_debug_data(param3), line, column, param1, param2)
            elseif display_type == "line" then
                gr:DrawLine(get_colour_for_render_debug_data(param3), line, column, param1, param2)
            elseif display_type == "point" then
                gr:DrawPoint(get_colour_for_render_debug_data(param3), line, column)
            elseif display_type == "circle" then
                gr:DrawCircle(get_colour_for_render_debug_data(param3), line, column, param1)
            end
                --DrawTriangle(gr, colour, 10, 10, 15, 20, 20, 15)
        end
    end
end

