-- Forlorn Fox main file
-- Copyright © 2014-2020 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
gulp.LuaStateCodeMode = "CLIENT"

-- uncomment to get a small amount of debug information about level loading
-- gulp.DebugLevelLoading = true

-- engine version check
local required_engine = 0.86
if gulp.forlorn_fox_engine_version < required_engine then
    print("*******************************************")
    print("Engine too old - needs " .. required_engine .. ", currently "..gulp.forlorn_fox_engine_version)
    print("*******************************************")
    os.exit(1)
end

local check_debugger = require("debug_select")
check_debugger("ui")

main_common = {}

if argk['--OPEN_STD_CONSOLE'] then -- and gulp.platform ~= "iPhone" then
    if argk['SERVER'] then
        require("server_CLI_commands")
    end
    -- allow commands from stdin
    gulp.lua_main:open_console()
end

main_common.hotswap_reloader = require("utilities/hotswap_reloader")
if argk['--HOTSWAP_RELOADER'] then
    main_common.hotswap_reloader.enable() 
end

-- Check if keepalive disable command is passed in
gulp.enableKeepAlive = true
if argk['--NOKEEPALIVE'] then
    -- turn off the keep alive
    gulp.enableKeepAlive = false
end

require("game_configuration")

require("helper")


function main_common.decode_command_line_arguments()
    
    -- arg[0] = program, arg[1] = first parameter, arg[2] = second parameter, etc.
    --
    -- NOTE: We don't use #arg because (according to Lua 5.2 reference "the length 
    -- of a table t is only defined if the table is a sequence, that is, the set of 
    -- its positive numeric keys is equal to {1..n} for some integer n."
    -- "Note, however, that non-numeric keys do not interfere with whether a table 
    -- is a sequence."

    -- CLIENT mode must supply IP and PORT
    -- SERVER mode must supply PORT
    -- SIZE must specify a size (as a glyph size multiplier - 1 is very small, 5 is pretty big)


    -- set the gulp variables with default values
    gulp.NetworkError = false
    gulp.gameName = "My World"
    
    if arg[1] == nil then
        -- redundant info
        --print("No command line args passed in, using defaults")
    else
        local client_index = argk['CLIENT']
        local server_index = argk['SERVER']
        if client_index then
            gulp.target_IPAddr = arg[client_index+1]
            gulp.target_port = arg[client_index+2]
        elseif server_index then
            --gulp.prefs.network_port = arg[server_index+1]
            gulp.server = true
        end
        
        local size_index = argk['SIZE']
        if size_index then
            gulp_cpp.set_screen_size(arg[size_index+1])
        end 
        
        local map_index = argk['MAP']
        if map_index then
            gulp.start_map_name = arg[map_index+1]
        end
                
        if not gulp.nogui and argk['--show-renderers'] then
            local num_r = gulp_cpp.RendererInfo.number_renderers
            local current_renderer = gulp_cpp.RendererInfo(gulp.app:get_SDL_renderer(), 0)
            --debugMessage(string.format("Current Renderer: %s (%s)", current_renderer.name, render_flag_decode(current_renderer)) )
            for i = 0, num_r-1 do
                local rend = gulp_cpp.RendererInfo(nil, i)
                local current = ""
                if current_renderer.name == rend.name then
                    current = "ACTIVE"
                end
                debugMessage(string.format("Renderer %i: %s (%s) %s", i, rend.name, render_flag_decode(rend), current))
            end
        end
        
        if argk['--verbose-engine'] then
                gulp.app:verbose_engine(true)
        end
        
    end
end

local LuaMainThreaded = require("utilities/LuaMainThreaded")

function gulp.create_new_game(game_name)
    local new_game_index = #gulp.lua_game_world + 1
    
    print("New game: "..game_name.." "..new_game_index)
    
    gulp.ui_queue[new_game_index] = gulp_cpp.LuaStateQueue("ui_queue_"..new_game_index)
    gulp.server_queue[new_game_index] = gulp_cpp.LuaStateQueue("server_queue_"..new_game_index)
    gulp.lua_game_world[new_game_index] = LuaMainThreaded("scripts/game_world_server/server_main.lua")
    gulp.lua_game_world[new_game_index]:run("server_load", gulp.ui_queue[new_game_index], gulp.server_queue[new_game_index])
    gulp.lua_game_world[new_game_index]:run_in_thread("server_thread", string.format("game_%d", new_game_index))
    
    local server_info = { qcommand = "start", start_map_name= gulp.start_map_name }
    
    server_info.LAN_broadcast = false -- not gulp.prefs.inhibit_broadcast
    server_info.internet_broadcast = false -- gulp.prefs.allow_internet_broadcast
    server_info.port = gulp.prefs.network_port + new_game_index
    server_info.enableKeepAlive = gulp.enableKeepAlive
    
    -- send the start command to the new game
    gulp.server_queue[new_game_index]:send(server_info)
    
    -- add the game to the list of known games
    -- gulp.game_list[new_game_index] = { game_name=game_name, port=server_info.port }
    
    -- send the new game info to the connected clients
    -- gulp.server_queue['control']:send { qcommand="update_game_list" }
    
    
    -- this is used to cause local client to connect to this server
    -- not an issue in our new server/client model but leave in place, this
    -- would be required for any potential 'offline' mode
    gulp.enable_client_networking = false
    
    -- this is used to shut down a local server at game end if necessary
    -- again not really relevant now, but might be wanted for offline mode
    gulp.local_server_started = true
        
    -- todo
    -- Add a new menu here that shows that the server is running, 
    -- and allows the player to join their own server game if they choose
    -- or just watch the game

    -- launch a player
    -- avatar_menu()
    
    return server_info.port
    
end    

local names_generator = require('utilities/names_generator')


function gulp.HandleRequestGame(message_in)
    
    local port
    if #gulp.lua_game_world == 0 then
        -- generate a game code name
        local game_name = message_in.game_name      -- allow for game name to be selected from client
        if not game_name or game_name == "" then
            game_name = names_generator.GenerateGameName()
        end

        -- actually create the game
        port = gulp.create_new_game(game_name)
        
    else
        -- always the first game
        port = gulp.prefs.network_port + 1
    end
    
    local response = {
        qcommand = 'game_created',
        mt = 'game_created',
        control_id = message_in.control_id,
        game_name = "Central Cavern 1",
        port = port,
    }
    gulp.server_queue['control']:send(response)
        
end

--
-- This is called just before the App crashes
--
function gulp.quit()
    if gulp.server then
        print("gulp.guit()")
    end
    -- notice pairs - includes 'control' and numerical indexes
    for _, q in pairs(gulp.server_queue) do
        q:send({ qcommand = "stop_thread" })
    end
    
    -- wait up to 0.5 for the server thread to respond
    for _ = 1,10 do
        gulp_cpp.SDL_Delay(50)
            
        -- not the most elegant structure. Could use goto continue ::continue::
        local found_running = false
        if gulp.controller and gulp.controller:IsRunning() then
            found_running  = true
        end
        for _,v in pairs(gulp.lua_game_world) do
            if v:IsRunning() then
                found_running = true
            end
        end
        if not found_running then
            break
        end
    end
    
    -- print out threads that are still blocked
    for _,v in pairs(gulp.lua_game_world) do
        if v:IsRunning() then
            error("Thread ".. v:GetThreadName() .. " still running on shutdown")
        end
    end
    if gulp.controller and gulp.controller:IsRunning() then
        error("Thread ".. gulp.controller:GetThreadName() .. " still running on shutdown")
    end

    --print("Server Game World thread return value", gulp.lua_game_world:WaitThread())
    -- Not strictly necessary, but we do it anyway...        
    if gulp.controller then gulp.controller:WaitThread() end
    
end

--[[
-------------------------------------------------------------------
--]]

if argk['SERVER'] and gulp.nogui then
    require('main_server_ui_thread')
else
    -- if people use --nogui with the client, then it will not work :-)
    require('main_client_ui_thread')
end


