-- Forlorn Fox
-- Copyright © 2019 Rob Probin.
-- No redistribution without permission.
-- Ported from C++
--[[

#Provides info not shown in the normal game like/:
 - What this game is
 - What the version is
 - Who wrote the code
 - What website this is located on
 - Hello's to people
 - Software Library/Framework credits (e.g. SDL v2.0)
--]]

require("strict")
local class = require("middleclass")
local SimpleImage = require("utilities/SimpleImage")

local TitleScreen = class('TitleScreen')
require('graphics/colours')

-- Flash rate was in ZX Spectrum attribute data, here it will stay...
-- http://www.worldofspectrum.org/faq/reference/48kreference.htm
local ZX_Spectrum_Flash_Rate_in_ms = ((1000/50)*16);	-- 32 frames for cycle


local STATE_screen_zoom = 0
local STATE_scroll1 = 1
--local STATE_show_cs_logo = 2


local default_paper = "BLUE"
if gulp.app_name == "Egg Tag" or gulp.app_name == "Eyebeast Escape" then
    default_paper = "BLACK"
end

local local_viewport = nil

function current_print_area_height_chars() return local_viewport:lines(); end
function current_print_area_width_chars() return local_viewport:columns(); end


function TitleScreen:initialize(renderer, version_string, viewport)
    self.viewport = viewport
    local_viewport = viewport
if gulp.app_name == "Forlorn Fox" then
    self.screen1 = SimpleImage(gulp_cpp.LoadPath("graphics/ff_title1.png"):c_str(), renderer, viewport)
    self.screen2 = SimpleImage(gulp_cpp.LoadPath("graphics/ff_title2.png"):c_str(), renderer, viewport)
else
    self.screen1 = SimpleImage(gulp_cpp.LoadPath("graphics/eggtag_title1.png"):c_str(), renderer, viewport)
    self.screen2 = SimpleImage(gulp_cpp.LoadPath("graphics/eggtag_title2.png"):c_str(), renderer, viewport)
end
    self.cover = SimpleImage(gulp_cpp.LoadPath("graphics/ff_authors.png"):c_str(), renderer, viewport)
    self.flash = false
    self.flash_accum = 0
    self.flash_count = 0
    self.scale = 100
    self.state = STATE_screen_zoom
    self.current_line = 0
    self.last_print_area_height = current_print_area_height_chars()
    self.version = version_string

	--std::vector<DisplayListBase*> display_list;

	self.cover:set_scale_to_fit_screen(70)
	self:_init_display_list();
end

function TitleScreen:destroy_display_list()
    self:delete_display_list()
end


local DisplayListImage = class('DisplayListImage')
function DisplayListImage:initialize(image_to_draw, alt_flash_image)
    self.image = image_to_draw
    self.alt = alt_flash_image
end

function DisplayListImage:draw(gr, line, flash, first_line, last_line)
    local img = self.image
    if flash and self.alt then
        img = self.alt
    end
    local end_line = line + img:get_display_height_chars()
    
    if line < last_line and end_line > first_line then
        img:draw_horizontally_centered(gr, line)
    end
    
    return end_line
end


local DisplayListText = class('DisplayListText')
function DisplayListText:initialize(text_to_display, ink, paper, flashing)
    self.text = text_to_display
    self.fg = get_simple_colour_by_name(ink)
    self.bg = get_simple_colour_by_name(paper or default_paper)
    self.is_flashing = flashing
end

function DisplayListText:draw(gr, line, flash, first_line, last_line)
	if flash and self.is_flashing then
		gr:set_fg_colour(self.bg)
		gr:set_bg_colour(self.fg)
	else
		gr:set_fg_colour(self.fg)
		gr:set_bg_colour(self.bg)
	end
    
	local end_line = line + 1
	if line < last_line and end_line > first_line then
        -- adjust column (argument is an offset)
        local len = self.text:len() -- Stop warning: should never be bigger than (2^31)-1
        --double width = GetWindowWidth()/text_character_width_in_pixels;
        local width = current_print_area_width_chars()
        local column = (width - len) / 2
        gr:go_to(line, column)
        gulp_cpp.print_string(gr, self.text)
	end
    
	return end_line
end

-- ========================================================================

local end_marker_on_screen = false
local end_marker_off_top = false

local DisplayListEndMarker = class('DisplayListEndMarker')
function DisplayListEndMarker:initialize()
		end_marker_on_screen = false
		end_marker_off_top = false   
end

function DisplayListEndMarker:draw(gr, line, flash, first_line, last_line)
    
	if line > last_line then
		end_marker_on_screen = false
		end_marker_off_top = false
	elseif line > first_line then
		end_marker_on_screen = true
		end_marker_off_top = false
	else	-- y is less than zero
		end_marker_on_screen = false
		end_marker_off_top = true
	end
	return line
end

function DisplayListEndMarker.is_end_marker_on_screen()
	return end_marker_on_screen
end

function DisplayListEndMarker.is_end_marker_off_top_off_screen()
	return end_marker_off_top
end


function TitleScreen:_init_display_list()
    self.display_list = {

        DisplayListImage(self.screen1, self.screen2),
        DisplayListText("", "BLACK"),
        DisplayListText("Welcome to "..gulp.app_name, "BRIGHT_WHITE"),
        DisplayListText("", "BLACK"),
        DisplayListText(" v" .. tostring(self.version) .. " ", "RED", "WHITE"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        --              "00000000011111111112222222222333"
        --              "12345678901234567890123456789012"
        DisplayListText("Design/Code/Graphics: Rob Probin", "BRIGHT_CYAN"),
        DisplayListText("                      Tony Park", "BRIGHT_CYAN"),
        --              "00000000011111111112222222222333"
        --              "12345678901234567890123456789012"
        DisplayListText("Additional Code: Kenny Osborne", "BRIGHT_CYAN"),
        DisplayListText("Graphics: Claire Probin", "BRIGHT_CYAN"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        
        DisplayListText("Copyright © 2014-2020", "GREEN"),
        DisplayListText("Please do not redistribute", "GREEN"),
        DisplayListText("", "BLACK"),
        DisplayListText("Click to continue", "GREEN", "BLACK", true),
        DisplayListText("Or wait to learn more", "GREEN", "BLACK"),

        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListImage(self.cover),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("More about "..gulp.app_name, "MAGENTA"),
        DisplayListText("http://robprobin.com", "MAGENTA"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),

        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),

        DisplayListText("Testing:", "MAGENTA"),
        DisplayListText("", "BLACK"),
        DisplayListText("Claire, Emily, Johnny, Peter", "MAGENTA"),
        DisplayListText("Seiko, Alex, Akira, Sophia", "MAGENTA"),
        DisplayListText("George, Crystabel, Zander", "MAGENTA"),

        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),

        DisplayListText("Additional thanks to:", "BRIGHT_MAGENTA"),
        DisplayListText("", "BLACK"),
        DisplayListText("www.libsdl.org", "MAGENTA"),
        DisplayListText("www.lua.org", "MAGENTA"),
        DisplayListText("lodev.org/lodepng", "MAGENTA"),
        DisplayListText("github.com/vinniefalco/LuaBridge", "MAGENTA"),
        DisplayListText("luafilesystem Kepler Project", "MAGENTA"),
        DisplayListText("Some Fonts based on DamienG", "MAGENTA"),
        DisplayListText("        <damieng.com/zx-origins>", "MAGENTA"),
        --              "00000000011111111112222222222333"
        --              "12345678901234567890123456789012"
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),

        DisplayListText("Hello to:", "BRIGHT_RED"),
        --              "00000000011111111112222222222333"
        --              "12345678901234567890123456789012"
        DisplayListText("Eve, Stu, Cath, Jess...", "BRIGHT_RED"),
        DisplayListText("", "RED"),
        DisplayListText("... and everyone else :-)", "BRIGHT_RED"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        --              "00000000011111111112222222222333"
        --              "12345678901234567890123456789012"
        DisplayListText("Development Babies", "YELLOW"),
        DisplayListText("!*! Findlay,Eilidh !*!", "YELLOW"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),
        DisplayListText("", "BLACK"),

        -- special end marker
        DisplayListEndMarker()
    }
end

function TitleScreen:delete_display_list()
	-- Delete all entries in display list...    
    self.display_list = nil
end

function TitleScreen:validate_current_line()
    
	-- we need to do this check at the top of update() as well
	local current_height = current_print_area_height_chars()
	if current_height ~= self.last_print_area_height then
		self.current_line = self.current_line / self.last_print_area_height
		self.current_line = self.current_line * current_height
		self.last_print_area_height = current_height
	end
	
	return current_height
end

function TitleScreen:print(gr)
    
    --gr->set_game_screen_offset(true);

	self:validate_current_line()
	
	--SDL_Colour c;
    --get_rgb_from_simple_colour(&c, "BLACK");
	--gr->clear_screen(c);

	-- start with the current scroll variable
	local line = self.current_line

    local first = self.viewport:window_start_line()
    local last = self.viewport:window_end_line()
    
    for _, element in ipairs(self.display_list) do
        line = element:draw(gr, line, self.flash, first, last)
    end
    
	-- @todo: get rid of flicker / tearing on title screen scroll
	-- Mantis Bug Tracker http://ls1.dyndns.org:8080/mantis/view.php?id=48

	--
	-- @todo: animated CS doesn't resize properly ... nor anything below.
	-- @todo: resize flickers when picture is on ... why?

	if DisplayListEndMarker.is_end_marker_off_top_off_screen() then
		self.current_line = line
	end
	if DisplayListEndMarker.is_end_marker_off_top_off_screen() or
	   DisplayListEndMarker.is_end_marker_on_screen() then
                
        for _, element in ipairs(self.display_list) do
            line = element:draw(gr, line, self.flash, first, last)
        end
	end
end

function TitleScreen:update(dt)
	--
	-- work out the flash flag
	--
	self.flash_accum = self.flash_accum + dt
	local flash_rate = ZX_Spectrum_Flash_Rate_in_ms / 1000.0
	if self.flash_accum > flash_rate then
		self.flash = not self.flash

		self.flash_accum = self.flash_accum - flash_rate
		if self.flash_accum > flash_rate then
			-- has overrun, not catching up
			self.flash_accum = 0
		end

		self.flash_count = self.flash_count + 1
	end

	--
	-- The various states
	--
    if self.state == STATE_screen_zoom then
        if self.flash_count > 4 then
            
            if self.scale > 50 then

                self.scale = self.scale - (50*dt/0.65)	-- scale to 50 in 0.65 second
                if self.scale <= 50 then
                    
                    self.scale = 50
                    self.state = STATE_scroll1
                end
                self.screen1:set_scale_to_fit_screen(self.scale)
                self.screen2:set_scale_to_fit_screen(self.scale)
                self.current_line = (current_print_area_height_chars() - self.screen1:get_display_height_chars()) / 2
            end
        end
    elseif self.state == STATE_scroll1 then
        
        -- we need to do this check at the top of print() as well, for paranoia sakes
        local current_height = self:validate_current_line()
        self.current_line = self.current_line - (current_height*dt) / 10
        
    else
        self.state = STATE_screen_zoom
    end
end

--debugMessage(gulp.app:return_copyright())

return TitleScreen
