-- Forlorn Fox
-- Copyright © 2015 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

--
-- End users can optionally implement the follow extra methods:
--
--   on_close()
--   text_complete(text)
--   final_validation(text)       - should return true or false
--   character_validation(char, text, cursor_pos)     - should return true or false
--
require("strict")
local class = require("middleclass")

local ASCIITextEntry = class("ASCIITextEntry")

function ASCIITextEntry:initialize(line, column, initial_ascii_text)
    self._text = initial_ascii_text or ""
    self._original = initial_ascii_text or ""
    self._line = line or 1
    self._column = column or 1
    self._finished = false
    self._cursor_flash_time = 0
    self._cursor_flash_mode = false
    self._cursor_pos = #self._text
    
    gulp._delegate_list:add_delegate(self)
    gulp.add_user_movement_target(self)
end

function ASCIITextEntry:_text_before_cursor()
    return self._text:sub(1, self._cursor_pos)
end

function ASCIITextEntry:_text_after_cursor()
    return self._text:sub(self._cursor_pos+1, #self._text)
end

function ASCIITextEntry:draw()
    local fg = self._foreground or gulp.colour.WHITE.simple_colour
    local bg = self._background or gulp.colour.BLACK
    
    ANSI_reset_attributes()
    ANSI_set_fg_colour(fg)
    ANSI_set_bg_colour(bg)
    
    goto_position(self._line, self._column)
    ANSI_print(self:_text_before_cursor())
    
    -- display cursor
    if self._cursor_flash_mode then
        ANSI_print("_")
    else
        ANSI_print(" ")
    end
    
    ANSI_print(self:_text_after_cursor())
end

function ASCIITextEntry:close()
    if not self._finished then
        gulp.remove_user_movement_target(self)
        gulp._delegate_list:drop_delegate(self)
        self._finished = true
        if type(self.on_close) == "function" then
            self:on_close()
        end
    end
end

function ASCIITextEntry:update(dt)
    self._cursor_flash_time = self._cursor_flash_time + dt
    
    if self._cursor_flash_time > 0.4 then
        self._cursor_flash_mode = not self._cursor_flash_mode
        self._cursor_flash_time = self._cursor_flash_time - 0.4
        
        -- if overrun a lot, reset
        if self._cursor_flash_time > 1.5 then
            self._cursor_flash_time = 0
        end
    end
end

function ASCIITextEntry:get_result()
    return self._text
end

function ASCIITextEntry:bad_beep()
    gulp.sound.play_tink()      -- temporary sound for texting
end

function ASCIITextEntry:keypressed(sym, scancode, modifier, repeated)
    if back_key(sym) then
        self._text = self._original     -- cancel any edits
        self:close()
    elseif sym == SDLK.RETURN or sym == SDLK.KP_ENTER then
        -- activate
        local good_text = true
        if self.final_validation then
            good_text = self:final_validation(self._text)
        end
        if not good_text then
            self:bad_beep()
        else
            if self.text_complete then
                self:text_complete(self._text)
            end
            self:close()
        end
    elseif sym >= 32 and sym < 127 then
        
        local good_text = true
        if self.character_validation then
            good_text = self:character_validation(sym, self._text, self._cursor_pos)
        end
        if good_text then
            self._text = self:_text_before_cursor() .. string.char(sym) ..
                self:_text_after_cursor()
            self._cursor_pos = self._cursor_pos + 1
        else
            self:bad_beep()
        end
    
    elseif sym == SDLK.BACKSPACE then
        -- backwards
        if self._cursor_pos > 0 and #self._text > 0 then
            self._text = self:_text_before_cursor():sub(1, -2) .. 
                self:_text_after_cursor()
            self._cursor_pos = self._cursor_pos - 1
        else
            self:bad_beep()
        end
        
    elseif sym == SDLK.DELETE then
        -- forward delete
        if self._cursor_pos < #self._text then
            self._text = self:_text_before_cursor() .. 
                self:_text_after_cursor():sub(2)
        else
            self:bad_beep()
        end
    elseif sym == SDLK.LEFT then
        if self._cursor_pos > 0 then
            self._cursor_pos = self._cursor_pos - 1
        else
            self:bad_beep()
        end
    elseif sym == SDLK.RIGHT then
        if self._cursor_pos < #self._text then
            self._cursor_pos = self._cursor_pos + 1
        else
            self:bad_beep()
        end
        
    elseif sym == SDLK.END or sym == SDLK.DOWN then
        self._cursor_pos = #self._text
    
    elseif sym == SDLK.HOME or sym == SDLK.UP then
        self._cursor_pos = 0
    
    else
        -- print("ascii_text_entry", sym)
        -- we ignore it anyway...
    end
    return true     -- handled
end

return ASCIITextEntry

-- Should really use:
-- SDL_SetTextInputRect
-- SDL_TextInputEvent
-- SDL_StartTextInput
-- SDL_StopTextInput
-- SDL_TEXTINPUT and SDL_TEXTEDITING events
-- SDL_IsTextInputActive
-- 
-- https://wiki.libsdl.org/Tutorials/TextInput
--
-- This would allow entry of composite characters (accented, etc.)
-- However:
--  (a) We don't have accented or Japanese/Chinese characters yet.
--  (b) We only want this for entry of IP addresses and port numbers at the moment (and maybe simple hosts)
--  (c) ...
--
-- For speed, we just implementing a quite method.
--
-- Other notes (from https://wiki.libsdl.org/CategoryKeyboard)
-- SDL_Keycode values are mapped to the current layout of the keyboard and 
-- correlate to an SDL_Scancode
-- SDL_Scancode values are used to represent the physical location of a 
-- keyboard key on the keyboard.
--[[
The main difference between these two values comes into play in situations 
where the operating system is mapping the physical keyboard keys to different 
virtual letters. For instance, when a standard QWERTY keyboard is mapped to a 
German QWERTZ layout, the "Y" key will generate events with a scancode of 
SDL_SCANCODE.Y, but a keycode of SDLK.z. Which one to use is left to the 
application: scancodes are suited in situations where controls are layout-
dependent (eg. the "WASD" keys as left-handed arrow keys), whereas keycodes 
are better suited to situations where controls are character-dependent (eg. 
the "I" key for Inventory). 
--]]
