-- Forlorn Fox
-- Copyright © 2015 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

--
-- End users can optionally implement the follow extra methods:
--
--   on_close()
--   text_complete(utf8_text)
--   final_validation(utf8_text)       - should return true or false
--   character_validation(char, utf8_text, cursor_pos)     - should return true or false
--
-- Remember: Number of characters  is not  number of bytes

-- @todo: switching off the window and back while in composite mode entering
-- multkey characters will freeze the input until a single-key entry is pressed.
-- Is this an SDL2 bug? Is there something we should be doing?

require("strict")
local class = require("middleclass")
local utf8 = require("utilities/gulp_utf8")
local text_sizes = require("utilities/text_sizes")

local UnicodeTextEntry = class("UnicodeTextEntry")

local number_of_text_entries = 0

function UnicodeTextEntry:initialize(line, column, initial_utf8_text, colour)
    --print("SDL_IsTextInputActive", gulp_cpp.SDL_IsTextInputActive())
    self._original = initial_utf8_text or ""
    self._text  = self._original
    self._fg_colour = colour or gulp.colour.WHITE.simple_colour
    
    --[[self._text = utf8.to32(self._original)
    -- convert into a table of utf8 strings
    for k,v in ipairs(self._text) do
        self._text[k] = to_utf8(v)
    end--]]
    self._line = line or 1
    self._column = column or 1
    self._finished = false
    self._cursor_flash_time = 0
    self._cursor_flash_mode = false
    -- NOTES about cursor position
    -- ===========================
    -- Because UTF-8 text can be longer than one byte (= Lua string index)
    -- there are slightly more complex rules to remember about what it means.
    --
    -- KEY RULE
    -- self._cursor_pos is at the index _first byte_ of the UTF 8 character 
    -- codepoint of the text after the cursor.
    --
    -- SECOND RULE:
    -- self._cursor_pos is a string index (i.e. byte index) not a character
    -- index. (This is also of self.composition_cursor when it's valid).
    --
    -- REASON FOR CHOICE
    -- This is was mainly decided because the utf8.get_previous_index() and 
    -- utf8.get_next_index() functions return the index of first character.
    --
    -- EXAMPLES:
    -- self._cursor_pos must therefore be 1 if the cursor if before all text.
    -- self._cursor_pos = #self._text+1 must be the end of the string (with no 
    --   characters after the cursor).
    --
    -- self._cursor_pos-1 must be the index of last byte of the UTF-8 
    --   character before the cursor.
    -- if self._cursor_pos at #self.text must be the start of a one byte UTF-8 
    --   codepoint at the end of the string.
    --
    -- ERRORS:
    -- self._cursor_pos less than 1 are invalid. 
    -- self._cursor_pos greater than #text+1 are invalid

    self:cursor_to_end()
    
    gulp._delegate_list:add_delegate(self)
    gulp.add_user_movement_target(self)
    
    gulp_cpp.SDL_StartTextInput()
    number_of_text_entries = number_of_text_entries + 1
    
    local r = text_sizes:ScreenCells_to_SDL_Rect(self._line, self._column, text_sizes:columns()-self._column, 2) -- whole screen width, 2 lines?
    gulp_cpp.SDL_SetTextInputRect(r)   -- @todo fix this so that it's at a sensible place'

end

function UnicodeTextEntry:textinput(text)
    --print("textinput", text)
    self.composition_text = nil
    if utf8.valid(text) then
        local sym_table = utf8.to32(text)
        for _,sym in ipairs(sym_table) do
            local good_text = true
            if self.character_validation then
                good_text = self:character_validation(sym, self._text, self._cursor_pos)
            end
            if good_text then
                self._text = self:_text_before_cursor() .. to_utf8(sym) ..
                    self:_text_after_cursor()
                -- need to change the cursor position - currently it's before the character inserted
                self:move_cursor_forward()
            else
                self:bad_beep()
            end
        end
    else
        self:bad_beep()
        print("invalid text input", text)
        print("bytes:")
        for i in c:gmatch"." do
            print("", i:byte())
        end
    end
    
    return true     -- handled
end

function UnicodeTextEntry:textediting(text, start, length)
    -- Update the composition text.
    -- Update the cursor position.
    -- Update the selection length (if any).
    
    --print("textediting", text, start, length)
    if text == "" then
        self.composition_text = nil
        
    elseif utf8.valid(text) then
        self.composition_text = text
        -- NOTE about +1: Lua indexes are 1 based, not zero based
        self.composition_cursor = utf8.return_byte_index(text, start+1)
        self.composition_selection_len = length
    else
        self.composition_text = '?'
        self.composition_cursor = 2 -- this after the string '?'
        self.composition_selection_len = 0
    end
    
    return true     -- handled
end


function UnicodeTextEntry:cursor_to_end()
    self._cursor_pos = #self._text+1
end

function UnicodeTextEntry:cursor_to_start()
    self._cursor_pos = 1
end

-- returns if successfully moved
function UnicodeTextEntry:move_cursor_forward()
    -- already at end?
    if self._cursor_pos > #self._text then
        -- already at end
        return false
    end
    
    -- move one forward - or to end if not any more
    local new_index = utf8.get_next_index(self._text, self._cursor_pos)
    if new_index then
        self._cursor_pos = new_index
    else
        self:cursor_to_end()
    end
    return true
end

-- returns if successfully moved
function UnicodeTextEntry:move_cursor_backward()
    -- if at start already...
    if self._cursor_pos <= 1 then
        return false
    end
    
    local new_index = utf8.get_previous_index(self._text, self._cursor_pos)
    if new_index then
        self._cursor_pos = new_index
    else
        self:cursor_to_start()
    end
    return true
end


function UnicodeTextEntry:_text_before_cursor()
    return self._text:sub(1, self._cursor_pos-1)
end

function UnicodeTextEntry:_text_after_cursor()
    return self._text:sub(self._cursor_pos, #self._text)
end


function UnicodeTextEntry:draw(gr)
    text_sizes:set_viewport(gr)
    local wrap = gr:get_wrap()
    gr:wrap(true)

    -- Ideally need to detect how many lines it is...
    gr:FillRect(get_SDL_Color_by_name('BLACK'), self._column-0.1, self._line-0.1, text_sizes:columns()+0.1, self._line+2.1)
    
    gr:DrawRect(get_SDL_Color_by_name('WHITE'), self._column-0.1, self._line-0.1, text_sizes:columns()+0.1, self._line+2.1)
    
    local fg = self._foreground or self._fg_colour
    local bg = self._background or gulp.colour.BLACK
    
    ANSI_reset_attributes()
    ANSI_set_fg_colour(fg)
    ANSI_set_bg_colour(bg)
    
    goto_position(self._line, self._column)
    ANSI_print(self:_text_before_cursor())
    
    if self.composition_text then
        ANSI_print(self.composition_text:sub(1, self.composition_cursor-1))
    end
    
    -- display cursor
    if self._cursor_flash_mode then
        ANSI_print("_")
    else
        ANSI_print(" ")
    end
    
    if self.composition_text then
        ANSI_print(self.composition_text:sub(self.composition_cursor))
    end
    
    ANSI_print(self:_text_after_cursor())
    gr:wrap(wrap)
end

function UnicodeTextEntry:close()
    if not self._finished then
        number_of_text_entries = number_of_text_entries - 1
        if number_of_text_entries == 0 then
            gulp_cpp.SDL_StopTextInput()
        end
        gulp.remove_user_movement_target(self)
        gulp._delegate_list:drop_delegate(self)
        self._finished = true
        if type(self.on_close) == "function" then
            self:on_close()
        end
    end
end

function UnicodeTextEntry:update(dt)
    self._cursor_flash_time = self._cursor_flash_time + dt
    
    if self._cursor_flash_time > 0.4 then
        self._cursor_flash_mode = not self._cursor_flash_mode
        self._cursor_flash_time = self._cursor_flash_time - 0.4
        
        -- if overrun a lot, reset
        if self._cursor_flash_time > 1.5 then
            self._cursor_flash_time = 0
        end
    end
end

function UnicodeTextEntry:get_result()
    return self._text
end

function UnicodeTextEntry:bad_beep()
    gulp.sound.play_tink()      -- temporary sound for texting
end

function UnicodeTextEntry:keypressed(sym, scancode, modifier, repeated)
    if self.composition_text then
        return true
    end
    if back_key(sym) then
        self._text = self._original     -- cancel any edits
        self:close()
    elseif sym == SDLK.RETURN or sym == SDLK.KP_ENTER then
        -- activate
        local good_text = true
        if self.final_validation then
            good_text = self:final_validation(self._text)
        end
        if not good_text then
            self:bad_beep()
        else
            if self.text_complete then
                self:text_complete(self._text)
            end
            self:close()
        end
    elseif sym >= 32 and sym < 127 then
        -- do nothing here
    
    elseif sym == SDLK.BACKSPACE then
        -- backwards delete
        local old_index = self._cursor_pos
        if self:move_cursor_backward() then
            -- since we've moved the cursor back, the character to delete is 
            -- now ahead of the cursor position and the old index is where
            -- the good text starts
            self._text = self:_text_before_cursor() .. self._text:sub(old_index)
        else
            self:bad_beep()
        end
        
    elseif sym == SDLK.DELETE then
        -- forward delete        
        local old_index = self._cursor_pos
        if self:move_cursor_forward() then
            -- since we've moved the cursor forward, the character to delete is 
            -- now behind of the cursor position and the old index is one 
            -- ahead of where the good text ends. However, if we delete now
            -- then the cursor will be illegal (because character sizes are 
            -- not the same). Therefore we need to move backwards
            -- first. We can do this by resetting the cursor position back. 
            -- First we take a copy of the new index.
            local new_index = self._cursor_pos
            self._cursor_pos = old_index
            
            self._text = self:_text_before_cursor() .. self._text:sub(new_index)
        else
            self:bad_beep()
        end
    
    elseif sym == SDLK.LEFT then
        if not self:move_cursor_backward() then
            self:bad_beep()
        end
    
    elseif sym == SDLK.RIGHT then
        if not self:move_cursor_forward() then
            self:bad_beep()
        end
        
    elseif sym == SDLK.END or sym == SDLK.DOWN then
        self:cursor_to_end()
    
    elseif sym == SDLK.HOME or sym == SDLK.UP then
        self:cursor_to_start()
    
    else
        -- print("unicode_text_entry", sym)
        -- we ignore it anyway...
    end
    return true     -- handled
end

return UnicodeTextEntry
