-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

local Building = require("objects/building")
local Tree = require("objects/tree")
local GoldCoin = require('objects/gold_coin')
require('utils')
require('graphics/colours')
require('graphics/glyph_sets')


local utf8 = require("utilities/gulp_utf8")

function update_floor_fonts(map, maze)
    if map.fg_areas ~= nil then
        for _,entry in pairs(map.fg_areas) do
            if entry.font ~= nil then
                for l = entry.start_line, entry.end_line do
                    for c = entry.start_column, entry.end_column do
                        local glyph = maze:get_glyph(l, c)
                        if glyph >= 0x20 and glyph < 0x80 then
                            local new_glyph = get_glyph(entry.font, glyph)
                            maze:update_glyph(l, c, new_glyph)
                        end
                    end
                end
            end
        end
    end
end


function load_colours(map, maze)

    local fg = map.fg
    if fg then
        fg = get_simple_colour_by_name(fg)
    else
        fg = gulp.colour.RED.simple_colour
    end

    local bg = map.bg
    if bg then
        bg = get_SDL_Color_by_name(bg)
    else
        bg = get_SDL_Color_by_name("GREEN")
    end

    maze:set_default_maze_colours(fg, bg)
    
    if map.fg_areas ~= nil then
        for _,entry in pairs(map.fg_areas) do
            for l = entry.start_line, entry.end_line do
                for c = entry.start_column, entry.end_column do
                    maze:set_maze_colours(l, c, get_simple_colour_by_name(entry.colour), bg)
                end
            end
        end
    end
end


--
-- Read through the map looking for characters representing start positions for exits
-- (and maybe objects at a later date)
--
local function load_objects_from_maze(map, maze, character_table, action_function)

    if character_table == nil then
        return
    end

    local finder = gulp_cpp.MapFinder(maze)
    -- candidates is a table, key = unicode integer, value = utf8 string
    local candidates = {}
    for k, _ in pairs(character_table) do
        local target = k
        if type(k) == "string" then
            target = from_utf8(k)
        end
        candidates[target] = k
        finder:add_target(target)
    end

    while true do
        local found = finder:find_anywhere()
        if not found then
            break;
        end

        local line = finder.line
        local column = finder.column
        local unicode_integer = maze:get_glyph(line, column)
        local c = candidates[unicode_integer]   -- c must exist... we've just found it

        local element = character_table[c]
        
        if element ~= nil then
            action_function(element, line, column, c)

            local replacement_char = ' '
            local fg_colour
            if type(element) == 'table' and element['map_character'] ~= nil then
                replacement_char = element.map_character
                fg_colour = element['fg'] or map.fg or 'WHITE'
            end
            if type(replacement_char) == "string" then
                replacement_char = from_utf8(replacement_char)
                fg_colour = map.fg or 'WHITE'
            end

            -- wipe it from the map, and the view.
            map[line+1][column+1] = replacement_char
            maze:set_glyph(line, column, replacement_char, 0, 1, 1);
            local bg_colour = map.bg or 'BLACK'
            if gulp.LuaStateCodeMode == 'CLIENT' then
                -- get_SDL_Color_by_name crashes server...!
                maze:set_maze_colours(line, column, get_simple_colour_by_name(fg_colour), get_SDL_Color_by_name(bg_colour))
            end
        end


    end

end


function load_fixed_objects(map, maze)
    -- map is our data structure
    -- maze is PresentationMaze in C++

    local objects = map.objects

    if objects then
        -- do fully specified objects
        for _, object in ipairs(objects) do
            if object['object_type'] == 'building' then
                -- this object is a building
                table.insert(map.auto.objects, Building(map, object['object_data']))
            elseif object['object_type'] == 'tree' then
                -- this object is a building
                table.insert(map.auto.objects, Tree(map, object['object_data']))
            else
                print("Unknown object type ") --.. object[object_type]
            end
        end
    end
end

local function load_start_light(map, line, column, light)
    if map.auto.start_lights == nil then 
        map.auto.start_lights = {}
    end
    
    local group = light.group
    if map.auto.start_lights[group] == nil then 
        map.auto.start_lights[group] = {}
    end
    
    local start_light = {}
    for k,v in pairs(light) do start_light[k]=v end
    start_light.line = line
    start_light.column = column
    -- adjust the glyph location so it lines up on map correctly
    start_light.glyph = new_maze_display_element(map.auto.maze, line-0.5, column-0.5, light.glyphs[light.status], 1)

    
    table.insert(map.auto.start_lights[group], start_light)
end

local function load_exit_pad(map, line, column, pad)
    if map.auto.exit_pads == nil then 
        map.auto.exit_pads = {}
    end
    
    local group = pad.group
    if map.auto.exit_pads[group] == nil then 
        map.auto.exit_pads[group] = {}
    end
    
    local exit_pad = {}
    for k,v in pairs(pad) do exit_pad[k]=v end
    exit_pad.line = line
    exit_pad.column = column
    
    table.insert(map.auto.exit_pads[group], exit_pad)
end

function load_on_map_objects(map, maze)
    -- look for map specified objects
    local objects = map.objects
    local server = (gulp.LuaStateCodeMode == 'SERVER')
    
    if objects and objects.on_map then
        
        local f = function(element, line, column, character)
            -- set up the object
            line = line + 0.5
            column = column + 0.5
            if type(element) == 'table' then
                if element.name == 'exit_pad' then
                    load_exit_pad(map, line, column, element)
                elseif element.name == 'start light' then
                    load_start_light(map, line, column, element)
                end
            elseif element == 'gold coin' then
                if server then map.auto.pickables:add_to_floor(line, column, GoldCoin(element), 'map_generation') end
            else
                print("Unknown object type:", element)
            end
        end
        
        load_objects_from_maze(map, maze, objects.on_map, f)
    end
end
--[[
function remove_pickable_objects(map, maze)
    -- look for map specified objects
    local objects = map.objects
    
    if objects.on_map then
        
        local f = function(element, line, column)
        end
        
        load_objects_from_maze(map, maze, objects.on_map, f)
    end
end
--]]


--
-- Read through the map looking for characters representing start positions for exits
-- (and maybe objects at a later date)
--
function load_exits_from_maze(map, maze)

    local f = function(element, line, column, character)
        if element.line == nil then
            -- set up the exit
            element.line = line
            element.column = column
            element.group = character
        else
            -- there is already an exit for this glyph - make a new copy at this new location
            local new_element = deep_copy(element)
            new_element.line = line
            new_element.column = column
            table.insert(map.exits, new_element)
            
        end
        
    end
        
    load_objects_from_maze(map, maze, map.exits, f)
    
end

function load_doors_from_maze(map, maze)

    local f = function(element, line, column, character)
        if element.line == nil then
            -- set up the exit
            element.line = line
            element.column = column
            local state = map.conditionals[element.conditional]
            element.map_character = element[state]
        else
            -- there is already an exit for this glyph - make a new copy at this new location
            local new_element = deep_copy(element)
            new_element.line = line
            new_element.column = column
            local state = map.conditionals[new_element.conditional]
            new_element.map_character = new_element[state]
            table.insert(map.doors, new_element)
            
        end
        
        map.auto.wall_detect_type[element.closed] = true
        map.auto.wall_detect_type[element.open] = false
        map.auto.autowall_like[element.closed] = true
        map.auto.autowall_like[element.open] = true
        
    end
        
    load_objects_from_maze(map, maze, map.doors, f)
    
end

--
-- Read through the map looking for characters representing start positions for mobs
-- (and maybe objects at a later date)
--
function load_mobs_from_maze(map, maze)

    local f = function(element, line, column)
        --local success, mob = pcall(dofile, gulp_cpp.LoadPath("scripts/" .. element['file'] .. ".lua"):str())
        local success, file = pcall(loadfile, gulp_cpp.LoadPath("scripts/" .. element['file'] .. ".lua"):str())

        if success == false then
            -- if it failed, the return is an error string
            print(file)
        else
            -- otherwise, the mob file is returned, call that to get the mob object
            local mob = file(map.auto.filename)
            if mob ~= nil then
                table.insert(map.auto.mobs, mob)
                mob:set_location(line, column, "create")
                mob:set_visible(true)

                if element.speed ~= nil then mob:set_speed(element.speed) end
            end
        end

    end
        
    load_objects_from_maze(map, maze, map.mobs, f)
    
end


function remove_mobs_from_maze(map, maze)

    local f = function(element, line, column)
    end
        
    load_objects_from_maze(map, maze, map.mobs, f)
    
end



local function set_input_wall_features(map, map_char, unicode_value_or_table, wall_type, autowall_glyphset_index, influencing_glyph)

    local value = from_utf8(map_char)
    if value == 0 then
        -- never map zeros!
        return;
    end

    -- allow utf8 string as well...
    if type(unicode_value_or_table) == "string" then
        unicode_value_or_table = from_utf8(unicode_value_or_table)
    end
    
    -- look for animated glyph list and record all glyph which starts at the 2nd element
    local glyph_list = {}
    if type(unicode_value_or_table) == "table" then
        for i = 2, #unicode_value_or_table do
            table.insert(glyph_list, unicode_value_or_table[i])
        end
        
    else
        -- single glyph case (e.g. not an animated list)
        glyph_list[1] = unicode_value_or_table
    end
    
    -- map char should be utf8 string, but we allow integers
    if type(map_char) == "string" then
        map_char = from_utf8(map_char)
    end


    -- We always set the glyph mapping up
    map.auto.glyph_mappings[map_char] = unicode_value_or_table

    -- record all glyph
    for _, unicode_glyph_point in pairs(glyph_list) do
        
        -- we set the default values based on a space
        map.auto.autowall_like[unicode_glyph_point] = false     
        map.auto.autowall[unicode_glyph_point] = 0              -- zero means no autowall, a value indicates the glyph set index
        map.auto.wall_detect_type[unicode_glyph_point] = false

        if wall_type == "Wall" then
            
            map.auto.wall_detect_type[unicode_glyph_point] = true
            map.auto.autowall_like[unicode_glyph_point] = true

        elseif wall_type == "AutoWall"  then

            map.auto.autowall_like[unicode_glyph_point] = true
            map.auto.autowall[unicode_glyph_point] = autowall_glyphset_index
            map.auto.wall_detect_type[unicode_glyph_point] = true

        elseif wall_type == "Gate" then

            map.auto.wall_detect_type[unicode_glyph_point] = true

        elseif wall_type == "Space_InfluenceAutoWall" then

            map.auto.autowall_like[influencing_glyph] = true

        else    -- spaces, and space-like characters

            -- Specifics:
            -- =========
            -- unknown types, treat as spaces
            -- type == "Space" - do nothing
            -- type == "Mob"
            --      glyph blanking should be in features in map file, not here
            --      glyph_mappings[value] = ' ';
            -- type == "Object"
            --      glyph blanking should be in features in map file, not here
            --      glyph_mappings[value] = ' ';
        end

    end
end

--[[
function is_map_feature(map, char)
    if map.features then
        for _,v in ipairs(map.features) do
            local map_char = v[1]
            if char == map_char then
                return true
            end
        end
    else
        -- backwards compatibility
        if char == ' ' or char == '#' then
            return true
        end
    end

    return false
end
--]]

local wall_glyph_lookup_table_untransformed =
{         -- N,E,S,W
    0x1c, -- 0,0,0,0
    0x5d, -- 0,0,0,1
    0x3c, -- 0,0,1,0
    0x3d, -- 0,0,1,1
    0x5b, -- 0,1,0,0
    0x5e, -- 0,1,0,1
    0x3b, -- 0,1,1,0
    0x7f, -- 0,1,1,1
    0x7c, -- 1,0,0,0
    0x7d, -- 1,0,0,1
    0x7e, -- 1,0,1,0
    0x5f, -- 1,0,1,1
    0x7b, -- 1,1,0,0
    0x3f, -- 1,1,0,1
    0x1f, -- 1,1,1,0
    0x5c, -- 1,1,1,1
}
local wall_glyph_lookup_table_untransformed_Z_dungeon =
{         -- N,E,S,W
    0x14, -- 0,0,0,0
    0x12, -- 0,0,0,1
    0x10, -- 0,0,1,0
    0x0e, -- 0,0,1,1
    0x13, -- 0,1,0,0
    0x11, -- 0,1,0,1
    0x0f, -- 0,1,1,0
    0x0d, -- 0,1,1,1
    0x0c, -- 1,0,0,0
    0x0a, -- 1,0,0,1
    0x08, -- 1,0,1,0
    0x06, -- 1,0,1,1
    0x0b, -- 1,1,0,0
    0x09, -- 1,1,0,1
    0x07, -- 1,1,1,0
    0x05, -- 1,1,1,1
}



local function set_up_wall_detect_features(map)

    local autowall_type = wall_glyph_lookup_table_untransformed
    
    if map.features then -- and type(map.features) == "table" then
        
        map.auto.autowall_glyphset_count = 0
        map.auto.autowall_glyph_sets = {}
        map.auto.autowall_wall_chars = {}
        
        
        for _,v in ipairs(map.features) do
            local map_char = v[1]
            local unicode_value_or_table = v[2]
            local walltype = v[3]
            local select_autowall_set = v[4]
            local wall_modifier = v[5]
            local influencing_glyph = nil
            
            if walltype == 'AutoWall' then
                map.auto.autowall_glyphset_count = map.auto.autowall_glyphset_count + 1
                map.auto.autowall_glyph_sets[map.auto.autowall_glyphset_count] = select_autowall_set or map.autowall_glyph_set or "map" -- backwards compatible
                map.auto.autowall_wall_chars[map.auto.autowall_glyphset_count] = unicode_value_or_table     -- used for corner fills
                if wall_modifier == "Z" then
                    autowall_type = wall_glyph_lookup_table_untransformed_Z_dungeon
                end
            end
            
            if walltype == 'Space_InfluenceAutoWall' then
                influencing_glyph = v[4]
            end
            
            set_input_wall_features(map, map_char, unicode_value_or_table, walltype, map.auto.autowall_glyphset_count, influencing_glyph)
        end
    else
        -- these are the only ones that get mappings, by default

        map.auto.autowall_glyphset_count = 1
        map.auto.autowall_glyph_sets = {map.autowall_glyph_set or 'blackwall'}
        map.auto.autowall_wall_chars = {'#'}

        set_input_wall_features(map, "#", from_utf8('#'), "AutoWall", 1)
        set_input_wall_features(map, ' ', 32, "Space")
    end
    
    return autowall_type
end


local character_base_code = 0


local function set_autowall_glyph_set_base(map, autowall_type)

    for index = 1, map.auto.autowall_glyphset_count do
    
        local autowall_glyph_set = map.auto.autowall_glyph_sets[index]

        map.auto.wall_glyph_lookup_table_transformed[index] = {}

        -- We should insert all the autowall glyphs in the wall table...
        -- (Yes, this is a bit of a horrible order dependancy with
        --  clear_default_wall_features - just live with it)
        for i, offset_code in ipairs(autowall_type) do
            
            local unicode_value = get_glyph(autowall_glyph_set, offset_code)
            map.auto.autowall_like[unicode_value] = true
            map.auto.autowall[unicode_value] = index
            map.auto.wall_glyph_lookup_table_transformed[index][i] = unicode_value
            map.auto.wall_detect_type[unicode_value] = true
        
        end
    end
end

--[[
local function calculate_wall_glyph(map, line, column, lmax, cmax)

    local lookup_pointer = 1
    local maze = map.auto.maze
    
    if line == 0 or map.auto.autowall_like[maze:get_item(line-1, column)] then
        -- set north bit
        lookup_pointer = lookup_pointer + 0x08
    end

    if line == lmax or map.auto.autowall_like[maze:get_item(line+1, column)] then
        -- set south bit
        lookup_pointer = lookup_pointer + 0x02
    end

    if column == 0 or map.auto.autowall_like[maze:get_item(line, column-1)] then
        -- set west bit
        lookup_pointer = lookup_pointer + 0x01
    end

    if column == cmax or map.auto.autowall_like[maze:get_item(line, column+1)] then
        -- set east bit
        lookup_pointer = lookup_pointer + 0x04
    end

    return map.auto.wall_glyph_lookup_table_transformed[lookup_pointer]
end


local function get_autowall_glyph(map, line, column, lmax, cmax)

    local m = map.auto.maze:get_item(line, column)

    if map.auto.autowall[m] ~= 0 then
        m = calculate_wall_glyph(map, line, column, lmax, cmax)
    end

    return m
end


local function process_autowall_glyphs(map)
    local maze = map.auto.maze
    local lmax = maze:height()-1
    local cmax = maze:width()-1
    for line = 0, lmax do
        for column = 0, cmax do
            maze:set_item(line, column, get_autowall_glyph(map, line, column, lmax, cmax))
            -- map[line+1][column+1] = value
        end
    end
end
--]]

local corner_table_store = {}

local function get_corner_table(map, wall, space, glyph_set)
    --local glyph_set = map.autowall_glyph_set or "map"
    local corner_table = corner_table_store[glyph_set]
    if corner_table == nil then
        corner_table = {}
        for n = 0, 255 do
            -- raster scan order
            local NW = 0 ~= bit32.band(0x01, n)
            local N  = 0 ~= bit32.band(0x02, n)
            local NE = 0 ~= bit32.band(0x04, n)
            
            local W  = 0 ~= bit32.band(0x08, n)
            local E  = 0 ~= bit32.band(0x10, n)
            
            local SW = 0 ~= bit32.band(0x20, n)
            local S  = 0 ~= bit32.band(0x40, n)
            local SE = 0 ~= bit32.band(0x80, n)
            
            corner_table[n] = space     -- we are checking a space, so this is true
            
            -- override for specific cases
            if N and S and E and W then 
                corner_table[n] = wall 
            elseif N and S and W and not NW and not SW then 
                corner_table[n] =  get_glyph(glyph_set, 0x27)    
            elseif N and S and E and not NE and not SE then 
                corner_table[n] =  get_glyph(glyph_set, 0x25) 
            elseif N and E and W and not NE and not NW then 
                corner_table[n] =  get_glyph(glyph_set, 0x26) 
            elseif S and E and W and not SE and not SW then 
                corner_table[n] =  get_glyph(glyph_set, 0x28)
            elseif N and W and not NW then
                if S then
                    corner_table[n] =  get_glyph(glyph_set, 0x27)
                elseif E then
                    corner_table[n] =  get_glyph(glyph_set, 0x26)
                else
                    corner_table[n] =  get_glyph(glyph_set, 0x17)
                end
            elseif N and E and not NE then 
                if S then
                    corner_table[n] =  get_glyph(glyph_set, 0x25)
                elseif W then
                    corner_table[n] =  get_glyph(glyph_set, 0x26)
                else
                    corner_table[n] =  get_glyph(glyph_set, 0x16) 
                end
            elseif S and W and not SW then 
                if N then
                    corner_table[n] =  get_glyph(glyph_set, 0x27)
                elseif E then
                    corner_table[n] =  get_glyph(glyph_set, 0x28)
                else
                    corner_table[n] =  get_glyph(glyph_set, 0x18)
                end
            elseif S and E and not SE then 
                if N then
                    corner_table[n] =  get_glyph(glyph_set, 0x25)
                elseif W then
                    corner_table[n] =  get_glyph(glyph_set, 0x28)
                else
                    corner_table[n] =  get_glyph(glyph_set, 0x15)
                end
            end 
        end
        corner_table_store[glyph_set] = corner_table
    end
    return corner_table
end

local function fill_corner_gaps(map)
    
    for index = 1, map.auto.autowall_glyphset_count do
        
        local wall = map.auto.autowall_wall_chars[index]           
        local space = 0x20                      -- space character
        local corner_table = get_corner_table(map, wall, space, map.auto.
            autowall_glyph_sets[index])
        
        -- one less because we don't process the last line or column
        local lmax = map.height - 1
        local cmax = map.width - 1
        
        -- start at 2 because we don't process the first line (or column)
        for line = 2, lmax do
            
            local line_pre = map[line-1]
            local line_this = map[line]
            local line_next = map[line+1]

            -- set up variables to shift along at the start of the loop
            local column_this = line_this[1]
            local column_next = line_this[2]
            for column = 2, cmax do
                -- sometimes when there is only one corner gap to fill in, we still want the 3 sided 
                -- fill in glyph, because the 2 sided fill in glyph affects wall glyph selection
                -- for adjacent walls even if those walls are not next to a gap that needs filling
                
                -- we shift along the three local column on the current line since 
                -- we have to evaluate them anyway,
                local column_pre = column_this  -- from last time
                column_this = column_next -- from last time
                column_next = line_this[column+1]
                
                -- Don't see a way to get away from this 'if' because
                -- We couldn't do this because the 'non-spaces' can be several different
                -- things therefore hard to encode into a smallish table
                -- 
                -- Since we are aiming for a game with run mazes (mostly wall fills, 
                -- unlike trinity that's mostly spaces) also optimise by 
                -- skip the map fetch logic for upper and lower
                -- walls for mostly filled mazes.

                if column_this == space then
                    --[[
                    local N = line == 1         or line_pre[column] == wall
                    local S = line == lmax      or line_next[column] == wall
                    local W = column == 1       or line_this[column-1] == wall
                    local E = column == cmax    or line_this[column+1] == wall

                    local NW = line == 1        or column == 1      or line_pre[column-1] == wall
                    local NE = line == 1        or column == cmax   or line_pre[column+1] == wall
                    local SW = line == lmax     or column == 1      or line_next[column-1] == wall
                    local SE = line == lmax     or column == cmax   or line_next[column+1] == wall
                    --]]
                    
                    -- raster scan order
                    -- faster way of doing this?
                    local NW = line_pre [column-1] == wall and 0x01 or 0
                    local N  = line_pre [column  ] == wall and 0x02 or 0
                    local NE = line_pre [column+1] == wall and 0x04 or 0
                    
                    local W  = column_pre == wall and 0x08 or 0
                    local E  = column_next == wall and 0x10 or 0
                    
                    local SW = line_next[column-1] == wall and 0x20 or 0
                    local S  = line_next[column  ] == wall and 0x40 or 0
                    local SE = line_next[column+1] == wall and 0x80 or 0
                    
                    local composite_of_cells = N+S+E+W+NW+NE+SW+SE
                    line_this[column] = corner_table[composite_of_cells]
                end
            end
        end

        local glyph_set = map.auto.autowall_glyph_sets[index]
        
        map.auto.autowall_like[get_glyph(glyph_set, 0x15)] = true
        map.auto.autowall_like[get_glyph(glyph_set, 0x16)] = true
        map.auto.autowall_like[get_glyph(glyph_set, 0x17)] = true
        map.auto.autowall_like[get_glyph(glyph_set, 0x18)] = true
        map.auto.autowall_like[get_glyph(glyph_set, 0x25)] = true
        map.auto.autowall_like[get_glyph(glyph_set, 0x26)] = true
        map.auto.autowall_like[get_glyph(glyph_set, 0x27)] = true
        map.auto.autowall_like[get_glyph(glyph_set, 0x28)] = true
    end

end

function tabulate_map(map)
    
    -- read wall features
    local autowall_type = set_up_wall_detect_features(map)

    -- set the glyph set - before we load the objects and set up the map,
    -- so they get the glyphs from the right set.
    set_autowall_glyph_set_base(map, autowall_type)

    -- now go though and convert the strings to tables of tables of int32 (unicode values)
    -- this is not required at clients because it already happened at the server
    for line, map_string in ipairs(map) do
        -- convert string into table of values
        local line_table = utf8.to32(map_string)
        for column, unicode in ipairs(line_table) do
            if map.auto.glyph_mappings[unicode] == nil then 
                -- handle plain text, allows us to write on the map
                line_table[column] = unicode
            else    
                -- there is a specified mapping
                line_table[column] = map.auto.glyph_mappings[unicode]
            end
        end
        map[line] = line_table
    end
    
    map.height = #map 
    map.width = #map[1] 
    
    local start_t = gulp_cpp.SDL_GetTicks()
    fill_corner_gaps(map)
    local time_to_fill = gulp_cpp.SDL_GetTicks() - start_t
    
    if gulp.DebugLevelLoading then
        print(gulp.LuaStateCodeMode, map.title, time_to_fill)
    end

    return map
end




function load_maze_assets(map_name)

    -- get rid of old data
    -- delete_map()     -- needs looking at - currently handled in load_map
    -- delete_objects()

    -- load the named map file
    -- load_map(maze_name, map_number)

    -- choose where we are storing the loaded data
    local map = get_map(map_name)


    --local SDL_GetTicks = gulp_cpp.SDL_GetTicks
    --local t1 = SDL_GetTicks()
    --local t2 = SDL_GetTicks()
    --print("T1", t2-t1)


    
    
    local maze = map.auto.maze

    -- read colours from map data and set up in current maze
    if gulp.LuaStateCodeMode == 'CLIENT' then
        load_colours(map, maze)
    end

    
    -- set up maze itself, as soon as we can
    maze:load_current_maze(map)

    -- this uses maze to get height and width ---- TOTAL CRAP!!!
    


    -- load the objects, mobs, exits
    -- for client mode, mobile items are provided by the server, so get rid of them
    if gulp.LuaStateCodeMode == 'CLIENT' then
         -- is this necessary? Client no longer loads mazes, gets them from server?
        remove_mobs_from_maze(map, maze)
    else
        load_mobs_from_maze(map, maze)      
    end
    
    load_on_map_objects(map, maze)
    load_fixed_objects(map, maze)
    load_exits_from_maze(map, maze)
    load_doors_from_maze(map, maze)
    
    -- make changes based on various things, on the view copy
    gulp_cpp.map_transform(maze, map.auto.autowall, map.auto.autowall_like, map.auto.wall_glyph_lookup_table_transformed)
    
    update_floor_fonts(map, maze)
    
    load_game_manager(map_name)
    

end


function maze_set_string(map_name, line, column, string)
    local map 
    local maze 
    
    if map_name then map = get_map(map_name) end
    
    if map then 
        maze = map.auto.maze
        for i = 1,#string do
            local c = string:byte(i)
            map[line+1][column+1] = c
            maze:update_glyph(line, column, c)
            column = column + 1
        end
    end
end


function set_conditional(map_name, conditional, value, new_glyph)
    
    local map = get_map(map_name)
    
    map.conditionals[conditional] = value

    -- notify the manager in case it wants to do something elsewhere in the map
    if map.manager_name then
        find_game_manager(map):map_conditional_notify(conditional, value, map_name)
    end
                
    -- update the floor glyph
    if new_glyph then
        set_map_glyph(map_name, line, column, new_glyph)
    end

    local message = { 
        mt = "conditional_update",
        map_name = map_name,
        conditional = conditional,
        value = value,
    }

    gulp_networking:broadcast_to_all_with_map_loaded(map_name, message)
end    

function set_map_glyph(map_name, line, column, glyph)
    
    local map = get_map(map_name)
    local maze = map.auto.maze
    
    map[line][column] = glyph
    maze:update_glyph(line-1, column-1, glyph)
    
    local message = {
        mt = 'glyph_update',
        map_name = map_name,
        line = line,
        column = column,
        glyph = glyph,
    }
    
    gulp_networking:broadcast_to_all_with_map_loaded(map_name, message)
end



function is_active_floor(map_name, line, column)
    
    local map = get_map(map_name)
    local maze = map.auto.maze
    
    local function handle_map_based_floor_effects(floor_effect)
        
        local command = floor_effect[1]
        local new_glyph = nil
        
        if command == 'set' or command == 'add' then
            
            local conditional, value = floor_effect[2], floor_effect[3]
            new_glyph = floor_effect[4]
            
            -- set the conditional
            if command == 'add' then
                set_conditional(map_name, conditional, map.conditionals[conditional] + value, new_glyph)
            else
                set_conditional(map_name, conditional, value, new_glyph)
            end
            
            -- consume floor effect
            floor_effect = nil
        end
        
        if command == 'timer' then
            if map.manager_name then
                local timer, action = floor_effect[2], floor_effect[3] 
                
                find_game_manager(map):map_timer_control(timer, action)
                
                if floor_effect[4] then
                    set_map_glyph(map_name, line, column, floor_effect[4])
                end
            end
        end
    end
    
    
    -- check if this map has any active floors
    if map.active_floor == nil then return nil end
    
    if line<0 or line>=maze:height() then return false end
    if column<0 or column>=maze:width() then return false end
    
    
    line = math.floor(line+1)
    column = math.floor(column+1)
    
    local glyph = map[line][column]
    local floor_effect = map.active_floor[glyph]
    
    -- handle map based floor effects here
    if floor_effect then
        -- this may consume the floor effect (set it to nil), or even alter what goes back to the player
        floor_effect = handle_map_based_floor_effects(floor_effect)
    end
    
    -- return player based floor effects to player
    return floor_effect
end    

function is_wall(map_name, line, column)
    local map = get_map(map_name)
    local maze = map.auto.maze
    if line<0 or line>=maze:height() then return true end
    if column<0 or column>=maze:width() then return true end
--[[    
    object solidity controlled by collision rects or map definitions for now
    
    local objects = map.auto.objects
    for _, object in pairs(objects) do
        -- check if we can walk through this object
        if object:is_solid(line, column) then return true end
    end    
--]]    
    return map.auto.wall_detect_type[maze:get_glyph(math.floor(line), math.floor(column))]
end

