-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
--[[

    Responsibile for loading all the maps that the server requires to be loaded
    
--]]

require('strict')
require('utils')
class = require('middleclass')


local ClientMapLoader = class('ClientMapLoader')

gulp.maps = {}

function ClientMapLoader:load_map_file(name, file)
    
    print("client load_map_file("..name..")")

    if file == nil then
        print("FATAL - no file data in client's load_map_file")
        error("FATAL - no file data in client's load_map_file")
    end
    
    local success = true    -- assume load success
    
    if success then
        local map
        map = file
        
        map.filename = name
        
        local viewport_width = map.viewport_width or 32 
        local viewport_height = map.viewport_height or 21
        
        map.auto.maze = gulp_cpp.PresentationMaze(viewport_width, viewport_height)
        
        -- no mobs or pickables at the client end
        map.auto.mobs = {}
        map.auto.pickables = {}
        map.auto.objects = {}

        -- set a default game manager to be loaded
        if map.manager_name == nil then map.manager_name = 'default_game_manager' end
        if map.manager_file == nil then map.manager_file = map.manager_name end
        
        if map.auto.maze == nil or map.auto.maze == 0 then
            print("FATAL - no maze allocated in load_map_file")
            error("FATAL - no maze allocated in load_map_file")
        else
            gulp.maps[name] = map
        end

    end


end

function ClientMapLoader:form_map_request_message(map)
    return { 
        mt = "map_request",
        map_name = map,
    }
end

function ClientMapLoader:form_drop_map_message(map)
    return { 
        mt = "drop_map",
        map_name = map,
    }
end

 
function ClientMapLoader:request_neighbouring_maps(map_name)
    
    if not check_map_exists(map_name) then return false end
    
    local function mangle_map_name(exit, instance_extension)
        local new_map_name = exit.file
        
        -- Common_area is used for the end of multiplater games to omit the extension so
        -- players can exit the game back to the lobby (or wherever).
        if exit.mode ~= 'common_area' then
            -- If we have an instance extension already, then just add it,
            if instance_extension ~= "" then
                new_map_name = new_map_name..instance_extension
            
            -- else if we are starting a singleplayer game, mangle with the player id.
            -- The start of multiplayer games is handled by the multiplayer game lobby code.
            elseif exit.mode == 'singleplayer' then
                if gulp.localplayer then
                    new_map_name = new_map_name.."_^_"..gulp.localplayer.id
                else
                    debugMessage("Tried to create singleplayer game with no player")
                end
            end
        end
        
        return new_map_name
    end
    
    -- demangle map_name to get any instance extension
    local instance_extension = ""
    local s,e = map_name:find("_^_")
    if s ~= nil then
        instance_extension = map_name:sub(s)
    end

    
    -- also drops non-neighbouring maps
    
    local neighbouring_maps = { [map_name] = true }
    local new_map_name
    
    -- preload levels to north, south, east and west
    local exit_names = { 'exit_north', 'exit_south', 'exit_east', 'exit_west' }
    for _,exit in ipairs(exit_names) do
        -- can't enter special mode maps by north, south, east, west
        new_map_name = gulp.maps[map_name][exit]
        if new_map_name then
            neighbouring_maps[new_map_name] = true
        end
    end
    
    -- preload any levels with other exits
    if gulp.maps[map_name].exits ~= nil then
        for _, exit in pairs(gulp.maps[map_name].exits) do
            if exit.file then
                new_map_name = mangle_map_name(exit, instance_extension)
                neighbouring_maps[new_map_name] = true
            end
        end
    end
    
    -- preload any other levels that the level specifies
    if gulp.maps[map_name].preload ~= nil then
        for _, exit in pairs(gulp.maps[map_name].preload) do
            if exit.file then
                new_map_name = mangle_map_name(exit, instance_extension)
                neighbouring_maps[new_map_name] = true
            end
        end
    end

    -- debug
    --print("request_neighbouring_maps("..map_name..")")
    
    local message
    -- drop maps we have loaded but are no longer next to
    for map,_ in pairs(gulp.maps) do
        if not neighbouring_maps[map] then 
            -- debug
            print("      drop "..map.. "("..map_name..")")
            self:DropMap(map)
            
            -- tell the server we are dropping the map
            -- this will allow the server to garbage collect it if no other client is using it
            message = self:form_drop_map_message(map)
            gulp_networking:SendToServer(message)
        end
    end
    
    -- requests maps we are next to but don't have loaded already
    for map,_ in pairs(neighbouring_maps) do
        if not check_map_exists(map) then 
            -- debug
            --print("   request "..map)
            message = self:form_map_request_message(map)
            gulp_networking:SendToServer(message)
        end
    end
    
end


function ClientMapLoader:LoadInitialMap()
    local file = gulp.client_initial_map_file
    local map_name = gulp.client_initial_map_name
    
    -- load the map file we were given on connection
    self:load_map_file(map_name, file)
        
    -- load all the maze assets
    load_maze_assets(map_name)
end





function ClientMapLoader:initialize()

end

function ClientMapLoader:DropAllMaps()
    for name,map in pairs(gulp.maps) do
        self:DropMap(name)
    end
end


function ClientMapLoader:DropMap(map_name)
    
    if not check_map_exists(map_name) then return end
    local map = get_map(map_name)
    
    for _, object in ipairs(map.auto.objects) do
        object:finish()
    end
    
    -- drop the display elements for all remote players on this map
    -- but we don't drop the players themselves
    for id,player in all_remote_players_on_map(map_name) do
        player:drop_glyphs()
    end
    
    -- drop the mobs completely
    for id,mob in pairs(map.auto.mobs) do
        mob:finish()
    end
    
    map.auto.mobs = {}
    map.auto.objects = {}
    
    collectgarbage()
    
    local manager_name = map.manager_name
    gulp.maps[map_name] = nil
        
    -- must drop the game manager after the map is deleted,
    -- or it will be seen as still in use
    drop_game_manager(manager_name)

end

function ClientMapLoader:SwitchMap(player_id, map_name)
    if gulp.DebugLevelLoading then
        print("SwitchMap("..map_name..")")
    end
    
    -- in the client, we only ever have one map loaded at a time
    -- self:DropAllMaps()
    -- self:LoadMap(map_name)
    
    -- if the map hasn't been loaded from the server, just refuse the switch
    if not check_map_exists(map_name) then return false end
    
    -- this is the local player changing map - set up the graphics and localplayer
    if gulp.localplayer then        -- require check to handle startup
        gulp.localplayer.map_name = map_name
        gulp.localplayer:switch_glyph_to_map(map_name) 
    end     
    
    local map = get_map(map_name)
    if(map) then
        local manager = find_game_manager(map)
        if(manager) then
            manager:arrange_screen()
        end
    end

    -- no need to cause a draw on the game manager of the map that we moved to
    -- because this will happen when we get send the game_manager state from 
    -- the server
    
    -- request neighbouring maps from server so they are pre-loaded for when we use an exit
    self:request_neighbouring_maps(map_name)
    
    return true
end

return ClientMapLoader