-- Forlorn Fox
-- Copyright © 2020 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Caves of Chaos game manager
--

require('strict')
require('utils')
require('status_manager')
local class = require('middleclass')
local GameManager = require('map_controllers/game_manager')

local CavesManager = class('CavesManager', GameManager)

local combat_menu = require('map_controllers/caves/combat_menu')

require('map_control')


--[[

Notes:

Created by caves11 map when a player enters it. Need to worry about the garbage collection of caves11, 
because that will destroy this manager even if it is managing other maps.

When to reset / respawn?

How to generate the rest of the maps in the caves?

In the original, players cannot escape from the start room, should this be the case in this version?

If a player goes into combat with a monster, we should have a way of indicating that to other players 
who are not yet in combat

If a player dies, have them transferred to a hospital and attended by the nurse, as per the summerhill game

--]]


function CavesManager:initialize()
    
    debugMessage("CavesManager:initialize()")
    
    GameManager:initialize()
    gulp._delegate_list:add_delegate(self)
    
    -- this is called when someone enters caves...
    -- here we should proc gen the whole cave system?
    
    self.combats = {}
    self.next_combat_id = 1         -- this will loop... in theory
    
    -- don't do garbage collection until the caves have been disturbed!
    self.disturbed = false
    
    self.maps_in_caves = {'caves/caves11'}
    
end

function CavesManager:finish()
    gulp._delegate_list:drop_delegate(self)
    debugMessage("CavesManager:finish()")
end

function CavesManager:add_map(map_name)
    for _,n in ipairs(self.maps_in_caves) do
        if n == map_name then return end
    end
    table.insert(self.maps_in_caves, map_name)
end


function CavesManager:update(dt)
    if gulp.LuaStateCodeMode == 'SERVER' then    
        self:server_update(dt)
    else
        self:client_update(dt)
    end
end

function CavesManager:server_update(dt)    
    -- need to check all maps for things like respawn etc.
    
    -- for all our maps...
    -- if there are a lot of maps in caves then we might want to maintain a list of those with players on
    -- to keep processing time here reasonable
    
    -- proposed
    -- check to see if we create a new combat (player approaches mob)
    -- check to see if players join an existing combat
    -- check to see if a mob joins an existing combat
    -- handle combats
    -- check to see if a combat is over, release survivors
    
    -- can run the whole of the above on a per map basis
    
    local function nearby(player, mob)
        local d1 = mob.line - player.line
        local d2 = mob.column - player.column
        local d = math.sqrt(d1*d1 + d2*d2)
        return d < 2
    end
    
    local function new_combat(map_name, player, mob)
        local combat = {}
        local combat_id = self.next_combat_id
        self.combats[combat_id] = combat
        self.next_combat_id = self.next_combat_id + 1
        
        combat.combat_id = combat_id
        combat.map_name = map_name
        combat.players = { player }
        combat.mobs = { mob }
        combat.line = mob.line
        combat.column = mob.column
        
        debugMessage("Starting combat: player "..player.name.." to combat "..combat.combat_id.." state "..tostring(player.combat_state))
        
        player.combat_id = combat_id
        player.combat_state = 'fight'
        
        mob.combat_id = combat_id
        mob.combat_state = 'fight'
        
        mob:start_combat(combat_id, self.my_name)
        
        -- need to send messages to clients!!
    
        local message = { 
            mt = "game_control",
            manager = self.my_name,
            command = "caves_start_combat",
            combat_id = combat.combat_id,
            mob_name = mob.avatar_data.avatar_type,
            mob_hp = mob.hp,
            mob_meter_time = mob.mob_meter.time,
            player_hp = player.hp,
            graphics_bg = get_map(map_name).bg,
            description = mob:get_start_combat_description(),
            time_to_first_hit = mob.time_to_hit,
            -- picture = -- specify the glyph to show in the menu
        }
        player.socket_ref:SendMessage(message)
        

        return combat
    end
    
    local function end_combat(combat)
        self.combats[combat.combat_id] = nil        -- don't use table.remove because other combat's IDs would be changed!
        
        for _, mob in pairs(combat.mobs) do
            mob:release()
            mob.combat_id = nil
            mob.combat_state = nil
        end
        for _,player in pairs(combat.players) do
            debugMessage("End combat")
            player.combat_id = nil
            player.combat_state = nil
        end
        
        -- need to send messages to clients!!
    end
    
    local function add_player_to_combat(player, combat)
        
        debugMessage("Adding player "..player.name.." to combat "..combat.combat_id)
        
        table.insert(combat.players, player)
        player.combat_id = combat.combat_id
        
        -- need to send messages!!
        local mob = combat.mobs[1]          -- revisit if adding multiple mobs...!
        local message = { 
            mt = "game_control",
            manager = self.my_name,
            command = "caves_start_combat",
            combat_id = combat.combat_id,
            mob_name = mob.avatar_data.avatar_type,
            mob_hp = mob.hp,
            mob_meter_time = mob.mob_meter.time,
            graphics_bg = get_map(combat.map_name).bg,
            description = mob:get_start_combat_description(),
            time_to_first_hit = mob.time_to_hit,
            -- picture = -- specify the glyph to show in the menu
        }
        player.socket_ref:SendMessage(message)
        
        
        for _,p in pairs(combat.players) do
            local message = { 
                mt = "game_control",
                manager = self.my_name,
                command = "caves_message",
                description = player.name.." joined the battle!"
            }
            p.socket_ref:SendMessage(message) 
        end
        
    end
    
    local function player_ran_from_combat(player, combat)
        for i,p in pairs(combat.players) do
            if p == player then
                table.remove(combat.players, i)
            end
        end
        
        player.combat_state = 'running'
        player.combat_id = nil
        
        -- need to send messages!!
        for _,p in pairs(combat.players) do
            local message = { 
                mt = "game_control",
                manager = self.my_name,
                command = "caves_message",
                description = player.name.." ran from the battle!"
            }
            p.socket_ref:SendMessage(message) 
        end
        
    end

    local function player_running_from_combat(player)
        player.after_combat_timer = player.after_combat_timer - dt
        if player.after_combat_timer < 0 then
            player.combat_state = nil
        end
    end

    local function player_dying(player)
        player.after_combat_timer = player.after_combat_timer - dt
        if player.after_combat_timer < 0 then
            -- player finished dying and is now dead
            local combat = self.combats[player.combat_id]
            for i,p in pairs(combat.players) do
                if p == player then
                    table.remove(combat.players, i)
                end
            end
            
            debugMessage("Player dying")
            
            player.combat_state = nil
            player.combat_id = nil
            
            -- teleport the player to the hospital
            player.hp = 1
            if player.level > 1 then player.level = player.level - 1 end
            
            -- set up an exit for the player to take 
            -- take_exit runs client-side, and so does not act immediately
            -- this causes a race condition between take_exit() happening and the player re-entering combat,
            -- since we just set combat_state and combat_id for the player to nil
            -- to prevent this, we have an extremely nasty hack:
            player.map_name = 'caves/caves_infirmary'
            -- here we preset the map name to what the take_exit call is going to set it to
            -- we are setting it server side so we don't actually affect what take_exit() does
            -- although we are relying on take_exit working 100% of the time and actually going 
            -- where we said it should
            
            local exit_to_take = { file = 'caves/caves_infirmary', destination = '2' }
            player:take_exit(exit_to_take)
            
        end
    end
    
    

    --[[
    local function add_mob_to_combat(mob, combat)
        debugMessage("No support for multi mob battles")
        --table.insert(combat.mobs, mob)
        --mob.combat_id = combat.combat_id
        -- need to send messages!!
    end
    --]]
    
    local function win_combat(combat)
        for _,p in pairs(combat.players) do
            p.combat_state = 'won'
            local message = { 
                mt = "game_control",
                manager = self.my_name,
                command = "caves_won_combat",
            }
            p.socket_ref:SendMessage(message) 
        end
    end
        
    local no_players_in_caves = true
    
    --local map_name = self.map_name
    --for(all maps in caves) do
    for _,map_name in pairs(self.maps_in_caves) do
        local map = get_map(map_name)
        
        for id,player in all_players_on_map(map_name) do
            
            -- found at least one player in the caves
            self.disturbed = true
            no_players_in_caves = false
            
            -- process the players
            if player.combat_state == 'running' then
                player_running_from_combat(player)
            elseif player.combat_state == 'dying' then
                player_dying(player)
            else
                -- check to see if we create a new combat (player approaches mob)
                if player.combat_id == nil then 
                    for _,mob in pairs(map.auto.mobs) do
                        if mob.combat_id == nil then
                            if nearby(player, mob) then
                                new_combat(map_name, player, mob)
                            end
                        end
                    end
                    -- check to see if players join an existing combat, if they didn't form a new one
                    if player.combat_id == nil then
                        for _,combat in pairs(self.combats) do
                            if nearby(player, combat) then
                                add_player_to_combat(player, combat)
                            end
                        end
                    end
                end
            end
        end
        
        -- check to see if mobs join an existing combat
        --[[
        for id,mob in pairs(map.auto.mobs) do
            if mob.combat_id == nil then
                for _,combat in pairs(self.combats) do
                    if nearby(mob, combat) then
                        add_mob_to_combat(mob, combat)
                    end
                end
            end
        end
        --]]
                        
        -- handle combats
        for combat_id, combat in pairs(self.combats) do
            -- check players run from combat
            for _,player in pairs(combat.players) do
                if player.id == nil then
                    -- player seems to have logged off
                    player_ran_from_combat(player, combat)
                    player.combat_state = nil
                elseif player.combat_state == 'run' then
                    player.after_combat_timer = 2
                    player_ran_from_combat(player, combat)
                elseif player.combat_state == 'die' then
                    player.after_combat_timer = 4
                    player.combat_state = 'dying'
                end
            end
            
            -- could check for mobs running from combat, 
            -- but they can't run from combat yet so don't bother
            
            -- if all the mobs (currently only 1) are dead, the battle was won
            if #combat.mobs == 0 then
                win_combat(combat)
            end
            
            -- check for combats that are over
            if #combat.players == 0 or #combat.mobs == 0 then
                end_combat(combat)
            end
            
        end
        
    end    
    
    if no_players_in_caves and self.disturbed then
        -- reset caves for the next run
        debugMessage("Garbage collecting in caves manager")
        gulp.map_loader:map_set_garbage_collect(self.maps_in_caves)
        self.maps_in_caves = {}
    end
    
end

function CavesManager:client_update(dt)    
end

function CavesManager:process_game_control_message(message)    
    if gulp.LuaStateCodeMode == 'SERVER' then    
        self:server_process_game_control_message(message)
    else
        self:client_process_game_control_message(message)
    end
end


function CavesManager:player_fight(message)
    -- fight who?
    -- currently we only support one enemy, so this isn't an issue
    -- in future the message would tell us
    
    local combat = self.combats[message.combat_id]
    if combat then 
        local mob = combat.mobs[1]
        local player = get_player_by_id(message.player_id)
        
        if mob and player then
            
            local damage = player:hit_mob(mob)
            mob:take_damage(damage)
            
            -- define carrier message for descriptions
            local _message = { 
                mt = "game_control",
                manager = self.my_name,
                command = "caves_message",
            }
            
            local description, index = player:get_hit_description(mob, damage)
            local other_description = player:get_other_hit_description(mob, index, damage)
            
            _message.description = description
            player.socket_ref:SendMessage(_message) 
            
            _message.description = other_description
            for _,p in pairs(combat.players) do
                if p.id ~= message.player_id then
                    --print("opid:"..other_player_id.." mpid:"..message.player_id)
                    p.socket_ref:SendMessage(_message) 
                end
            end
        
            if mob:is_dead() then
            
                local death_description = mob:get_death_description(player, damage)
                local other_death_description = mob:get_other_death_description(player, damage)

                _message.description = death_description
                player.socket_ref:SendMessage(_message) 
            
                _message.description = other_death_description
                for _,p in pairs(combat.players) do
                    if p.id ~= message.player_id then
                        p.socket_ref:SendMessage(_message) 
                    end
                end
            end
        end
    end
end


function CavesManager:player_run(message)
    local player = get_player_by_id(message.player_id)
    player.combat_state = 'run'
end


function CavesManager:server_process_game_control_message(message)    
    if message.command == 'caves_combat_run' then
        self:player_run(message)
    elseif message.command == 'caves_combat_fight' then
        self:player_fight(message)
    end
end

function CavesManager:client_process_game_control_message(message)    
    if message.command == 'caves_start_combat' then
        -- show the combat menu
        self.combat_menu = combat_menu:new(self.my_name, message)
        self.combat_menu:show_message(message.description)
    elseif message.command == 'caves_mob_attack' then
        self.combat_menu:mob_attack(message)
    elseif message.command == 'caves_message' then
        self.combat_menu:show_message(message.description)
    elseif message.command == 'caves_won_combat' then
        self.combat_menu:won_combat()
    elseif message.command == 'caves_mob_hp_update' then
        -- work to do here if there are ever multiple mobs in the fight...
        self.combat_menu:mob_hp_update(message.mob_hp)
    elseif message.command == 'caves_player_hp_update' then
        -- work to do here if there are ever multiple mobs in the fight...
        self.combat_menu:player_hp_update(message.hp)
    end
end

function CavesManager:broadcast_to_all_players_in_combat(combat_id, message)
    local combat = self.combats[combat_id]
    if combat then
        for _,player in pairs(combat.players) do
            player.socket_ref:SendMessage(message)
        end
    end
end

function CavesManager:broadcast_to_all_players_in_combat_except(combat_id, message, player_id)
    local combat = self.combats[combat_id]
    if combat then
        for _,player in pairs(combat.players) do
            if player.id ~= player_id then
                player.socket_ref:SendMessage(message)
            end
        end
    end
end

function CavesManager:mob_died(combat_id, network_id)
    
    local combat = self.combats[combat_id]
    
    local mob_to_remove = nil
    for i,mob in pairs(combat.mobs) do
        if mob.network_id == network_id then
            mob_to_remove = i
        end
    end
    if mob_to_remove then table.remove(combat.mobs, mob_to_remove) end
    
    -- open the exit doors
    set_conditional(combat.map_name, 'exit_doors', 'open')
    
end

    

return CavesManager
