-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

--[[
    
    This is the game manager for the Coin Rush map
    
    This manager runs the coin rush game across up to 4 players (server + 3 clients)
    
    It behaves differently depending on whether it is running in the server or
    client:
        * display is the same in each
        * server keeps count of coins for all players, via notify() calls
          from Player and PlayerBase classes
        * notify() calls are ignored in clients  
        * server updates clients with current coin counts whenever they change  
        * both server and client detect game end and display game end menu 
    
--]]

require('strict')
require('utils')
require('status_manager')
local GoldCoin = require('objects/gold_coin')
local class = require('middleclass')
local GameManager = require('map_controllers/game_manager')
local chase_lights = require('controllers/ChaseLights')
local CollectableCounter = require('map_controllers/collectable_counter')


local CoinRushManager = class('CoinRushManager', GameManager)

local total_coins = 169
local win_coins = 85
local gold_coin_glyph = to_utf8(get_glyph('blackwall', 159))

function CoinRushManager:initialize()
    GameManager:initialize()
    
    gulp._delegate_list:add_delegate(self)
    
    self.coin_counter = CollectableCounter("_crm_coin_count")
    
    self.game_over = false
    
    self:reset_coin_counts()
    
end



function CoinRushManager:finish()
    gulp._delegate_list:drop_delegate(self)
end

function CoinRushManager:get_coin_count_table()
    
    local cc = {0,0,0,0}
    local total = 0
    local game_group = nil
    
    if gulp.LuaStateCodeMode == 'SERVER' then
        -- find a player playing this game
        for _,player in all_players_on_map(self.map_name) do
            game_group = player.game_group
            break
        end
    else
        game_group = gulp.localplayer.game_group
    end
    
    if game_group ~= nil then
        for position_in_group, id in pairs(game_group) do
            cc[position_in_group] = self.coin_counter:get_count_by_id(id)
            total = total + cc[position_in_group]
        end
    end
        
    return cc, total
end

    

function CoinRushManager:draw()
    if self.need_to_draw then
    
        clear_status_line(0)
        clear_status_line(1)
        clear_status_line(2)
        
        local cc = self:get_coin_count_table()
        
        set_status_line(1, 0, cc[1] .. gold_coin_glyph)
        set_status_line(2, 0, cc[4] .. gold_coin_glyph)
        set_status_line(1, 31-string.len(cc[2]), gold_coin_glyph .. cc[2])
        set_status_line(2, 31-string.len(cc[3]), gold_coin_glyph .. cc[3])
        
        if self.game_over then
            set_status_line(0, 9.5, gold_coin_glyph .. ' GAME OVER ' .. gold_coin_glyph)
        else 
            set_status_line(0, 9.5, gold_coin_glyph .. ' Coin Rush ' .. gold_coin_glyph)
            set_status_line(0, 0, 'Left' .. gold_coin_glyph .. total_coins-(cc[1]+cc[2]+cc[3]+cc[4]))
            local win_string = win_coins .. gold_coin_glyph .. 'Win'
            set_status_line(0, 32-string.len(win_string), win_string)
        end
        
        self.need_to_draw = false
    end

end


function CoinRushManager:show_game_over_menu()
    
    local winning_amount = 0
    local cc = self:get_coin_count_table()
    
    for _,value in ipairs(cc) do
        if value > winning_amount then
            winning_amount = value
        end
    end

   
    local function switch_to_lobby()

        local line_modifier = 0
        for player = 1, 4 do
            if gulp.localplayer.game_group[player] == gulp.localplayer.id then
                line_modifier = player
            end
        end


        local exit_to_take = { file='coin_rush_lobby', 
                    destination_line = 10 + line_modifier,
                    destination_column = 16,
            }
        
        gulp.localplayer:take_exit(exit_to_take)
        
    end
    

    -- show the end of game menu
    local menu = Menu{
        
        line = 4,
        column = 1,
        
        background = gulp.colour.BLUE.simple_colour, 
        foreground = gulp.colour.BRIGHT_YELLOW.simple_colour,
        border_style = 2,
        {        "\x1b[32m     We have a winner!!     " },
        { "" },
        { "" },
        { "\x1b[33m\x1b[1m       EXIT TO LOBBY", switch_to_lobby },
    }
        
    for key,value in ipairs(cc) do
        if value == winning_amount then
            -- this is (one of) the player(s) who got the winning amount
            local winner = {         "\x1b[32m        " .. gold_coin_glyph .. " Player " .. key .. " " .. gold_coin_glyph}
            menu:insert_item(3, winner)
        end
    end
            
end
            

function CoinRushManager:reset_coin_counts()
    
    for id,player in all_players() do
        
        self.coin_counter:set_count_by_id(id, 0)
    end
    
    if gulp.LuaStateCodeMode == 'SERVER' then 
        -- update the clients with the new coin counts
        local message = self:make_coin_count_message()
        gulp_networking:broadcast_to_all_on_game_controller(self.my_name, message)
    end
    
end


function CoinRushManager:update(dt)
    
    if self.game_over == false then
    
        local cc, total = self:get_coin_count_table()
        
        if total >= win_coins then
            -- enough coins picked, game is over
            self.game_over = true
            
            if gulp.LuaStateCodeMode == 'SERVER' then
                hold_all_mobs('coin_rush')
            else 
                -- might not be coin rush only in the case that this controller is running on 
                -- the server, but the server player is not themselves playing coin rush
                gulp.localplayer:hold()
                self:show_game_over_menu()
            end
            
        end
    end
    
end

function CoinRushManager:make_coin_count_message()
    local message = {
        mt = 'game_control',
        command = 'coin_count',
        manager = self.my_name,
        counts = {},
    }
    
    for id,player in all_players() do
        message.counts[id] = self.coin_counter:get_count(player)
    end
    
    return message
end

function CoinRushManager:send_game_manager_status(socket_ref)
    local message = self:make_coin_count_message()
    socket_ref:SendMessage(message)
end

function CoinRushManager:notify(player_id, action)
    -- no action in client mode, the server will be taking care of this for us
    if gulp.LuaStateCodeMode ~= 'SERVER' then return end
    
    -- BELOW, WE ONLY RUN ON THE SERVER
    
    if action == 'pick_up_coin' then
        self.coin_counter:change_count_by_id(player_id, 1)
    end
    
    if action == 'drop_coin' then
        self.coin_counter:change_count_by_id(player_id, -1)
    end
    
    -- update the clients with the new coin counts
    local message = self:make_coin_count_message()
    gulp_networking:broadcast_to_all_on_game_controller(self.my_name, message)
    print_table(message)

end


function CoinRushManager:process_game_control_message(message)
    
    if message.manager == self.my_name then
        if message.command == 'coin_count' then
            for id,count in pairs(message.counts) do
                self.coin_counter:set_count_by_id(id, count)
                print_table(message)
            end
            self:cause_drawing()
        end
    end
end             


return CoinRushManager


