-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

--[[
    
    This is the default game manager
    
    It is loaded if there is no game manager specified in a map file
    
    Currently, it displays the title of the current map at the top of the screen,
    and keeps and displays a count of the coins collected by players.
    
        
    For an example of a more integrated game manager, see coin_rush_manager.lua
    
    
--]]


require('strict')
require('utils')
require('status_manager')
local GoldCoin = require('objects/gold_coin')
local class = require('middleclass')
local GameManager = require('map_controllers/game_manager')
local CollectableCounter = require('map_controllers/collectable_counter')

local DefaultGameManager = class('DefaultGameManager', GameManager)

local gold_coin_glyph = to_utf8(get_glyph('blackwall', 159))

function DefaultGameManager:initialize()
    GameManager:initialize(self)
    gulp._delegate_list:add_delegate(self)
    self.coin_counter = CollectableCounter("_dgm_coin_count")
end

function DefaultGameManager:finish()
    gulp._delegate_list:drop_delegate(self)
end



function GameManager:draw()
    --set_status_line(1, 5, "L:" .. gulp.localplayer.line)
    --set_status_line(2, 5, "C:" .. gulp.localplayer.column)
    if self.need_to_draw then
        
        -- need to draw the status lines using the absolute draw list, or in a better way than that
        
        --[[
        -- set the map title
        local available_width = gulp.STATUS_COLUMNS
        
        local player = gulp.localplayer
        local title = get_map(player.map_name).title
        clear_status_line(0)
        local start = math.floor((available_width - #title)/2)
        if start < 0 then start = 0 end
        set_status_line(0, start, title)
        
        if gulp.localplayer.playing_tag then
            set_status_line(0, 0, "Tag! ")
            set_status_line(0, available_width-5, " Tag!")
        end
        
        clear_status_line(1)
        clear_status_line(2)
        
        set_status_line(1, 0, self.coin_counter:get_count_by_id('client1') .. gold_coin_glyph)
        set_status_line(2, 0, self.coin_counter:get_count_by_id('client4') .. gold_coin_glyph)
        
        local cc1 = gold_coin_glyph .. self.coin_counter:get_count_by_id('client2')
        local cc2 = gold_coin_glyph .. self.coin_counter:get_count_by_id('client3')
        
        set_status_line(1, available_width-string.len(cc1), cc1)
        set_status_line(2, available_width-string.len(cc2), cc2)
        --]]
        self.need_to_draw = false
        
    end
    
    --print_on_map(10, 10, "hello")
    --print_on_map(11, 10, "hello", 'forgotten')
    --print_on_map(12, 10, "hello", 'blood', gulp.colour.RED.simple_colour)
    --print_on_map(13, 10, "hello", 'blood', nil, get_SDL_Color_by_name('RED'))
    --print_on_map(14, 10, "hello", 'scribe', gulp.colour.RED.simple_colour, get_SDL_Color_by_name('GREEN'))
end

function DefaultGameManager:make_coin_count_message(socket_ref)
    local message = {
        mt = 'game_control',
        manager = self.my_name,
        command = 'coin_count',
        counts = {},
    }
    
    for id,player in all_players() do
        message.counts[id] = self.coin_counter:get_count(player)
    end
    
    return message
end

function DefaultGameManager:send_game_manager_status(socket_ref)
    local message = self:make_coin_count_message()
    socket_ref:SendMessage(message)
end

function DefaultGameManager:notify(player_id, action)
    -- no action in client mode, the server will be taking care of this for us
    if gulp.LuaStateCodeMode ~= 'SERVER' then return end
    
    -- BELOW, WE ONLY RUN ON THE SERVER
    
    if action == 'pick_up_coin' then
        self.coin_counter:change_count_by_id(player_id, 1)
    end
    
    if action == 'drop_coin' then
        self.coin_counter:change_count_by_id(player_id, -1)
    end
    
    -- update the clients with the new coin counts
    local message = self:make_coin_count_message()
    gulp_networking:broadcast_to_all_on_game_controller(self.my_name, message)
    
end


function DefaultGameManager:process_game_control_message(message)
    
    if message.manager == self.my_name then
        if message.command == 'coin_count' then
            for id,count in pairs(message.counts) do
                self.coin_counter:set_count_by_id(id, count)
            end
            self:cause_drawing()
        end
    end
    
end             


return DefaultGameManager
