-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

--[[
    
    This is the game manager for Eyebeast Escape
    
--]]

require('strict')
require('utils')
require('status_manager')
local class = require('middleclass')
local GameManager = require('map_controllers/game_manager')

local EyebeastEscapeManager = class('EyebeastEscapeManager', GameManager)

local number_of_levels = 3

function EyebeastEscapeManager:initialize()
    
    GameManager:initialize()
    gulp._delegate_list:add_delegate(self)
    
    self.running = false
    
    self.level_timer_running = false
    self.level_timer = 0
end


-- SERVER side
function EyebeastEscapeManager:stop_managed_game(player_id, map_being_left)
    self.running = false
    
    -- save the game data
    gulp.db:save_eyebeast_escape_progress(self.player_id, map_being_left)
    
    -- move the player to lobby
    local exit_to_take = { file = 'singleplayer_lobby', destination = '$', mode='common_area' }
    local player = get_player_by_id(self.player_id)
    player:take_exit(exit_to_take)
end

-- SERVER side
function EyebeastEscapeManager:start_managed_game(player_id)
        
    if self.running then
        debugMessage("Eyebeast Escape Manager already running")
        return
    end
    
    self.running = true
    self.player_id = player_id
    
    local player = get_player_by_id(player_id)
    
    
    -- load any data required from db
    local game_map_file = gulp.db:get_eyebeast_escape_progress(player_id)
    if game_map_file == nil then 
        game_map_file = 'eyebeast_escape/highscore'
    end
    
    debugMessage("Starting Eyebeast Escape on level "..game_map_file)
    
    self.wr, self.pb = gulp.db:get_eyebeast_escape_level_times(player_id, game_map_file)
    
    -- load the map (based on loaded data...?) and send it down to the client 
    local game_map = game_map_file
    
    -- mangle the game_map name for a multiplayer instance
    game_map = game_map.."_^_"..player_id
    
    -- now we have the game map, we can set it for the manager and do the post map init
    self.map_name = game_map
    self:post_map_initialisation()
    

    -- actually load the level
    gulp.map_loader:LoadMap(game_map)
    local map_message = gulp.map_loader:form_map_message(game_map)
    
    -- send the game map to the client
    local socket_ref = player.socket_ref
    socket_ref:SendMessage(map_message) 
    gulp_networking:SendMobsAndPickables(socket_ref, game_map)
    player:register_map(game_map)
    
    -- send a start game notice
    local message =  
    { 
        mt = 'game_control', 
        command = 'start_game', 
        manager = self.my_name, 
        player_id = player_id, 
    }
    socket_ref:SendMessage(message) 

    -- set up an exit for the player to take (based on loaded data...?)
    local exit_to_take = { file = game_map_file, destination = '1', mode='singleplayer' }
    player:take_exit(exit_to_take)

    self:reset_level_timer()
    
    if game_map_file == 'eyebeast_escape/highscore' then
        self:send_highscore_table(player_id)
    else
        self:send_wr_pb()
    end
    
end

-- SERVER side
function EyebeastEscapeManager:notify(player_id, action)
    local player = get_player_by_id(player_id)
    if action=='died' then
        local map = get_map(player.map_name)
        if map.reset_all_mobs_on_death then
            -- reset all the mobs!
            for _, mob in pairs(map.auto.mobs) do
                mob:reset()
            end
            --debugMessage("Holding mobs during reset")
            hold_all_mobs(player.map_name)
        end
        
        self:reset_level_timer()
    end
    
    if action=='reset' then
        --debugMessage("Releasing mobs after reset")
        release_all_mobs(player.map_name)
    end
    
end

function EyebeastEscapeManager:send_wr_pb()
    local message =  
    { 
        mt = 'game_control', 
        command = 'wr_pb', 
        manager = self.my_name, 
        wr = self.wr, 
        pb = self.pb,
    }
    
    local player = get_player_by_id(self.player_id)
    local socket_ref = player.socket_ref
    socket_ref:SendMessage(message) 
end

function EyebeastEscapeManager:send_highscore_table(player_id)
    
    local highscores = {}
    for level = 1, number_of_levels do
        local wr, pb = gulp.db:get_eyebeast_escape_level_times(player_id, "eyebeast_escape/level"..level)
        table.insert(highscores, level, { wr, pb })
    end
    
    local message =  
    { 
        mt = 'game_control', 
        command = 'highscores', 
        manager = self.my_name, 
        highscores = highscores,
    }
    
    local player = get_player_by_id(self.player_id)
    local socket_ref = player.socket_ref
    socket_ref:SendMessage(message) 
end

function EyebeastEscapeManager:switch_map(player_id, new_map_name)
    self.map_name = new_map_name
    
    -- disable draw of timers until we have received new maze data
    self.timer_l = nil
    
    local ret_val = GameManager.switch_map(self, player_id, new_map_name)
    
    -- timer reset on map switch is server side only
    if gulp.LuaStateCodeMode == 'SERVER' then
        -- reset timings
        self:reset_level_timer()
        
        -- demangle the map name for load times
        local map_name = demangle_singleplayer_map_name(new_map_name)
        
        if map_name == 'highscore' then
            self:send_highscore_table(player_id)
        else
            -- load the times for just the new level
            self.wr, self.pb = gulp.db:get_eyebeast_escape_level_times(player_id, map_name)
            self:send_wr_pb()
        end
    end
    
    return ret_val
end

function EyebeastEscapeManager:reset_level_timer()
    self:send_timer_message('reset', 0)
    self.level_timer_running = false
    self.level_timer = 0
end


function EyebeastEscapeManager:send_timer_message(action, time)
    local message =  
    { 
        mt = 'game_control', 
        command = 'timer', 
        manager = self.my_name, 
        action = action, 
        time = time,
    }
    
    local player = get_player_by_id(self.player_id)
    local socket_ref = player.socket_ref
    socket_ref:SendMessage(message) 
end


function EyebeastEscapeManager:map_timer_control(timer, action)
    
    if action == 'start' then
        if self.level_timer_running == false then
            self.level_timer_running = true
            self:send_timer_message(action, self.level_timer)
            print("timer started")
        end
    elseif action == 'stop' then
        if self.level_timer_running == true then
            self.level_timer_running = false
            self:send_timer_message(action, self.level_timer)
            print("timer stopped")
        
            if self.level_timer < self.pb then
                -- save the time if it is a new pb
                local map_name = demangle_singleplayer_map_name(self.map_name)
                gulp.db:save_eyebeast_escape_time(self.player_id, map_name, self.level_timer)
            end
        end
    end
end

function EyebeastEscapeManager:update(dt)
    if self.level_timer_running then
        self.level_timer = self.level_timer + dt
    end
end

function EyebeastEscapeManager:draw()
    -- client side only
    -- does the server even call draw...?
    if gulp.LuaStateCodeMode == 'SERVER' then return end
    
    local map_name = demangle_singleplayer_map_name(self.map_name)
    
    -- no special drawing for exit or win
    if map_name == 'exit' or map_name == 'win' or map_name == 'highscore' then return end
    
    -- this is a normal level, draw the timer
    if self.timer_l then
        print_on_map(self.timer_l, self.timer_c, string.format("   %.1f ", self.level_timer))
    end
    
end


-- CLIENT side
function EyebeastEscapeManager:handle_escape_key()
    local bg = gulp.colour.BRIGHT_WHITE.simple_colour
    local fg = gulp.colour.BRIGHT_GREEN.simple_colour
    if math.random(2) == 1 then
        local swap = bg
        bg = fg
        fg = swap
    end
    
    local function quit_to_lobby()
        debugMessage("client quit to lobby")
        self.running = false
        
        local message =  {
            mt = 'game_control',
            command = 'quit_to_lobby',
            manager = self.my_name,
            player_id = self.player_id,
            map_name = gulp.localplayer.map_name,
        }
        
        gulp_networking:SendToServer(message)    
        
        --gulp.save_and_quit_game()
    end
    
    self._escape_menu = Menu{
        background = bg, 
        foreground = fg,
        border_style = 2,
        { "You hit Escape? " },
        { "" },
        { "  SAVE AND QUIT ", quit_to_lobby },
        { "  Back to game ", "this is not a function so just exit"},
    }
    
    return true
end

function EyebeastEscapeManager:process_game_control_message(message)
    
    if message.manager == self.my_name then
        if gulp.LuaStateCodeMode == 'SERVER' then
            -- SERVER side
            if message.command == 'quit_to_lobby' then
                debugMessage("server quit to lobby")
                
                -- demangle the map name for save
                local map_name = message.map_name
                local s,e = map_name:find("_^_")
                if s ~= nil then
                    map_name = map_name:sub(1, s-1)
                end
                
                self:stop_managed_game(self.player_id, map_name)

            end
        else
            -- CLIENT side
            if message.command == 'start_game' then
                debugMessage("client start game")
                self.running = true
                self.player_id = message.player_id
            end
            
            if message.command == 'timer' then
                debugMessage("timer message: ".. message.action..message.time)
                if message.action == 'start' then
                    self.level_timer = 0
                    self.level_timer_running = true
                elseif message.action == 'stop' then
                    -- update client time with what the server recorded - for fairness...
                    self.level_timer = message.time
                    self.level_timer_running = false
                elseif message.action == 'reset' then
                    self.level_timer = 0
                    self.level_timer_running = false
                end
            end
            
            if message.command == 'wr_pb' then
                self.wr = message.wr
                self.pb = message.pb
    
                local map = get_map(self.map_name)
                
                if map.timer_location then
                    self.timer_l = map.timer_location[1]
                    self.timer_c = map.timer_location[2]
                
                    if self.wr == 999.9 then 
                        print_on_map(self.timer_l, self.timer_c, "WR:--.- ")
                    else
                        print_on_map(self.timer_l, self.timer_c, string.format("WR:%02.1f ", self.wr))
                    end
                    
                    self.timer_l = self.timer_l + 1
                    
                    if self.pb == 999.9 then 
                        print_on_map(self.timer_l, self.timer_c, "PB:--.- ")
                    else
                        print_on_map(self.timer_l, self.timer_c, string.format("PB:%02.1f ", self.pb))
                    end
                
                    -- update for the actual timer
                    self.timer_l = self.timer_l + 1
                end
            end
            
            if message.command == 'highscores' then
                -- draw the highscore table
                for level = 1, number_of_levels do
                    local wr, pb = message.highscores[level][1], message.highscores[level][2]
                    print_on_map(level*4,   6, "Level "..level)
                    local wrs = "World Record:  "
                    local pbs = "Personal Best: "
                    
                    -- fp number got slightly mangled in transit, compare with 999.9 failed...
                    wrs = wrs..(wr > 999 and "--.-" or string.format("%.1f ", wr))
                    pbs = pbs..(pb > 999 and "--.-" or string.format("%.1f ", pb))
                    
                    print_on_map(level*4+1, 6, wrs)
                    print_on_map(level*4+2, 6, pbs)
                end    
            end
            
        end
    end
end

return EyebeastEscapeManager


