-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

--[[
    
    This is the game manager for game lobbies
    
    This manager runs a game lobby across up to 4 players 
    
--]]

require('strict')
require('utils')
require('status_manager')

local class = require('middleclass')
local GameManager = require('map_controllers/game_manager')
local chase_lights = require('controllers/ChaseLights')


local LobbyManager = class('LobbyManager', GameManager)

local gold_coin_glyph = to_utf8(get_glyph('blackwall', 159))

function LobbyManager:initialize()
    GameManager:initialize()
    
    gulp._delegate_list:add_delegate(self)
    
    --self.map_name = map_name
    
end



function LobbyManager:finish()
    gulp._delegate_list:drop_delegate(self)
end

   

function LobbyManager:draw()
    if self.need_to_draw then
    
        clear_status_line(0)
        clear_status_line(1)
        clear_status_line(2)
        
        
        if self.counting_down then
            local digit = tostring(math.ceil(self.countdown_time))
            local char = 's'
            if digit == '1' then char = '!' end
            set_status_line(0, 4, gold_coin_glyph .. " Starting in " .. digit .. " second" .. char .. "! " .. gold_coin_glyph)
            maze_set_string(self.map_name, 7, 16, tostring(math.ceil(self.countdown_time)))
        else    
            set_status_line(0, 4, gold_coin_glyph .. " Let's Play " .. get_map(self.map_name).game_name .. " " .. gold_coin_glyph)
            maze_set_string(self.map_name, 7, 16, ' ')
        end
        
        self.need_to_draw = false
    end

end


function LobbyManager:check_for_player_on_pad(pad)
    
    for id,player in all_players_on_map(self.map_name) do
        
        local line_delta = pad.line - player.line        
        local column_delta = pad.column - player.column 
        local pad_threshold = 0.5
        
        if  line_delta > -pad_threshold and line_delta < pad_threshold and 
            column_delta > -pad_threshold and column_delta < pad_threshold then
            -- a player is on the pad!
            return id
        end
    end
    
    return nil
end

            
        
function LobbyManager:switch_start_light(group, column, status)
    
    
    -- check for map is nil required in case any extra messages arrive 
    -- from the server after we have switched map away from coin_rush_lobby
    if not check_map_exists(self.map_name) then return end
    
    local map = get_map(self.map_name)
    
    local lights = map.auto.start_lights[group]
    for _, light in ipairs(lights) do
        if light.column == column then
            -- switch the local light - if the status has changed
            if light.status ~= status then
                light.status = status
                if gulp.LuaStateCodeMode == 'CLIENT' then
                    -- switch light glyphs
                    light.glyph:update_glyph(light.glyphs[status])
                else
                    -- server mode - switch the client lights
                    local message = {
                        mt = 'game_control',
                        manager = self.my_name,
                        command = 'start_light',
                        group = light.group,
                        column = light.column,
                        status = light.status,
                        map_name = self.map_name,
                    }
                    gulp_networking:broadcast_to_all_on_map(self.map_name, message)
                end
            end
        end
    end
end


function LobbyManager:monitor_pad_group(group)
    
    local players = {}
    local map = get_map(self.map_name)
    local game_start = true
    
    for _, pad in ipairs(map.auto.exit_pads[group]) do
        -- if there is no player stood on this pad, exit this loop
        local player = self:check_for_player_on_pad(pad)
        if player == nil then
            self:switch_start_light(group, pad.column, 'off')
            game_start = false
        else
            self:switch_start_light(group, pad.column, 'on')
            table.insert(players, player)
        end
    end
    
    if game_start == false then return nil end
    return players
end

function LobbyManager:exit_level(id, exit_to_take)
    
    local player = get_player_by_id(id)
    if player then
        player:take_exit(exit_to_take)
    end

end

function LobbyManager:send_start_game(id, players_in)
    
    if id == 'server' or id == 'solo' then
        -- server player -nothing to do
        return
    end
    
    local message =  {
        mt = 'game_control',
        command = 'start_game',
        manager = self.my_name,
        players = players_in,
    }
    
    local rp = get_player_by_id(id)
    rp.socket_ref:SendMessage(message)

end

local multiplayer_game_index = 0

function LobbyManager:server_start_game(game_group)

    local map = get_map(self.map_name)
    local pads = map.auto.exit_pads[self.group_size]
    local game_map = pads[1].file                       -- query the first pad of the group
    
    -- mangle the game_map name for a multiplayer instance
    multiplayer_game_index = multiplayer_game_index + 1
    game_map = game_map.."_^_"..multiplayer_game_index
    
    -- actually load the level
    gulp.map_loader:LoadMap(game_map)
    local map_message = gulp.map_loader:form_map_message(game_map)
    
    local exit_to_take = { file=game_map }
    
    -- record the player group for this particular 
    
    release_all_mobs(game_map)
    
    for key,player_id in ipairs(game_group) do
        local player = get_player_by_id(player_id)
        
        -- record game map and game group for each player at server
        player.game_map = game_map
        player.game_group = game_group
        
        -- send the game map to the client
        local socket_ref = player.socket_ref
        socket_ref:SendMessage(map_message) 
        gulp_networking:SendMobsAndPickables(socket_ref, game_map)
        player:register_map(game_map)
        
        -- let the clients know the game is starting, and who is playing
        -- send before we take the exit!
        self:send_start_game(player_id, game_group)
        
        -- take the exit for each player
        exit_to_take.destination = tostring(key)
        self:exit_level(player_id, exit_to_take)
    end
    
end

function LobbyManager:client_start_game(players)
    self:abort_countdown()
    gulp.localplayer.game_group = players
end



function LobbyManager:switch_chase_light(status)
    
    -- check in case called in response to an old message after the map is dropped, or similar
    if not check_map_exists(self.map_name) then return end
    
    local map = get_map(self.map_name)
    
    -- happens in server at game end, not sure why current_chase_light is nil, requires investigation
    if self.current_chase_light == nil then return end
    
    -- start lights can be nil during the garbage collection for coin_rush_lobby straight
    -- after a client enters coin_rush. This check is a bit of a hack...
    if map.auto.start_lights ~= nil then
        local lights = map.auto.start_lights[5]     -- group 5 is the chase lights
        local light = lights[self.current_chase_light]
        light.status = status
        if gulp.LuaStateCodeMode == 'CLIENT' then
            light.glyph:update_glyph(light.glyphs[status])
        end
    end
    
end

function LobbyManager:make_start_countdown_message()
    return  {
                mt = 'game_control',
                command = 'start_countdown',
                time = self.countdown_time,
                manager = self.my_name,
            }
end


function LobbyManager:start_countdown(t)
    self.counting_down = true
    self.countdown_time = t     -- seconds
    self:cause_drawing()
    self.current_chase_light = 2
    self:switch_chase_light('on')
    self.chase_light_accumulator = 0
    self.next_chase_light = { 2, 3, 5, 1, 8, 4, 6, 7 }
    
    
    
    if gulp.LuaStateCodeMode == 'SERVER' then
        local message = self:make_start_countdown_message()
        gulp_networking:broadcast_to_all_on_map(self.map_name, message)
    else
        -- sounds on client side only
        gulp.sound.sonar('reset')
    end
    
end

function LobbyManager:abort_countdown()
    self.counting_down = false
    self:cause_drawing()
    self:switch_chase_light('off')
    
    if gulp.LuaStateCodeMode == 'SERVER' then
        local message = {
            mt = 'game_control',
            command = 'abort_countdown',
            manager = self.my_name,
        }
        gulp_networking:broadcast_to_all_on_map(self.map_name, message)
    end
end

function LobbyManager:monitor_countdown(dt)
    
    if self.counting_down == true then
        local old_ceil_time = math.ceil(self.countdown_time)
        self.countdown_time = self.countdown_time - dt
        if self.countdown_time < 0 then
            -- turn off countdown just in case the server isn't actually going to do a level switch here
            self.counting_down = false
            self:cause_drawing()
            self:switch_chase_light('off')
            if gulp.LuaStateCodeMode == 'SERVER' then
                -- clients start in response to a message from server
                self:server_start_game(self.players_counting_down)
            end
        elseif math.ceil(self.countdown_time) < old_ceil_time then
            self:cause_drawing()
        end
        
        -- do the chase lights
        self.chase_light_accumulator = self.chase_light_accumulator + dt 
        
        while self.chase_light_accumulator > 0.125 do
            self.chase_light_accumulator = self.chase_light_accumulator - 0.125
        
            self:switch_chase_light('off')
            self.current_chase_light = self.next_chase_light[self.current_chase_light]
            if self.current_chase_light > 8 then self.current_chase_light = 1 end
            self:switch_chase_light('on')
            
            if gulp.LuaStateCodeMode == 'CLIENT' then
                gulp.sound.sonar()
            end
            
        end
        
    end
end


function LobbyManager:monitor_pads()
    
    -- check whether players are stood on the start game pads
    
    
    -- first check the coin_rush_lobby map exists
    if not check_map_exists(self.map_name) then return end

    if self.players_counting_down == nil then
        -- countdown not running
        -- loop through all groups, checking for game start
        for group_size = 1,4 do
            local players = self:monitor_pad_group(group_size)
            if players ~= nil then 
                -- start a multiplayer player game
                self.players_counting_down = players
                self.group_size = group_size
                --self:reset_coin_counts(players)
                self:start_countdown(5)
            end
        end
    else
        -- countdown running
        -- monitor this group of pads, checking for abort countdown
        local players = self:monitor_pad_group(self.group_size)
        if players == nil then 
            -- reset everything
            self.players_counting_down = nil
            self.group_size = nil
            self:abort_countdown()
        end
    end
end  
    
    

function LobbyManager:update(dt)
    
    if gulp.LuaStateCodeMode == 'SERVER' then
        self:monitor_pads()
    end
    
    self:monitor_countdown(dt)
    
    
end

function LobbyManager:send_game_manager_status(socket_ref)
    
    if self.counting_down then
        local message = self:make_start_countdown_message(self.countdown_time)
        socket_ref:SendMessage(message)
    end
    
end

function LobbyManager:process_game_control_message(message)
    
    if message.manager == self.my_name then
        if message.command == 'start_light' then        
            self:switch_start_light(message.group, message.column, message.status)
        elseif message.command == 'start_countdown' then
            self:start_countdown(message.time)
        elseif message.command == 'abort_countdown' then
            self:abort_countdown()
        elseif message.command == 'start_game' then
            self:client_start_game(message.players)
        end
    end
end             


return LobbyManager


