-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

--[[
    
    This is the game manager for the Maze Race map
    
    This manager runs the Maze Race game 
    
--]]



require('strict')
require('utils')
require('status_manager')
local class = require('middleclass')
local GameManager = require('map_controllers/game_manager')
local CollectableCounter = require('map_controllers/collectable_counter')


local MazeTagManager = class('MazeTagManager', GameManager)

local gold_coin_glyph = to_utf8(get_glyph('blackwall', 159))

local win_score = 5


function MazeTagManager:initialize()
    GameManager:initialize()
    
    gulp._delegate_list:add_delegate(self)
    
    self.score_counter = CollectableCounter("_maze_tag_score")
    
    self.game_over = false
    
    self:reset_score()
    
end



function MazeTagManager:finish()
    gulp._delegate_list:drop_delegate(self)
end

function MazeTagManager:get_score_table()

    local score = {0,0,0,0}
    local highest = 0
    
    local game_group = nil
    
    if gulp.LuaStateCodeMode == 'SERVER' then
        -- find a player playing this game
        for _,player in all_players_on_map(self.map_name) do
            game_group = player.game_group
            break
        end
    else
        game_group = gulp.localplayer.game_group
    end
    
    if game_group ~= nil then
        for position_in_group, id in pairs(game_group) do
            score[position_in_group] = self.score_counter:get_count_by_id(id)
            if score[position_in_group] > highest then highest = score[position_in_group] end
        end
    end
    
    return score, highest
end

    

function MazeTagManager:draw()
    if self.need_to_draw then
    
        clear_status_line(0)
        clear_status_line(1)
        clear_status_line(2)
        
        local score = self:get_score_table()
        
        set_status_line(1, 0, score[1] .. gold_coin_glyph)
        set_status_line(2, 0, score[4] .. gold_coin_glyph)
        set_status_line(1, 31-string.len(score[2]), gold_coin_glyph .. score[2])
        set_status_line(2, 31-string.len(score[3]), gold_coin_glyph .. score[3])
        
        if self.game_over then
            set_status_line(0, 9.5, gold_coin_glyph .. ' GAME OVER ' .. gold_coin_glyph)
        else 
            set_status_line(0, 10, gold_coin_glyph .. gulp.maps[self.map_name].title .. gold_coin_glyph)
        end
        
        self.need_to_draw = false
    end

end


function MazeTagManager:show_game_over_menu()
    
    local winning_amount = 0
    local score = self:get_score_table()
    
    for _,value in ipairs(score) do
        if value > winning_amount then
            winning_amount = value
        end
    end

   
    local function switch_to_lobby()

        local line_modifier = 0
        for player = 1, 4 do
            if gulp.localplayer.game_group[player] == gulp.localplayer.id then
                line_modifier = player
            end
        end


        local exit_to_take = { file='maze_tag_lobby', 
                    destination_line = 10 + line_modifier,
                    destination_column = 16,
            }
        
        gulp.localplayer:take_exit(exit_to_take)
        
    end
    

    -- show the end of game menu
    local menu = Menu{
        
        line = 4,
        column = 1,
        
        background = gulp.colour.BLUE.simple_colour, 
        foreground = gulp.colour.BRIGHT_YELLOW.simple_colour,
        border_style = 2,
        {        "\x1b[32m     We have a winner!!     " },
        { "" },
        { "" },
        { "\x1b[33m\x1b[1m       EXIT TO LOBBY", switch_to_lobby },
    }
        
    for key,value in ipairs(score) do
        if value == winning_amount then
            -- this is (one of) the player(s) who got the winning amount
            local winner = {         "\x1b[32m        " .. gold_coin_glyph .. " Player " .. key .. " " .. gold_coin_glyph}
            menu:insert_item(3, winner)
        end
    end
            
end
            

function MazeTagManager:reset_score()
    
    for id,player in all_players() do
        
        self.score_counter:set_count_by_id(id, 0)
    end
    
    if gulp.LuaStateCodeMode == 'SERVER' then 
        -- update the clients with the new coin counts
        local message = self:make_status_message()
        gulp_networking:broadcast_to_all_on_game_controller(self.my_name, message)
    end
    
    
end


function MazeTagManager:update(dt)
    
    if self.game_over == false then
    
        local score, highest = self:get_score_table()
        
        if highest >= win_score then
            --  game is over
            self.game_over = true
            
            --hold_all_mobs('coin_rush')
            
            if gulp.LuaStateCodeMode == 'CLIENT' then
                gulp.localplayer:hold()
                self:show_game_over_menu()
            end
            
        end
    end
    
end

function MazeTagManager:make_status_message()
    local message = {
        mt = 'game_control',
        command = 'status',
        manager = self.my_name,
        map_name = self.map_name,
        counts = {},
    }
    
    for id,player in all_players() do
        message.counts[id] = self.score_counter:get_count(player)
    end
    
    return message
end

function MazeTagManager:send_game_manager_status(socket_ref)
    local message = self:make_status_message()
    socket_ref:SendMessage(message)
end

function MazeTagManager:notify(player_id_in, action, map_name_in)
    
    if gulp.LuaStateCodeMode == 'CLIENT' then 
        local message = {
            mt = 'game_control',
            manager = self.my_name,
            command = 'notify',
            act = action,
            player_id = player_id_in,
            map_name = self.map_name,
        }
        
        print(action)
        
        gulp_networking:SendToServer(message) 
        
        return 
    end
    
    -- BELOW, WE ONLY RUN ON THE SERVER
    local player_id = player_id_in
    
    if action == 'reached_exit' then
        
        --print("Player ",player_id, " reached exit on map", map_name_in)
        
        
        --local next_map = { 
        --    ['maze_tag_2_player']='maze_tag_2_player_b',
        --    ['maze_tag_2_player_b']='maze_tag_2_player_c',
        --    ['maze_tag_2_player_c']='maze_tag_2_player',
        --}
        
        local entrances = {'1','2','3','4',}
        
        local map = get_map(self.map_name)
        self.map_name = map.next_map..self.instance_extension
        
        self.score_counter:change_count_by_id(player_id, 1)
        --self.map_name = next_map[self.map_name]
        local message = self:make_status_message()
        gulp_networking:broadcast_to_all_on_game_controller(self.my_name, message)
        
        local player_at_exit = get_player_by_id(player_id)
        
        for player_number = 1, 4 do
            
            if player_at_exit.game_group[player_number] ~= nil  then
                local player = get_player_by_id(player_at_exit.game_group[player_number])
                local exit_to_take = { file=self.map_name, destination = entrances[player_number], }
                player:take_exit(exit_to_take)
            end
        
        end
    end
    
    self:cause_drawing()

end


function MazeTagManager:process_game_control_message(message)
    
    if message.manager == self.my_name then
        if message.command == 'status' then
            self.map_name = message.map_name
            for id,count in pairs(message.counts) do
                self.score_counter:set_count_by_id(id, count)
            end
            self:cause_drawing()
        elseif message.command == 'notify' then
            if message.act == 'reached_exit' and message.map_name == self.map_name then
                self:notify(message.player_id, message.act, message.map_name)
            end
        end
    end
end             


return MazeTagManager


