-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

--[[
    
    This is the game manager for the Maze Race map
    
    This manager runs the Maze Race game 
    
--]]



require('strict')
require('utils')
require('status_manager')
local class = require('middleclass')
local GameManager = require('map_controllers/game_manager')


local SlimeEscapeManager = class('SlimeEscapeManager', GameManager)

-- definitions for slime within the map data
local slime = get_glyph('blackwall', 0x75)  -- needs to match glyph in map file
local slimy_wall = 1000000                  -- can be anything that we wouldn't normally see in the map


local function initialise_slime_front(map_name)
    
    local map = get_map(map_name)
    local maze = map.auto.maze
    local slime_map = map.slime_map
    
    local height = maze:height()
    local width = maze:width()
    
    if map.slime_front_table == nil then map.slime_front_table = {} end
    
    for c = 1, width do
        for l = 1, height do
            if map[l][c] == slime then
                table.insert(map.slime_front_table, {l,c})
            end
        end
    end
end

local function initialise_slime_map(map_name)
    local map = get_map(map_name)
    local maze = map.auto.maze
    
    map.slime_map = {}
    
    for k,v in ipairs(map) do
        map.slime_map[k] = {}
    end
end



function SlimeEscapeManager:initialize()
    GameManager.initialize(self)
    
    gulp._delegate_list:add_delegate(self)
    
    self.accumulator = 0
    
end

function SlimeEscapeManager:post_map_initialisation()
    -- called after construction of the game manager, once the map name has been set
    GameManager.post_map_initialisation(self)
    
    if gulp.LuaStateCodeMode == 'SERVER' then    
        
        initialise_slime_map(self.map_name)
        initialise_slime_front(self.map_name)
    
    end
end


function SlimeEscapeManager:finish()
    gulp._delegate_list:drop_delegate(self)
end

--[[
function SlimeEscapeManager:draw()
    

end
--]]



function SlimeEscapeManager:update(dt)
    
    if gulp.LuaStateCodeMode == 'SERVER' then    
    
        self.accumulator = self.accumulator + dt
        
        local slime_step = 0.5
        
        if self.accumulator > slime_step then
            self.accumulator = self.accumulator - slime_step
            
            self:grow_slime()
        end
    
    end
end


local north_bit = 0x8
local east_bit = 0x4
local south_bit = 0x2
local west_bit = 0x1

local ne_bit = 0x8
local nw_bit = 0x4
local se_bit = 0x2
local sw_bit = 0x1

local slime_glyph_lookup_table =
{   --   N,E,S,W              glyph  NE,NW,SE,SW              glyph
    --[[ 0,0,0,0 --]] [0x0] = get_glyph('blackwall',0x75),
    --[[ 0,0,0,1 --]] [0x1] = get_glyph('blackwall',0x67),
    --[[ 0,0,1,0 --]] [0x2] = get_glyph('blackwall',0x56),
    --[[ 0,0,1,1 --]] [0x3] = {                    ['mask'] = 0x1,
                                  --[[ 0,0,0,0 --]]   [0x0] = get_glyph('blackwall',0x54),
                                  --[[ 0,0,0,1 --]]   [0x1] = get_glyph('blackwall',0x52),     },
    --[[ 0,1,0,0 --]] [0x4] = get_glyph('blackwall',0x65),
    --[[ 0,1,0,1 --]] [0x5] = get_glyph('blackwall',0x74),
    --[[ 0,1,1,0 --]] [0x6] = {                    ['mask'] = 0x2,
                                  --[[ 0,0,0,0 --]]   [0x0] = get_glyph('blackwall',0x53),
                                  --[[ 0,0,1,0 --]]   [0x2] = get_glyph('blackwall',0x50),     },
    --[[ 0,1,1,1 --]] [0x7] = {                    ['mask'] = 0x3,
                                  --[[ 0,0,0,0 --]]   [0x0] = get_glyph('blackwall',0x45),
                                  --[[ 0,0,0,1 --]]   [0x1] = get_glyph('blackwall',0x49),     
                                  --[[ 0,0,1,0 --]]   [0x2] = get_glyph('blackwall',0x68),   
                                  --[[ 0,0,1,1 --]]   [0x3] = get_glyph('blackwall',0x51),     },                            
    --[[ 1,0,0,0 --]] [0x8] = get_glyph('blackwall',0x76),
    --[[ 1,0,0,1 --]] [0x9] = {                    ['mask'] = 0x4,
                                  --[[ 0,0,0,0 --]]   [0x0] = get_glyph('blackwall',0x64),
                                  --[[ 0,1,0,0 --]]   [0x4] = get_glyph('blackwall',0x72),     },
    --[[ 1,0,1,0 --]] [0xa] = get_glyph('blackwall',0x73),
    --[[ 1,0,1,1 --]] [0xb] = {                    ['mask'] = 0x5,
                                  --[[ 0,0,0,0 --]]   [0x0] = get_glyph('blackwall',0x46),
                                  --[[ 0,0,0,1 --]]   [0x1] = get_glyph('blackwall',0x69),     
                                  --[[ 0,1,0,0 --]]   [0x4] = get_glyph('blackwall',0x59),   
                                  --[[ 0,1,0,1 --]]   [0x5] = get_glyph('blackwall',0x62),     },  
    --[[ 1,1,0,0 --]] [0xc] = {                    ['mask'] = 0x8,
                                  --[[ 0,0,0,0 --]]   [0x0] = get_glyph('blackwall',0x63),
                                  --[[ 1,0,0,0 --]]   [0x8] = get_glyph('blackwall',0x70),     },
    --[[ 1,1,0,1 --]] [0xd] = {                    ['mask'] = 0xc,
                                  --[[ 0,0,0,0 --]]   [0x0] = get_glyph('blackwall',0x55),
                                  --[[ 0,1,0,0 --]]   [0x4] = get_glyph('blackwall',0x79),     
                                  --[[ 1,0,0,0 --]]   [0x8] = get_glyph('blackwall',0x58),   
                                  --[[ 1,1,0,0 --]]   [0xc] = get_glyph('blackwall',0x71),     },
    --[[ 1,1,1,0 --]] [0xe] = {                    ['mask'] = 0xa,
                                  --[[ 0,0,0,0 --]]   [0x0] = get_glyph('blackwall',0x44),
                                  --[[ 0,0,1,0 --]]   [0x2] = get_glyph('blackwall',0x48),     
                                  --[[ 1,0,0,0 --]]   [0x8] = get_glyph('blackwall',0x78),   
                                  --[[ 1,0,1,0 --]]   [0xa] = get_glyph('blackwall',0x60),     }, 
    --[[ 1,1,1,1 --]] [0xf] = {                    ['mask'] = 0xf,
                                  --[[ 0,0,0,0 --]]   [0x0] = get_glyph('blackwall',0x66),
                                  --[[ 0,0,0,1 --]]   [0x1] = get_glyph('blackwall',0x6d),     
                                  --[[ 0,0,1,0 --]]   [0x2] = get_glyph('blackwall',0x6a),   
                                  --[[ 0,0,1,1 --]]   [0x3] = get_glyph('blackwall',0x4a), 
                                  --[[ 0,1,0,0 --]]   [0x4] = get_glyph('blackwall',0x6c),
                                  --[[ 0,1,0,1 --]]   [0x5] = get_glyph('blackwall',0x4d),     
                                  --[[ 0,1,1,0 --]]   [0x6] = get_glyph('blackwall',0x57),   
                                  --[[ 0,1,1,1 --]]   [0x7] = get_glyph('blackwall',0x43), 
                                  --[[ 1,0,0,0 --]]   [0x8] = get_glyph('blackwall',0x6b),
                                  --[[ 1,0,0,1 --]]   [0x9] = get_glyph('blackwall',0x47),     
                                  --[[ 1,0,1,0 --]]   [0xa] = get_glyph('blackwall',0x4b),   
                                  --[[ 1,0,1,1 --]]   [0xb] = get_glyph('blackwall',0x40), 
                                  --[[ 1,1,0,0 --]]   [0xc] = get_glyph('blackwall',0x4c),
                                  --[[ 1,1,0,1 --]]   [0xd] = get_glyph('blackwall',0x42),     
                                  --[[ 1,1,1,0 --]]   [0xe] = get_glyph('blackwall',0x41),   
                                  --[[ 1,1,1,1 --]]   [0xf] = get_glyph('blackwall',0x61),    },
}





function SlimeEscapeManager:grow_slime()

    local map = get_map(self.map_name)
    local maze = map.auto.maze
    
    local height = maze:height()
    local width = maze:width()
    
    local new_slime_table = {}
    local function new_slime(l,c)
        if new_slime_table[l] == nil then
            new_slime_table[l] = {}
        end
        new_slime_table[l][c] = true
    end
    
    local slime_front_table = map.slime_front_table
    local slime_map = map.slime_map
    
    --print_table(slime_front_table)
    
    -- for each location in the slime front...
    for element,location in pairs(slime_front_table) do
        
        local function check_new_slime(l,c)
            if (slime_map[l][c] ~= slime and slime_map[l][c] ~= slimy_wall) then
                new_slime(l,c)
            end
        end        
        
        local l, c = location[1], location[2]
        
        if c~=1 then check_new_slime(l, c-1) end
        if c~=width then check_new_slime(l, c+1) end
        if l~=1  then check_new_slime(l-1, c) end
        if l~=height  then check_new_slime(l+1, c) end
    
    end
    
    local function slime_can_grow(l, c)
        -- whether slime can grow depends on the wall states in the real map (not slime map)
        return not map.auto.wall_detect_type[map[l][c]]
    end
    
    for line, line_table in pairs(new_slime_table) do
        for column, _ in pairs(line_table) do
        
            -- update the lua map representation and slime front
            -- only put slime that filled space into slime front, not slimy walls
            if slime_can_grow(line,column) then
                table.insert(slime_front_table, {line,column})
            end
            
            if map[line][column] == 32 then
                slime_map[line][column] = slime
            else
                slime_map[line][column] = slimy_wall      -- prints new slime glyph underneath 
            end
            
            local function slimy_north(l,c) return l==1      or slime_map[l-1][c] == slime or slime_map[l-1][c] == slimy_wall end
            local function slimy_south(l,c) return l==height or slime_map[l+1][c] == slime or slime_map[l+1][c] == slimy_wall end
            local function slimy_east(l,c)  return c==width  or slime_map[l][c+1] == slime or slime_map[l][c+1] == slimy_wall end
            local function slimy_west(l,c)  return c==1      or slime_map[l][c-1] == slime or slime_map[l][c-1] == slimy_wall end
            
            local function slimy_nw(l,c) return l==1      or c==1     or slime_map[l-1][c-1] == slime or slime_map[l-1][c-1] == slimy_wall end
            local function slimy_se(l,c) return l==height or c==width or slime_map[l+1][c+1] == slime or slime_map[l+1][c+1] == slimy_wall end
            local function slimy_ne(l,c) return l==1      or c==width or slime_map[l-1][c+1] == slime or slime_map[l-1][c+1] == slimy_wall end
            local function slimy_sw(l,c) return l==height or c==1     or slime_map[l+1][c-1] == slime or slime_map[l+1][c-1] == slimy_wall end
            
            
            -- calculate glyph for new slime location in c++ presentation map, and all surrounding it
            for l = line-1, line+1 do if l >= 1 and l <= height then        -- loop with sanitised, calculated control variable
                for c = column-1, column+1 do if c >= 1 and c <= width then
                    if slime_map[l][c]==slime or slime_map[l][c]==slimy_wall then
                    
                        local map_bits = 0
                        local glyph
                        
                        if slimy_north(l,c) then map_bits = bit32.bor(map_bits, north_bit) end
                        if slimy_south(l,c) then map_bits = bit32.bor(map_bits, south_bit) end
                        if slimy_east(l,c)  then map_bits = bit32.bor(map_bits, east_bit) end
                        if slimy_west(l,c)  then map_bits = bit32.bor(map_bits, west_bit) end
                        
                        local table_entry = slime_glyph_lookup_table[map_bits]
                        
                        if type(table_entry) == 'number' then
                            glyph = table_entry
                        else
                            map_bits = 0
                        
                            if slimy_nw(l,c)then map_bits = bit32.bor(map_bits, nw_bit) end
                            if slimy_se(l,c)then map_bits = bit32.bor(map_bits, se_bit) end
                            if slimy_ne(l,c)then map_bits = bit32.bor(map_bits, ne_bit) end
                            if slimy_sw(l,c)then map_bits = bit32.bor(map_bits, sw_bit) end
                        
                            map_bits = bit32.band(map_bits, table_entry['mask'])
                        
                            glyph = table_entry[map_bits]
                        end
                        
                        if slime_map[l][c]==slime then
                            
                            local message = {
                                mt = 'game_control',
                                manager = self.my_name,
                                command = 'slime_set_glyph',
                                line = l-1,
                                column = c-1,
                                glyph = glyph,
                            }
                            gulp_networking:broadcast_to_all_on_game_controller(self.my_name, message)

                            
                        else
                            
                            local message = {
                                mt = 'game_control',
                                manager = self.my_name,
                                command = 'slimy_wall_set_glyph',
                                line = l-1,
                                column = c-1,
                                glyph = glyph,
                            }
                            gulp_networking:broadcast_to_all_on_game_controller(self.my_name, message)
                            
                        end
                    end
                end end     -- terminate sanitised loop construct
            end end
        end
    end

    -- prune the slime front...
    for element,location in pairs(slime_front_table) do
        
        local l = location[1]
        local c = location[2]
        
        -- if this slime front is now surround, it isn't the front any more
        if  (c==1      or slime_map[l][c-1] == slime or slime_map[l][c-1] == slimy_wall) and 
            (c==width  or slime_map[l][c+1] == slime or slime_map[l][c+1] == slimy_wall) and 
            (l==1      or slime_map[l-1][c] == slime or slime_map[l-1][c] == slimy_wall) and
            (l==height or slime_map[l+1][c] == slime or slime_map[l+1][c] == slimy_wall) then
            table.remove(slime_front_table, element)
        end
    end


end

local slimy_wall_mdles = {}


function SlimeEscapeManager:process_game_control_message(message)
    
    local map = get_map(self.map_name)
    local maze = map.auto.maze
    
    local l = message.line
    local c = message.column
    local glyph = message.glyph

    if message.manager == self.my_name then
        if message.command == 'slime_set_glyph' then        
            maze:set_glyph(l, c, glyph, 0, 1, 1)
        elseif message.command == 'slimy_wall_set_glyph' then        
        
            --slimy wall, create new maze display element or update existing
            if slimy_wall_mdles[l] == nil then
                slimy_wall_mdles[l] = {}
            end
            
            if slimy_wall_mdles[l][c] == nil then
                slimy_wall_mdles[l][c] = new_maze_display_element(maze, l, c, glyph, 0)
            else
                slimy_wall_mdles[l][c]:update_glyph(glyph)
            end
        
        
        end
    end
end             

function SlimeEscapeManager:map_conditional_notify(conditional, value)
    -- checks doors - if you override and want doors to work, remember to call this too
    local map = get_map(self.map_name)
    local maze = map.auto.maze
    local slime_map = map.slime_map

    for _, door in pairs(map.doors) do
        if door.conditional == conditional then
            
            local l, c = door.line+1, door.column+1
            
            set_map_glyph(self.map_name, l, c, door[value])
            
            -- handle slime growth at opened doors
            if value == 'open' and slime_map[l][c] == slimy_wall then
                table.insert(map.slime_front_table, {l, c})
            end
                
        end
    end
 end
          


return SlimeEscapeManager


