-- Forlorn Fox
-- Copyright © 2015 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Handle online tracker
--
-- Created by Kenny Osborne September 2015
-- For Forlorn Fox

-- configuration parameters
local online_server_address = "http://forlorn.wubbleu.co.uk/php"
local registerScript = "/registerServer.php"
--local registerScript = "/registerServerDebug.php" -- Use for debugging the register request
local unRegisterScript = "/unregisterServer.php"
local pollScript = "/serverPoll.php"

-- Poll times for server messages
local broadcast_time = 15 -- 15 seconds
local poll_interval = 15 -- 15 seconds 
local register_retry_time = 60 -- 60 seconds

local msg_contains = "Forlorn Fox Game Server"

-- dependancies
require("strict")
local class = require("middleclass")
local copas = require("copas")
module = false  -- work around to make ltn12.lua work properly under Lua 5.2 and strict.lua
--local synchttp = require("socket.http").request
local asynchttp = require("copas.http").request
local lunajson = require ("thirdparty/json/lunajson")

-- isolate the rest in a local table that's returned
local Internetb = {}
Internetb.server_time = 0

-- Message queue for copas sending 
local sendMessageQueue = {}

-- Error table
local trackerErrorStrings = {}

-- Function to add message to the queue for sending
function Internetb.sendMessage(message)
    table.insert(sendMessageQueue, message)
end

-- Copas thread that will pop the message and send over the network    
function sendInternetMessageThread()
    while true do
        while #sendMessageQueue ~= 0 do
            local message = table.remove(sendMessageQueue, 1) -- remove the oldest item from the queue
            local destination = message.address
            local parmData = message.param
            print("** Send Internet broadbast")
            local res, err = asynchttp(destination, parmData)
            --print (res) -- Debug print of server response
            -- If we got data from the server
            if res then
                local server_result = lunajson.decode(res)
                local request_result = server_result["result"]
                -- if error 
                if request_result == "false" then

                    local err_code = server_result["error_code"]
                    gulp.ui_queue:send( 
                        {   qcommand = "internet_broadcast_error", 
                            error_code = err_code,
                            error_string = Internetb.getErrorTextFromCode(err_code)
                        })
                      
                    -- Set to retry interval
                    broadcast_time = register_retry_time   
                    
                    -- ADDED 21Jan2016 bu Rob
                    -- only try once
                    Internetb._broadcast_enabled = false
                    break 
                
                elseif request_result == "true" then
                    if Internetb.registered_with_tracker == false then   
                      Internetb.registered_with_tracker = true
                                          
                      -- Send message to ui to ensure error prompt removed
                      message.qcommand = 'internet_broadcast_now_registered'
                      gulp.ui_queue:send(message)
                    end
                    
                    -- Set to poll interval
                    broadcast_time = poll_interval 
                end
            end    
        end 
        
        if Internetb._broadcast_enabled == false then
            break
        end     
        copas.sleep(0) -- use 0 just to other copas threads to run  
    end 
    Internetb._message_thread_created = false
end

-- Register with the online tracker and start the copas thread
function Internetb.start_server_game_broadcast(port)    
    Internetb.port = port
    local addressString = string.format("%s%s", online_server_address, registerScript)  
    local paramString = string.format("game_id=%s&game_port=%d&game_name=%s&game_version=%s", gulp.server_identifier, port, gulp.gameName, gulp.forlorn_fox_version)
    
    -- Build error message strings
    Internetb.buildServerErrorStrings()
    
    -- Build the request and then add to queue for sending 
    local message = {address = addressString, param = paramString} 
    Internetb.sendMessage(message)
    
    -- Allow the loop to run
    Internetb._broadcast_enabled = true
    
    -- Not registered yet so clear flag
    Internetb.registered_with_tracker = false
                 
    -- Assume the register will fail and set to retry time             
    broadcast_time = poll_interval

    -- Start thread for sending messages
    if not Internetb._message_thread_created then
        local mythread = copas.addthread(sendInternetMessageThread)
        name_network_item(mythread, "addthread - sendInternetMessageThread")

        Internetb._message_thread_created = true
    end
end

-- Deregister from the online tracker
function Internetb.stop_server_game_broadcast()
    -- if the server polling is running then turn it off
    if Internetb._broadcast_enabled == true then
        local addressString = string.format("%s%s", online_server_address, unRegisterScript)
        local paramString = string.format("game_id=%s", gulp.server_identifier)
        
        -- Build the request and then add to queue for sending 
        local message = {address = addressString, param = paramString} 
        Internetb.sendMessage(message)
    
        Internetb._broadcast_enabled = false
    end
end

-- Main loop for timing when to send updates to the tracker
function Internetb.broadcast_server_update(dt)
    if Internetb._broadcast_enabled == true then
        -- we update the server every 15 seconds
        Internetb.server_time = Internetb.server_time + dt
       
        -- Time for the poll
        if Internetb.server_time >= broadcast_time then
          
          local addressString = string.format("%s%s", online_server_address, pollScript)
          local paramString = string.format("game_id=%s", gulp.server_identifier)
        
          -- If not registered then change the message to the register message
          if Internetb.registered_with_tracker == false then 
            addressString = string.format("%s%s", online_server_address, registerScript)  
            paramString = string.format("game_id=%s&game_port=%d&game_name=%s&game_version=%s", gulp.server_identifier, Internetb.port, gulp.gameName, gulp.forlorn_fox_version)
          end   
            
          -- Build the request and then add to queue for sending 
          local message = {address = addressString, param = paramString} 
          Internetb.sendMessage(message)
            
          -- reset the time for the next poll 
          Internetb.server_time = 0
        end
    end
end


function Internetb.getErrorTextFromCode(error_code)
    if error_code <= #trackerErrorStrings and error_code >= 1 then
        return trackerErrorStrings[error_code]
    else
        return string.format("Unknown Error Code: 0x%x", error_code)
    end
end

-- Build error table
function Internetb.buildServerErrorStrings()
    trackerErrorStrings[1] = "Error removing game from server"
    trackerErrorStrings[2] = "Error updating poll"
    trackerErrorStrings[3] = "Error with database"
    trackerErrorStrings[4] = "Registering game failed"
    trackerErrorStrings[5] =  string.format("Check port %d is open", Internetb.port)
    trackerErrorStrings[6] = "Failed to get IP address"
    trackerErrorStrings[7] = "Invalid post data"
    trackerErrorStrings[8] = "Not registered with server"
end

-- Error codes
-- echo ('{ "result" : "false","error_text" : "error_removing_server","error_code" : "0x1" }'); 
-- echo ('{ "result" : "false", "error_text" : "error_updating_polltime", "error_code" : "0x2" }'); 
-- echo ('{ "result" : "false", "error_text" : "error_database_not_connected", "error_code" : "0x3" }'); 
-- echo ('{ "result" : "false","error_text" : "register_server_failed", "error_code" : "0x4" }'); 
-- echo ('{ "result" : "false","error_text" : "port_not_open", "error_code" : "0x5" }'); 
-- echo ('{ "result" : "false","error_text" : "failed_getting_ipaddress", "error_code" : "0x6" }'); 
-- echo ('{ "result" : "false","error_text" : "invalid_post_data", "error_code" : "0x7" }'); 
-- echo ('{ "result" : "false", "error_text" : "error_game_not_registered_with_server", "error_code" : 8 }');

return Internetb