-- Forlorn Fox
-- Copyright © 2015 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Keep Alive
-- Network keep alive 

require("networking/network_serializer")
require('strict')
class = require('middleclass')

--local gulpNetworkingDebuggingEnabled = 1

local keepAlive = class('keepAlive')

local PING_INTERVAL_TIME = 10 -- 10 seconds
local PING_TIMEOUT = 60 -- 60 seconds as checks every -- Timeout value for ping response from client 
    
local showingLostServerMenu = false
local showingNetworkError = false

function keepAlive:initialize(network)
    self.ping_time = 0
    self.ServerListOfClientPingResponseTimeouts = {}    -- [socket] = timeremaining List of clients the server is looking to respond
    self.ClientPingResponseTimeout = PING_TIMEOUT       -- Timeout used by the client to detect server has stopped responding
    self.CheckServerConnection = false
    self.network = network
    
    showingLostServerMenu = false;  -- Reset flag to allow the missing server menu to be shown again
    showingNetworkError = false; -- Reset network error flag for menu
end

-- Periodic task 
function keepAlive:tick(dt)
    -- we broadcast an update to the network every PING_INTERVAL_TIME seconds
    self.ping_time = self.ping_time + dt
    
    if self.ping_time >= PING_INTERVAL_TIME then
        self.ping_time = 0
            
        if gulp.LuaStateCodeMode == "CLIENT" then
            if self.CheckServerConnection == true then
                self:checkForMissingServer(dt)
            end    
        else
            self:sendServerPing(dt)
            self:checkForMissingClients(dt)
        end    
    end   
    
    -- Check for network errors
    if gulp.LuaStateCodeMode == "CLIENT" then
        self:checkForNetworkErrors()
    end
end

-- Check for network error being flagged
function keepAlive:checkForNetworkErrors()
    if gulp.isNetworkError() then
        -- If not already showing the menu then show it
        if showingNetworkError == false then
            
            print("KeepAlive - Network error")
    
            self._network_error_menu = Menu{
                background = gulp.colour.BRIGHT_RED.simple_colour, 
                foreground = gulp.colour.BRIGHT_WHITE.simple_colour,
                border_style = 2,
                { "Network Error!" },
                { " QUIT to Menu ", function() gulp.save_and_quit_game() showingNetworkError = false end },
            }
            
            -- Stop us showing the network error again
            showingNetworkError = true
        end    
    end    
end

-- Stop looking for the server connection dropping 
function keepAlive:stopCheckingForMissingServer()
    self.CheckServerConnection = false
     
    if gulpNetworkingDebuggingEnabled then          
        print ("KeepAlive - stopCheckingForMissingServer")
    end    
end

-- Start looking for the server connection going missing
function keepAlive:startCheckingForMissingServer()
    self.CheckServerConnection = true
     
    if gulpNetworkingDebuggingEnabled then          
        print ("KeepAlive - startCheckingForMissingServer")
    end    
end


function keepAlive:checkForMissingServer(dt)
    if self.ClientPingResponseTimeout <= 0 then 
        
        -- If not already showing the menu then show it
        if showingLostServerMenu == false then
            
        print("KeepAlive - Server missing")
    
            self._escape_menu = Menu{
                background = gulp.colour.BRIGHT_RED.simple_colour, 
                foreground = gulp.colour.BRIGHT_WHITE.simple_colour,
                border_style = 2,
                { "Server not responding!" },
                { "  QUIT ", function() gulp.save_and_quit_game() showingLostServerMenu = false end},
            }
            
            -- Stop us showing the missing sever menu again
            showingLostServerMenu = true
        end    
    else
        -- Decrement the remaining time by the poll interval  
        self.ClientPingResponseTimeout = (self.ClientPingResponseTimeout-PING_INTERVAL_TIME)
        
        if gulpNetworkingDebuggingEnabled then    
            print("KeepAlive - Server has Remaining time to respond: ", self.ClientPingResponseTimeout)
        end
    end   
end    
 
 -- Function to extend waiting on server to respond from the missing server menu
function keepAlive:waitOnServerToRespond()
    self.ClientPingResponseTimeout = PING_TIMEOUT

    -- Reset the flag so we show the missing server menu again
    showingLostServerMenu = false
    
    if gulpNetworkingDebuggingEnabled then          
        print ("KeepAlive - waitOnServerToRespond")
    end    
end 

-- Function resets client timeout as it has responded to the server  using the socket as id 
function keepAlive:resetServerPingTimeout()
    self.ClientPingResponseTimeout = PING_TIMEOUT

    showingLostServerMenu = false
    
    if gulpNetworkingDebuggingEnabled then          
        print ("KeepAlive - resetServerPingTimeout")
    end    
end 

function keepAlive:sendServerPing(dt)
    if gulpNetworkingDebuggingEnabled then          
        print ("KeepAlive - SERVER - keep alive tick")
    end    
    local message = { 
        mt = "server_ping",
        dt = dt
    }
    
    self.network:broadcast(message)
end    
    
 -- Function checks all registered client for ping response timeout   
function keepAlive:checkForMissingClients(dt)
    for socket_ref, timeoutRemaining in pairs(self.ServerListOfClientPingResponseTimeouts) do   
        if timeoutRemaining <= 0 then 
            print("KeepAlive - Socket: ", socket_ref, "Client missing")
                           
            -- Remove the remote player as there connection has failed
            HandleClientUnregisterMessage(socket_ref, "dummy as not used")
        else
            -- Decrement the remaining time by the poll interval  
            self.ServerListOfClientPingResponseTimeouts[socket_ref] = (timeoutRemaining-PING_INTERVAL_TIME)
            
            if gulpNetworkingDebuggingEnabled then     
                print("KeepAlive - Socket: ", socket_ref, " Remaining time: ", timeoutRemaining)
            end
        end     
    end
end    

 -- Function add the client using it's socket as id to the list we expect to ack the ping   
function keepAlive:addClientForServerToCheck(socket_ref)
    self.ServerListOfClientPingResponseTimeouts[socket_ref] = PING_TIMEOUT
    
    if gulpNetworkingDebuggingEnabled then          
        print ("KeepAlive - addClientToCheck: ", socket_ref)
    end    
end 

 -- Function removes the client using it's socket as id to the list we expect to ack the ping   
function keepAlive:removeClientForServerToCheck(socket_ref)
    self.ServerListOfClientPingResponseTimeouts[socket_ref] = nil
    
    if gulpNetworkingDebuggingEnabled then          
        print ("KeepAlive - removeClientToCheck: ", socket_ref)
    end   
end 

 -- Function resets client timeout as it has responded to the server  using the socket as id 
function keepAlive:resetClientPingTimeout(socket_ref)
    self.ServerListOfClientPingResponseTimeouts[socket_ref] = PING_TIMEOUT
    
    if gulpNetworkingDebuggingEnabled then          
        print ("KeepAlive - resetPingTimeout: ", socket_ref)
    end    
end 

return keepAlive
