-- Find local clients on the network
-- 
-- Ideas from Mark Steelman here:
-- http://coronalabs.com/blog/2014/09/23/tutorial-local-multiplayer-with-udptcp/
-- http://www.iana.org/assignments/multicast-addresses/multicast-addresses.xhtml
-- http://stackoverflow.com/questions/236231/how-do-i-choose-a-multicast-address-for-my-applications-use

-- configuration parameters
local multicast_address = "239.255.96.69"   -- might need to change for conflicts?
local use_port = 10000
local multicast_port = use_port
local broadcast_port = use_port
local broadcast_time = 1.5
local msg_contains = "Forlorn Fox Game Server"

-- dependancies
require("strict")
local socket = require( "socket" )
local class = require("middleclass")
require("networking/network_serializer")


-- isolate the rest in a local table that's returned
local LANb = {}
LANb.server_time = 0

function LANb.set_broadcast_port(port_num)
    multicast_port = port_num
    broadcast_port = port_num
end

function LANb.start_server_game_broadcast(port)
    LANb._broadcast_enabled = true
    local msg_lines = {
        game = msg_contains,
        game_version = gulp.forlorn_fox_version,
        game_id = gulp.server_identifier,
        -- names the same for backwards compatability.
        -- Can be removed at some point.
        ver = gulp.forlorn_fox_version,
        id = gulp.server_identifier,
        
        game_name = gulp.gameName,
        game_port = port
        }
    
        --print("broadcast - server_identifier", msg_lines.id)
        --print("broadcast - server version", msg_lines.ver)

    LANb.server_broadcast_msg = encode_for_network(msg_lines)
end

function LANb.stop_server_game_broadcast()
    LANb._broadcast_enabled = false
end

function LANb.broadcast_server_update(dt)
    if LANb._broadcast_enabled then
        -- we broadcast an update to the network every 1.5 seconds
        LANb.server_time = LANb.server_time + dt
        if LANb.server_time >= broadcast_time then
            LANb.server_time = 0
            
            if LANb.send == nil then

                local udp_error
                LANb.send, udp_error = socket.udp4() -- when does this get cleaned up? ever?
                if LANb.send then
                -- commented out becaue ... send never blocks
                --    LANb.send:settimeout( 0 )  --this is important (see notes below)
                else
                    LANb._broadcast_enabled = false
                    print("Problem opening udp socket for server broadcast:")
                    if udp_error then
                        print("  " .. udp_error)
                    end
                    return
                end
            end
            -- get a local copy
            local send = LANb.send

            print("** Send LAN multicast")
            local msuccess, multicast_error = send:sendto(LANb.server_broadcast_msg, multicast_address, multicast_port )
            if not msuccess then
                print("Error sending to multicast:")
                print("  " .. multicast_error)
            end
            
            --not all devices can multicast so it's a good idea to broadcast too
            --however, for broadcast to work, the network has to allow it
            local success, broadcast_error = send:setoption( "broadcast", true )  --turn on broadcast
            if not success then
                print("Error setting broadcast to true:")
                print("  " .. broadcast_error)
            end
             
             print("** Send LAN broadcast")
            success, broadcast_error = send:sendto(LANb.server_broadcast_msg, "255.255.255.255", broadcast_port )
            if not success then
                print("Error sending to broadcast:")
                print("  " .. broadcast_error)
            end
            success, broadcast_error = send:setoption( "broadcast", false )  --turn off broadcast
            if not success then
                print("Error setting broadcast to false:")
                print("  " .. broadcast_error)
            end
        end
    end
end

-- From Wikipedia
-- Google Public DNS operates recursive name servers for public use at the 
-- following IP addresses:[5] 8.8.8.8 and 8.8.4.4 for IPv4 service, as well 
-- as 2001:4860:4860::8888 and 2001:4860:4860::8844, for IPv6 access.
--
-- Mark Steelman originally used "74.125.115.104" which is googles website.
-- Aparrently it doesn't matter as this it not looked up.
local function getIP()
    local s = socket.udp4()  --creates a UDP object
    s:setpeername( "74.125.115.104", 80 )  --Google website
    local ip, sock = s:getsockname()
    --print( "myIP:", ip, sock )
    return ip
end

local listener = class("listener")

-- get new listener
function listener:initialize()
    self.servers_by_id = {}
    self.servers_by_order = {}
    local listen = socket.udp4()
    
    listen:setsockname( multicast_address, multicast_port )  --this only works if the device supports multicast

    local name = listen:getsockname()
    if ( name ) then  --test to see if device supports multicast
        listen:setoption( "ip-add-membership", { multiaddr=multicast_address, interface = getIP() } )
    else  --the device doesn't support multicast so we'll listen for broadcast
        listen:close()  --first we close the old socket; this is important
        listen = socket.udp4()  --make a new socket
        listen:setsockname( getIP(), broadcast_port )  --set the socket name to the real IP address
    end

    listen:settimeout( 0 )  --move along if there is nothing to hear
    
    self._listen = listen
    self.active = true
end

-- call listen() regularly
function listener:listen()
    if not self.active then 
        return
    end
    print("**Listen for LAN broadcast")
    local data, ip, port = self._listen:receivefrom()
    --print( "data: ", data, "IP: ", ip, "port: ", port )
    if data and data:find(msg_contains, 1, true) then
        local t = decode_from_network_string(data)
        
        -- backwards compatability code
        -- if new names don't exist translate from old data to new data
        if not t.game_version then
            t.game_version = t.ver
        end
        if not t.game_id then
            t.game_id = t.id
        end
        
        -- add local data
        t.game_ip_address = ip
        t.UDP_client_broadcast_port = port      -- this is not the game port
        
        t.received_last_update = gulp.time
        
        -- add into list if not see already
        local existing_server_id = self.servers_by_id[t.game_id]
        if not existing_server_id then
            --print( "I hear a server", ip, port, "Version", t.game_version)
            table.insert(self.servers_by_order, t)
            self.servers_by_id[t.game_id] = #self.servers_by_order
        else
            -- update data with info from new poll!
            self.servers_by_order[existing_server_id] = t
        end
        
    end
end

function listener:get_server_list()
    return self.servers_by_order
end

-- call terminate_listen when you are done watching to listen
function listener:terminate_listen()
    self._listen:close()  --never forget to close the socket!
    self.active = false
end

LANb.listener = listener

return LANb