-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
--
--
-- Usage:
--
--  a_string = encode_for_network(a_table)
--
--  a_table = decode_from_network_string(a_string)
--
--  a_table = decode_from_network_string(a_string, 23)
--
--  table_out = decode_from_network_incremental_new()
--  for line in a_string:gmatch("(.-)\x0a") do
--      decode_from_network_incremental_add_line(table_out, line)
--  end
--
--[[

KEY DESIGN CRITERIA:
    * Fast to encode/decode in pure Lua - this is the primary design criteria
    * Robust to any string data
    * Robust resequencing of data
    
NOT DESIGNED FOR:
    * Not designed for size.


FORMAT:
    Basically a mix of line based YAML and JSON for associate arrays (tables 
    in Lua), i.e. '{}'. 
    
    We recognise all types based on their first character.
    
    We encode all strings with troubling characters (mainly : and \x0a).
    
    Should be easy to port the encoder or decoder into C++ if necessary for
    speed.


FUTURE:
    Maybe errors shouldn't cause any noise, and instead the line should be
    silently ignored?
    
    

MORE NOTES:

Ignore string keys starting with underscore.

Keys: Encode strings, numbers, boolean
Values: Encode strings, numbers, tables, boolean 

Data is big endian (network byte order) - it's all in text anyway :-)

Unlimited level nested tables only.. Cyclic references not supported. 
(Cycles will local the encoder)


POSSIBLE PROBLEMS FOR BINARY:

 * Resynchronisation from an error condition is hard in binary, but a bit better in 
   text (not much). Conclusion: avoid errors :-)
 * Lots of test cases required for the packed binary - therefore dropped.
 * The packed binary isn't that packed, e.g. strings, floats - therefore dropped
    - But it is designed for SPEED of DECODING
 * Packing floats would be relative easy
 * Packing strings would require some some of updating dictionary - should we just zlib
 * If we are using Zlib (which we have in libpng), then surely we don't need binary encoding?

Choices:
 * Do we care about size? (ANSWER: No)
 * Do we care more about speed of encode and decode in Lua? (ANSWER: Yes)
 * Could we ignore the second thing by using C/C++ encoder/decoder? 
    (ANSWER: Or avoiding by putting networking in a seperate thread).
  
  
Things to consider:

Maybe printf will make '.' into ',' because of some wierd locale? 
    (ANSWER: CHECK ON FILE LOAD)

http://www.cplusplus.com/reference/clocale/setlocale/
  "But, independently of this system locale, on start, all C programs have 
  the "C" locale set, which is a rather neutral locale with minimal locale 
  information that allows the result of programs to be predictable"
  
  "On program startup, the locale selected is the "C" locale, which is the 
  same as would be set by calling setlocale(LC_ALL,"C")."
  
http://binglongx.wordpress.com/2009/07/01/locale-in-c-and-c/
  "All C programs (and C++ programs, as one of C++’s efforts is to be best 
  compatible with C) have the “C” locale in stock, which is a neutral locale 
  with minimal locale information across all compilers/systems, and not 
  affected by the current system locale or current program’s global locale."
  
https://stdcxx.apache.org/doc/stdlibug/24-3.html
  "Default locale. As a developer, you may never require internationalization 
  features, and thus you may never call std::setlocale(). If you can safely 
  assume that users of your applications are accommodated by the classic US 
  English ASCII behavior, you have no need for localization. Without even 
  knowing it, you always use the default locale, which is the US English 
  ASCII locale."

http://stackoverflow.com/questions/571359/how-do-i-set-the-proper-initial-locale-for-a-c-program-on-windows
  "At program startup, the global locale is the "C" or "Classic" locale."


--]]

-------------------------------------------------------------------------
-- The encoder
-------------------------------------------------------------------------

-- define ahead of time
local encode_table = nil

local string_replacement_chars = {
    ['"'] = '\\x22', -- reencode end delimiter
    ['\x0a'] = '\\n', -- reencode line end
    ['\x0d'] = '\\r', -- reencode CR
    [':'] = '\\x3a', -- reencode seperator
    ['\\'] = '\\\\' -- reencode the escape character
}

local function encode_string(out, s)
    -- check if the string has any protected characters in it
    
    local s_modified, count = s:gsub('(["\x0a:\\])', string_replacement_chars)
    if count ~= 0 then
        table.insert(out, '"')
        table.insert(out, s_modified)
        table.insert(out,  '"')
    else
        -- no special character, just put it in raw
        table.insert(out, "'")
        table.insert(out, s)
        table.insert(out, "'")
    end
end

local function encode_string_string(out, k, v)
    if k:sub(1,1) == "_" then
        return
    end
    encode_string(out, k)
    table.insert(out, ":")
    encode_string(out, v)
    table.insert(out, "\x0a")    
end
local function encode_string_simple(out, k, v)
    if k:sub(1,1) == "_" then
        return
    end
    encode_string(out, k)
    table.insert(out, ":")
    table.insert(out, tostring(v))
    table.insert(out, "\x0a")    
end
local function encode_string_table(out, k, v)
    if k:sub(1,1) == "_" then
        return
    end
    encode_string(out, k)
    table.insert(out, ":{")
    encode_table(out, v)
    table.insert(out, "}\x0a")    
end

local function encode_simple_string(out, k, v)
    table.insert(out, tostring(k))
    table.insert(out, ":")
    encode_string(out, v)
    table.insert(out, "\x0a")    
end
local function encode_simple_simple(out, k, v)
    table.insert(out, tostring(k))
    table.insert(out, ":")
    table.insert(out, tostring(v))
    table.insert(out, "\x0a")    
end
local function encode_simple_table(out, k, v)
    table.insert(out, tostring(k))
    table.insert(out, ":{")
    encode_table(out, v)
    table.insert(out, "}\x0a")    
end




local encode_funcs = {
    ["string"] = {
         ["string"] = encode_string_string,
         ["number"] = encode_string_simple,
         ["table"] = encode_string_table,
         ["boolean"] = encode_string_simple,
    },
    ["number"] = { 
         ["string"] = encode_simple_string,
         ["number"] = encode_simple_simple,
         ["table"] = encode_simple_table,
         ["boolean"] = encode_simple_simple,
    },
    ["boolean"] = { 
         ["string"] = encode_simple_string,
         ["number"] = encode_simple_simple,
         ["table"] = encode_simple_table,
         ["boolean"] = encode_simple_simple,
    }
}

encode_table = function(out, table_in)
    for k,v in pairs(table_in) do
        local func = encode_funcs[type(k)][type(v)]
        if func then
            func(out, k, v)
        end
    end
end

function encode_for_network(table_in)
    local table_of_strings = {}
    encode_table(table_of_strings, table_in)
    
    -- end marker must start on seperate line
    if #table_of_strings == 0 then
        table.insert(table_of_strings, "\x0a")
    end
    -- make end marker    
    table.insert(table_of_strings, "...\x0a")
    return table.concat(table_of_strings)
end


-------------------------------------------------------------------------
-- The decoder
-------------------------------------------------------------------------


local function true_function()
    -- don't even check string is true (case sensoitive or insensitive)
    return true
end
local function false_function()
    -- don't even check string is true (case sensoitive or insensitive)
    return false
end
local _substitute_slash_table = {
    ['n'] = '\x0a',
    ['r'] = '\x0d',
    ['\\'] = '\\',
    --['"'] = '"',     -- we don't produce this because it would break the decoder ... we produce \x22
    --[' '] = ' ',
    -- could add all the ones here: http://www.yaml.org/refcard.html
}

local function _substitute_slash_hex_function(capture)
    return string.char(tonumber(capture, 16))
end

local function escaped_string_decode_function(data, s, e)
    if data:sub(e, e) == '"' then
        data = data:sub(s+1,e-1)
    else
        -- need to strip the whitespace off the end?
        data = data:match('^"(.-)"%s-[:\x0a]?', s)
    end
    if data then
        data = data:gsub('\\x(%x%x)', _substitute_slash_hex_function)
        data = data:gsub('\\(.)', _substitute_slash_table)
    end
    return data
end

local function unescaped_string_decode_function(data, s, e)
    if data:sub(e, e) == "'" then
        return data:sub(s+1,e-1)
    else
        -- need to strip the whitespace off the end?
        return data:match("^'(.-)'%s-[:\x0a]?", s)
    end
end
local function number_decode_function(data, s, e)
    return tonumber(data:sub(s,e))
end
local function infinity_decode_function(data, s, e)
    return 1/0
end
local function nan_decode_function(data, s, e)
    return 0/0
end

local type_starts_with  -- define ahead
local decoder_line_loop -- define ahead of use

local function decode_leading_spaces(data, s, e) -- , maximum)
    --data = data:match('^%s-([^%s]-:)', s)  
    local i = data:find('[^%s]', s)
    if i == nil or i > e then
        return nil
    end
    local first_char_of_value = data:sub(i,i)
    local decode_function = type_starts_with[first_char_of_value]
    if decode_function then
        return decode_function(data, i, e)
    end
    --elseif first_char_of_value == '{' then
    --    -- ignore rest of line
    --    local output, e = decoder_line_loop(data, i+1, maximum)
    --    return output
    --else
    return nil
    --end
end

type_starts_with =
{
    ['t'] = true_function,    -- true
    ['f'] = false_function,   -- false
    ['"'] = escaped_string_decode_function,
    ["'"] = unescaped_string_decode_function, -- means no ' or LF or :
    ['0'] = number_decode_function,
    ['1'] = number_decode_function,
    ['2'] = number_decode_function,
    ['3'] = number_decode_function,
    ['4'] = number_decode_function,
    ['5'] = number_decode_function,
    ['6'] = number_decode_function,
    ['7'] = number_decode_function, 
    ['8'] = number_decode_function,
    ['9'] = number_decode_function,
    ['-'] = number_decode_function,   -- negative numbers
    ['+'] = number_decode_function,   -- show we allow this as well?
    ['.'] = number_decode_function,   -- fractional number - not standard, but allow
    ['i'] = infinity_decode_function,   -- inf
    ['n'] = nan_decode_function,   -- nan
    [' '] = decode_leading_spaces,
    ['\t'] = decode_leading_spaces,
    --['\x0a'] = function(data, s, e) return s+1 end
}

-- ASSUMPTIONS:
-- Normally no leading Whitespace
-- All types start with unique character.
-- NOTES:
-- very similar to YAML


local function decode_key_value(out, s, i, maximum)
    local err = false
    
    -- we know the LF exists because we've check for LF...
    -- don't allow raw LF in any string
    local line_end = s:find("\x0a", i, true)
    
    -- don't allow unescaped : in keys (for the moment)
    local seperator_index = s:find(":", i, true)
    if not seperator_index or seperator_index > line_end then
        --local close_table = s:find("^%s*}%s-\x0a", i)
        local close_table = s:find("^%s*}", i)
        if close_table == i then
            return -1, line_end+1
        end
        local empty_line = s:find("^%s-\x0a", i)
        if empty_line == i then
            return false, line_end+1
        end
        return true, line_end+1
    end
    local value_start = seperator_index +1
    local key_end = seperator_index - 1
    
    -- safe to assume at least 1 byte, again
    local decode_function = type_starts_with[s:sub(i,i)]
    if decode_function then
        local k = decode_function(s, i, key_end) --, maximum);
        if k ~= nil then
            local first_char_of_value = s:sub(value_start, value_start)
            decode_function = type_starts_with[first_char_of_value]
            if decode_function then
                local v = decode_function(s, value_start, line_end-1) --, maximum)
                if v ~= nil then
                    out[k] = v
                    return false, line_end+1
                --else
                    -- might be space {
                    -- could try to decode here..
                end
            elseif first_char_of_value == '{' then
                -- NOTE: Must be immediately after ':' otherwise will not be accepted
                -- ignore rest of line
                local output, e = decoder_line_loop(s, value_start+1, maximum)
                out[k] = output
                return false, e
            end
        end
    end
    
    return true, line_end+1
end

function decoder_line_loop(string_in, i, maximum)
    local output = {}
    while i < maximum do
        local err
        err, i = decode_key_value(output, string_in, i, maximum)
        if err then
            -- can be unrecoverable error
            -- or end of block marker
            -- either way we return what we can without error
            break
        end
    end
    return output, i
end

-- Assumes that error checking and correction has been done
-- But there might still be misformed data...
function decode_from_network_string(string_in, start_position)
    
    local i = start_position or 1
    -- Find the document terminator
    -- NOTE: Character LF cannot appear in strings (it's escaped)
    local s, e = string_in:find("\x0a...", i, true)
    if s == nil then
        return nil, i
    end
    
    local output = decoder_line_loop(string_in, i, s)
    return output, e+1
end



-- create a table to allow add_line to work
function decode_from_network_incremental_new()
    local t = {}
    t._table_from_network_current_table = t
    return t
end

-- s = line as string with no newline (LF or CRLF)
-- t = table under construction
function decode_from_network_incremental_add_line(t, s)

    -- don't allow unescaped : in keys (for the moment)
    local seperator_index = s:find(":", 1, true)
    if not seperator_index then
        -- Generally we ignore lines that don't contain a seperator
        -- e.g. illegal lines or empty lines
        -- Except for table close lines or table terminators
        if s:find("^%.%.%.") then
            t._table_from_network_current_table = nil
            return true
        end
        local close_table = s:find("^%s*}")
        if close_table then
            local current = t._table_from_network_current_table
            if current._decode_from_network_previous_table then
                -- we are not at the top level, step back one
                t._table_from_network_current_table = current._decode_from_network_previous_table
                current._decode_from_network_previous_table = nil -- erase this, we have finished at this level
            end
        end
        return false
    end
    local value_start = seperator_index +1
    local key_end = seperator_index - 1
    
    -- safe to assume at least 1 byte, again
    local decode_function = type_starts_with[s:sub(1,1)]
    if decode_function then
        local k = decode_function(s, 1, key_end)
        if k ~= nil then
            local first_char_of_value = s:sub(value_start, value_start)
            decode_function = type_starts_with[first_char_of_value]
            if decode_function then
                local v = decode_function(s, value_start, #s)
                if v ~= nil then
                    local out = t._table_from_network_current_table
                    out[k] = v
                    return false
                --else
                    -- might be space {
                    -- could try to decode here..
                end
            elseif first_char_of_value == '{' then
                -- NOTE: Must be immediately after ':' otherwise will not be accepted
                -- ignore rest of line
                local new_table = {}
                local out = t._table_from_network_current_table
                out[k] = new_table
                new_table._decode_from_network_previous_table = out
                t._table_from_network_current_table = new_table
                -- there might be stuff afterwards (for good or for bad)
                decode_from_network_incremental_add_line(t, s:sub(value_start+1))
            end
        end
    end
    
    return false
end

--
-- Do a test for the correct numeric locale (for decimal points)
--
local numeric_type = os.setlocale(nil, "numeric")
if numeric_type ~= "C" then
    print("Numeric type is not C it's", numeric_type)
    os.exit(1234)
end


-- UNIT TESTS
-- Run this file from a stock Lua 5.2 with this uncommented.
--
--[=[

-- from http://stackoverflow.com/questions/8722620/comparing-two-index-tables-by-index-value-in-lua
-- This is not clever enough to find matching table keys
-- i.e. this will return false
--   recursive_compare( { [{}]:1 }, { [{}]:1 } )
-- but this is unusual enough for me not to care ;)
-- It can also get stuck in infinite loops if you use it on 
-- an evil table like this:
--     t = {}
--     t[1] = t
local function recursive_compare(t1,t2)
  -- Use usual comparison first.
  if t1==t2 then return true end
  -- We only support non-default behavior for tables
  if (type(t1)~="table") then return false end
  if (type(t2)~="table") then return false end
  
  -- They better have the same metatables
  local mt1 = getmetatable(t1)
  local mt2 = getmetatable(t2)
  if( not recursive_compare(mt1,mt2) ) then return false end

  -- Check each key-value pair
  -- We have to do this both ways in case we miss some.
  -- TODO: Could probably be smarter and not check those we've 
  -- already checked though!
  for k1,v1 in pairs(t1) do
    local v2 = t2[k1]
    if( not recursive_compare(v1,v2) ) then return false end
  end
  for k2,v2 in pairs(t2) do
    local v1 = t1[k2]
    if( not recursive_compare(v1,v2) ) then return false end
  end

  return true  
end

local function print_table(table_to_print, indent)
    local indent = indent or 0
    for k, v in pairs(table_to_print) do
        local formatting = string.rep("  ", indent) .. tostring(k) .. ": "
        if type(v) == "table" then
            print(formatting)
            print_table(v, indent+1)
        elseif type(v) == 'boolean' or type(v) == 'function' then
            print(formatting .. tostring(v))      
        else
            print(formatting .. v)
        end
    end
end

local function verify_table(s1, t1, t1out, expected_t1, my_end)
    ---[[        <<for testing just endcode and decode>>
    local equal = recursive_compare(t1, t1out)
    if expected_t1 then
        equal = not equal and recursive_compare(t1out, expected_t1)
    end
    if not equal or my_end > #s1 then
        print("success?", equal)
        print("end, length:", my_end, #s1)
        print("TABLE IN")
        print_table(t1, 2)
        print("STRING OUT")
        print(s1)
        print("TABLE OUT")
        print_table(t1out, 2)
        os.exit(1)
    end
    --]]
end

local function test_table_serial(s1, t1, expected_t1)
    local t1out, my_end = decode_from_network_string(s1)
    verify_table(s1, t1, t1out, expected_t1, my_end)
end

local function test_table_line_table(s1, t1, expected_t1)
    -- we need to split this into lines
    local t1out = decode_from_network_incremental_new()
    for line in s1:gmatch("(.-)\x0a") do
        decode_from_network_incremental_add_line(t1out, line)
    end
    local my_end = #s1
    verify_table(s1, t1, t1out, expected_t1, my_end)
end

local function test_table(t1, expected_t1)
    local s1 = encode_for_network(t1)
    test_table_serial(s1, t1, expected_t1)
    test_table_line_table(s1, t1, expected_t1)
end

local function test_hand_created_string(s1, expected_t1, i)
    local t1out, my_end = decode_from_network_string(s1, i)
    ---[[    <<for testing just encode and decode>>
    local equal = recursive_compare(expected_t1, t1out)
    if not equal or my_end > #s1 then
        print("success?", equal)
        print("end, length:", my_end, #s1)
        print("INPUT STRING")
        print(s1)
        print("Start Pos", i)
        print("TABLE OUT")
        print_table(t1out,2)
        os.exit(1)
    end
    --]]
    return my_end
end

local start_time = os.clock()

local test_loops = 100
for i = 1, test_loops do
test_table{}
test_table{ "a", "b", "c" }
test_table{ a=22, b=23, c="cats on a roof" }
test_table{ [false]=true, [true]=false }
test_table{ [':\\\n"']=true, b=false, c="ca\\t:s o\n\n\n\\\\n a roof" }
test_table{
    [true] = "true",
    [false] = "false",
    }
test_table{
    [true] =  1,
    [false] = 2,
}
test_table{
    3.1,
    3.2,
    3.4,
    34e56,
    -1,
    99e-1
}
test_table({
    _private = "not encoded",
    _private2 = 1,    
    public = "quack"
}, {public = "quack"})
test_table({
    _private = {},
    public = true
}, {public = true})
test_table{
    [1] = 1,
    [true] = 2,
    [false] = 3,
    [4] = true,
    [5] = false,
    [6] = "six",
    [7] = {},
    eight = 8,
    ["9"] = "nine",
    ["10ten"] = {},
    ["eleven"] = false,
    ["twelve"] = true
}
test_table{
    [true] =  {},
    [false] = { "hi" },
}
test_table{" xxx\x0a:::\x0a  "}
-- deep nesting
test_table{ {}, { "a", "b", "c"}, { { "one", { "two", "three" } } } }


-- multiple strings concatenated
local t2 = {"1", "2", "3"}
local t3 = {"a", "b", "c"}
local t4 = {"x", "y", "z"}
local s2 = encode_for_network(t2)
s2 = s2 .. encode_for_network(t3)
s2 = s2 .. encode_for_network(t4)
local my_end = test_hand_created_string(s2, t2)
my_end = test_hand_created_string(s2, t3, my_end)
my_end = test_hand_created_string(s2, t4, my_end)

-- spaces
test_hand_created_string("   'hello':   1\x0a...\x0a", {hello=1})
test_hand_created_string("'goodbye'  :2   \x0a...\x0a", {goodbye=2})
test_hand_created_string(" 'test'  :{  \x0a  \x0a 'fred' : 23  \x0a  }   \x0a...  \x0a", {test={fred=23}})

-- test CR and LF combos
test_table{['\r\n\r\n\x0d\x0a'] = " xxx\x0a:::\x0a \x0d \r\n "}

end
local total_time = os.clock() - start_time
print("TOTAL RUN TIME ALL TESTS", total_time/test_loops)
print("Frame rate max." , test_loops/total_time)
--]=]

