-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
require("strict")
require('utils')
local class = require("middleclass")
local copas = require("copas")
--local chat = require("player/chat")
local fkey = require("utilities/function_key_monitor")
local ServerRemotePlayer = require("player/ServerRemotePlayer")

local ClientCount = 0
local RegisteredClientList = {}         -- [socket] = id
gulp.ControlSocketList = {}            -- [control_id] = skt

local ServerReceiveThreadExistsList = {}

function GulpNetworking:cleanupClients()
    for socket_ref,_ in pairs(RegisteredClientList) do  
        self.keepAlive:removeClientForServerToCheck(socket_ref)        
        socket_ref:AddSocketToClose()
        RegisteredClientList[socket_ref] = nil
        ServerReceiveThreadExistsList[socket_ref] = nil
        socket_ref = nil
    end
end    

function GulpNetworking:print_server_debug_info()
    print("Connected Clients:")
    for sock, id in pairs(RegisteredClientList) do
        print(" ", sock, id)
    end
    print("ServerReceiveThreads:")
    for sock, id in pairs(ServerReceiveThreadExistsList) do
        print(" ", sock, id)
    end
end

function GulpNetworking:broadcast(message)
    for socket_ref,_ in pairs(RegisteredClientList) do 
        --socket_ref:send(message)
        socket_ref:SendMessage(message)
        if gulpNetworkingDebuggingEnabled then 
            print("Relay message to socket:", socket_ref, message)
        end  
    end
end

function GulpNetworking:broadcast_to_all_except(excluded_socket, message)
    for socket_ref,id in pairs(RegisteredClientList) do 
        if socket_ref ~= excluded_socket then
            socket_ref:SendMessage(message) -- No need to re-encode as already done
            if gulpNetworkingDebuggingEnabled then 
                print("Relay message to socket:", socket_ref, message)
            end  
        end
    end
end
            
function GulpNetworking:broadcast_to_all_on_map_except(excluded_socket, map_name, message)
    for socket_ref,id in pairs(RegisteredClientList) do 
        if socket_ref ~= excluded_socket then
            local rp = get_player_by_id(id)
            if rp then 
                if map_name == rp.map_name then
                    socket_ref:SendMessage(message) -- No need to re-encode as already done
                    if gulpNetworkingDebuggingEnabled then 
                        print("Relay message to socket:", socket_ref, message)
                    end  
                end
            end    
        end
    end
end

function GulpNetworking:broadcast_to_all_on_map(map_name, message)
    self:broadcast_to_all_on_map_except(nil, map_name, message)
end

function GulpNetworking:broadcast_to_all_with_map_loaded_except(excluded_socket, map_name, message)
    for socket_ref,id in pairs(RegisteredClientList) do 
        if socket_ref ~= excluded_socket then
            local rp = get_player_by_id(id)
            if rp then 
                -- send to everyone who has the map loaded
                if rp.loaded_maps[map_name] then
                    socket_ref:SendMessage(message) -- No need to re-encode as already done
                    if gulpNetworkingDebuggingEnabled then 
                        print("Relay message to socket:", socket_ref, message)
                    end  
                end
            end    
        end
    end
end

function GulpNetworking:broadcast_to_all_with_map_loaded(map_name, message)
    self:broadcast_to_all_with_map_loaded_except(nil, map_name, message)
end


function GulpNetworking:broadcast_to_all_on_game_controller(controller_name, message)
    for socket_ref,id in pairs(RegisteredClientList) do 
        local rp = get_player_by_id(id)
        if rp then 
            if controller_name == get_map(rp.map_name).manager_name then
                socket_ref:SendMessage(message)
                if gulpNetworkingDebuggingEnabled then 
                    print("Relay message to socket:", socket_ref, message)
                end  
            end
        end    
    end
end


function GulpNetworking:SendItemAcrossNetwork(socket_ref, item)
    
    local message = item:serialise()
    
    socket_ref:SendMessage(message)
    
    if gulpNetworkingDebuggingEnabled then
        print(socket_ref, message)
    end
end


function GulpNetworking:SendMobsAndPickables(socket_ref, map_name)
    -- send all the mobs from the current map to the id specified
    
    local map = get_map(map_name)
    
    if socket_ref ~= nil then
        -- we found a socket, walk the mob table
        for key,mob in pairs(map.auto.mobs) do 
            if not mob.network_id then mob:set_network_id(key) end
            self:SendItemAcrossNetwork(socket_ref, mob)
        end
        
        -- now send the pickables
        map.auto.pickables:serialise_and_send(socket_ref)   
    else
        print("Null socket - not sending mobs and pickables")
    end
end
            
function GulpNetworking:SendPlayersAndGameStatus(player_in, map_name)
    local map = get_map(map_name)
    local socket_ref = player_in.socket_ref
    
    if socket_ref ~= nil then
        
        -- send the player locations for all players which are on this map
        for _, player in all_players_on_map_except(map_name, player_in) do
            socket_ref:SendMessage(player:form_position_message('switch_map'));
        end
        
        -- and we must send the game controller status for display
        gulp.game_managers[map.manager_name]:send_game_manager_status(socket_ref)
        
    else
        print("Null socket - not sending players or game status")
    end
end


function GulpNetworking:SendPickedItem(socket_ref, item)
    if socket_ref ~= nil then
        local message = { 
            mt = "picked",
            item = item:serialise()         
        }
        socket_ref:SendMessage(message);
        
        if gulpNetworkingDebuggingEnabled then
            print(socket_ref, message)
        end
    end
end
        


function GulpNetworking:SendDeath(socket_ref)
    if socket_ref ~= nil then
        local message = { 
            mt = "die",
            string = "Yes,really"
        }
      
        socket_ref:SendMessage(message);
       
        if gulpNetworkingDebuggingEnabled then
            print(socket_ref, message)
        end
    end
end
        
function GulpNetworking:broadcast_item(object) 
    for socket_ref,id in pairs(RegisteredClientList) do 
        local rp = get_player_by_id(id)
        if rp.loaded_maps[object.map_name] then
            self:SendItemAcrossNetwork(socket_ref, object)
        end
    end
end


function GulpNetworking:SendServerGameEnd(socket_ref)
    
    if socket_ref ~= nil then
    
        local message = { 
            mt = "server_exit",
            string = "Yes,really"
            }
      
        socket_ref:SendMessage(message);
            
        if gulpNetworkingDebuggingEnabled then
            print(socket_ref, message)
        end
    end
end


function HandleDroppedItemMessage(skt, message)
    local line = message.line
    local column = message.column

    local item_type = message.item.item_type  

    if (line) and (column) then
        if RegisteredClientList[skt] ~= nil then
            local rp = get_player_by_id(RegisteredClientList[skt])
            if rp ~= nil then
                rp:processDroppedItem(line, column, item_type)
            end
        end
    end
end    

function GulpNetworking:disableServer()   
    if self.Servers then
        for _,server in pairs(self.Servers) do 
            copas.removeserver(server)
        end        
    end
    print("GulpNetworking:disableServer()")
    self.Servers  = {}

end  

local socket = require "socket"
 
local server_addresses = { "0.0.0.0", "::"} -- previously "*"

function GulpNetworking:enableServer(ServerPort, mode)

    self.Port = ServerPort

    local success = false

    -- Bind the host and port
    for _, server_address in ipairs(server_addresses) do
        
        -- @todo: check if server has multiple interfaces if this works - see socket.lua from LuaSocket
        -- this only returns one for each IP address type. That could be a prpblem.
        local server, err = socket.bind(server_address, self.Port)    -- shouldn't block (copas.bind is just an alias for socket.bind)
        
        if gulpNetworkingDebuggingEnabled then print(self.Port, server, server_address) end
        if (server) then
            table.insert(self.Servers, server)
            --print("GulpNetworking Server Started:", server)  

            if gulpNetworkingDebuggingEnabled then        
                local ip, port = server:getsockname()    
                print("IP: ", ip)
                print("Port: ", port)
                print("for", server_address)
            end

            -- Register the server 
            if mode == 'control' then
                copas.addserver(server, ServerReceiveControlFunction)
                if gulpNetworkingDebuggingEnabled then print("add control server", tostring(ServerReceiveControlFunction) .. " " .. server_address) end
            else
                copas.addserver(server, ServerReceiveFunction)
                if gulpNetworkingDebuggingEnabled then print("add game server", tostring(ServerReceiveFunction) .. " " .. server_address) end
            end
            
            name_network_item(ServerReceiveFunction, "addserver - ServerReceiveFunction for " .. tostring(server) .. " " .. server_address)

            -- Register the message send thread
            local mythread
            if mode == 'control' then
                mythread = copas.addthread(ServerSendControlMessageThread) 
                if gulpNetworkingDebuggingEnabled then print("addthread ServerSendControlMessageThread()", tostring(mythread)  .. " " .. server_address) end
            else
                mythread = copas.addthread(ServerSendMessageThread) 
                if gulpNetworkingDebuggingEnabled then print("addthread ServerSendMessageThread()", tostring(mythread) .. " " .. server_address) end
            end
            name_network_item(mythread, "addthread - server SendMessageThread for " .. tostring(server) .. " " .. server_address)            
            debugMessage("Started server " .. server_address  .. " on port " .. ServerPort)
            success = true
        else
            print("socket.bind error", err, "for", server_address)
        end 
    end
    
    return success
end

local function send_map(socket_ref, map_name)
    local message_out = gulp.map_loader:form_map_message(map_name)
    socket_ref:SendMessage(message_out) 
end

function handle_client_request_connect(skt, message)
    local return_message
    local our_reason
    
    if not message.magic or message.magic ~= "FORLORN FOX CONNECT" then
        our_reason = "magic"
    end
    if not message.version or message.version ~= gulp.forlorn_fox_version then
        our_reason = "version"
    end
    
    if our_reason then
        return_message = {
            mt = "reject_connect",
            reason = our_reason,
            server_version = gulp.forlorn_fox_version,
        }
    else
        return_message = {
            mt = "accept_connect",
            magic = "Forlorn Fox Connect",
            server_version = gulp.forlorn_fox_version,
        }
    end
    
    -- send an initial map during connection
    send_map(skt, gulp.server_initial_map_name)
    
    skt:SendMessage(return_message)
end


function find_next_available_client_id()
    local id = 0
    
    local found = false
    
    while not found do
        id = id + 1
        for k,v in pairs(RegisteredClientList) do
            if v == ('client' .. id) then
                found = true
            end
        end
        -- if the id already exists, then we haven't found one we can use...
        found = not found
    end
    
    return 'client' .. id
end

function form_remote_ghost_message(player)
    return {
        mt = "add_remote_ghost_id",
        id = player.id,
        avatar = player.avatar,
        colour = player.colour,
        line = player.line,
        column = player.column,
        map_name = player.map_name,
        tagged = player.tagged,
        tool = player.tool,
        tool_orientation = player.tool_orientation,
        always_process_at_client = true,
    }
end


function HandleClientRegisterPlayerMessage(skt, RegisterMessage)
    --Get the unique player id    
    local ClientID = find_next_available_client_id()
    
    -- Create the remote player for this client
    ServerRemotePlayer(skt, RegisterMessage, ClientID)
    RegisteredClientList[skt] = ClientID
    
    
    local ResponseMessage = { mt = "accept_register", player_id = ClientID }

    -- Send the message
    skt:SendMessage(ResponseMessage)

    -- Send all current clients out to this new client too
    for _,id in pairs(RegisteredClientList) do 
        local remote_player = get_player_by_id(id)
        
        local ClientRegisteredMessage = form_remote_ghost_message(remote_player)
        skt:SendMessage(ClientRegisteredMessage)

        --print("Register existing player ", id, " with new player ", ClientID)
    end
    
        
    -- For all registed clients need to relay this message to them 
    local remote_player = get_player_by_id(ClientID)
    
    local ClientRegisteredMessage = form_remote_ghost_message(remote_player)
    gulp_networking:broadcast_to_all_except(skt, ClientRegisteredMessage)

    
    -- Add the client socket to the list of pings and set to the timeout
    gulp_networking.keepAlive:addClientForServerToCheck(skt)

    if gulpNetworkingDebuggingEnabled then 
        print("Total clients registered: ", ClientCount)

        -- Print out the list of registed clients
        for key,value in pairs(RegisteredClientList) do print(key,value) end
    end   

    ResponseMessage = { mt = "registration_complete", player_id = ClientID }

    -- Send the message
    skt:SendMessage(ResponseMessage)


end

function UnregisterClient(skt, networking_environment)
    
    if (RegisteredClientList[skt] ~= nil) then
        local player = get_player_by_id(RegisteredClientList[skt])
        RegisteredClientList[skt] = nil
        ServerReceiveThreadExistsList[skt] = nil
        print("HandleClientUnregisterMessage")
        
        if player then 
            
            local ClientUnregisteredMessage = { 
                mt = "remove_remote_ghost_id",
                id = player.id,
                }
                
            networking_environment:broadcast(ClientUnregisteredMessage)
            
            player:remove() 
        end
        ClientCount = ClientCount - 1
    end
    
    -- we know we can not longer talk to this client so remove from keep alive check
    networking_environment.keepAlive:removeClientForServerToCheck(skt)
end


function HandleClientUnregisterMessage(skt, UnregisterMessage)
    
    UnregisterClient(skt, gulp_networking)

    if gulpNetworkingDebuggingEnabled then 
        print("Unregister called")

        -- Print out the list of registed clients  
        for key,value in pairs(RegisteredClientList) do print(key,value) end
    end 
    
end

function HandleMoveMessage(skt, actionMove)  
    
    local player_map = actionMove.map_name

    -- If the message is complete then pass on to process
    if (player_map) then
        if RegisteredClientList[skt] ~= nil then
            local rp = get_player_by_id(RegisteredClientList[skt])
            
            rp:processPosition(actionMove)
    
            gulp_networking:broadcast_to_all_with_map_loaded_except(skt, player_map, actionMove)
            
        end
    end 
end

function server_handle_player_tagged_message(skt, message)
    local tagged_player_id = message.tagged_player
    local untagged_player_id = message.untagged_player
    
    if RegisteredClientList[skt] ~= nil then
        local trp = get_player_by_id(tagged_player_id)
        local utrp = get_player_by_id(untagged_player_id)
    
        trp:process_tagged('tag')
        utrp:process_tagged('untag')
    
        gulp_networking:broadcast_to_all_with_map_loaded(trp.map_name, message)
    end
end

function server_handle_player_speech_message(skt, message)
    local player_id = message.player
    
    if RegisteredClientList[skt] ~= nil then
        local p = get_player_by_id(player_id)
        gulp_networking:broadcast_to_all_with_map_loaded(p.map_name, message)
    end
end


local function server_process_game_control_message(socket_ref, message)
    local player_id = message.player_id
    find_player_game_manager(player_id):process_game_control_message(message)
end


local function HandlePlayerActionMessage(socket_ref, message)
    if (message) then
        local rp = get_player_by_id(RegisteredClientList[socket_ref])  
        rp:processAction(message.action)
        -- LOOP HERE TO UPDATE OTHER CLIENTS WITH ACTION?
      end
end

local function server_process_play_sound_message(socket_ref, message)
    gulp_networking:broadcast_to_all_on_map_except(socket_ref, message.map, message)
end

local function server_process_play_named_sound_message(socket_ref, message)
    gulp_networking:broadcast_to_all_on_map_except(socket_ref, message.map, message)
end

local function server_process_chat_message(socket_ref, message)
    gulp_networking:broadcast(message)
    
    if gulpNetworkingDebuggingEnabled then 
        print("server_process_chat_message: ", message.message) 
    end
end

local function server_process_server_ping_ack_message(socket_ref, message)
    
    -- Reset the timeout for this client to respond to the server ping
    gulp_networking.keepAlive:resetClientPingTimeout(socket_ref)
    
    if gulpNetworkingDebuggingEnabled then 
        print("server_process_server_ping_ack_message") 
    end
end

local function server_process_map_request(socket_ref, message_in)
    local map_name = message_in.map_name
    send_map(socket_ref, map_name)
    gulp_networking:SendMobsAndPickables(socket_ref, map_name)
    
    local rp = get_player_by_id(RegisteredClientList[socket_ref])  
    rp:register_map(map_name)
end

local function server_process_drop_map(socket_ref, message_in)
    local map_name = message_in.map_name
    local rp = get_player_by_id(RegisteredClientList[socket_ref])  
    rp:drop_map(map_name)
end

local function HandleFunctionKeyMessage(socket_ref, message_in)
    process_network_fkey(message_in.key)
end

local function HandlePlayerControlMessage(socket_ref, message_in)
    if message_in.command == 'take_managed_exit' then
        if message_in.exit.mode == 'start' then
            start_managed_game_for_player(message_in.exit.manager_file, message_in.id)
        elseif  message_in.exit.mode == 'exit' then
            stop_managed_game_for_player(message_in.exit.manager_file, message_in.id, message_in.map)
        end
        
    end
    
end

        
        

-- this table allow specific dispatch of message types (mt)
local _in_game_server_receivers =
{
    register_player = HandleClientRegisterPlayerMessage,
    request_connnect = handle_client_request_connect,
    unregister = HandleClientUnregisterMessage,
    game_control = server_process_game_control_message,
    player_move = HandleMoveMessage,
    player_action = HandlePlayerActionMessage,
    dropped = HandleDroppedItemMessage,
    play_sound = server_process_play_sound_message,
    play_named_sound = server_process_play_named_sound_message,
    player_tagged = server_handle_player_tagged_message,
    player_speech = server_handle_player_speech_message,
    chat_message = server_process_chat_message,
    server_ping_ack = server_process_server_ping_ack_message,
    map_request = server_process_map_request,
    drop_map = server_process_drop_map,
    function_key = HandleFunctionKeyMessage,
    player_control = HandlePlayerControlMessage,

}

local _out_game_server_receivers = 
{
}

local function HandleRegisterForControl(skt, message_in)
    
    local ControlID = #gulp.ControlSocketList + 1
    
    gulp.ControlSocketList[ControlID] = skt
    
    -- Add the client socket to the list of pings and set to the timeout
    -- no keepalive on control port for now
    -- gulp_control_networking.keepAlive:addClientForServerToCheck(skt)

    print("Handle registration for control, ControlID: "..ControlID)

    local ResponseMessage = { mt = "control_registration_complete", control_id = ControlID }

    -- Send the message
    skt:SendMessage(ResponseMessage)

end

local function HandleTestMessage(socket_ref, message_in)
    print(message_in.text)
    
    local response = {
        mt = 'test_response',
        text = 'Tony Test Reponse',
    }
    
    socket_ref:SendMessage(response)
end



local function HandleRequestGame(socket_ref, message_in)
    -- pass the message to the ui level that is running the game servers
    print("HandleRequestGame")
    message_in.qcommand = 'request_game'
    gulp.ui_queue:send(message_in)
end

local function HandleGetGameList(socket_ref, message_in)
    -- send the requester the games list
    print("HandleGetGameList")
    local response = { mt="game_list", game_list = gulp.game_list }
    socket_ref:SendMessage(response)
end

local _control_server_receivers = 
{
    register_for_control = HandleRegisterForControl,
    test_message = HandleTestMessage,
    request_game = HandleRequestGame,
    get_game_list = HandleGetGameList,
}

-- make it accessible for local
gulp.server_receivers = _in_game_server_receivers

function server_receiver_set_in_game_mode()
    gulp.server_receivers = _in_game_server_receivers
end

function server_receiver_set_out_game_mode()
    gulp.server_receivers = _out_game_server_receivers
end

function server_receiver_set_control_mode()
    gulp.server_receivers = _control_server_receivers
end

local NetworkTransport = require("networking/network_transport")

local function _RunServerReceiveFunction(skt, network)
    local success = skt:setoption('tcp-nodelay', true)

    if not success then
        print("Couldn't set TCP-Nodelay")
    end
        
    local abstracted_socket = NetworkTransport(skt)
    
    table.insert(network.clientSockets, abstracted_socket)

    ServerReceiveThreadExistsList[abstracted_socket] = true

    remember_socket_for_debug("server_receive", skt)
  --In a loop and only exits when player unregisters or connection dropped
  local message = decode_from_network_incremental_new()
  
  while true do
    
    local reqdata, err, partial = copas.receive(skt, '*l')    -- read just a line
    -- only get partial back after an error

    if gulpNetworkingDebuggingEnabled then 
      print("Gulpnetworking socket: ", skt)
      print("Gulpnetworking Received Message: ", reqdata) 
      print("Gulpnetworking partial: ", partial)
      print("Gulpnetworking error: ", err)
    end
    
    if ServerReceiveThreadExistsList[abstracted_socket] == nil then
        -- our socket has been deleted due to server quit - terminate without reporting error
        print("terminating ServerReceiveFunction()")
        return
    end
    
    -- Check for the online tracker port check message and filter out as not an error
    -- and socket will be closed after message is sent.
    if (reqdata) then
        -- @todo: should fix this to be a better method rather than throwing everything away!
        if string.find(reqdata, "OnlineTracker_Check_Port_Message") then
            skt:close() -- Close this as the connection is terminated by the server already and is preventing us starting a new game
            return 
        end  
    end
    
    -- Check the connection didn't have an error
    if (err) then
      print ("ServerReceiveFunction Copas receive error stopping err was: ", err) 
      
      abstracted_socket:HandleNetworkErrorServer("ServerReceive", network)
      return
    end
    
    -- Check for partial message recevied
    -- Will probably never get here because err will be set.
    --if (partial) then
    --        print ("ServerReceiveFunction Copas receive parial") 
    --end
    
    -- I don't think we need this if-reqdata test, because if it's nil
    -- then we must have an error message
    -- if reqdata then
        local finished = decode_from_network_incremental_add_line(message, reqdata)
        if finished and message.mt then
            local dispatcher = gulp.server_receivers[message.mt]
            if dispatcher then
                dispatcher(abstracted_socket, message)
            end
            message = decode_from_network_incremental_new()
        end
    --end
    
  end -- while loop
end

function ServerReceiveFunction(skt)
    print("+Game Connection from", skt, skt:getpeername())
    _RunServerReceiveFunction(skt, gulp_networking)
end

function ServerReceiveControlFunction(skt)
    print("+Control Connection from", skt, skt:getpeername())
    _RunServerReceiveFunction(skt, gulp_control_networking)
end


