-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Building

require('strict')
local class = require('middleclass')
require('graphics/glyph_sets')

local Building = class('Building')
local building_counter = 0

function Building:initialize(map, building_data)
    
    gulp._delegate_list:add_delegate(self)
    self.line = building_data.line
    self.column = building_data.column
    self.height = building_data.height
    self.width = building_data.width
    --self.maze = maze
    --self.player = player
    self.inside = false       
    self.done = false
    
    building_counter = building_counter + 1
    self.building_number = building_counter

    -- create the glyphs we need in C++
    self:create_glyphs(map.auto.maze)
    
    -- sort out our initial graphical state
    -- we can't do this because it uses player state that hasn't been set up on map switching
    -- seems that waiting for the first 'real' update is not a problem anyway
    --self:update(0)

    local collision_rects = 
    {
        --[[
        { centre = {0.9,self.width/2}             , size = {0.1,self.width/2-0.5}  } ,     -- top or back
        { centre = {self.height/2+0.5,0.5}            , size = {self.height/2-0.5,0.1} } ,     -- left
        { centre = {self.height/2+0.5,self.width-0.5} , size = {self.height/2-0.5,0.1} } ,     -- right
        { centre = {self.height-0.1,2*self.width/5-0.5} , size = {0.1,self.width/5-0.25}  } ,     -- bottom or front
        { centre = {self.height-0.1,4*self.width/5-0.5} , size = {0.1,self.width/5-0.25}  } ,     -- bottom or front
    --]]
        { line = 0.5, column = 0.25,              size = { line = 1/2, column = self.width - 0.5 } },   -- top or back
        { line = 0.5, column = 0.25,              size = { line = self.height - 0.5, column = 1/2 } },  -- left
        { line = 0.5, column = self.width - 0.75, size = { line = self.height - 0.5, column = 1/2 } },  -- right
        { line = self.height - 0.25, column = 0.25,      size = { line = 1/2, column = (self.width/5*2)-0.25 } },        -- bottom/front
        { line = self.height - 0.25, column = (self.width/5*3), size = { line = 1/2, column = (self.width/5*2)-0.25 } }        -- bottom / front
    
    }


    if gulp.LuaStateCodeMode == 'CLIENT' then
        self.collision_refs = {}
        for key, rect in ipairs(collision_rects) do
            self.collision_refs[key] = "Building-"..self.building_number.."-"..key
            local object =            
            {
                name = "Building-"..self.building_number,
                line = self.line,
                column = self.column,
                map_name = map.filename
            }
            register_for_collisions(self.collision_refs[key], object, rect)
        end
    end
end

function Building:get_building_inside_glyph(line, column)
    local glyph
    
    if line == 0 then glyph = 32        -- return a space where the roof used to be
    elseif line == 1 then glyph = 45+column 
    elseif line == self.height then glyph = 109+column 
    else glyph = 77+column end

    return get_glyph("townmap", glyph)
end

function Building:get_building_roof_glyph(line, column)
    local glyph
    
    if line == 0 then glyph = 7+column      -- return a space where the roof used to be
    elseif line == self.height-1 then glyph = 71+column 
    elseif line == self.height then glyph = 103+column 
    else glyph = 39+column end

    return get_glyph("townmap", glyph)
end

function Building:create_glyphs(maze)
    
    self.glyphs = { ['inside'] = {}, ['roof'] = {} }
    for line = 0, self.height do    -- start at zero - this is the extra line for the roof which is actually outside
        self.glyphs['inside'][line] = {}
        self.glyphs['roof'][line] = {}
        
        local inside_layer = 1
        local outside_layer = 1000
        
        if line == 1 then 
            inside_layer = 100
        end
        
        if line == self.height then 
            outside_layer = 100
        end
        
        for column = 1, self.width do
            
            self.glyphs['inside'][line][column] = 
                new_maze_display_element(maze, line+self.line-1, column+self.column-1, 
                                            self:get_building_inside_glyph(line, column), inside_layer)
            
            self.glyphs['roof'][line][column] = 
                new_maze_display_element(maze, line+self.line-1, column+self.column-1,
                                            self:get_building_roof_glyph(line, column), outside_layer)
                                        
        end
    end
        
end


function Building:update_player_view_layer(player)
    
    local layer = top_layer
    if self.inside == true then layer = 800 end
        
    for line = self.line-1, self.height+self.line do
        for column = self.column, self.width+self.column do
            player:set_view_layer(line, column, layer)
        end
    end
end


function Building:delete_glyphs()
    
    for line = 0, self.height do
        for column = 1, self.width do
            if gulp.LuaStateCodeMode == 'CLIENT' then
                self.glyphs['inside'][line][column]:hide()
                self.glyphs['roof'][line][column]:hide()
            end
            self.glyphs['inside'][line][column] = nil
            self.glyphs['roof'][line][column] = nil
        end
    end
        
end

-- a finisher method
function Building:_finish()
    gulp._delegate_list:drop_delegate(self)
    self:delete_glyphs()
    
    if gulp.LuaStateCodeMode == 'CLIENT' then
        for _, ref in pairs(self.collision_refs) do
            deregister_for_collisions(ref)
        end
    end
    
end

function Building:finish()
    self.done = true
end


function Building:update(dt)
    -- use this as a regular check to see if the player has entered us
    
    if self.done == true then 
        self:_finish()
    else
        
        local inside = false
        local player = gulp.localplayer
        
        if player then
            local line = player.line + player.collision_rect.centre.line
            local column = player.column + player.collision_rect.centre.column
        
            if line >= self.line + 1 and column >= self.column and
                line < self.line+self.height and column < self.column+self.width then
                -- player is inside
                inside = true
            end
        end
            
        if inside ~= self.inside then
            self.inside = inside
            self:update_player_view_layer(player)
        end
    end
    
end


return Building
