-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
require('strict')
require('utils')
local class = require('middleclass')

-- need objects if they are to be pickable across the network
local GoldCoin = require('objects/gold_coin')
local Hole = require('objects/hole')

-- debug mode to show buried items
local show_buried = false

--------------------------------------------------------------
--
-- None-class function

function place_item(item, line, column, map_name)
    if not is_wall(map_name, line, column) then
        if gulp.LuaStateCodeMode == 'CLIENT' then
            -- drop item across network to server pickable list
            gulp_networking:SendDroppedItem(line, column, item) 
        else
            -- drop items to server pickable list
            get_map(map_name).auto.pickables:add_to_floor(line, column, item)   
        end
    end
end


--------------------------------------------------------------
--
-- Helper class

local ObjectInMazeHolder = class('ObjectInMazeHolder')

function ObjectInMazeHolder:initialize(object, line, column, map_name, view_depth, buried)
    self.line = line 
    self.column = column 
    self.map_name = map_name
    self.view_depth = view_depth
    self.buried = buried
    
    self.object = object
    self.network_id = nil
    
    local maze = get_maze(self.map_name)
    local glyph = self.object.glyph
    if show_buried and self.buried then self.object.glyph = get_glyph('spectrum', 66) end
    self.img = new_hidden_mob_display_element(maze, self.object.glyph, self.line, self.column, self.view_depth, get_simple_colour_by_name('BRIGHT_WHITE'))
end

function ObjectInMazeHolder:get_object()
    return self.object
end

-- client side
function ObjectInMazeHolder:show()
    if not self.buried or show_buried then
        self:local_show()
    end
end

-- client side
function ObjectInMazeHolder:local_show()
    self.img:set_layer(self.view_depth)
    self.img:set_glyph(self.object.glyph)
    self.img:show()
end

-- server side
function ObjectInMazeHolder:network_show()
    
    local message = { 
        mt = "item_show",
        map_name = self.map_name,
        network_id = self.network_id,
        view_depth = self.view_depth,
        glyph = self.object.glyph,
    }

    gulp_networking:broadcast_to_all_with_map_loaded(self.map_name, message)          
end

function ObjectInMazeHolder:process_show_message(message)   
    
    self.view_depth = message.view_depth
    if message.glyph ~= self.object.glyph then
        self.object.glyph = message.glyph
        self.img:set_glyph(self.object.glyph)
    end
    
    self:local_show()
end             

function ObjectInMazeHolder:local_hide()
    self.img:hide()
end

function ObjectInMazeHolder:network_hide()--view_depth)
    local message = { 
        mt = "item_hide",
        map_name = self.map_name,
        network_id = self.network_id,
    }
    
    gulp_networking:broadcast_to_all_with_map_loaded(self.map_name, message)          
end

function ObjectInMazeHolder:process_hide_message(message)
    -- object_show tags already stripped
    -- id already processed
    self:local_hide()
end             


function ObjectInMazeHolder:set_network_id(id)
    self.network_id = id
end

function ObjectInMazeHolder:serialise()

    local message = {
        mt = "floor_item",
        object = self.object:serialise(),
        network_id = self.network_id,
        line = self.line,
        column = self.column,
        map_name = self.map_name,
        view_depth = self.view_depth,
        buried = self.buried,
        }

    return message
end

function new_network_object(message)
    
    local item_type = message.object.item_type
    local line = message.line
    local column = message.column
    local map_name = message.map_name
    local view_depth = message.view_depth
    local buried = message.buried
    local object = nil
    
    if item_type == 'gold_coin' then
        -- handle a gold coin
        object = GoldCoin("gold coin")
    elseif item_type == 'hole' then
        -- handle a gold coin
        object = Hole("hole")
    else
        print("Tried to create new network object of unknown type " .. item_type)
    end
   
    if object ~= nil and column ~= nil and line ~= nil then
        object = ObjectInMazeHolder(object, line, column, map_name, view_depth, buried)
        object:show()      
    end
    
    return object
end

--------------------------------------------------------------
--
-- This is the main class

local PickableManager = class('PickableManager')

function PickableManager:initialize(map_name)
    self._line_data = {}
    self.map_name = map_name 
    
    if gulp.LuaStateCodeMode == 'SERVER' then
        self.next_network_id = 0
    else
        self.next_network_id = nil  -- @todo: Check what the client is doing with this!
    end
    
        
end

function PickableManager:serialise_and_send(socket_ref)
    -- loop through our data structure, serialising and sending each item
    for _, line_data in pairs(self._line_data) do
        for _, column_data in pairs(line_data) do
            for _, item in pairs(column_data) do
                
                gulp_networking:SendItemAcrossNetwork(socket_ref, item)
                
            end
        end
    end
end 

function PickableManager:delete_all()
    self._line_data = {}
end

function PickableManager:add_to_floor(line, column, object, mode, buried)
    
    -- default the view depth, make sure holes always underneath treasure
    local view_depth = 2    
    if object.item_type == 'hole' then view_depth = 1 end
    
    if buried == nil then buried = false end
    
    -- create a holder for this object
    local holder = ObjectInMazeHolder(object, 
                    line - object.center_height_offset, 
                    column - object.center_width_offset, 
                    self.map_name, view_depth, buried)
    
    
    -- we don't deal with centers, for this approximation
    local integer_line = math.floor(line)
    local integer_column = math.floor(column)
    
    local ld = self._line_data[integer_line]
    if ld == nil then
        self._line_data[integer_line] = { [integer_column] = { holder } }
    else
        if ld[integer_column] then
            -- adjust view depth to keep this new item on top of the stack - if it is not a hole
            if object.item_type ~= 'hole' then 
                holder.view_depth = holder.view_depth + #ld[integer_column]
            end
            
            table.insert(ld[integer_column], holder)
        else
            ld[integer_column] = { holder }
        end
    end
    
    -- show the object at the calculated view depth, 
    if gulp.LuaStateCodeMode == 'CLIENT' then
        holder:show()
    else
        -- set the network_id of the item
        holder:set_network_id(self.next_network_id)
        self.next_network_id = self.next_network_id + 1
        
        -- Finally broadcast to network after view_depth has been set up
        -- if we are not being called during map generation.
        -- Map generation is handled by a call to SendMobsAndPickables() to the client
        -- immediately after map generation / transmit has finished, if the map was 
        -- generated for a client. This ensures the client already has a map to add 
        -- the pickable to.
        if mode ~= 'map_generation' then
            gulp_networking:broadcast_item(holder)
        end
    end
end

function PickableManager:check_for_buried_treasure(line, column)
    -- check up to 4 cells overlapped by this hole
    line = line - 0.5
    column = column - 0.5
    local start_line = math.floor(line)
    local end_line = math.ceil(line)
    local start_column = math.floor(column)
    local end_column = math.ceil(column)
    
    for l = start_line, end_line do
        local line_data = self._line_data[l]
        if line_data then
            for c = start_column, end_column do
                local data = line_data[c]
                if data then
                    for _, item in pairs(data) do
                        if math.abs(item.line - line) < 0.4 and math.abs(item.column - column) < 0.4 then
                            if item.buried then
                                item.buried = false
                                -- debug - if we were showing buried, switch back to the real glyph, always a coin right now
                                if show_buried then item.object.glyph = get_glyph('blackwall',159) end
                                item:network_show()
                            end
                        end
                    end
                end
            end
        end
    end
end

function PickableManager:get_distance_to_metallic_treasure(line, column, maximum_distance)
    line = line - 0.5
    column = column - 0.5
    local start_line = math.floor(line-maximum_distance)
    local end_line = math.ceil(line+maximum_distance)
    local start_column = math.floor(column-maximum_distance)
    local end_column = math.ceil(column)+maximum_distance
    
    local minimum = maximum_distance

    for l = start_line, end_line do
        local line_data = self._line_data[l]
        if line_data then
            for c = start_column, end_column do
                local data = line_data[c]
                if data then
                    for _, item in pairs(data) do
                        if item.object.metallic then
                            local line_delta = math.abs(item.line - line)
                            local column_delta = math.abs(item.column - column)
                            if line_delta < maximum_distance and column_delta < maximum_distance then
                                minimum = math.min(math.max(line_delta, column_delta), minimum)
                            end
                        end
                    end
                end
            end
        end
    end
    return minimum
end



-- return a table (might be empty) or nil
function PickableManager:_get_holders_in_cell(line, column)
    local ld = self._line_data[math.floor(line)]
    if ld == nil then
        return nil
    end
    return ld[ math.floor(column) ]
end

--function PickableManager:list_of_reachable(line, column) --, width, height)
    --[[
    local x1 = column - width
    local x2 = column + width
    local y1 = line - height
    local y2 = line + height
    
    if x1 < 0 then x1 = math.floor(x1) else x1 = math.ceil(x1) end
    if x2 < 0 then x2 = math.floor(x1) else x2 = math.ceil(x1) end
    if y1 < 0 then y1 = math.floor(y1) else y1 = math.ceil(y1) end
    if y2 < 0 then y2 = math.floor(y2) else y2 = math.ceil(y2) end
    --]]
    
--end

-- pick up functions are server side only!!

-- nasty dependency on the order buried objects are added to the list... fix sometime
function PickableManager:pick_up_first(line, column) -- , width, height)
    local holders = self:_get_holders_in_cell(line, column)
    
    if holders and #holders > 0 then
        for ref = #holders, 1, -1 do
            local holder = holders[ref] 
            if holder.object.pickable and not holder.buried then
                holder:network_hide()
                table.remove(holders, ref)
                return holder:get_object()
            end
        end
    end
    
    return nil
end

-- not updated to handled buried objects
function PickableManager:pick_up_specific_by_name(line, column, name) -- , width, height, name)
    local holders = self:_get_holders_in_cell(line, column)
    
    if holders and #holders > 0 then
        for k,holder in ipairs(holders) do
            local object = holder:get_object()
            if object.name == name then
                holder:network_hide()
                table.remove(holders, k)
                return object
            end
        end
    end
    return nil    
end

-- not updated to handled buried objects
function PickableManager:pick_up_specific_by_object(line, column, object) -- width, height, object)
    local holders = self:_get_holders_in_cell(line, column)
    
    if holders and #holders > 0 then
        for k,holder in ipairs(holders) do
            local obj = holder:get_object()
            if obj == object then
                holder:network_hide()
                table.remove(holders, k)     -- remove from holders list
                return obj
            end
        end
    end
    return nil    
end

return PickableManager

