-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Player
-- client side only code

require('strict')
class = require('middleclass')
require('map_control')
require('utils')
require('library/SDL_keycode')

Menu = require("menu")

-- need objects if they are to be pickable
local GoldCoin = require('objects/gold_coin')
local Hole = require('objects/hole')

require('gulp_networking')
require('map_controllers/game_manager')
require('objects/buried_item_manager')

local player_game_menu = require('player/player_game_menu')
local PlayerAvatar = require("player/PlayerAvatar")

local MouseMove = require("player/mouse_move")
local SpeechBubble = require('objects/speech_bubble')



local Player = class('Player', PlayerAvatar)

function Player:initialize(avatar, colour, id)
    
    local tagged = false
    PlayerAvatar.initialize(self, avatar, colour, id, gulp.client_initial_map_name, tagged)
    
    
    gulp._delegate_list:add_delegate(self)
 
    local maze = get_maze(self.map_name)
    
    self.direction_action = Player.direction_action_none
    self.fire_action = Player.fire_action_none
    self.selected_fire_action = Player.fire_action_none
    
    self.emote_action = nil
    
    self.dl = 0
    self.dc = 0
    self.last_dl = 0
    self.last_dc = 0
    self.last_animation_mode = 's_st'
    self.animation_time = 0
    
    self.current_line_offset = 0
    self.current_column_offset = 0

    self.height = maze:height()
    self.width = maze:width()
    
    self:calculate_offsets()

    -- If 'playing_tag' is changed to true, make sure that tag markers are correctly displayed
    -- for server and client when they log on. This may be slightly awkward to do during
    -- initialisation, since PlayerAvatar checks localplayer when showing a tag marker,
    -- and localplayer doesn't exist during initialisation of Player
    self.playing_tag = false
    --self:change_tag_game_mode(true)
    
    self.items_held = {}
    self.held = false
    
    
    self.ping_accumulator = 0
    self.ping_counter = 0
    self.last_ping_distance_from_server = 10        -- something high
    
    
    self.max_dt = 0.01
    
    self.speed_multiplier = 1.25
    
end

function Player:move_to_start_location()
    
    local line, column
    
    line = self.spawn_line
    column = self.spawn_column

    if line ~= nil and column ~= nil then
        self.line = line
        self.column = column
    
    end
    
    self:calculate_offsets()
    self:update_location()
    self.visibile_to_mobs = true

    -- Send the position to the server
    -- debugMessage("Sending position of start location "..self.line..self.column)
    gulp_networking:sendPlayerPosition(self.column, self.line, self.map_name, "reset")

end

function Player:update_location()
    self:update_glyphs_location()
end


-- a finisher method
function Player:remove()
    gulp._delegate_list:drop_delegate(self)
    PlayerAvatar.remove(self)
end



function Player:drop_item(item_key, line, column)
    place_item(self.items_held[item_key], line, column, self.map_name)
    table.remove(self.items_held, item_key)
end    

function Player:die()
    
    -- drop items
    for key,_ in ipairs(self.items_held) do
        if --[[math.random(0,1) ==--]] 1 then
            -- drop this item
            
            local item_line, item_column
            repeat
                item_line = math.floor(self.line) + math.random(-1, 1) + 0.5        -- counteract centrist drawing algorithm
                item_column = math.floor(self.column) + math.random(-1, 1) + 0.5
            until not is_wall(self.map_name, item_line, item_column)
                
            self:drop_item(key, item_line, item_column)
        end
    end
    
    -- move to respawn point
    self:move_to_start_location()
    
end     

function Player:set_view_layer(line, column, layer)
    local maze = get_maze(self.map_name)
    maze:set_view_layer(line, column, layer)
end



function sign(x)
  return (x<0 and -1) or 1
end

local MOVE_MAP = 0x0F
    
local     GAME_UP_ACTION = 0x01
local     GAME_DOWN_ACTION = 0x02
local     GAME_LEFT_ACTION = 0x04
local     GAME_RIGHT_ACTION = 0x08
local     GAME_FIRE_ACTION = 0x10
local     GAME_MENU_ACTION = 0x20
local     GAME_EMOTE_ACTION = 0x40


local key1_table = 
{
    GAME_UP_ACTION,
    GAME_DOWN_ACTION,
    GAME_LEFT_ACTION,
    GAME_RIGHT_ACTION,
}
local key2_table = 
{
    [GAME_UP_ACTION] = { 0, 0, GAME_LEFT_ACTION, GAME_RIGHT_ACTION },
    [GAME_DOWN_ACTION] = { 0, 0, GAME_LEFT_ACTION, GAME_RIGHT_ACTION },
    [GAME_LEFT_ACTION] = { 0, 0, GAME_UP_ACTION, GAME_DOWN_ACTION },
    [GAME_RIGHT_ACTION] = { 0, 0, GAME_UP_ACTION, GAME_DOWN_ACTION },
}


function Player:ai_toggle()
    if self.ai == nil then
        local ai = {}
        ai.tick = 0
        ai.time = 0.25 * math.random(8)
        ai.key1 = key1_table[math.random(4)]
        ai.key2 = key2_table[ai.key1][math.random(4)]
        ai.move_map = ai.key1 + ai.key2
        
        self.ai = ai
    else
        self.ai = nil
    end
end


function Player:ai_actions(dt)
    
    local ai = self.ai
    
    ai.tick = ai.tick + dt
    
    if ai.tick > ai.time then
        -- change direction
        
        if math.random(3) == 1 then
            -- change key2
            ai.key2 = key2_table[ai.key1][math.random(4)]
        else
            -- change key1 
            ai.key1 = key1_table[math.random(4)]
            ai.key2 = key2_table[ai.key1][math.random(4)]
        end
            
        ai.move_map = ai.key1+ai.key2    
        
        ai.tick = 0
        ai.time = 0.25 * math.random(8)
    end
    
    self:move(ai.move_map)
    
end


function Player:update(dt)
        
    if self.ai ~= nil then
        self:ai_actions(dt)
    end
        
    self:direction_action(dt, self.dl, self.dc, self.direction_type)
    self:fire_action(dt)
    
    -- animation
    self:select_animation_frame(dt)    
    
    -- Send the position to the server
    gulp_networking:sendPlayerPosition(self.column, self.line, self.map_name)
    
    -- speech
    if self.emote_action then
        self:emote_action(dt)
    end
    
end



function Player:escape_key()
    
    local gm = find_game_manager(get_map(self.map_name))
    
    if gm == nil or gm:handle_escape_key() == false then
        local bg = gulp.colour.BRIGHT_WHITE.simple_colour
        local fg = gulp.colour.BRIGHT_RED.simple_colour
        if math.random(2) == 1 then
            local swap = bg
            bg = fg
            fg = swap
        end
        
        self._escape_menu = Menu{
            background = bg, 
            foreground = fg,
            border_style = 2,
            { "You hit Escape? " },
            { "" },
            { "  SAVE AND QUIT ", function() gulp.save_and_quit_game() end},
            { "  Back to game ", "this is not a function so just exit"},
        }
    end
end

local player_key_actions = {
    [SDLK.ESCAPE] = Player.escape_key
}

function Player:keypressed(sym)
    local action = player_key_actions[sym]
    if action then
        action(self, sym)
        return true
    end
    return false
end

function Player:switch_map(map_name)
    if not find_local_game_manager():switch_map(self.id, map_name) then
        return false
    end
    
    self.spawn_line = self.line
    self.spawn_column = self.column
    
    local maze = get_maze(self.map_name)
    self.height = maze:height()
    self.width = maze:width()
    
    return true
end


function Player:pick_up_item(item)
    table.insert(self.items_held, item)
    gulp.sound.play_tink()
end 


function Player:process_pick_item_message(item_in)
    
    local item_type = item_in.item_type
    
    if item_type == 'gold_coin' then
        -- handle a gold coin
        local item = GoldCoin("gold coin")
        self:pick_up_item(item)
    else
        print("Tried to network pick unknown object " .. item_type)
    end
end


--------------------
-- PLAYER ACTIONS --
--------------------

require("bit32")


local move_map_lookup = 
{
    [GAME_UP_ACTION]    = { ['dl'] = -1,  ['dc'] = 0,     ['type'] = 'row' }, 
    [GAME_DOWN_ACTION]  = { ['dl'] = 1,   ['dc'] = 0,     ['type'] = 'row' }, 
    [GAME_LEFT_ACTION]  = { ['dl'] = 0,     ['dc'] = -1,  ['type'] = 'column' }, 
    [GAME_RIGHT_ACTION] = { ['dl'] = 0,     ['dc'] = 1,   ['type'] = 'column' }, 
    [GAME_UP_ACTION+GAME_LEFT_ACTION]   
                        = { ['dl'] = -0.7, ['dc'] = -0.7, ['type'] = 'both' }, 
    [GAME_UP_ACTION+GAME_RIGHT_ACTION]  
                        = { ['dl'] = -0.7, ['dc'] = 0.7,  ['type'] = 'both' }, 
    [GAME_DOWN_ACTION+GAME_LEFT_ACTION]
                        = { ['dl'] = 0.7,  ['dc'] = -0.7, ['type'] = 'both' }, 
    [GAME_DOWN_ACTION+GAME_RIGHT_ACTION] 
                        = { ['dl'] = 0.7,  ['dc'] = 0.7,  ['type'] = 'both' }, 
}

function Player:move(move_map)
    --print("Player:action ", action)
    
    --local move_map = bit32.band(action, MOVE_MAP)
    
    if self.held or move_map_lookup[move_map] == nil then
        self.direction_action = Player.direction_action_none
        self.dl = 0
        self.dc = 0
    else
        local entry = move_map_lookup[move_map]
        self.dl = entry['dl'] * self.speed_multiplier
        self.dc = entry['dc'] * self.speed_multiplier
        self.direction_type = entry['type']
        self.direction_action = Player.action_move
        
    end
end

local pe = require("graphics/particle_effects")
local pe_obj = nil

function Player:fire(action)

    if bit32.band(action, GAME_FIRE_ACTION) ~= 0 then
        self.fire_action = self.selected_fire_action
        --[[
        if pe_obj == nil then
            pe_obj = pe.new_magic(get_maze(self.map_name), self.line, self.column)
        else
            pe_obj:move_source(self.line, self.column)
        end
        --]]
    else
        -- cancel actions that continue until key is released
        self.fire_action = self.fire_action_none
    end
    
    -- Send the action on the network, will just return if not in network mode
    --gulp.network_send_player_action(action)   

    if bit32.band(action, GAME_MENU_ACTION) ~= 0 then
        -- don't need to delay to update() to run menu
        player_game_menu(self)
    end

    if bit32.band(action, GAME_EMOTE_ACTION) ~= 0 then
        self.emote_action = Player.start_emote
        --self:action_emote()
    end

end


function Player:action_move(dt, dl, dc, direction_type)

    
    -- slow player down during tool use
    if self.tool and self.tool ~= 'sword' then
        dl = dl / 2
        dc = dc / 2
    end
    
    if self.tagged then
        dl = dl * 1.1
        dc = dc * 1.1
    end
    
    dl = dl * dt * self.avatar_data.speed
    dc = dc * dt * self.avatar_data.speed
    
    -- check if dl and dc need adjusting
    local _
    _, dl, dc = get_collision("Player", self.map_name, self.line, self.column, dl, dc) 
    
    local line = self.line + dl
    local column = self.column + dc

    local avatar_edge_adjust_column = (self.avatar_data.gr_size_x / 2) 
    local avatar_edge_adjust_line = (self.avatar_data.gr_size_y / 2)  

    local move = false

    local top_edge =    line - avatar_edge_adjust_line
    local bottom_edge = line + avatar_edge_adjust_line
    local left_edge =   column - avatar_edge_adjust_column
    local right_edge =  column + avatar_edge_adjust_column



local function vertical_adjust(ahead_edge)        
    
    local finished = false
    local adjusted_right = false

    while not finished do
        
        local wall_straight = is_wall(self.map_name, ahead_edge, column)
        local wall_left = is_wall(self.map_name, ahead_edge, left_edge)
        local wall_right = is_wall(self.map_name, ahead_edge, right_edge)
        
        if wall_straight then 
            -- wall in that direction - nothing doing
            finished = true
        elseif wall_left and wall_right then
            -- gap is too small
            finished = true
        elseif wall_left then
            column = column + 0.05
            left_edge = left_edge + 0.05
            right_edge = right_edge + 0.05
            adjusted_right = true
        elseif wall_right then
            column = column - 0.05
            left_edge = left_edge - 0.05
            right_edge = right_edge - 0.05
            
            -- if we previously adjusted right, and now try to adjust left, 
            -- then the gap is too small!
            if adjusted_right then
                finished = true
            end
        else
            -- no wall straigh, left or right, we can move!
            move = true
            finished = true
        end
        
    end
end

local function horizontal_adjust(ahead_edge)        
    
    local finished = false
    local adjusted_down = false

    while not finished do
        
        local wall_straight = is_wall(self.map_name, line, ahead_edge)
        local wall_top = is_wall(self.map_name, top_edge, ahead_edge)
        local wall_bottom = is_wall(self.map_name, bottom_edge, ahead_edge)
        
        if wall_straight then 
            -- wall in that direction - nothing doing
            finished = true
        elseif wall_top and wall_bottom then
            -- gap is too small
            finished = true
        elseif wall_top then
            line = line + 0.05
            top_edge = top_edge + 0.05
            bottom_edge = bottom_edge + 0.05
            adjusted_down = true
        elseif wall_bottom then
            line = line - 0.05
            top_edge = top_edge - 0.05
            bottom_edge = bottom_edge - 0.05
            
            -- if we previously adjusted down, and now try to adjust up, 
            -- then the gap is too small!
            if adjusted_down then
                finished = true
            end
        else
            -- no wall straigh, top or bottom, we can move!
            move = true
            finished = true
        end
        
    end
end


local function diagonal_adjust(ahead_line, ahead_column)        
    
    local wall_straight = is_wall(self.map_name, ahead_line, ahead_column)
    local wall_vertical = is_wall(self.map_name, ahead_line, column)
    local wall_horizontal = is_wall(self.map_name, line, ahead_column)
    
    if wall_straight then 
        -- wall in that direction - can we slip or slide somewhere?
        if not wall_vertical then
            -- we can slide vertically
            column = self.column
            line = self.line + dl*1.4
            move = true
            --vertical_adjust(ahead_line)
        elseif not wall_horizontal then
            -- we can slide horizontally
            line = self.line                
            column = self.column + dc*1.4
            move = true
            --horizontal_adjust(ahead_column)
        else
            -- no sliding allowed
        end
    elseif wall_vertical and wall_horizontal then
        -- gap is too small
    elseif wall_vertical then
        -- we can slide horizontally
        line = self.line
        column = self.column + dc*1.4
        move = true
        --horizontal_adjust(ahead_column)
    elseif wall_horizontal then
        -- we can slide vertically
        column = self.column
        line = self.line + dl*1.4
        move = true
        --vertical_adjust(ahead_line)
    else
        -- no wall straight, top or bottom, we can move!
        move = true
    end

end


    
    -- switch mazes if there is an exit and matching room
    if self:check_exits(dl, dc, line, column) then
        -- do nothing, check_exits() has already done the work if necessary
    elseif top_edge < 0 then
        self:move_off_maze('up')
    elseif left_edge < 0 then
        self:move_off_maze('left')
    elseif bottom_edge >= self.height  then
        self:move_off_maze('down')
    elseif right_edge >= self.width then
        self:move_off_maze('right')
--[[    
    elseif not is_wall(self.map_name, bottom_edge, left_edge) and
           not is_wall(self.map_name, bottom_edge, right_edge) and
           not is_wall(self.map_name, top_edge, left_edge) and
           not is_wall(self.map_name, top_edge, right_edge) then
        move = true
--]]    
    elseif dl > 0 and dc == 0 then
        -- moving straight downwards
        vertical_adjust(bottom_edge)
        
    elseif dl < 0 and dc == 0 then
        -- moving straight upwards
        vertical_adjust(top_edge)
    
    elseif dl == 0 and dc > 0 then
        -- moving straight right
        horizontal_adjust(right_edge)
        
    elseif dl ==0 and dc < 0 then
        -- moving straight right
        horizontal_adjust(left_edge)
    
    elseif dl > 0 and dc < 0 then 
        -- moving diagonally down, left
        diagonal_adjust(bottom_edge, left_edge)
    
    elseif dl > 0 and dc > 0 then 
        -- moving diagonally down, right
        diagonal_adjust(bottom_edge, right_edge)
    
    elseif dl < 0 and dc < 0 then 
        -- moving diagonally up, left
        diagonal_adjust(top_edge, left_edge)
    
    elseif dl < 0 and dc > 0 then 
        -- moving diagonally up, right
        diagonal_adjust(top_edge, right_edge)
        
    end
    
    if move then
        self.line = line
        self.column = column
        self:calculate_maze_offset(direction_type)
    end
    
    

    
    self:update_location()
    
    local tool_orientation = 'down'
    if dl < 0 then
        tool_orientation = 'up'
    elseif dc > 0 then
        tool_orientation = 'right'
    elseif dc < 0 then
        tool_orientation = 'left'
    end
   
    self:shift_tool(tool_orientation)
    
    
end


---------------------------------------------------------------
-- CODE FOR FINDING AND HANDLING EXITS FROM THE CURRENT MAZE --
---------------------------------------------------------------

function Player:find_top_bottom_entrance()
    -- find an entrance in the new map which matches the line
    local exits = get_map(self.map_name).exits
    if exits ~= nil then
        for _, v in pairs(exits) do
            if v.line == math.floor(self.line) then
                -- found a matching entrance
                self.column = v.column + self.avatar_data.gr_center_x
                self:calculate_offsets()
            end
        end
    end
end
        
function Player:find_left_right_entrance()
    -- find an entrance in the new map which matches the line
    local exits = get_map(self.map_name).exits
    if exits ~= nil then
        for _, v in pairs(exits) do
            if v.column == math.floor(self.column) then
                -- found a matching entrance
                self.line = v.line + self.avatar_data.gr_center_y
                self:calculate_offsets()
            end
        end
    end
end

function Player:move_off_maze(direction)
    
    local map_id 
    
    if direction == 'up' then
        map_id = get_map(self.map_name).exit_north
        if map_id ~= nil then
            if self:switch_map(map_id) then
                self.line = self.height - 1.5
                self:find_top_bottom_entrance()
                self.spawn_line = self.line
                self.spawn_column = self.column
            end
        end
    elseif direction == 'down' then
        map_id = get_map(self.map_name).exit_south
        if map_id ~= nil then
            if self:switch_map(map_id) then
                self.line = 0.5
                self:find_top_bottom_entrance()
                self.spawn_line = self.line
                self.spawn_column = self.column
            end
        end
    elseif direction == 'right' then
        map_id = get_map(self.map_name).exit_east
        if map_id ~= nil then
            if self:switch_map(map_id) then
                self.column = 0.5
                self:find_left_right_entrance()
                self.spawn_line = self.line
                self.spawn_column = self.column
            end
        end
    elseif direction == 'left' then
        map_id = get_map(self.map_name).exit_west
        if map_id ~= nil then
            if self:switch_map(map_id) then
                self.column = self.width - 1.5
                self:find_left_right_entrance()
                self.spawn_line = self.line
                self.spawn_column = self.column
            end
        end
    end
    
    self:calculate_offsets()
    
end

function Player:take_exit(exit)
    local map = get_map(self.map_name)
    local new_map
    local map_id = exit.file
    
    -- check for exit to manager
    if exit.manager_file ~= nil then
        -- allow the game manager to handle the level transition
        -- server needs to deal with this
        gulp_networking:sendPlayerTookManagedExit(exit, self.map_name)
        return
    end
    
    if exit.mode == 'singleplayer' then
        map_id = map_id.."_^_"..self.id
    end
    
    --Store the current map and current location 
    local current_map = self.map_name;
    local current_line = self.line;
    local current_column = self.column;
    
    if map_id == nil then
        -- exit to somewhere else on this map
        new_map = map
    else
        if not self:switch_map(map_id) then return end
        new_map = get_map(self.map_name)
    end
                        
    if exit.destination ~= nil then
        -- we have a specific destination target character
        -- search the exit list of the new map to find that character
        if new_map.exits ~= nil and new_map.exits[exit.destination] ~= nil then
            -- we found the destination
            self.line = new_map.exits[exit.destination].line + 0.5 --self.avatar_data.gr_center_y
            self.column = new_map.exits[exit.destination].column + 0.5 --self.avatar_data.gr_center_x
            
            -- Play sound for map exit if there is one 
            if (exit.sound ~= nil) then
                local sound = gulp.sound[exit.sound]
                if sound then
                    -- Play on the current map at the current location 
                    play_and_transmit_located_sound(sound, 8, current_line, current_column, current_map)
                    
                    -- Play on this on the exit map locally and remote at the new postion 
                    play_and_transmit_located_sound(sound, 8, self.line, self.column, self.map_name)
                end
            end        
        else
            error("FATAL: Target destination doesn't exist ".. tostring(exit.destination))
        end
    
    elseif exit.destination_line ~= nil then
        self.line = exit.destination_line + 0.5 --self.avatar_data.gr_center_y
        
        if exit.destination_column ~= nil then
            self.column = exit.destination_column + 0.5 --self.avatar_data.gr_center_x
            -- we were given both line and column
            -- nothing more to do
        else
            -- we have a line but no column, so find an entrance on the line
            self:find_top_bottom_entrance()
        end
    else
        if exit.destination_column ~= nil then
            -- we have no line, but we have a column
            -- find a side entrance
            self.column = exit.destination_column + 0.5 --self.avatar_data.gr_center_x
            self:find_left_right_entrance()
        else
            -- we have no destination line or column - give up
            if map_id == nil then
                -- not a problem, this is just here to be a possible entrance
                -- or it might be a special exit that the game control is in charge of...
                if exit.notification then
                    find_local_game_manager():notify(self.id, exit.notification, self.map_name )
                    return true     -- exit has been dealt with
                else
                    return false
                end
            else
                error("FATAL: No destination specified for exit " .. exit.group .. " in " .. map_id)
            end
        end
    end
    
    self.spawn_line = self.line
    self.spawn_column = self.column
    
    -- work out the appropriate offsets 
    self:calculate_offsets()
    self:update_location()
    
    return true
end    

function Player:check_exits(dl, dc, line, column)
    local map = get_map(self.map_name)
    if map.exits == nil then return false end
    
    for k, v in pairs(map.exits) do
        
        local exit_line = v.line
        local exit_column = v.column
        local found_exit = false
        
        -- exits are map cells, work out a floating point exit 'centre'
        -- if it is at the edge of the map, make the centre at the map edge, 
        -- otherwise it is the centre of the exit cell
        local exit_centre_line
        local exit_centre_column
        
        if exit_line == 0 then
            exit_centre_line = 0
        elseif exit_line == self.height - 1 then
            exit_centre_line = self.height
        else
            exit_centre_line = exit_line + 0.5
        end
        
        if exit_column == 0 then
            exit_centre_column = 0
        elseif exit_column == self.width - 1 then
            exit_centre_column = self.width
        else
            exit_centre_column = exit_column + 0.5
        end
        
        -- now work out how far we are from the exit centre, 
        -- and how far we were before this move
        local vertical_distance_from_exit_centre = math.abs(line - exit_centre_line)
        local horizontal_distance_from_exit_centre = math.abs(column - exit_centre_column)
        local last_vertical_distance_from_exit_centre = math.abs(self.line - exit_centre_line)
        local last_horizontal_distance_from_exit_centre = math.abs(self.column - exit_centre_column)
        
        -- if we are close to the exit location, and moving towards it, then take the exit
        if  horizontal_distance_from_exit_centre <= 0.5 and 
            vertical_distance_from_exit_centre <= 0.5 and
            (last_vertical_distance_from_exit_centre > vertical_distance_from_exit_centre or
            last_horizontal_distance_from_exit_centre > horizontal_distance_from_exit_centre) then
            found_exit = true
        end
        
        if found_exit then
            if v.conditional then
                if v.condition ~= map.conditionals[v.conditional] then 
                    return false 
                end
            end
            return self:take_exit(v)
        end
    end
    
    return false
end

function Player:process_player_control_message(message)
        
    if message.command == 'take_exit' then
        self:take_exit(message.exit)
    elseif message.command == 'detector' then
        self.last_ping_distance_from_server = message.distance
    elseif message.command == 'collision' then
        self.line = message.line
        self.column = message.column
    end
    
end

function Player:change_tag_game_mode(playing)   -- argument is true / false
    
    self.playing_tag = playing
    
    -- set up the fire action
    if self.playing_tag == true then
        self.selected_fire_action = Player.action_tag
    else
        if self.selected_fire_action == Player.action_tag then
            self.selected_fire_action = Player.fire_action_none
        end
    end
    
    if self.playing_tag then
        -- make sure someone is tagged!
        local any_tag = false
        for _,player in all_players() do
            if player.tagged then
                any_tag = true
            end
        end
        
        if not any_tag then
            self.tagged = true
        end
        
        -- if we are tagged, show our marker 
        if self.tagged then
            self:show_tag_marker()
        end
        
        -- find any remote players that are tagged, and show their tags too
        for _,player in all_remote_players() do
            if player.tagged then
                player:show_tag_marker()
            end
        end
    else
        -- we we are not playing, definitely hide the marker - we might have stopped playing while we were tagged
        self:hide_tag_marker()
        
        -- also check the remote players for whether they are tagged, and hide their tags
        for _,player in all_remote_players() do
            if player.tagged then
                player:hide_tag_marker()
            end
        end
    end

    -- cause a map manager draw to update the status lines with the game status
    gulp.game_managers[get_map(self.map_name).manager_name]:cause_drawing()
    

end



function Player:select_tool(tool)
    
    local tool_action_table = {
        spade       = Player.action_dig,
        detector    = Player.action_detect,
        flash       = Player.action_flash,
        colourator  = Player.action_switch_maze_colours,
        teleport    = Player.action_teleport_to_exit,
        sword       = Player.action_jab,
    }
    
    
    if (tool == 'sword') and not self.avatar_data.has_sword then
        self.tool = nil
    else
        self.tool = tool
    end
    
    if self.tool == nil then
        self.selected_fire_action = Player.fire_action_none
    else
        self.selected_fire_action = tool_action_table[tool]
    end
    
    if self.avatar_data.has_sword then
        local sword = (tool == 'sword')
        self.avatar_data.compound_glyph_visible['scabbard'] =  not sword
        self.avatar_data.compound_glyph_visible['arm'] = not sword
        self.avatar_data.compound_glyph_visible['sword'] = sword
        self:set_compound_glyph_visibilities()
    end
    
    
    if tool ~= 'sword' then
        self:shift_tool()           -- no arg leaves orientation unchanged but alters glyph to new tool
    end

    self:show()
    self:broadcast_position()   -- causes network update of tool glyph
    
end


--------------------------------------------------------
--  FIRE ACTIONS                                      --
-- Several different actions and test functions here  --
--------------------------------------------------------

function Player:action_jab(dt)
    -- only jab once...?
    if self.thrusting then
        -- thrust finished?
        self.thrust_time = self.thrust_time + dt
        if self.thrust_time > self.avatar_data.compound_animation_stationary_time['thrust'] then
            -- thrust is finished, swith glyphs
            -- but don't set thrusting to false, or another thrust will be allowed to start, 
            -- and we want the player to lift the spacebar before that can happen
            self.avatar_data.compound_glyph_visible['sword'] = true
            self.avatar_data.compound_glyph_visible['thrust'] = false
            self:set_compound_glyph_visibilities()
        end
    else
        -- start the thrust
        self.thrusting = true
        self.thrust_time = 0
        
        -- switch sword subglyph off, switch jab subglyph on
        self.avatar_data.compound_glyph_visible['sword'] = false
        self.avatar_data.compound_glyph_visible['thrust'] = true
        self:set_compound_glyph_visibilities()
    end
    
end

function Player:action_dig(dt)
    -- a hole is an item... place it in front of the player 
    -- that is, 0.5 away in the direction they were last moving
    local line = self.line + (7*self.dl)
    local column = self.column + (7*self.dc)
    
    place_item(Hole('hole'), line, column, self.map_name)
    
    -- only dig one hole...
    self.fire_action = Player.fire_action_none
end


function Player:action_detect(dt)
    
    if not self.detect_mode then
        -- send detect mode message
        self.detect_mode = true
        gulp_networking:sendPlayerAction('detect_mode')
    end
    
    self.ping_accumulator = self.ping_accumulator + dt
    
    if self.ping_accumulator > 0.25 then
        self.ping_accumulator = self.ping_accumulator - 0.25
        
        -- do the actual ping
        self:do_metal_detector_ping(self.last_ping_distance_from_server)
        
    end
end

function Player:do_metal_detector_ping(distance)
    if self.ping_counter > distance then
        self.ping_counter = 0
        if distance < 1 then
            play_sound(gulp.sound.detector_ping[4], 8)
        elseif distance < 2 then
            play_sound(gulp.sound.detector_ping[3], 6)
        elseif distance < 3 then
            play_sound(gulp.sound.detector_ping[2], 4)
        else
            play_sound(gulp.sound.detector_ping[1], 2)
        end
    end
    
    self.ping_counter = self.ping_counter + 1
end        

function Player:action_tag(dt)

    -- check for player nearby 
    local nearby_player_ids = find_other_nearby_players(self, 1)
    
    -- if we are not near any other player, bail out here
    if #nearby_player_ids ~= 0 then 
    
        if self.tagged then
            -- we are tagged - we are trying to give the tag away
            -- pick a nearby player at random (could pick closest? doesn't matter)
            local tagged_player_id = random_entry(nearby_player_ids)
            --print ("Tagging player ", tagged_player)
            gulp_networking:SendTaggedPlayers(tagged_player_id, self.id)
            self.tagged = false
            self:hide_tag_marker()
        else
            -- we aren't tagged - we are trying to steal the tag
            -- check if any of the nearby players has the tag
            for _,nearby_player_id in pairs(nearby_player_ids) do
                if get_player_by_id(nearby_player_id).tagged == true then
                    --print ("Stealing tag from player ", untagged_player)
                    gulp_networking:SendTaggedPlayers(self.id, nearby_player_id)
                    self.tagged = true
                    self:show_tag_marker()
                end
            end
        end
    end
    
    self.fire_action = Player.fire_action_none
    
end

function Player:start_emote(dt)

    local untagged_texts = 
    {
        "Over here!",
        "Can't catch me!",
        "You're so slow!",
        "Tee Hee!",
        "Can you see me?",
    }
    
    local tagged_texts = 
    {
        "I'm gonna get you!",
        "No escape!",
        "Come back!",
        "I see you!!!",
    }
    
    local text
    
    if self.tagged then
        text = random_entry(tagged_texts)
    else
        text = random_entry(untagged_texts)
    end
        
    gulp_networking:SendSpeech(self.id, text)
    
    self.emote_time = 0
    self.emote_action = Player.check_emote
    
end

function Player:check_emote(dt)
    self.emote_time = self.emote_time + dt
    if self.emote_time > 2.0 then
        gulp_networking:SendSpeech(self.id, "")
        self.emote_action = nil
    end
end

    


function Player:hold()
    self.direction_action = Player.direction_action_none
    self.held = true
end

function Player:release()
    self.held = false
end

function Player:direction_action_none(dt)
end

function Player:fire_action_none(dt)
    if self.detect_mode then
        -- send detect mode message
        self.detect_mode = false
        gulp_networking:sendPlayerAction('detect_mode_cancel')
    end
    
    if self.thrusting then
        self.thrusting = false
        self.avatar_data.compound_glyph_visible['sword'] = true
        self.avatar_data.compound_glyph_visible['thrust'] = false
        self:set_compound_glyph_visibilities()
    end
    
        
end

function Player:action_flash(dt)
    -- this is buggy - will need fixed if we ever want to 
    -- actually use it
    
    local fg, bg
    local maze = get_maze(self.map_name)

    if not self.flash_on then
        -- turning flash on
        fg = get_simple_colour_by_name('BLACK')
        bg = get_SDL_Color_by_name('BRIGHT_WHITE')
        self.flash_on = true
        self.flash_on_time = 0
    else
        
        -- test the time and bail out if we haven't expired the flash yet
        self.flash_on_time = self.flash_on_time + dt
        if self.flash_on_time < 0.5 then return end
            
        -- turning flash off again
        fg = get_simple_colour_by_name(get_map(self.map_name).fg)
        bg = get_SDL_Color_by_name(get_map(self.map_name).bg)
        self.flash_on = false
        self.fire_action = Player.fire_action_none
    end

    local flash_pattern =
    {
        {' ',' ',' ','#','#','#','#','#',' ',' ',' '},
        {' ',' ','#','#','#','#','#','#','#',' ',' '},
        {' ','#','#','#','#','#','#','#','#','#',' '},
        {'#','#','#','#','#','#','#','#','#','#','#'},
        {'#','#','#','#','#','#','#','#','#','#','#'},
        {'#','#','#','#','#','#','#','#','#','#','#'},
        {'#','#','#','#','#','#','#','#','#','#','#'},
        {'#','#','#','#','#','#','#','#','#','#','#'},
        {' ','#','#','#','#','#','#','#','#','#',' '},
        {' ',' ','#','#','#','#','#','#','#',' ',' '},
        {' ',' ',' ','#','#','#','#','#',' ',' ',' '}
    }

    for line = self.line-5, self.line+5 do
        for column = self.column-5, self.column+5 do
            if line >= 0 and line <= self.height and column >= 0 and column <= self.width then
                if flash_pattern[line-self.line+6][column-self.column+6] == '#' then
                    maze:set_maze_colours(line, column, fg, bg)
                end
            end
        end
    end

end

function Player:action_los_test()
    local maze = get_maze(self.map_name)
    
    -- only do this action once per press of fire 
    self.fire_action = Player.fire_action_none
    
    local SDL_GetTicks = gulp_cpp.SDL_GetTicks
    
    local t1 = SDL_GetTicks()
    local los = maze:line_of_sight(self.line+0.5, self.column+0.5, 10.5, 16.5)
    local t2 = SDL_GetTicks()

    print ("LOS=", los, " in ", t2-t1, "ms")
    
    --maze:set_default_maze_colours(get_simple_colour_by_name('WHITE'),get_SDL_Color_by_name('BLACK'))
    
end


function Player:action_switch_maze_colours()
    local colour1, colour2
    local colour_table = {}
    local colour_name_table = {}
    
    local i = 1
    local k,v
    for k,v in pairs(gulp.colour) do
        colour_table[i] = v
        colour_name_table[i] = k
        i= i+1
    end
    
    repeat
        
        colour1 = random_entry(colour_table).simple_colour
        colour2 = random_entry(colour_name_table)
    until colour1 ~= colour2 and colour1 ~= nil and colour2 ~= nil
    
    get_maze(self.map_name):set_default_maze_colours(colour1,get_SDL_Color_by_name(colour2))

    -- only do this action once per press of fire 
    self.fire_action = Player.fire_action_none

end

function Player:action_test_iterators()
    display_playerlist_ex("central_cavern", get_player_by_id(2))
    
    -- only do this action once per press of fire 
    self.fire_action = Player.fire_action_none
end

function Player:action_teleport_to_exit()
    self.line = 20.5
    self.column = 15.5
    self:action_move(0.01, 0.1, 0, 'row')
    -- only do this action once per press of fire 
    self.fire_action = Player.fire_action_none
end



-------------------------------------------------------------------
-- SCROLLING ALGORITHMS                                          --
-- The rest of this file is a selection of scrolling algorithms  --
-- These may be selected per level, or as a player preference,   --
-- or else we'll just pick one                                   --
-------------------------------------------------------------------

-- This is the basic function. It does not try to scroll. Used at start
-- of game and respawn to set the offsets appropriately.
function Player:calculate_offsets()
    local maze = get_maze(self.map_name)
    local available_columns = maze:GetAvailableLevelWidth()
    
    if self.width < available_columns then
        self.current_column_offset = (self.width - available_columns) / 2
    else
        self.current_column_offset = self.column - (available_columns/2)
        if self.current_column_offset + available_columns > self.width then
            self.current_column_offset = self.width - available_columns
        end
        if self.current_column_offset < 0 then
            self.current_column_offset = 0
        end
    end
    
    local available_lines = maze:GetAvailableLevelHeight()
    
    if self.height < available_lines then
        self.current_line_offset = (self.height - available_lines) / 2
    else
        self.current_line_offset = self.line - (available_lines/2)
        if self.current_line_offset + available_lines > self.height then
            self.current_line_offset = self.height - available_lines
        end
        if self.current_line_offset < 0 then
            self.current_line_offset = 0
        end
    end
    
    maze:set_offset(self.current_line_offset, self.current_column_offset)
end





local scroll_speed = 20

function Player:calculate_maze_offset(offset_type)
    
    local maze = get_maze(self.map_name)
    
    if offset_type ~= 'column' then
        -- do row calcs
        local available_lines = maze:GetAvailableLevelHeight()
        
        if self.height < available_lines then
            self.current_line_offset = (self.height - available_lines) / 2
        else
            local screen_cent_y = available_lines / 2
            local view_adjust_y = (self.line - self.current_line_offset) - screen_cent_y;
            view_adjust_y = view_adjust_y / scroll_speed
            self.current_line_offset = self.current_line_offset + view_adjust_y
            
            if self.current_line_offset + available_lines > self.height then
                self.current_line_offset = self.height - available_lines
            end
            
            if self.current_line_offset < 0 then
                self.current_line_offset = 0
            end
        end
    end
    
    if offset_type ~= 'row' then
        -- do column calcs
        local available_columns = maze:GetAvailableLevelWidth()
        
        if self.width < available_columns then
            self.current_column_offset = (self.width - available_columns) / 2
        else
            local screen_cent_x = available_columns / 2
            local view_adjust_x = (self.column - self.current_column_offset) - screen_cent_x;
            view_adjust_x = view_adjust_x / scroll_speed
            self.current_column_offset = self.current_column_offset + view_adjust_x
            
            if self.current_column_offset + available_columns > self.width then
                self.current_column_offset = self.width - available_columns
            end
            if self.current_column_offset < 0 then
                self.current_column_offset = 0
            end
        
        end
    end
    
    maze:set_offset(self.current_line_offset, self.current_column_offset)
end





return Player
