-- Forlorn Fox
-- Copyright © 2015 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Client Remote Player
-- Used on client and server sides to display the player's avatars

-- todo avatar display not needed at server, but server remote player is based on this class
-- and some facilities still used at server, like the player list. 

require('strict')
require('utils')
class = require('middleclass')

local PlayerBase = require("player/PlayerBase")
local PlayerAvatar = class('PlayerAvatar', PlayerBase)
local avatars = require("avatars/avatar_loader")
require("player/PlayerList")

local SpeechBubble = require('objects/speech_bubble')


-- Creates the player
function PlayerAvatar:initialize(avatar, colour, id, map_name, tagged, tool, tool_orientation)
    
    PlayerBase.initialize(self, avatar, colour, id, map_name, tagged)

    -- avatar specific stuff (set a default if the avatar was not found)
    self.avatar_data = avatars[avatar]
    if self.avatar_data == nil then self.avatar_data = avatars.fox end

    -- Create graphical elements, providing this remote player is on a map that exists in memory
    if check_map_exists(map_name) then 
        local maze = get_maze(map_name)
        self:create_glyphs(maze)
    else
        self.glyphs = {}    -- must be an empty table, not nil
    end
    
end


function PlayerAvatar:die()
    -- this is a trap - this function should be implemented by derived classes 
    -- and the base version shouldn't be called
    print("ERROR - PlayerAvatar:die() was called")
end

-- used in several function in this file
local tool_glyph_offsets = {
    up = { 
        line_offset     = -0.3,
        column_offset   = 0,
        layer           = -1,                      -- behind player
    },
    down = { 
        line_offset     = 0.3,
        column_offset   = 0,
        layer           = 1,                      
    },
    left = { 
        line_offset     = 0,
        column_offset   = -0.3,
        layer           = 1,                      
    },
    right = { 
        line_offset     = 0,
        column_offset   = 0.3,
        layer           = 1,                      
    },
}

local tool_glyphs = {
    spade = {
        up      = get_glyph("mobs", 11),
        down    = get_glyph("mobs", 12),
        right   = get_glyph("mobs", 13),
        left    = get_glyph("mobs", 14),
    },
    detector = {
        up      = get_glyph("mobs", 43),
        down    = get_glyph("mobs", 44),
        right   = get_glyph("mobs", 45),
        left    = get_glyph("mobs", 46),
    },
    colourator = {
        up      = get_glyph("mobs", 20),
        down    = get_glyph("mobs", 21),
        right   = get_glyph("mobs", 22),
        left    = get_glyph("mobs", 23),
    },
}

function PlayerAvatar:create_glyphs(maze)
    
    self:drop_glyphs()
    
    local function define_glyph(line_offset, column_offset, unicode_point, layer)
        return
        {
            ['line_offset']     = line_offset,
            ['column_offset']   = column_offset,
            ['display_element'] = new_hidden_mob_display_element(maze, unicode_point, 
                                    self.line+line_offset, self.column+column_offset, 
                                    layer, self.colour),
        }
    end
    
    local tool_glyph = 32   -- a space is used by default - could be anything, since it isn't drawn unless self.tool is set
    
    if self.tool then
        if tool_glyphs[self.tool] then
            tool_glyph = tool_glyphs[self.tool][self.tool_orientation]
        end
    end
    
    local tool_glyph_line_offset =      tool_glyph_offsets[self.tool_orientation].line_offset
    local tool_glyph_column_offset =    tool_glyph_offsets[self.tool_orientation].column_offset
    local tool_glyph_layer =            tool_glyph_offsets[self.tool_orientation].layer
    
    ---[[
    self.glyphs.avatar = define_glyph(-self.avatar_data.gr_center_y, -self.avatar_data.gr_center_x, 
                                    self.avatar_data.graphic_unicode_point, 100)
                                
    -- set tag glyphs up as compounds
    self.tagglyphs = {}
    self.tagglyphs.tag1 = self.glyphs.avatar.display_element:new_compound_glyph(get_glyph("map", 0x92), 100, false,-1, 0)
    self.tagglyphs.tag2 = self.glyphs.avatar.display_element:new_compound_glyph(get_glyph("map", 0xa1), 100, false, 0,-1)
    self.tagglyphs.tag3 = self.glyphs.avatar.display_element:new_compound_glyph(get_glyph("map", 0xa2), 100, false, 0, 0)
    self.tagglyphs.tag4 = self.glyphs.avatar.display_element:new_compound_glyph(get_glyph("map", 0xa3), 100, false, 0, 1)
    self.tagglyphs.tag5 = self.glyphs.avatar.display_element:new_compound_glyph(get_glyph("map", 0xb2), 100, false, 1, 0)
    
    --]]
    --[[
    local glyphs = 
    {
        avatar =    define_glyph(-self.avatar_data.gr_center_y, -self.avatar_data.gr_center_x, 
                                    self.avatar_data.graphic_unicode_point, 100),
        tag1 =      define_glyph(-1-self.avatar_data.gr_center_y, -self.avatar_data.gr_center_x, 
                                    get_glyph("map", 0x92), 100),
        tag2 =      define_glyph(-self.avatar_data.gr_center_y, -1-self.avatar_data.gr_center_x, 
                                    get_glyph("map", 0xa1), 100),
        tag3 =      define_glyph(-self.avatar_data.gr_center_y, -self.avatar_data.gr_center_x, 
                                    get_glyph("map", 0xa2), 100),
        tag4 =      define_glyph(-self.avatar_data.gr_center_y, 1-self.avatar_data.gr_center_x, 
                                    get_glyph("map", 0xa3), 100),
        tag5 =      define_glyph(1-self.avatar_data.gr_center_y, -self.avatar_data.gr_center_x, 
                                    get_glyph("map", 0xb2), 100),
        --tool =      define_glyph(   tool_glyph_line_offset-self.avatar_data.gr_center_y, 
        --                            tool_glyph_column_offset-self.avatar_data.gr_center_x, 
        --                            tool_glyph, tool_glyph_layer),
    }
    --]]
    -- load avatar's compound glyphs
    if self.avatar_data.compound_glyphs then
        self.glyphs.avatar.compound_glyphs = {}
        for _,value in ipairs(self.avatar_data.compound_glyphs) do
            self.glyphs.avatar.compound_glyphs[value.name] = glyphs.avatar.display_element:new_compound_glyph
                    (self.avatar_data.compound_animation_set[value.name]['s_st'][1], 
                     self.avatar_data.compound_animation_layer[value.name]['s_st'],
                     self.avatar_data.compound_glyph_visible[value.name], 
                     self.avatar_data.compound_animation_line_offset[value.name]['s_st'],
                     self.avatar_data.compound_animation_column_offset[value.name]['s_st'])
        end
    end
    
                
            
        

    -- create a compound glyph for the tool
    self.tool_compound = self.glyphs.avatar.display_element:new_compound_glyph(tool_glyph, tool_glyph_layer, true, tool_glyph_line_offset, tool_glyph_column_offset)

    --self.glyphs = glyphs
    
    self:set_colour(self.colour)
    
    
    self:show()

end


function PlayerAvatar:set_colour(colour)
    self.colour = colour
    self.glyphs.avatar.display_element:set_fg_colour(colour)
    
    -- better to do colour and orientation by rotating and colouring a grey / white figure, rather than with other glyphs
    --[[
    if self.avatar_data.orientation_glyphs then
        local maze = get_map(self.map_name).auto.maze
        self:change_glyphs(maze, colour, self.orientation)
    end
    --]]
end

function PlayerAvatar:set_orientation(orientation)
    self.orientation = orientation
    
    -- better to do colour and orientation by rotating and colouring a grey / white figure, rather than with other glyphs
    --[[
    if self.avatar_data.orientation_glyphs then
        local maze = get_map(self.map_name).auto.maze
        self:change_glyphs(maze, self.colour, orientation)
    end
    --]]
end


function PlayerAvatar:set_layer(layer)
    
    -- does not work with animation, tools
    
    for _, glyph in pairs(self.glyphs) do
        glyph.display_element:set_layer(layer)
    end
end


function PlayerAvatar:set_compound_glyph_visibilities()
    if self.avatar_data.compound_glyphs then
        for key,value in ipairs(self.avatar_data.compound_glyphs) do
            if self.avatar_data.compound_glyph_visible[value.name] then
                self.glyphs.avatar.display_element:show_compound_glyph(key)
            else
                self.glyphs.avatar.display_element:hide_compound_glyph(key)
            end
        end
    end
end



function PlayerAvatar:move_animation(move_time, mode)
    
    local frame
    
    if mode:sub(2) == '_st' then        -- a stationary mode
        
        local frames = self.avatar_data.animation_length[mode]
        local time_per_frame = (self.avatar_data.time_for_stationary_animation_cycle / frames)
        frame = math.floor((move_time / time_per_frame) % frames) + 1
        
    else
        
        local distance = move_time * self.avatar_data.speed    -- because sqrt(dc^2 + dl^2) always == 1 it works for diagonal too
        local frames = self.avatar_data.animation_length[mode]
        local distance_per_frame = (self.avatar_data.walk_distance_per_animation_cycle / frames)
        frame = math.floor((distance / distance_per_frame) % frames) + 1
        
    end
    
    self.glyphs.avatar.display_element:set_glyph(self.avatar_data.animation_sets[mode][frame])

    if self.avatar_data.compound_glyphs then
        for key,value in ipairs(self.avatar_data.compound_glyphs) do
            
            local name = value.name
            
            if mode:sub(2) == '_st' then
                
                local frames = self.avatar_data.compound_animation_length[name][mode]
                local cycle_time = self.avatar_data.compound_animation_stationary_time[name] or self.avatar_data.time_for_stationary_animation_cycle
                local time_per_frame = (cycle_time / frames)
                frame = math.floor((move_time / time_per_frame) % frames) + 1
                
            else
                
                local distance = move_time * self.avatar_data.speed    -- because sqrt(dc^2 + dl^2) always == 1 it works for diagonal too
                local frames = self.avatar_data.compound_animation_length[name][mode]
                local cycle_distance = self.avatar_data.compound_animation_distance[name] or self.avatar_data.walk_distance_per_animation_cycle
                local distance_per_frame = (cycle_distance / frames)
                frame = math.floor((distance / distance_per_frame) % frames) + 1
                
            end
            
            self.glyphs.avatar.display_element:set_compound_glyph(key, self.avatar_data.compound_animation_set[name][mode][frame])
            
        end
    end

end

function PlayerAvatar:direction_change(new_direction)
    -- time on the current animation direction is reset
    self.animation_time = 0
    
    -- reorder subglyphs based on new direction
    if self.avatar_data.compound_glyphs then
        for key,value in ipairs(self.avatar_data.compound_glyphs) do
            self.glyphs.avatar.display_element:set_compound_glyph_layer(key, self.avatar_data.compound_animation_layer[value.name][new_direction])
            self.glyphs.avatar.display_element:set_compound_glyph_offsets(key, 
                                    self.avatar_data.compound_animation_line_offset[value.name][new_direction],
                                    self.avatar_data.compound_animation_column_offset[value.name][new_direction])
        end
    end
end


function PlayerAvatar:select_animation_frame(dt)
        
    local dl = self.dl
    local dc = self.dc
    local last_dl = self.last_dl
    local last_dc = self.last_dc
    local change_direction = false
    
    -- detect change of direction
    -- if column is changing, then disregard line changes - means we use east/west animation for diagonals
    if dl ~= 0 and dc == 0 then
        if last_dl == 0 or (sign(last_dl) ~= sign(dl)) then
            change_direction = true
        end
    elseif dc ~= 0 then
        if last_dc == 0 or (sign(last_dc) ~= sign(dc)) then
            change_direction = true
        end
    end
    
    -- detect animation mode
    local animation_mode = self.last_animation_mode
    if animation_mode:sub(2) ~= '_st' then
        animation_mode = animation_mode .. '_st'
    end
    
    if dc > 0 then 
        animation_mode = 'e'
    elseif dc < 0 then 
        animation_mode = 'w'
    elseif dl > 0 then 
        animation_mode = 's'
    elseif dl < 0 then 
        animation_mode = 'n'
    end
    
    if change_direction then
        self:direction_change(animation_mode)
    end
    
    -- update move distance - either zeroed by direction change, or not
    self.animation_time = self.animation_time + math.abs(dt)
    
    -- a zero walk distance will display the stationary glyph
    self:move_animation(self.animation_time, animation_mode)

    self.last_dl = dl
    self.last_dc = dc
    self.last_animation_mode = animation_mode
    
end

function PlayerAvatar:shift_tool(tool_orientation)
    
    if not table_is_empty(self.glyphs) then
        local tool_glyph_set = tool_glyphs[self.tool]
        
        if tool_glyph_set == nil then
            -- no glyphs for this tool - hide the compound glyph
            self.glyphs.avatar.display_element:hide_compound_glyph(self.tool_compound)            
        else
            self.glyphs.avatar.display_element:show_compound_glyph(self.tool_compound)            
        
            if tool_orientation == nil then
                -- if no orientation passed in, keep old direction but still run the update
                -- allows for changing tool
                tool_orientation = self.tool_orientation
                self.tool_orientation = nil
            end

            if self.tool_orientation ~= tool_orientation then
                self.tool_orientation = tool_orientation
                --self.glyphs.tool.line_offset = tool_glyph_offsets[tool_orientation].line_offset-self.avatar_data.gr_center_y
                --self.glyphs.tool.column_offset = tool_glyph_offsets[tool_orientation].column_offset-self.avatar_data.gr_center_x
                --self.glyphs.tool.display_element:set_glyph(tool_glyph_set[tool_orientation])
                --self.glyphs.tool.display_element:set_layer(tool_glyph_offsets[tool_orientation].layer)
                
                self.glyphs.avatar.display_element:set_compound_glyph(self.tool_compound, tool_glyph_set[tool_orientation])
                self.glyphs.avatar.display_element:set_compound_glyph_layer(self.tool_compound, tool_glyph_offsets[tool_orientation].layer)
                self.glyphs.avatar.display_element:set_compound_glyph_offsets(self.tool_compound, 
                    tool_glyph_offsets[tool_orientation].line_offset, tool_glyph_offsets[tool_orientation].column_offset)
            end
        end
    end
end

        

-- Hide the user from the screen
function PlayerAvatar:hide()
    for _,v in pairs(self.glyphs) do
        v.display_element:hide()
    end
end


-- Show the user 
function PlayerAvatar:show()
    
    self.glyphs.avatar.display_element:show()
    if self.tagged then
        self:show_tag_marker()
    else
        self:hide_tag_marker()
    end
    --[[
    if tool_glyphs[self.tool] then
        self.glyphs.tool.display_element:show()
    else
        self.glyphs.tool.display_element:hide()
    end
       --]]
end

function PlayerAvatar:hide_tag_marker()
    if not table_is_empty(self.glyphs) then 
        
        
        self.glyphs.avatar.display_element:hide_compound_glyph(self.tagglyphs.tag1)
        self.glyphs.avatar.display_element:hide_compound_glyph(self.tagglyphs.tag2)
        self.glyphs.avatar.display_element:hide_compound_glyph(self.tagglyphs.tag3)
        self.glyphs.avatar.display_element:hide_compound_glyph(self.tagglyphs.tag4)
        self.glyphs.avatar.display_element:hide_compound_glyph(self.tagglyphs.tag5)
        --[[
        self.glyphs.tag1.display_element:hide()
        self.glyphs.tag2.display_element:hide()
        self.glyphs.tag3.display_element:hide()
        self.glyphs.tag4.display_element:hide()
        self.glyphs.tag5.display_element:hide()
        --]]
    end
end

function PlayerAvatar:show_tag_marker()
    -- only show the tag marker if the local player (not necessarily this player) is playing tag
    -- cope with this function being called during the construction of localplayer
    if gulp.localplayer == nil then
        return
    end
    
    if gulp.localplayer.playing_tag and not table_is_empty(self.glyphs) then 
        self.glyphs.avatar.display_element:show_compound_glyph(self.tagglyphs.tag1)
        self.glyphs.avatar.display_element:show_compound_glyph(self.tagglyphs.tag2)
        self.glyphs.avatar.display_element:show_compound_glyph(self.tagglyphs.tag3)
        self.glyphs.avatar.display_element:show_compound_glyph(self.tagglyphs.tag4)
        self.glyphs.avatar.display_element:show_compound_glyph(self.tagglyphs.tag5)
        --[[
        self.glyphs.tag1.display_element:show()
        self.glyphs.tag2.display_element:show()
        self.glyphs.tag3.display_element:show()
        self.glyphs.tag4.display_element:show()
        self.glyphs.tag5.display_element:show()
        --]]
    end
end

-- drop glyphs for this player
function PlayerAvatar:drop_glyphs()
    self.glyphs = {}
    if self.speech_bubble then 
        self.speech_bubble:close() 
        self.speech_bubble = nil
    end
end

-- Remove the user 
function PlayerAvatar:remove()
    self:hide()
    self:drop_glyphs()
    
    PlayerBase.remove(self)
end



function PlayerAvatar:process_tagged(mode)
    
    PlayerBase.process_tagged(self, mode)
    
    if mode == 'tag' then
        self:show_tag_marker()
    else
        self:hide_tag_marker()
    end
end

function PlayerAvatar:process_speech(text)
    if self.speech_bubble then
        self.speech_bubble:close()
        self.speech_bubble = nil
    end
    
    if text ~= "" and self.map_name == gulp.localplayer.map_name then
        self.speech_bubble = SpeechBubble(text, self, 'right', 'top', self.colour)
    end
end        


function PlayerAvatar:update_glyphs_location()
    
    for _,v in pairs(self.glyphs) do
        v.display_element:set_line(self.line + v.line_offset)
        v.display_element:set_column(self.column + v.column_offset)
        
        
    end
    
    if self.speech_bubble then
        --self.speech_bubble:update_glyphs_location(self.line - self.current_line_offset, self.column - self.current_column_offset)
        --self.speech_bubble:update_glyphs_location(self.line, self.column)
        self.speech_bubble:moved()
    end
    
    
end


-- Updates the player postion on the maze   
function PlayerAvatar:processPosition(message)
    
    PlayerBase.processPosition(self, message)
    
    
    -- set layer here, depending on the line the player is on
    
    
    -- We send the player position of all players on a client's new level in 
    -- SendPlayersAndGameStatus() when the client changes map. Therefore, we 
    -- may need to construct a display element here for a client that has just 
    -- appears on our map. 
    
    if self.map_name == gulp.localplayer.map_name then
        if table_is_empty(self.glyphs) or message.switch_map then
            self:switch_glyph_to_map(gulp.localplayer.map_name)
        end
    else
        self:hide()
        self:drop_glyphs()
    end
    
    self:shift_tool()
    self:update_glyphs_location()
    
end  



function PlayerAvatar:switch_glyph_to_map(map_name)
    self:hide()
    local maze = get_maze(map_name)
    self:create_glyphs(maze)
    if self.tagged then
        self:show_tag_marker()
    end
end
    
    

return PlayerAvatar
