-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Client Remote Player
-- Used on client and server sides to display the player's avatars

-- todo avatar display not needed at server, but server remote player is based on this class
-- and some facilities still used at server, like the player list. 

require('strict')
require('utils')
class = require('middleclass')

local PlayerBase = class('PlayerBase')
require("player/PlayerList")
avatars = require('avatars/avatar_loader')

-- Creates the player
function PlayerBase:initialize(avatar, colour, id, map_name, tagged, tool, tool_orientation)
    
    -- Store the player data
    self.avatar = avatar
    self.colour = colour
    self:set_id(id)
    self.map_name = map_name
    self.tagged = tagged
    self.tool = tool
    self.tool_orientation = tool_orientation

    self.visibile_to_mobs = false
    
    -- set some defaults in case the map isn't loaded or doesn;t specify the spawn point
    self.spawn_line = 10.5
    self.spawn_column = 16.5
    
    if check_map_exists(map_name) then 
        local map = get_map(map_name)
        if map.spawn_line and map.spawn_column then 
            self.spawn_line = map.spawn_line
            self.spawn_column = map.spawn_column
        end
    end
    
    self.line = self.spawn_line
    self.column = self.spawn_column

    -- start with tool facing downwards
    if not self.tool_orientation then
        self.tool_orientation = 'down'
    end
   
    -- register for collisions if necessary
    self.collision_rect = deep_copy(avatars[self.avatar].collision_rect) or {size={line=1/4,column=1/2}} -- default size
    
    -- adjust avatar data for players - mobs and players have different ways of positioning... WHY?
    self.collision_rect.centre = {line = 0.4, column = 0} 
    
    register_for_collisions("Player", self, self.collision_rect)
end

function PlayerBase:set_id(id)
    
    if self.id ~= nil then
        remove_from_playerlist(self)
    end
        
    self.id = id
    
    if self.id ~= nil then
        add_to_playerlist(self)
    end
    
end    

function PlayerBase:die()
    -- this is a trap - this function should be implemented by derived classes 
    -- and the base version shouldn't be called
    print("ERROR - PlayerBase:die() was called")
end

-- Remove the user 
function PlayerBase:remove()
    deregister_for_collisions("Player")
    remove_from_playerlist(self)
end



function PlayerBase:process_tagged(mode)
    if mode == 'tag' then
        self.tagged = true
    else
        self.tagged = false
    end
end



-- Updates the player postion on the maze   
function PlayerBase:processPosition(message)
    --print("PlayerBase:processPosition")
    
    self.line = message.line
    self.column = message.column
    self.map_name = message.map_name
    self.tagged = message.tagged
    self.tool = message.tool
    self.tool_orientation = message.tool_orientation  
    
    if message.spawn_line ~= nil then
        self.spawn_line = message.spawn_line
        self.spawn_column = message.spawn_column
        print("PlayerBase:processPosition()", self.spawn_line, self.spawn_column, self.map_name)
    end
    
end  

function PlayerBase:form_position_message(mode)
    local message = { 
        mt = "player_move",
        column = self.column,
        line = self.line,
        map_name = self.map_name,
        id = self.id,
        tagged = self.tagged,
        tool = self.tool,
        tool_orientation = self.tool_orientation,
    }
    
    if mode == 'broadcast' then
        message.always_process_at_client = true
    
    elseif mode == 'switch_map' then
        message.always_process_at_client = true
        message.switch_map = true
        message.spawn_line = self.new_spawn_line
        message.spawn_column = self.new_spawn_column
        debugMessage("PlayerBase:form_position_message()", message.spawn_line, message.spawn_column, message.map_name)
    
    elseif mode == 'reset' then
        message.always_process_at_client = true
        message.reset = true
    
    end
    
    return message
end


function PlayerBase:broadcast_position(mode)
    -- send all other players our position
    -- used when a player changes map, etc
    local message
    
    mode = mode or 'broadcast'
    
    if gulp.LuaStateCodeMode == 'CLIENT' then
        -- just tell the server
        message = self:form_position_message(mode)
        gulp_networking:SendToServer(message)
        gulp_networking:UpdateLastSentPosition(self)
    else
        for _, player in all_players_except(self) do
            if player.socket_ref then 
                message = self:form_position_message(mode)
                player.socket_ref:SendMessage(message);
            end
        end
    end
end    



return PlayerBase
