-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Player List
-- Contains a list of all the players (local and remote) and
-- some methods for accessing the lits



require('strict')
class = require('middleclass')


-- put guards on adding to this list so only the types this code knows about get added to it
local PlayerList = {}

function add_to_playerlist(player)
    --table.insert(PlayerList, player)
    if PlayerList[player.id] == nil then
        PlayerList[player.id] = player
        --print("Inserted player " .. player.id .. " into PlayerList")
    else
        print("ERROR: player " .. player.id .. " is already in PlayerList")
    end
    --display_playerlist()
end

function remove_from_playerlist(player)
    if player.id then
        PlayerList[player.id] = nil
        --display_playerlist()
    end
end
 
function get_player_by_id(id)
    if id and PlayerList[id] then
        return PlayerList[id]
    end
end

function does_player_exist(id)
    if id and PlayerList[id] then
        return true
    else
        return false
    end
end


--[[

The operation of the following functions is a bit convoluted, so here is an explanation.

find_nearby() is a utility function which takes 4 args
1) location to search nearby to
2) a list of objects / locations to check if they are nearby the search location
3) the maximum distance away to count as 'nearby'
4) the iterator which tells find_nearby() how to step through the list of targets in the second argument

Usually for an iterator, you pass in the thing to be iterated. This is what find_nearby() does,
it passes arg 2 (the list to be iterated) into arg 4 (the iterator)

For the player list though, the thing to be iterated is implicit in the iterator used (it always
iterates PlayerList) and the arg to the iterator is used to control how the iteration proceeds, by
potentially excluding certain retults.

So in our functions below where we are using PlayerList iterators as args to find_nearby(), we don't
pass the list of things to find, we instead pass what we want to exclude from the searches - nothing
in the case of find_nearby_players() and the player who serves as the search location in the case
of find_other_nearby_players()

--]]
function find_nearby_players(location, distance)
    return find_nearby(location, nil, distance, all_players)
end

function find_other_nearby_players(player, distance)
    return find_nearby(player, player, distance, all_players_except)
end


----------------------------------------
-- wrappers for iterators on PlayerList

function all_players()
    -- no need for iterator - uses next directly
    return next, PlayerList, nil
end

function all_players_except(player)
    return _all_players_except_iterator, player, nil
end

function all_players_on_map(map_name)
    return _all_players_on_map_iterator, map_name, nil
end

function all_players_on_map_except(map_name, player)
    return _all_players_on_map_except_iterator, {map_name, player}, nil
end

function all_remote_players()
    return _all_remote_players_iterator, nil, nil
end

function all_remote_players_except(player)
    return _all_remote_players_except_iterator, player, nil
end

function all_remote_players_on_map(map_name)
    return _all_remote_players_on_map_iterator, map_name, nil
end

function all_remote_players_on_map_except(map_name, player)
    return _all_remote_players_on_map_except_iterator, {map_name, player}, nil
end


---------------------------------------------
-- The iterators themselves

function _all_players_except_iterator(player, i)
    i = next(PlayerList, i)
    while i and (PlayerList[i] == player) do
        -- we are the excluded player
        i = next(PlayerList, i)
    end
    
    if i then
        return i, PlayerList[i]
    end
end

function _all_players_on_map_iterator(map_name, i)
    i = next(PlayerList, i)
    while i and  PlayerList[i].map_name ~= map_name do
        -- we are not on the correct map
        i = next(PlayerList, i)
    end
    
    if i then
        return i, PlayerList[i]
    end
end



function _all_remote_players_iterator(_, i)
    i = next(PlayerList, i)
    while i and PlayerList[i]:isInstanceOfByName("Player") do
        -- we are the local player
        i = next(PlayerList, i)
    end
    
    if i then
        return i, PlayerList[i]
    end
end

function _all_remote_players_except_iterator(player, i)
    i = next(PlayerList, i)
    while i and (PlayerList[i]:isInstanceOfByName("Player") or PlayerList[i] == player) do
        -- we are the local player or we are the excluded player
        i = next(PlayerList, i)
    end
    
    if i then
        return i, PlayerList[i]
    end
end

function _all_remote_players_on_map_iterator(map_name, i)
    i = next(PlayerList, i)
    while i and (PlayerList[i]:isInstanceOfByName("Player") or PlayerList[i].map_name ~= map_name) do
        -- we are the local player, or we are not on the correct map
        i = next(PlayerList, i)
    end
    
    if i then
        return i, PlayerList[i]
    end
end


function _all_players_on_map_except_iterator(data, i)
    i = next(PlayerList, i)
    while i and (PlayerList[i].map_name ~= data[1] or PlayerList[i] == data[2]) do
        -- or we are not on the correct map, or we are the excluded player
        i = next(PlayerList, i)
    end
    
    if i then
        return i, PlayerList[i]
    end
end

function _all_remote_players_on_map_except_iterator(data, i)
    i = next(PlayerList, i)
    while i and (PlayerList[i]:isInstanceOfByName("Player") or 
                 PlayerList[i].map_name ~= data[1] or 
                 PlayerList[i] == data[2]) do
        -- we are the local player, or we are not on the correct map, or we are the excluded player
        i = next(PlayerList, i)
    end
    
    if i then
        return i, PlayerList[i]
    end
end


--------------------------------------
-- debug

function display_playerlist()
    for k, v in all_players() do
        print("Player ", k, " = ", v, v.avatar)
    end
end

function display_playerlist_ex(map_name, player)
    
    print("ALL PLAYERS")
    for k, v in all_players() do
        print("  Player ", k, " = ", v)
    end
    
    print("ALL PLAYERS ON "..map_name)
    for k, v in all_players_on_map(map_name) do
        print("  Player ", k, " = ", v)
    end
    
    print("ALL REMOTE PLAYERS")
    for k, v in all_remote_players() do
        print("  Player ", k, " = ", v)
    end
    
    print("ALL REMOTE PLAYERS EXCEPT "..tostring(player))
    for k, v in all_remote_players_except(player) do
        print("  Player ", k, " = ", v)
    end
    
    print("ALL REMOTE PLAYERS ON "..map_name)
    for k, v in all_remote_players_on_map(map_name) do
        print("  Player ", k, " = ", v)
    end
    
    print("ALL PLAYERS ON "..map_name.." EXCEPT "..tostring(player))
    for k, v in all_players_on_map_except(map_name, player) do
        print("  Player ", k, " = ", v)
    end
    
    print("ALL REMOTE PLAYERS ON "..map_name.." EXCEPT "..tostring(player))
    for k, v in all_remote_players_on_map_except(map_name,player) do
        print("  Player ", k, " = ", v)
    end
    
    print(" ")
end


