-- Forlorn Fox
-- Copyright © 2015 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

require('utils')

--[[

    CARVE PATHS FROM NUMBERS
    
    The procedural generation code converts all numbers to spaces after generation. 
    Carving paths from numbers allows paths to avoid each other during generation, 
    or specifically join, or whatever.

    For this to work nicely, path generation should be parallel, though. So we use
    types such as 'maze' and 'last maze', where 'maze' just adds the path details
    to a list, and 'last_maze' generates all the listed paths together.

--]]

local function generate_direct_path(level, path_number, start, finish)
    if start.c > finish.c then start, finish = finish, start end
    
    local l, c, adjust
    
    if start.l > finish.l then adjust = -1 else adjust = 1 end
    l = start.l
    
    for c = start.c,finish.c do
        level[l][c] = path_number
        
        if l ~= finish.l then 
            l = l + adjust
            level[l][c] = path_number
        end
    end
end

local directions = {
        { line = -1, column =  0 },      -- up
        { line =  1, column =  0 },      -- down
        { line =  0, column = -1 },      -- left
        { line =  0, column =  1 },      -- right
    }

local alldirections = {
        { line = -1, column =  0 },      -- up
        { line =  1, column =  0 },      -- down
        { line =  0, column = -1 },      -- left
        { line =  0, column =  1 },      -- right
        { line = -1, column = -1 },      -- 
        { line =  1, column =  1 },      -- 
        { line = -1, column =  1 },      -- 
        { line =  1, column = -1 },      -- 
    }

local path_current_list = {}
local path_finished = {}
local path_history = {}
local path_joined = {}
local path_min = {}
local path_max = {}


local function list_maze_path(path_number, start, finish, path)
    
    path_current_list[path_number] = start
    path_finished[path_number] = false
    path_history[path_number] = {}
    
    path_current_list[path_number+300] = finish
    path_finished[path_number+300] = false
    path_history[path_number+300] = {}
    
    path_joined[path_number] = false
    
    path_min[path_number] = path.min
    path_min[path_number+300] = path.min
    path_max[path_number] = path.max
    path_max[path_number+300] = path.max
end


local required = {}

local function get_valid_directions(path_number, level)
    
    
    
    local function allow_growth(path_number, line, column)
        
        if line < path_min[path_number].l then return false end
        if line > path_max[path_number].l then return false end
        if column < path_min[path_number].c then return false end
        if column > path_max[path_number].c then return false end
        
        local cell = level[line][column]
        
        -- if the next cell on is blank, we are fine
        if cell == 'blank' then return true end 
        
        -- if the cell is a wall in a predefined chunk, we can go next to it
        if cell == '#' then return true end 
        
        -- if the path is not joined to it's counterpart, then join it
        local join_counterpart = false
        if path_number < 300 then
            if cell == path_number + 300 then join_counterpart = true end
        else
            if cell == path_number - 300 then join_counterpart = true end
            path_number = path_number - 300
        end
        
        if join_counterpart then
            if not path_joined[path_number] then
                required = {l=line,c=column}
                return true 
            else
                return percent_chance(10)
            end
        end
        
        if cell == path_number then return percent_chance(10) end

        
        
        -- if the cell is the destination, we definitely want to allow growth!
        -- make sure the destination cells are surrounded by '#' so the surrounding cells don't prevent growth
        --if path_end_list[path_number].l == line and path_end_list[path_number].c == column then return true end
        
        -- allow paths to loop, rarely
        -- this really is very rare, because although paths will join end to end 10% of the time,
        -- if there are two cells occupied by path to be jon it is only 1% successful, and if all 
        -- 3 cells to be checked are occupied by the other path, we will only join 0.1% of the time...
        -- this could be handled better!
        --if cell == path_number or cell == path_number+1000 then return percent_chance(10) end
        
        -- allow paths to join another path, rarely
        --if type(cell) == 'number' then return percent_chance(10) end
        
        return false
        
    end
    
    
    
    -- check the 4 directions from the current location
    local line = path_current_list[path_number].l
    local column = path_current_list[path_number].c
    
    local allowed_directions = {}
    
    -- we can only grow to a blank location
    for index,direction in ipairs(directions) do
        allowed_directions[index] = (level[line+direction.line][column+direction.column] == 'blank')
    end
    
    -- now check the blank locations to see if we are growing too near to something
    for index,allowed in ipairs(allowed_directions) do
        if allowed then
            
            -- two ahead lookup
            -- generates mazes without touching corners
            -- further checking of all 3 elements in advance of this, direct ahead, and the corners ahead
            -- if we find something we can join directly ahead, allow this
            -- if we find something we cannot join ahead, then disallow this
            --[[
            if directions[index].line == 0 then
                -- direction is horizontal
                local c = column + 2*directions[index].column
                if not allow_growth(path_number, line, c) then
                    allowed_directions[index] = false
                elseif not allow_growth(path_number, line+1, c) or not allow_growth(path_number, line-1, c) then
                    allowed_directions[index] = false
                end
            else
                -- direction is vertical
                local l = line + 2*directions[index].line
                if not allow_growth(path_number, l, column) then
                    allowed_directions[index] = false
                elseif not allow_growth(path_number, l, column+1) or not allow_growth(path_number, l, column-1) then
                    allowed_directions[index] = false
                end
            end
            --]]
            -- one ahead lookup
            -- allows mazes to touch corners
            -- need to check all 3 directions from candidate to see if we join something illegal, because
            -- we don't check the corners as we advance
                        
            if directions[index].line == 0 then
                -- direction is horizontal
                local column_mod = directions[index].column
                if not allow_growth(path_number, line, column + 2*column_mod) then
                    allowed_directions[index] = false
                elseif not allow_growth(path_number, line+1, column + column_mod) or not allow_growth(path_number, line-1, column + column_mod) then
                    allowed_directions[index] = false
                end
            else
                -- direction is vertical
                local line_mod = directions[index].line
                if not allow_growth(path_number, line + 2*line_mod, column) then
                    allowed_directions[index] = false
                elseif not allow_growth(path_number, line + line_mod, column+1) or not allow_growth(path_number, line + line_mod, column-1) then
                    allowed_directions[index] = false
                end
            end
        
        
        
        end
    end
   
    return allowed_directions
end
    
    
    

local function generate_maze_paths(level)
    
    local function paths_growing()
        for _, finished in pairs(path_finished) do
            if not finished then return true end
        end
        return false
    end
    
    
    while paths_growing() do
    
        for path_number, specification in pairs(path_current_list) do
        
            if not path_finished[path_number] then
                
                -- grow path in random, valid direction
                -- generate list of valid directions
        
                -- need to make sure we don't cut off other paths from their exits, or allow path crossing
                -- joining another path terminates the branch, but allows other branches to carry on growing
                -- if path is joined to other path and other path reaches our exit, then we are finished
                
                -- could be fractal if we generate a junction whenever possible, instead of only after 
                -- backtracking when we get stuck
                
                local allowed_directions = get_valid_directions(path_number, level)
                local line = path_current_list[path_number].l
                local column = path_current_list[path_number].c
                
                -- pick a random entry from the allowed directions
                -- there must be a better way to do this :-)
                local allowed_count = 0
                for x = 1,4 do
                    if allowed_directions[x] ~= false then
                        allowed_count = allowed_count + 1
                    end
                end
                

                if allowed_count == 0 then
                    -- backtrack and try again elsewhere
                    -- mark as not growable
                    level[line][column] = level[line][column] + 1000 
                    
                    
                    if #path_history[path_number] == 0 then
                        path_finished[path_number] = true
                    else
                        path_current_list[path_number] = table.remove(path_history[path_number])
                    end
                    
                    --[[
                    -- find an adjacent cell with the path number, set current location to that
                    -- flaky method, should pick one at random, although often there will be only 1 choice anyway
                    local found = false
                    for x = 1,4 do
                        if level[line+directions[x].line][column+directions[x].column] == path_number then
                            path_current_list[path_number].l = line+directions[x].line
                            path_current_list[path_number].c = column+directions[x].column
                            found = true
                        end
                    end
                    if not found then  end
                    --]]
                    
                
                else
                    
                    if #required == 0 then
                        local direction
                        local r = math.random(allowed_count)
                        allowed_count = 0
                        for x = 1,4 do
                            if allowed_directions[x] ~= false then
                                allowed_count = allowed_count + 1
                                if allowed_count == r then
                                    direction = x
                                end
                            end
                        end
                
                        line = line+directions[direction].line
                        column = column+directions[direction].column 
                    else
                        line = required.l
                        column = required.c
                    end
                    
                    -- actually carve a path
                    level[line][column] = path_number
                    
                    table.insert(path_history[path_number], {l=line, c=column})
                    
                    --if allowed_directions[x] == 'extend' then
                    --    level[line+directions[direction].line*2][column+directions[direction].column*2] = path_number
                    --end
                    
                    path_current_list[path_number].l = line
                    path_current_list[path_number].c = column
                    
                end
                
                
                -- after growth, check for finished status
                -- check the finish location, is any path adjacent?
                -- if this path is adjacent, then we are finished, OR if any path this path touches is adjacent, 
                -- we are finished
            end
        end
    end
end

                
                
        
    

function generate_connections(level, map)
    
    -- for each path...
    for path_number,path in ipairs(map.procedural.paths) do
        
        -- find the two ends of the path in the map
        local l, c
        local start, finish
        
        for l,line_data in ipairs(level) do
            for c, data in ipairs(line_data) do
                if data == path.letter then
                    if start == nil then
                        start = {l=l,c=c}
                    else
                        finish = {l=l,c=c}
                    end
                end
            end
        end
    
        if path.path_type == 'direct' then
            generate_direct_path(level, path_number, start, finish)
        elseif path.path_type == 'maze' then
            list_maze_path(path_number, start, finish, path)
        elseif path.path_type == 'last_maze' then
            list_maze_path(path_number, start, finish, path)
            generate_maze_paths(level)
        end
    
    end
     
end


