-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
require('strict')

local playback_frequency = gulp.sound.manager:get_sample_frequency()

local function _sine_wave_flatten_backtrack(sound)
    if #sound == 0 then return end
    
    local last_val = sound[#sound]
    -- unlikely but test
    if last_val == 0 then return end
    
    if last_val > 0 then
        -- look for the first non-positive sample, i.e. the half wave start
        for i = #sound, 1, -1 do
            if sound[i] > 0 then
                sound[i] = 0
            else
                break
            end
        end
    else
        -- look for the first non-negative sample, i.e. the half wave start
        for i = #sound, 1, -1 do
            if sound[i] < 0 then
                sound[i] = 0
            else
                break
            end
        end
    end
end


local function create_sine_wave(frequency, duration, volume)
    local sound = {}
    local number_samples = playback_frequency * duration
    
    -- increment in playback samples, e.g. 0.00997732 for 440Hz, 0.00004535 for 2Hz, 0.25 for 11025Hz
    -- once this accumulates 1, then we have a cycle.
    local increment = frequency / playback_frequency;
    
    -- convert from cycles to radians
    increment = 2 * math.pi * increment;
    
    local radians = 0
    
    for i = 1, number_samples do
        sound[i] = volume * math.sin(radians)
        radians = radians + increment
    end
    
    _sine_wave_flatten_backtrack(sound)
    return sound
end


-- Flatten the end wave so that it doesn't appear as a high frequency spike at the end
-- Very short (i.e. half wave) sounds will be silenced.
--
-- The square wave functions finish at a random point through the waveform ...
-- this is likely to result in a glitch at the end.
-- This function fixes that.
local function _square_wave_flatten_backtrack(sound)
    if #sound == 0 then return end
    
    local last_val = sound[#sound]
    
    for i = #sound, 1, -1 do
        if last_val == sound[i] then
            sound[i] = 0
        else
            break
        end
    end
end

-- frequency in Hz
-- duration in seconds
-- volume from 0 to 1
local function create_square_wave(frequency, duration, volume)
    local sound = {}
    local number_samples = playback_frequency * duration
    
    -- increment in playback samples, e.g. 0.00997732 for 440Hz, 0.00004535 for 2Hz, 0.25 for 11025Hz
    local increment = frequency / playback_frequency;
    -- we want it to toggle twice as often for high and low... because half waveform is twice frequency
    increment = increment * 2
    
    local counter = 0
    local level = -volume
    
    for i = 1, number_samples do
        if counter > 1 then
            level = -level
            counter = counter - 1
        end        
        sound[i] = level
        counter = counter + increment
    end
    
    _square_wave_flatten_backtrack(sound)
    return sound
end


local function create_silence(duration_in_seconds)
    local sound = {}
    local number_samples = playback_frequency * duration_in_seconds
    for i = 1, number_samples do
        sound[i] = 0
    end
    return sound
end


local function append_sound(t1, t2)
    for i = 1, #t2 do
        t1[#t1+1] = t2[i]
    end
    return t1
end

local function copy_sound(t)
    local newt = {}
    for k,v in ipairs(t) do
        newt[k] = v
    end
    return newt
end

local function _normal_clip(value)
    if value > 1 then value = 1 end
    if value < 0 then value = 0 end
    return value
end


function add_ADSR_envelope(sound, attack_time_percentage, decay_time_percentage, sustain_level_percentage, release_time_percentage)
    -- ADSR is usually used in Sythesizers for hitting keys, so it's use here is slightly artificial. 
    -- http://en.wikipedia.org/wiki/Synthesizer#ADSR_envelope
    --
    -- amplitude
    --     ^
    --     |<-A-><D->   <R->
    --     |    /\  |   |  |
    --     |   /  \ |   |  |
    --     |  /    \|___|  |
    --     | /        ^  \ |
    --     |/       S |   \
    --     +---------------------> t
    
    -- normalise to 0-1
    local attack = attack_time_percentage / 100
    local decay = decay_time_percentage / 100
    local sustain_level = sustain_level_percentage /100
    local release = release_time_percentage / 100   

    attack = _normal_clip(attack)   
    decay = _normal_clip(decay)
    release = _normal_clip(release) 
    sustain_level = _normal_clip(sustain_level)

    -- what if these three add up to more than 100%?
    local sum = attack + decay + release
    if sum > 1 then
        -- assume sustain = 0
        attack = attack / sum
        decay = decay / sum
        release = release / sum
    end
    
    local sustain = 1 - (attack + decay + release)
    
    local sustain_start = attack + decay
    local release_start = sustain_start + sustain
    
    --local f = io.open("ADSR_output.txt", "w")
    local len = #sound
    for k,v in ipairs(sound) do
        local time_fraction = k / len
        local volume_fraction
        --local vtype
        if time_fraction > release_start then
            -- release goes from sustain down to zero (as time increases)
            volume_fraction = 1 - ( (time_fraction - release_start) / release)
            volume_fraction = sustain_level * volume_fraction
            --vtype = "release"
            
        elseif time_fraction > sustain_start then
            -- sustain is a constant reduction
            volume_fraction = sustain_level
            --vtype = "sustain"
            
        elseif time_fraction > attack then
            -- must be decay
            -- decay goes from max to sustain as time increases
            volume_fraction = 1 - ( (time_fraction - attack) / decay)
            volume_fraction = ((1 - sustain_level) * volume_fraction)
            volume_fraction = volume_fraction + sustain_level
            --vtype = "decay"
            
        else
            -- must be attack
            -- attack goes from 0 to max as time increases.
            volume_fraction = time_fraction / attack
            --vtype = "attack"
        end
        
        sound[k] = v * volume_fraction
        --f:write(string.format("%d %.2f %s\n", k, volume_fraction, vtype))
    end
    --f:close()
end

local function make_extra_sounds()
    --gulp.sound.manager:play_sound(gulp.sound.tink)
    --local duration_in_seconds = 1.0
    --local square_table = create_square_wave(440, 0.95, 0.25)
    --gulp.sound.new_tink = gulp.sound.manager:add_mono_sound(square_table)
    --gulp.sound.manager:play_sound(gulp.sound.new_tink)
    
    --local square_table2 = create_square_wave(1440, 0.5, 0.25)
    --gulp.sound.new_tink2 = gulp.sound.manager:add_stereo_sound(square_table, square_table2)
    --gulp.sound.manager:play_sound(gulp.sound.new_tink2)
    
    if false then
        local square_table = create_square_wave(440, 0.5, 0.25)
        local silence = create_silence(0.5)
        local left = {}
        append_sound(left, square_table)
        append_sound(left, silence)
        append_sound(left, square_table)
        append_sound(left, silence)

        local right = {}
        append_sound(right, silence)
        append_sound(right, square_table)
        append_sound(right, silence)
        append_sound(right, square_table)
        
        gulp.sound.new_tink3 = gulp.sound.manager:add_stereo_sound(right, left)
        gulp.sound.manager:play_sound(gulp.sound.new_tink3) 
    end

    gulp.sound.new_tink = {}
    local sine_table = create_sine_wave(2000, 0.02, 1)
    add_ADSR_envelope(sine_table, 10,90,0,0)
    --add_ADSR_envelope(sine_table, 25,25,50,25)
    gulp.sound.new_tink[1] = gulp.sound.manager:add_mono_sound(sine_table)
    --gulp.sound.manager:play_sound(gulp.sound.new_tink1)

    local sine_table = create_sine_wave(2100, 0.02, 1)
    add_ADSR_envelope(sine_table, 10,90,0,0)
    --add_ADSR_envelope(sine_table, 25,25,50,25)
    gulp.sound.new_tink[2] = gulp.sound.manager:add_mono_sound(sine_table)

    local sine_table = create_sine_wave(2200, 0.02, 1)
    add_ADSR_envelope(sine_table, 10,90,0,0)
    --add_ADSR_envelope(sine_table, 25,25,50,25)
    gulp.sound.new_tink[3] = gulp.sound.manager:add_mono_sound(sine_table)

    gulp.sound.detector_ping = {}
    local sine_table = create_sine_wave(1000, 0.02, 1)
    add_ADSR_envelope(sine_table, 10,90,0,0)
    --add_ADSR_envelope(sine_table, 25,25,50,25)
    gulp.sound.detector_ping[1] = gulp.sound.manager:add_mono_sound(sine_table)

    local sine_table = create_sine_wave(1300, 0.04, 1)
    add_ADSR_envelope(sine_table, 10,90,0,0)
    --add_ADSR_envelope(sine_table, 25,25,50,25)
    gulp.sound.detector_ping[2] = gulp.sound.manager:add_mono_sound(sine_table)

    local sine_table = create_sine_wave(1600, 0.06, 1)
    add_ADSR_envelope(sine_table, 10,90,0,0)
    --add_ADSR_envelope(sine_table, 25,25,50,25)
    gulp.sound.detector_ping[3] = gulp.sound.manager:add_mono_sound(sine_table)

    local sine_table = create_sine_wave(1900, 0.08, 1)
    add_ADSR_envelope(sine_table, 10,90,0,0)
    --add_ADSR_envelope(sine_table, 25,25,50,25)
    gulp.sound.detector_ping[4] = gulp.sound.manager:add_mono_sound(sine_table)


    gulp.sound.play_tink = function()
        play_sound(gulp.sound.new_tink[math.random(3)], 6)
    end
    --[[
    gulp.sound.play_local_tink = function()
        play_sound_local_only(gulp.sound.new_tink[math.random(3)], 6)
    end
    --]]
    gulp.sound.play_tink()

    -- volume test code
    ---[[
    gulp.sound.sonar_volume = 8
    gulp.sound.sonar = function(command)
        if command == 'reset' then gulp.sound.sonar_volume = 8 end
        play_sound(gulp.sound.new_tink[1], gulp.sound.sonar_volume)
        gulp.sound.sonar_volume = gulp.sound.sonar_volume -1
        if gulp.sound.sonar_volume < 1 then gulp.sound.sonar_volume = 8 end
    end
    --[[
    while true do
        gulp.sound.sound_test()
        local x=1
        while x<25000 do x = x + 1 end
    end
    --]]
    
end

function attenuate_sound(volume, sound_line, sound_column)
    if sound_line ~= nil and gulp.localplayer ~= nil then    
        
        -- calculate distance
        local vd = math.abs(gulp.localplayer.line - sound_line)
        local hd = math.abs(gulp.localplayer.column - sound_column)
        local d = math.sqrt((vd*vd) + (hd*hd))
        
        -- attentuate sound
        while d > 3 do
            volume = volume - 1
            d = d / 2
        end
       
        -- always hear something
        if volume < 1 then volume = 1 end
    end
    
    return volume
end



-- in future, we should remove play_sound
-- all sounds should be named

function play_sound(sound, volume)
    gulp.sound.manager:play_sound(sound, volume)
end

function play_named_sound(sound, volume)
    if gulp.sound[sound] ~= nil then
        gulp.sound.manager:play_sound(gulp.sound[sound], volume)
    end
end

function transmit_located_sound(sound, volume, line, column, map)
    
    local message = {
        mt = 'play_sound',
        sound = sound,
        volume = volume,
        line = line,
        column = column,
        map = map,
    }
    
    gulp_networking:SendToServer(message)
    
end

function transmit_named_located_sound(sound, volume, line, column, map)
    
    local message = {
        mt = 'play_named_sound',
        sound = sound,
        volume = volume,
        line = line,
        column = column,
        map = map,
    }
    
    gulp_networking:SendToServer(message)
    
end


-- alias for located sound, when you want to use it for a non-located sound!
transmit_sound = transmit_located_sound -- (sound, volume)

function play_and_transmit_sound(sound, volume)
    play_sound(sound, volume)
    transmit_sound(sound, volume)
end

function play_located_sound(sound, volume, sound_line, sound_column, sound_map)
    -- don't play sounds for different maps
    if sound_map ~= gulp.localplayer.map_name then return end
    
    volume = attenuate_sound(volume, sound_line, sound_column)
    
    play_sound(sound,volume)
end

function play_named_located_sound(sound, volume, sound_line, sound_column, sound_map)
    -- don't play sounds for different maps
    if sound_map ~= gulp.localplayer.map_name then return end
    
    volume = attenuate_sound(volume, sound_line, sound_column)
    
    play_named_sound(sound,volume)
end


function play_and_transmit_located_sound(sound, volume, sound_line, sound_column, sound_map)
    play_located_sound(sound, volume, sound_line, sound_column, sound_map)
    transmit_located_sound(sound, volume, sound_line, sound_column, sound_map)
end


--[[
-- for menus and stuff :-)
function play_sound_local_only(sound, volume)
    gulp.sound.manager:play_sound(sound, volume)
end


function play_sound(sound, volume, sound_line, sound_column, sound_map, mode)
    
    -- If we have been given a position for this sound and have a localplayer,
    -- adjust the volume depending on the distance between the sound and the player.
    -- We could adjust this based on what is between the player and the sound on the map,
    -- e.g. walls attenuate sound. But for now, lets not :-)
        
        
    if sound_line ~= nil and gulp.localplayer ~= nil then    
        
        -- don't play sounds for different maps
        if sound_map ~= gulp.localplayer.map_name then return end
        
        -- calculate distance
        local vd = math.abs(gulp.localplayer.line - sound_line)
        local hd = math.abs(gulp.localplayer.column - sound_column)
        local d = math.sqrt((vd*vd) + (hd*hd))
        
        -- attentuate sound
        while d > 3 do
            volume = volume - 1
            d = d / 2
        end
       
        -- always hear something
        if volume < 1 then volume = 1 end
    end
    
    -- actually play the sound!
    gulp.sound.manager:play_sound(sound, volume)
    
    
    
    -- also need to transmit this sound across the network, providing this
    -- function was not called in response to a network request
    if mode ~= 'local' and gulp.LuaStateCodeMode then
    
        if sound_line == nil and gulp.localplayer ~= nil then
            -- this is a player sound
            -- add the player location into the message
            sound_line = gulp.localplayer.line
            sound_column =  gulp.localplayer.column
            sound_map = gulp.localplayer.map_name
        end
    
        local message = {
            mt = 'play_sound',
            sound = sound,
            volume = volume,
            line = sound_line,
            column = sound_column,
            map = sound_map,
        }
        
        if gulp.LuaStateCodeMode == 'SERVER' then
            if sound_map == nil then
                gulp_networking:broadcast(message)
            else
                gulp_networking:broadcast_to_all_on_map(sound_map, message)
            end
        else
            gulp_networking:SendToServer(message)
        end
    end
    
end
--]]
function load_sounds()
    --
    -- load our noises ... if they are loaded from file
    -- otherwise synthesise them
    --

    local normal_volume_event = 60
    local loud_volume_event = 120

    local m = gulp.sound.manager
    local s = gulp.sound
    local LoadPath = gulp_cpp.LoadPath

    
    gulp.sound.crunch_explosion = m:load_sound(LoadPath("sounds/Explosion_Crunch.wav"):str())
    
    gulp.sound.eb_eat1 = m:load_sound(LoadPath("sounds/eb_eating.wav"):c_str())
    gulp.sound.eb_eat2 = m:load_sound(LoadPath("sounds/eb_eating2.wav"):c_str())
    gulp.sound.eb_first_spot = m:load_sound(LoadPath("sounds/eb_first_spot.wav"):c_str())
    gulp.sound.eb_come_back1 = m:load_sound(LoadPath("sounds/eb_come_back.wav"):c_str())
    gulp.sound.eb_come_back2 = m:load_sound(LoadPath("sounds/eb_come_back2.wav"):c_str())
    gulp.sound.eb_gonna_eat = m:load_sound(LoadPath("sounds/eb_gonna_eat.wav"):c_str())

    local tink_table = create_square_wave(2000, 0.1, 0.25)
    gulp.sound.tink = gulp.sound.manager:add_mono_sound(tink_table)

    --[[
    crunch_explosion,
    eb_eat1,
    eb_eat2,
    eb_first_spot,
    eb_come_back1,
    eb_come_back2,
    eb_gonna_eat,
    tink,
    --]]
    
    -- Create room sounds 
    gulp.sound.thump = m:load_sound(LoadPath("sounds/thump.wav"):str())
    
    make_extra_sounds()

    -- make_all_notes()     -- takes ages
    --load_all_tracks()       -- alternate method also takes ages :-(
    --play_track(1)           -- for testing - means we've loaded the song
end


--
-- Music stuff
--
local notes_frequency = {
    [0] = 261.626, -- middle C (C4 on a piano)
    [1] = 277.183, -- C#/Db
    [2] = 293.665, -- D
    [3] = 311.127, -- D#/Eb
    [4] = 329.628, -- E
    [5] = 349.228, -- F
    [6] = 369.994, -- F#/Gb
    [7] = 391.995, -- G
    [8] = 415.305, -- G#/Ab
    [9] = 440.000, -- A
    [10] = 466.164, -- A#/Bb
    [11] = 493.883, -- B
    [12] = 523.251, -- C above middle C (C5 on a piano)
}

local notes_table = {}

local music_volume = 50
function make_note(note_number, duration)
    if not notes_table[duration] then
        notes_table[duration] = {}
    end
    local these_notes = notes_table[duration]
    if not these_notes[note_number] then
        local subnote = note_number % 12
        local octave = math.floor(note_number / 12)
        local freq = notes_frequency[subnote] * (2 ^ octave)
        local sine_table = create_sine_wave(freq, duration, music_volume)
        add_ADSR_envelope(sine_table, 10, 10, 70, 10)
        these_notes[note_number] = gulp.sound.manager:add_mono_sound(sine_table)
    end
end


function make_notes_at_length(duration)
    local these_notes = {}
    --for octave = -4,4 do
    for octave = -1,2 do
        for subnote = 1,12 do
            local freq = notes_frequency[subnote] * (2 ^ octave)
            local sine_table = create_sine_wave(freq, duration, music_volume)
            add_ADSR_envelope(sine_table, 10, 10, 70, 10)
            local note_index = (octave * 12) + subnote
            these_notes[note_index] = gulp.sound.manager:add_mono_sound(sine_table)
        end
    end
    notes_table[duration] = these_notes
end

function make_all_notes()
    print("Times for make_all_notes() - warning takes ages")
    print(gulp_cpp.SDL_GetTicks())
    make_notes_at_length(1/8)
    print(gulp_cpp.SDL_GetTicks())
    make_notes_at_length(1/4)
    print(gulp_cpp.SDL_GetTicks())
    make_notes_at_length(1/2)
    print(gulp_cpp.SDL_GetTicks())
    make_notes_at_length(1)
    print(gulp_cpp.SDL_GetTicks())
    make_notes_at_length(2)
    print(gulp_cpp.SDL_GetTicks())
    make_notes_at_length(4)
    print(gulp_cpp.SDL_GetTicks())
    print("End of make_all_notes()")
end

function play_track_table(tune)
    -- this is broken ... because it's a naive implementation of the playing function
    if tune and tune.songnote and tune.songlen then
        for i = 1, #tune.songnote do
            local note = tune.songnote[i]
            local len = tune.songlen[i]
            if note then
                local n = notes_table[len][note]
                if n then
                    play_sound(n, 6)
                end
            end
            sleep(len)
        end
    end
end

function make_all_notes_for_track(tune)
    if tune and tune.songnote and tune.songlen then
        for i = 1, #tune.songnote do
            local note = tune.songnote[i]
            local len = tune.songlen[i]
            if note then
                make_note(note, len)
            end
        end
    end
end


ABC_decoder = require('utilities/ABC_decoder')

local oh_well = "|c/B/A"

local tunes = { oh_well }
local tunes_cached = {}

function load_all_tracks()
    print(gulp_cpp.SDL_GetTicks())
    for track_number, tune in ipairs(tunes) do
        if not tunes_cached[track_number] then
            local track = ABC_decoder(tunes[track_number])
            tunes_cached[track_number] = track
            make_all_notes_for_track(track)
        end
    end
    print(gulp_cpp.SDL_GetTicks())
end


function play_track(track_number)
    if track_number < 1 or track_number > #tunes then
        track_number = 1
    end
    if not tunes_cached[track_number] then
        tunes_cached[track_number] = ABC_decoder(tunes[track_number])
    end
    play_track_table(tunes_cached[track_number])
end


-- NS_Log output when used by SDL_Log
--2020-05-22 13:37:58.078 Forlorn Fox[74128:6965596] INFO: Opened audio at 22050 Hz 16 bit stereo 1024 bytes audio buffer

function gulp.sound.print_info()
    local m = gulp.sound.manager
    local float_audio = ""
    if gulp_cpp.SDL_AUDIO_ISFLOAT(m.audio_format) then
        float_audio = " (float)"
    end
    local channel_type = "mono"
    if m.audio_channels > 2 then
        channel_type = "surround"
    elseif m.audio_channels > 1 then
        channel_type = "stereo"
    end

    print(string.format("Opened audio at %d Hz %d bit%s %s %d bytes audio buffer\n", 
        m.audio_rate,
        (bit32.band(m.audio_format, 0xFF)),
        float_audio,
        channel_type,
        m.audio_buffers))
end

--gulp.sound.print_info()
