--[[
ABC notation decoder written for Forlorn Fox 14/7/2016 by Rob Probin

References:
  

Information:
  C = middle C
  C, = octive below middle C
  c = octive above middle C
  c' = two octives above middle C
  B, = note below middle C
  CDEFGABcd
  lengths A/8 A/4 A/2 A/ A A2 A3 A4 A6 A7 A8 A12 A15 A16
  accidentals __A _A =A ^A ^^A   (Applies to that note to the end of the measure)
  z = rest
--]]
require('strict')

local note_table = { 
    C = 0, -- middle C
    D = 2,
    E = 4,
    F = 5,
    G = 7,
    A = 9,
    B = 11,

    c = 12,
    d = 14,
    e = 16,
    f = 17,
    g = 19,
    a = 21,
    b = 23,
}

local function do_note(music, c)
    local note = note_table[c]
    if note then
        local mod = music._flags.current_note_modifier
        if mod then
            music._flags.current_note_modifier = nil
            music._flags.note_accidentals[note] = mod
        else
            mod = music._flags.note_accidentals[note]
            if not mod then
                mod = 0
            end
        end
        note = note + mod
        table.insert(music.songnote, note)
        table.insert(music.songlen, 1)
    end
end

local function do_down_1_octive(music, c)
    local len = #music.songnote
    if len == 0 then    -- mistake in notaton
        return
    end
    music.songnote[len] = music.songnote[len] - 12
end

local function do_up_2_octives(music, c)
    local len = #music.songnote
    if len == 0 then    -- mistake in notaton
        return
    end
    music.songnote[len] = music.songnote[len] + 24
end

local function do_sharp(music, c)
    local mod = music._flags.current_note_modifier
    if mod then
        music._flags.current_note_modifier = mod + 1
    else
        music._flags.current_note_modifier = 1
    end
end

local function do_natural(music, c)
    music._flags.current_note_modifier = 0
end

local function do_flat(music, c)
    local mod = music._flags.current_note_modifier
    if mod then
        music._flags.current_note_modifier = mod - 1
    else
        music._flags.current_note_modifier = -1
    end
end

local function do_rest(music, c)
        table.insert(music.songnote, nil)
        table.insert(music.songlen, 1)
end

local function do_number(m, c)
    if c == "/" then
        m._flags.fraction = true
    elseif m._flags.in_number then
        m._flags.in_number = m._flags.in_number * 10 + string.byte(c) - string.byte("0")
    else
        m._flags.in_number = string.byte(c) - string.byte("0")
    end
end

local function do_barline(music, c)
    -- applies to this measre(=bar) only
    music._flags.note_accidentals = {}
end

local decode_char = {
    C = do_note,
    D = do_note,
    E = do_note,
    F = do_note,
    G = do_note,
    A = do_note,
    B = do_note,
    c = do_note,
    d = do_note,
    e = do_note,
    f = do_note,
    g = do_note,
    a = do_note,
    b = do_note,
    [","] = do_down_1_octive,
    ["'"] = do_up_2_octives,
    ["^"] = do_sharp,  
    ["="] = do_natural,
    ["_"] = do_flat,  
    z = do_rest,
    ["0"] = do_number,
    ["1"] = do_number,
    ["2"] = do_number,
    ["3"] = do_number,
    ["4"] = do_number,
    ["5"] = do_number,
    ["6"] = do_number,
    ["7"] = do_number,
    ["8"] = do_number,
    ["9"] = do_number,
    ["/"] = do_number,
--    " " = do_note,
    ["|"] = do_barline,
}

local function finish_number(m)
    local len = #m.songlen
    if len == 0 then    -- mistake in notaton
        return
    end
    local num = m._flags.in_number
    if num then
        if m._flags.fraction then
            num = 1 / num
        end
        m.songlen[len] = num
    elseif m._flags.fraction then
        m.songlen[len] = 0.5   -- same as /2
    end
    
    m._flags.fraction = nil
    m._flags.in_number = nil
end

local function ABC_decoder(music_string)
    local music = { 
        _flags = {note_accidentals = {}},
        songnote = {},
        songlen = {},
    }
    for c in music_string:gmatch"." do
        local f = decode_char[c]
        if f then
            if f ~= do_number then
                finish_number(music)
            end
            f(music, c)
        end -- otherwise ignore
    end

    finish_number(music)
    return music
end

--[[
-- for testing - Notice the accident holds in a bar
-- http://www.onlinesheetmusic.com/if-i-were-a-rich-man-p110718.aspx
local jetset = "| GFGF E2C2 | z2 EF GFGF  | EFGA _BABA | G2 z6 | _A2G2 _G2F2 | _EDCD E2z2 | _EDCD E2C2 | G2 z6 |"

local test = ABC_decoder(jetset)
for i = 1, #test.songnote do
    print(test.songnote[i], test.songlen[i])
end
--]]

return ABC_decoder
