-- Forlorn Fox
-- Copyright © 2019 Rob Probin.
-- No redistribution without permission.
-- Ported from C++

require("strict")
local class = require("middleclass")

local SimpleImage = class('SimpleImage')


function SimpleImage:initialize(name, renderer, viewport)
    self._viewport = viewport
	self.tex = self:_load_image(name, renderer)
	self:set_scale_to_fit_screen()
end

function SimpleImage:set_scale_to_fit_screen(required_scale_percentage)
	self.current_scale = required_scale_percentage or 100.0
	self:_rescale()
end

function SimpleImage:_rescale()
	self.last_print_area_width = gulp_cpp.GetWindowWidth()
	self.last_print_area_height = gulp_cpp.GetWindowHeight()
	
	local effective_image_width = self.image_width
	local effective_image_height = self.image_height

	-- main aspect ratio using width
	self.display_width = self.last_print_area_width * self.current_scale / 100.0
	local scale_ratio1 = self.display_width / effective_image_width

	-- main aspect ratio using height
	self.display_height = self.last_print_area_height * self.current_scale / 100.0
	local scale_ratio2 = self.display_height / effective_image_height

	if scale_ratio2 < scale_ratio1 then
		scale_ratio1 = scale_ratio2
	end

	self.display_height = self.image_height * scale_ratio1
	self.display_width = self.image_width * scale_ratio1
end


function SimpleImage:destroy_texture()
	gulp_cpp.SDL_DestroyTexture(self.tex)
    self.tex = nil
end

function SimpleImage:draw_horizontally_centered(gr, line)
	local effective_display_width = self.display_width

	local x = (gulp_cpp.GetWindowWidth() - effective_display_width) / 2
    local y = line * self._viewport:get_cell_size()
    self:_internal_draw_xy(gr, x, y)
end

function SimpleImage:draw_center(gr)
	local effective_display_width = self.display_width;
	local effective_display_height = self.display_height;

	local x = (gulp_cpp.GetWindowWidth() - effective_display_width) / 2
	local y = (gulp_cpp.GetWindowHeight() - effective_display_height) / 2

	self:_internal_draw_xy(gr, x, y)
end

function SimpleImage:draw(gr, line, column)
    local cell_size = self._viewport:get_cell_size()
	local x = cell_size * column
	local y = cell_size * line
	self:_internal_draw_xy(gr, x, y)
end

function SimpleImage:_internal_draw_xy(gr, x, y)

	if self.last_print_area_width ~= gulp_cpp.GetWindowWidth() or
	   self.last_print_area_height ~= gulp_cpp.GetWindowHeight() then
		self:_rescale()
	end
    
    local dstRect = gulp_cpp.SDL_Rect()
    --local dstRect = gulp_cpp.SDL_Rect.new(x, y, self.display_width, self.display_height)
    dstRect.x = x
    dstRect.y = y
    dstRect.w = self.display_width
    dstRect.h = self.display_height

	-- SDL_RenderCopyEx could do the rotation...
    if self.tex then
        gr:RenderCopy(self.tex, nil, dstRect)
    else
        error("gr:RenderCopy with no texture")
    end
end



function SimpleImage:_load_image(name, renderer)

	local load_surface = gulp_cpp.load_image(name)

	if load_surface == 0 then
		debugMessage("Error loading bitmap ", name)
		error(gulp_cpp.SDL_GetError())
	end

	-- now we convert the surface to our desired pixel format
	local format = gulp_cpp.get_SDL_PIXELFORMAT_ABGR8888()  -- desired texture format
	-- Rmask, Gmask, Bmask, Amask      -- masks for desired format
	-- bpp                -- bits per pixel for desired format, or fals
    -- or false + fail string
	local bpp, Rmask, Gmask, Bmask, Amask = gulp_cpp.SDL_PixelFormatEnumToMasks(format)
	if not bpp then
		debugMessage("SDL_PixelFormatEnumToMasks error: ", Rmask)
	end
    local get_h = gulp_cpp.get_SDL_Surface_height
    local get_w = gulp_cpp.get_SDL_Surface_width
    local h = get_h(load_surface)
    local w = get_w(load_surface)
    
	local converted = gulp_cpp.SDL_CreateRGBSurface(0, w, h, bpp, Rmask, Gmask,Bmask, Amask);
	if not converted then
		debugMessage("SDL_CreateRGBSurface error: ", gulp_cpp.SDL_GetError())
		error(gulp_cpp.SDL_GetError())
	end

	local err = gulp_cpp.SDL_BlitSurface(load_surface, nil, converted, nil)
	if err ~= 0 then
		debugMessage("SDL_BlitSurface<2> error: ", gulp_cpp.SDL_GetError())
		error(gulp_cpp.SDL_GetError())
	end
	gulp_cpp.SDL_FreeSurface(load_surface)

    self.image_width = get_w(converted)
    self.image_height = get_h(converted)

	-- create our texture
	local texture = gulp_cpp.SDL_CreateTextureFromSurface(renderer, converted)
	if texture == 0 then
		debugMessage("texture creation failed: ", gulp_cpp.SDL_GetError())
        error(gulp_cpp.SDL_GetError())
	else
		-- set blend mode for our texture
		local err = gulp_cpp.SDL_SetTextureBlendMode(texture, gulp.SDL_BlendMode.SDL_BLENDMODE_BLEND)
		if err ~= 0 then
			error("SimpleImage::_load_image() SDL_SetTextureBlendMode " .. gulp_cpp.SDL_GetError())
		end
	end

	gulp_cpp.SDL_FreeSurface(converted)

	return texture
end


function SimpleImage:get_display_width_pixels()
	return self.display_width
end

function SimpleImage:get_display_height_pixels()
	return self.display_height
end

function SimpleImage:get_display_width_chars()
	return self.display_width / self._viewport:get_cell_size()
end

function SimpleImage:get_display_height_chars()
	return self.display_height / self._viewport:get_cell_size()
end

local SimpleImage_unit_tests = false

if SimpleImage_unit_tests then
    local time = 0
    local state = 1
    local _functions = {}
    local test = nil
    function clrclip()
        gulp_cpp.SDL_RenderSetClipRect(gulp.app:get_SDL_renderer(), nil)
    end
    
    function SimpleImage:unit_test_update(dt)
        time = time - dt
        if time > 0 then
            return
        end
        if state == 1 then
            -- start a new cycle
            _functions = {}
            if test then
                test:destroy_texture()
            end
            test = SimpleImage(gulp_cpp.LoadPath("graphics/ff_title1.png"):str(), gulp.app:get_SDL_renderer(), 32)
            table.insert(_functions, function(gr) clrclip() test:draw(gr, 0, 0) end)
            time = 2
        elseif state == 2 then
            _functions = {}
            table.insert(_functions, function(gr) clrclip() test:draw(gr, 1, 1) end)            
            time = 2
        elseif state == 3 then
            _functions = {}
            table.insert(_functions, function(gr) clrclip() test:draw_center(gr, 1, 1) end)            
            time = 2
        elseif state == 4 then
            _functions = {}
            table.insert(_functions, function(gr) clrclip() test:draw_horizontally_centered(gr, 5) end)            
            time = 2
        else
            state = 0
        end
        state = state + 1
    end
    function SimpleImage:unit_test_draw(gr)
        for _, f in ipairs(_functions) do
            f(gr)
        end
    end
end

return SimpleImage
