-- Forlorn Fox
-- Copyright © 2015 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

require("strict")
local class = require("middleclass")

local dispatcher = class("dispatcher")


function dispatcher:initialize()
    self._delegate_list = {}
end

--
-- This is the action dispatcher - passes actions to each delegate until something consumes the action
-- 
function dispatcher:action_dispatcher(target, arg1, arg2, arg3, arg4)
    local delegates = self._delegate_list
    local i = #delegates
    local handled = false
    while i > 0 do
        local current = delegates[i]
        local f = current[target]
        -- has this object got a function to handle the case?
        if f then
            handled = f(current, arg1, arg2, arg3, arg4)    -- we pass the table/class in as well
            
            if(type(handled) ~= 'boolean') then
                print("An " .. tostring(current) .. ":".. target .. " returned non-bool; assuming true")
                handled = true
            end
            
            
            if handled then 
                break   -- get out if handled (nil&false = not handled)
            end
        end
        i = i - 1
    end
    
    return handled      
end

--
-- This is the passive dispatcher - dispatches passive system stuff to everything
-- in the dispatch list. Runs in the opposite direction to the action displatcher
-- 
function dispatcher:passive_dispatcher(target, arg1, arg2, arg3, arg4)
    local delegates = self._delegate_list
    local i = 1
    while i <= #delegates do
        local current = delegates[i]
        local f = current[target]
        -- has this object got a function to handle the case?
        if f then
            f(current, arg1, arg2, arg3, arg4)  -- we pass the table/class in as well
        end
        i = i + 1
    end
end

--
-- This is the passive dispatcher - dispatches passive system stuff to everything
-- in the dispatch list. Runs in the opposite direction to the action displatcher
-- 
function dispatcher:update_dispatcher(dt, arg1, arg2, arg3, arg4)
    local delegates = self._delegate_list
    local i = 1
    while i <= #delegates do
        local current = delegates[i]
        local f = current['update']
        -- has this object got a function to handle the case?
        if f then
            
            local max_dt = current.max_dt or dt
            local current_dt = dt
            
            while max_dt < current_dt do
            
                f(current, max_dt, arg1, arg2, arg3, arg4)  -- we pass the table/class in as well
                current_dt = current_dt - max_dt
            
            end
            
            if current_dt then
                f(current, current_dt, arg1, arg2, arg3, arg4)  -- we pass the table/class in as well
            end
            
        end
        i = i + 1
    end
end


--
-- These two functions allow input delegation
--
function dispatcher:add_delegate(delegate)
    
    -- don't add the same delegate twice
    local delegates = self._delegate_list
    local i = 1
    while i <= #delegates do
        if delegates[i] == delegate then
            return
        end
        i = i + 1
    end
    
    table.insert(self._delegate_list, delegate)
end


function dispatcher:drop_delegate(delegate)
    local delegates = self._delegate_list
    local i = 1
    while i <= #delegates do
        if delegates[i] == delegate then
            table.remove(delegates, i)
            break
        end
        i = i + 1
    end
end


return dispatcher
