-- Forlorn Fox
-- Copyright © 2015 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Written by Rob Probin

local lfs = require"lfs"

local file_utils = {}

function file_utils.strip_trailing_slash(s)
    if s:sub(-1) == "/" or s:sub(-1) == "\\" then
        s = s:sub(0, -2)
    end
    return s
end

function file_utils.attributes(f, aname, atable)
    return lfs.attributes (file_utils.strip_trailing_slash(f), aname, atable)
end

function file_utils.dir(path)
    return lfs.dir(file_utils.strip_trailing_slash(path))
end

function file_utils.mkdir(dirname)
    return lfs.mkdir(file_utils.strip_trailing_slash(dirname))
end

function file_utils.rmdir(dirname)
    return lfs.rmdir(file_utils.strip_trailing_slash(dirname))
end


function file_utils.print_attributes(path)
    for file in file_utils.dir(path) do
        if file ~= "." and file ~= ".." then
            local f = path..'/'..file
            print ("\t "..f)
            local attr = file_utils.attributes(f)
            assert (type(attr) == "table")
            if attr.mode == "directory" then
                file_utils.print_attributes (f)
            else
                for name, value in pairs(attr) do
                    print (name, value)
                end
            end
        end
    end
end


function file_utils.replicate_folders(source, dest)
    for file in file_utils.dir(source) do
        if file ~= "." and file ~= ".." then
            local s = source .. file
            local attr = file_utils.attributes(s)
            if type(attr) == "table" and attr.mode == "directory" then
                --print ("directory"..s)
                local d = dest .. file
                attr = file_utils.attributes(d)
                
                local success = true
                local err = "?"
                if type(attr) ~= "table" or attr.mode ~= "directory" then
                    success, err = file_utils.mkdir(d)
                    --print("mkdir" ,d)
                end
                if success then
                    file_utils.replicate_folders(s.."/", d.."/")
                else
                    print("Failure creating folder", d)
                    print("mkdir says", err)
                end
            end
        end
    end
end

local function create_path_if_doesnt_exist(f)
    local attr, err = file_utils.attributes(f)
    if attr == nil or attr.mode ~= "directory" then
        local success, err = file_utils.mkdir(f)
        if not success then
            print("Failed to create folder", f)
            print(err)
            return false
        end
    end
    return true
end

file_utils.create_path_if_doesnt_exist = create_path_if_doesnt_exist

function file_utils.create_save_folders()
    local save_path = gulp_cpp.SaveDataPath(""):str()
    local app_path = gulp_cpp.AppResourcePath(""):str()
    --print("Save Path", save_path)
    --print("App Path", app_path)
    
    local v = gulp.forlorn_fox_version
    if save_path:sub(-1-#v, -2) ~= v then
        print("Save Data Path doesn't contain version?")
        return
    end
    
    local base_path = save_path:sub(1, -2-#v)
    local attr, err = file_utils.attributes(base_path)
    if attr == nil or attr.mode ~= "directory" then
        print("create_save_folders: Save Data Path base is not directory?")
        if not attr then
            print(err)
        end
        return
    end
    
    -- now try creating the version folder, if it doesn't exist
    attr, err = file_utils.attributes(save_path)
    if attr == nil or attr.mode ~= "directory" then
        local success, err = file_utils.mkdir(save_path)
        if not success then
            print("Failed to create save version folder")
            print(err)
            return
        end
    end
    
    local success = create_path_if_doesnt_exist(save_path)
    if not success then return end
    
    -- let's create the save folder inside that
    local save_game_path = save_path.."saves"
    success = create_path_if_doesnt_exist(save_game_path)
    
    file_utils.replicate_folders(app_path, save_path)
end

-- use concat_path this one!
function file_utils.concat_paths(base, new)
    local output
    if gulp.platform == "Windows" then
        output = base .. '\\' .. new;
    else
        output = base .. '/' .. new;
    end
    
    return output
end

------------------------------------------------------------------------------
--
-- only intended for specific desktop builds
--
local checked_save_path_for_update = false
local update_folder_name = "game_versions"

function file_utils.UpdateDataPath(filename)    
    -- don't allow for mobile platforms
    if gulp.platform_type ~= "Desktop" then
        return nil
    end
    
    local save_path = gulp_cpp.SaveDataPath(""):str()
    
    local v = gulp.forlorn_fox_version
    if save_path:sub(-1-#v, -2) ~= v then
        print("Save Data Path doesn't contain version?")
        return nil
    end
    
    local base_path = save_path:sub(1, -2-#v)
    local update_base = base_path .. update_folder_name
    
    if not checked_save_path_for_update then
        local attr, err = file_utils.attributes(base_path)
        if attr == nil or attr.mode ~= "directory" then
            print("UpdateDataPath: Save Data Path base is not directory?")
            if not attr then
                print(err)
            end
            return nil
        end
    
        -- now try creating the update folder, if it doesn't exist
        attr, err = file_utils.attributes(update_base)
        if attr == nil or attr.mode ~= "directory" then
            local success, err = file_utils.mkdir(update_base)
            if not success then
                print("Failed to create update folder")
                print(err)
                return nil
            end
        end
        
        checked_save_path_for_update = true
    end
    
    return file_utils.concat_paths(update_base, filename)
end

function file_utils.isdir(item_path)
    
    local attr = file_utils.attributes(item_path)
    
    if not attr then 
        return false
    elseif type(attr) ~= "table" then
       return false
    end

    if attr.mode == "directory" then
        return true
    end
    
    return false
end



-- Deletes as much as possible.
-- but returns all errors
-- Aimed for folder/files only. Fails if contains links or others
function file_utils.remove_recursive(item_path, error_list)
    
    error_list = error_list or {}       -- populate one list
    local success1, err, err_code
    local attr
    attr, err, err_code = file_utils.attributes(item_path)
    if not attr then 
        table.insert(error_list, { err, err_code } )
        return nil, error_list
    elseif type(attr) ~= "table" then
        -- impossible!
        table.insert(error_list, { "lfs attribute returned neither table or nil", 1999 } )
        return nil, error_list
    end

    --
    -- simple case - it's just a file
    --
    if attr.mode == "file" then
        success1, err, err_code = os.remove(item_path)
        if not success1 then
            table.insert(error_list, { err, err_code } )
        end
        return success1, error_list
    elseif attr.mode ~= "directory" then
        table.insert(error_list, { "Only expected file or folder", 1998 } )
        return nil, error_list
    end
    
    -- 
    -- complex case - it's a folder
    --
    local success = true
    
    for file in file_utils.dir(item_path) do
        if file ~= "." and file ~= ".." then
            local f = item_path..'/'..file
                        
            -- since we created it, it will only be a directory or file ... no links, etc.
            -- remove recursive handles files and directories - so no 
            success1, error_list = file_utils.remove_recursive(f, error_list)
            if not success1 then
                success = nil
            end

        end
    end
    
    -- finally delete the directory, we since we know it's empty
    success1, err, err_code = file_utils.rmdir(item_path)
    if not success1 then
        success = nil
        table.insert(error_list, { err, err_code} )
    end

    return success, error_list
end


--local function check_folder_exists(dest)
--    return file_utils.attributes(dest, "mode") == "directory"
--end


function file_utils.is_folder(dest)
    return file_utils.attributes(dest, "mode") == "directory"
end
function file_utils.is_file(dest)
    return file_utils.attributes(dest, "mode") == "file"
end

function file_utils.concat_path(dest, fpn)
    -- don't add a seperator to an empty destination
    if dest == "" then
        return fpn
    end
    local last_char = dest:sub(-1)
    local first_char = fpn:sub(1,1)
    local last = last_char == "/" or last_char == "\\"
    local first = first_char == "/" or first_char == "\\"
    
    if (last and not first) or (first and not last) then        
        -- seperator in one or another
        return dest .. fpn
    elseif last and first then
        -- seperators on both .. 
        return dest .. fpn(2)
    else
        -- add a seperator
        return dest .. "/" .. fpn
    end
end

function file_utils.split_path(filepath)
    -- returns root-path, filename_with_extension, and extension
    -- Paul Kulchenko, https://stackoverflow.com/questions/5243179/what-is-the-neatest-way-to-split-out-a-path-name-into-its-components-in-lua#12188617
    return string.match(filepath, "(.-)([^\\/]-%.?([^%.\\/]*))$")
end

function file_utils.create_paths_if_necessary(dest, fp)
    local i = 1
    while true do
        i = fp:find("[\\/]", i)
        if not i then break end
        local item = fp:sub(1, i)
        local item_type = "directory"   -- don't do top level
        if item ~= "/" and item ~= "\\" and item ~= "" then
            item_type = file_utils.attributes(file_utils.concat_path(dest, item), "mode")
        end
        if item_type then
            if item_type ~= "directory" then
                return nil, "Item not a folder"
            end
        else
            -- Creates a new directory. The argument is the name of the new directory.
            -- Returns true if the operation was successful; in case of error, it returns nil plus an error string. 
            local err, str = file_utils.mkdir(file_utils.concat_path(dest, item))
            if not err then
                return err, str
            end
        end
        i = i + 1
    end
    
    return true
end


return file_utils
