-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- utf8 stuff
--

local utf8 = {}
  
--
-- Copy some key routines into the global function space
--
-- We don't want to overdo this, but for basic things it's fine.
to_utf8 = gulp_cpp.to_utf8
from_utf8 = gulp_cpp.from_utf8

-- copy these into the local namespace as well
utf8.to_utf8 = gulp_cpp.to_utf8
utf8.from_utf8 = gulp_cpp.to_utf8
utf8.byte = gulp_cpp.utf8_byte  -- similar to string.byte, but for utf8. NOTE: byte uses STRING indexes (not utf8 indexes) and may also overrun the end of the specified range to decode a utf8 code point...
utf8.char = gulp_cpp.utf8_char  -- similar to string.char, but for utf8 inputs

-- from http://lua-users.org/wiki/LuaUnicode
-- Utf8to32 - UTF8 decoding function
-- This function converts a lua string which contains UTF-8 encoded characters into 
-- a lua table with its corresponding unicode codepoints (UTF-32)
-- 
-- NOTE: This decodes 5 and 6 byte streams, but these will be rejected by gulp printing
--       since they are outside the Unicode range.
function utf8.slow_to32(utf8str)
    assert(type(utf8str) == "string")
    local res, seq, val = {}, 0, nil
    for i = 1, #utf8str do
        local c = string.byte(utf8str, i)
        if seq == 0 then
            table.insert(res, val)
            seq = c < 0x80 and 1 or c < 0xE0 and 2 or c < 0xF0 and 3 or
                  c < 0xF8 and 4 or --c < 0xFC and 5 or c < 0xFE and 6 or
                  error("invalid UTF-8 character sequence")
            val = bit32.band(c, 2^(8-seq) - 1)
        else
            val = bit32.bor(bit32.lshift(val, 6), bit32.band(c, 0x3F))
        end
        seq = seq - 1
    end
    table.insert(res, val)
    --table.insert(res, 0)      -- why was this inserting a zero anyway? is nil not good enough?
    return res
end

-- This one is approx. 30% faster than slow_to32
-- NOTICE, it terminates at a zero.
function utf8.new_to32(utf8str)
    local res, i = {}, 1
    local from_utf8 = gulp_cpp.from_utf8        -- avoid lookups
    local size = #utf8str
    repeat
        --local val, i = utf8_1byte(utf8str, i)
        local val, len = from_utf8(utf8str, i)
        i = i + len
        table.insert(res, val)
    until i > size
    
    return res
end

-- This is a version entirely implemented in C++
-- This is massively faster :-)
-- NOTICE, it terminates at a zero.
utf8.to32 = gulp_cpp.table_from_utf8_string


-- This is not necessary the same as the printed width for a monospace font normally.
-- There are non-printed characters, for example.
-- based on http://lua-users.org/wiki/LuaUnicode
--
function utf8.codepoints(utf8str)
    local _, count = string.gsub(utf8str, "[^\128-\193]", "")
    return count
end


-- returns index of next utf8 character or nil
function utf8.get_next_index(utf8str, index)
    index = index or 1
    return utf8str:find("[^\128-\193]", index+1)
end

-- returns index of previous utf8 character or nil
function utf8.get_previous_index(utf8str, index)
    index = index or (#utf8str)
    if index > #utf8str then index = #utf8str + 1 end
    while true do
        index = index - 1
        if index <= 0 then return nil end
        local c = utf8str:byte(index)
        if c < 128 or c > 193 then
            break
        end
    end
    return index
end


-- CONVERT FROM CHARACTER INDEX TO BYTE INDEX
-- Not the fastest, since we have to loop. but ok for occasional use.
--
-- char_index is the utf_8 character from (1 to n+1)
-- EXAMPLES:
-- 1 will give the byte of the first character (i.e. 1)
-- n+1 will give return just past the end (len+1)
-- ASSUMPTIONS:
-- 1. We assume that the text is valid UTF8. If it might not be
-- check with utf8.valid() first to ensure it is :-)
-- 2. char_index off the end, will, at the moment return the len+1. But this
-- is 'unsupported' behaviour.
function utf8.return_byte_index(utf8str, char_index)
    local index = 1
    while index and char_index > 1 do
        index = utf8.get_next_index(utf8str, index)
        char_index = char_index - 1
    end
    if index == nil then
        index = #utf8str+1
    end
    return index
end

-- CONVERT FROM BYTE INDEX TO CHARACTER INDEX
--function utf8.return_char_index(utf8str, byte_index)
--end


-- basic validation test
-- Does not check for: 
--   * Overlong encodings
--   * A 4-byte sequence (starting with 0xF4) that decodes to a value greater than U+10FFFF
--   * Doesn't check for surrogate halves
--   * Doesn't check for other invalid code-points in the Unicode planes
---    e.g. https://en.wikipedia.org/wiki/UTF-8#Invalid_code_points
--
-- Also see: https://en.wikipedia.org/wiki/UTF-8#Invalid_byte_sequences
function utf8.first_invalid(text)
    local len = 0
    local i = 1
    for c in text:gmatch"." do
        local v = c:byte()
        if len == 0 then
            -- must be valid start character
            if v >= 245 then
                -- these bytes are never valid
                return i
            elseif v >= 0xc2 then
                -- multi-byte valid
                if v >= 0xF0 then
                    len = 3
                elseif v >= 0xE0 then
                    len = 2
                else
                    len = 1
                end
            elseif v >= 0x80 then
                -- can't start with continuation byte
                return i
            end
            -- less than 0x80 is fine
        else
            if v >= 0x80 and v < 0xc0 then
                len = len - 1
            else
                -- must be continuation bytes
                return i
            end
        end
        i = i + 1
    end
    -- didn't fail...
    return nil
end

-- basic validation test ... returns true or false
-- Does not check for: 
--   * Overlong encodings
--   * A 4-byte sequence (starting with 0xF4) that decodes to a value greater than U+10FFFF
function utf8.valid(text)
    return utf8.first_invalid(text) == nil
end

--function utf8.fix_invalid(text)
--    if utf8.first_invalid(text) ~= nil then
--        -- not complete yet
--        text = "\xFF\xFD" --The replacement character "�" (U+FFFD)
--    end
--    return text
--end

-- Unicode unit tests
if true then
    assert(to_utf8(0xFFFD) == "\xEF\xBF\xBD") -- replacement char 0xFFFD encoded as UTF-8
    assert(to_utf8(-1) == to_utf8(0xFFFD))  -- -1 is an invalid unicode code point
    assert(to_utf8(32) == " ")
    assert(to_utf8(33) == "!")
    assert(to_utf8(0x7f) == "\x7F")
    assert(to_utf8(0x80) == "\xC2\x80") -- simple 2 code points
    assert(to_utf8(0xA2) == "\xC2\xA2") -- wikipedia 2 byte example
    assert(to_utf8(0x7FF) == "\xDF\xBF") -- end of 2 byte range
    assert(to_utf8(0x800) == "\xE0\xA0\x80") -- start of 3 byte range
    assert(to_utf8(0x20AC) == "\xE2\x82\xAC") -- wikipedia 3 byte example
    assert(to_utf8(0xFFFF) == "\xEF\xBF\xBF") -- end of 3 byte range
    assert(to_utf8(0x10000) == "\xF0\x90\x80\x80") -- start of 4 byte range
    assert(to_utf8(0x24B62) == "\xF0\xA4\xAD\xA2") -- wikipedia 4 byte example
    assert(to_utf8(0x10FFFF) == "\xF4\x8F\xBF\xBF") -- last valid unicode point
    
    -- right ... now we test the other way...
    assert(0xFFFD == from_utf8("\xEF\xBF\xBD")) -- replacement char 0xFFFD encoded as UTF-8
    assert(32 == from_utf8(" "))
    assert(33 == from_utf8("!"))
    assert(0x7F == from_utf8("\x7F"))
    assert(0x80 == from_utf8("\xC2\x80")) -- simple 2 code points
    assert(0xA2 == from_utf8("\xC2\xA2")) -- wikipedia 2 byte example
    assert(0x7FF == from_utf8("\xDF\xBF")) -- end of 2 byte range
    assert(0x800 == from_utf8("\xE0\xA0\x80")) -- start of 3 byte range
    assert(0x20AC == from_utf8("\xE2\x82\xAC")) -- wikipedia 3 byte example
    assert(0xFFFF == from_utf8("\xEF\xBF\xBF")) -- end of 3 byte range
    assert(0x10000 == from_utf8("\xF0\x90\x80\x80")) -- start of 4 byte range
    assert(0x24B62 == from_utf8("\xF0\xA4\xAD\xA2")) -- wikipedia 4 byte example
    assert(0x10FFFF == from_utf8("\xF4\x8F\xBF\xBF")) -- last valid unicode point
    
    -- also test the second parameter
    function _bytes_from_utf8(str) local _, r = from_utf8(str) return r end
    assert(3 == _bytes_from_utf8("\xEF\xBF\xBD")) -- replacement char 0xFFFD encoded as UTF-8
    assert(1 == _bytes_from_utf8(" "))
    assert(1 == _bytes_from_utf8("!"))
    assert(1 == _bytes_from_utf8("\x7F"))
    assert(2 == _bytes_from_utf8("\xC2\x80")) -- simple 2 code points
    assert(2 == _bytes_from_utf8("\xC2\xA2")) -- wikipedia 2 byte example
    assert(2 == _bytes_from_utf8("\xDF\xBF")) -- end of 2 byte range
    assert(3 == _bytes_from_utf8("\xE0\xA0\x80")) -- start of 3 byte range
    assert(3 == _bytes_from_utf8("\xE2\x82\xAC")) -- wikipedia 3 byte example
    assert(3 == _bytes_from_utf8("\xEF\xBF\xBF")) -- end of 3 byte range
    assert(4 == _bytes_from_utf8("\xF0\x90\x80\x80")) -- start of 4 byte range
    assert(4 == _bytes_from_utf8("\xF0\xA4\xAD\xA2")) -- wikipedia 4 byte example
    assert(4 == _bytes_from_utf8("\xF4\x8F\xBF\xBF")) -- last valid unicode point
    
    -- byte test
    assert(0x10FFFF == utf8.byte("no!\xF4\x8F\xBF\xBFzzzz", 4)) -- last valid unicode point
    assert(0x31 == utf8.byte("1zzzz!\xF4\x8F\xBF\xBF"))
    local r1, r2 = utf8.byte("\xE2\x82\xAC\xDF\xBFzzzz", 1, 5)
    assert(r1 == 0x20AC and r2 == 0x7FF)

    -- char test
    assert("1234" == utf8.char(0x31, 0x32, 0x33, 0x34))
    assert("\xEF\xBF\xBD !\x7F\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF"
        == utf8.char(0xFFFD,32,33,0x7F, 0x80, 0xA2, 0x7FF, 0x800, 0x20AC, 0xFFFF, 0x10000, 0x24B62, 0x10FFFF))
    
    local function comparetables(t1, t2)
        if #t1 ~= #t2 then return false end
        for i=1,#t1 do
            if t1[i] ~= t2[i] then return false end
        end
        return true
    end

    assert(comparetables({0xFFFD,32,33,0x7F, 0x80, 0xA2, 0x7FF, 0x800, 0x20AC, 0xFFFF, 0x10000, 0x24B62, 0x10FFFF}, utf8.slow_to32(
        "\xEF\xBF\xBD !\x7F\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF")))
    assert(comparetables({0xFFFD,32,33,0x7F, 0x80, 0xA2, 0x7FF, 0x800, 0x20AC, 0xFFFF, 0x10000, 0x24B62, 0x10FFFF}, utf8.new_to32(
        "\xEF\xBF\xBD !\x7F\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF"))) 
    assert(comparetables({0xFFFD,32,33,0x7F, 0x80, 0xA2, 0x7FF, 0x800, 0x20AC, 0xFFFF, 0x10000, 0x24B62, 0x10FFFF}, utf8.to32(
        "\xEF\xBF\xBD !\x7F\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF")))

    --require('utils')
    
    --print_table{3, 4, 5, ['hello']=true, FRED=4.5, GEORGE="string", FUNC=print_table}
    --print_indexed_table(utf8.slow_to32("\xEF\xBF\xBD !\x7F\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF"))
    
    --print(print_indexed_table(utf8.slow_to32("\xEF\xBF\xBD !\x7F\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF")))
    --print(print_indexed_table(utf8.new_to32 ("\xEF\xBF\xBD !\x7F\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF")))
    --print(print_indexed_table(utf8.to32     ("\xEF\xBF\xBD !\x7F\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF")))

    
    assert(comparetables({0xFFFD,32,33,0x7F, 0, 0x80, 0xA2, 0x7FF, 0x800, 0x20AC, 0xFFFF, 0x10000, 0x24B62, 0x10FFFF}, utf8.slow_to32(
        "\xEF\xBF\xBD !\x7F\x00\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF")))
    assert(comparetables({0xFFFD,32,33,0x7F, 0, 0x80, 0xA2, 0x7FF, 0x800, 0x20AC, 0xFFFF, 0x10000, 0x24B62, 0x10FFFF}, utf8.new_to32(
        "\xEF\xBF\xBD !\x7F\x00\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF"))) 
    assert(comparetables({0xFFFD,32,33,0x7F, 0, 0x80, 0xA2, 0x7FF, 0x800, 0x20AC, 0xFFFF, 0x10000, 0x24B62, 0x10FFFF}, utf8.to32(
        "\xEF\xBF\xBD !\x7F\x00\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF")))
    
    --print(print_indexed_table(utf8.slow_to32("\xEF\xBF\xBD !\x7F\x00\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF")))
    --print(print_indexed_table(utf8.new_to32 ("\xEF\xBF\xBD !\x7F\x00\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF")))
    --print(print_indexed_table(utf8.to32     ("\xEF\xBF\xBD !\x7F\x00\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF")))

    assert(comparetables({0xFFFD,32,33,0x7F, 0, 0x80, 0xA2, 0x7FF, 0x800, 0x20AC, 0xFFFF, 0x10000, 0x24B62, 0x10FFFF, 33}, utf8.slow_to32(
        "\xEF\xBF\xBD !\x7F\x00\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF!")))
    assert(comparetables({0xFFFD,32,33,0x7F, 0, 0x80, 0xA2, 0x7FF, 0x800, 0x20AC, 0xFFFF, 0x10000, 0x24B62, 0x10FFFF, 33}, utf8.new_to32(
        "\xEF\xBF\xBD !\x7F\x00\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF!")))    
    assert(comparetables({0xFFFD,32,33,0x7F, 0, 0x80, 0xA2, 0x7FF, 0x800, 0x20AC, 0xFFFF, 0x10000, 0x24B62, 0x10FFFF, 33}, utf8.to32(
        "\xEF\xBF\xBD !\x7F\x00\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF!")))

    assert(utf8.valid("01234\000555"))
    assert(utf8.valid(""))
    assert(utf8.valid("\0"))
    assert(not utf8.valid("\192")) -- invalid coding for 0-3F
    assert(not utf8.valid("\193")) -- invalid 0x40-7F
    assert(not utf8.valid("\245")) -- 4 byte invalid
    assert(not utf8.valid("\246")) -- 4 byte invalid
    assert(not utf8.valid("\247")) -- 4 byte invalid
    assert(not utf8.valid("\248")) -- 5 byte invalid
    assert(not utf8.valid("\249")) -- 5 byte invalid
    assert(not utf8.valid("\250")) -- 5 byte invalid
    assert(not utf8.valid("\251")) -- 5 byte invalid
    assert(not utf8.valid("\252")) -- 6 byte invalid
    assert(not utf8.valid("\253")) -- 6 byte invalid
    assert(not utf8.valid("\254"))
    assert(not utf8.valid("\255"))
    assert(not utf8.valid("\x80")) -- starts with continuation byte
    assert(not utf8.valid("\x8F")) -- starts with continuation byte
    assert(not utf8.valid("\xAF")) -- starts with continuation byte
    assert(not utf8.valid("\xBF")) -- starts with continuation byte
    assert(utf8.valid("\xF4\x8F\xBF\xBF"))
    assert(utf8.valid(" !\x7F\x00\xC2\x80\xC2\xA2\xDF\xBF\xE0\xA0\x80\xE2\x82\xAC\xEF\xBF\xBF\xF0\x90\x80\x80\xF0\xA4\xAD\xA2\xF4\x8F\xBF\xBF!"))
    assert(utf8.valid("\xDF\xBF"))
    assert(utf8.valid("\xEF\xBF\xBD")) -- replacement char 0xFFFD encoded as UTF-8
    assert(utf8.valid("Hello everyone! How about some \xE2\x82\xACEuros?"))
    assert(1 == utf8.return_byte_index("qw\xC2\xA2z\xEF\xBF\xBDx", 1))
    assert(2 == utf8.return_byte_index("qw\xC2\xA2z\xEF\xBF\xBDx", 2))
    assert(3 == utf8.return_byte_index("qw\xC2\xA2z\xEF\xBF\xBDx", 3))
    assert(5 == utf8.return_byte_index("qw\xC2\xA2z\xEF\xBF\xBDx", 4))
    assert(6 == utf8.return_byte_index("qw\xC2\xA2z\xEF\xBF\xBDx", 5))
    assert(9 == utf8.return_byte_index("qw\xC2\xA2z\xEF\xBF\xBDx", 6))
    assert(10 == utf8.return_byte_index("qw\xC2\xA2z\xEF\xBF\xBDx", 7))
    assert(10 == utf8.return_byte_index("qw\xC2\xA2z\xEF\xBF\xBDx", 8))
end


return utf8
