-- Zip Read
-- Copyright 2017 Rob Probin
-- No redistribution without permission.
--
-- This zip reader has some limitations:
-- 
-- * It's very simple - it does the minimum necessary.
-- * Similar to ISO/IEC 21320-1:2015, it also requires the following main restrictions of the ZIP file format:
--   * Files in ZIP archives may only be stored uncompressed, or using the "deflate" compression (i.e. compression method may contain the value "0" - stored or "8" - deflated).
--   * The encryption features are prohibited.
--   * The digital signature features are prohibited.
--   * The "patched data" features are prohibited.
--   * Archives may not span multiple volumes or be segmented.
--
-- Other links:
--   https://stackoverflow.com/questions/4802097/how-does-one-find-the-start-of-the-central-directory-in-zip-files
--   https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT
--   https://en.wikipedia.org/wiki/Zip_(file_format)
--   http://lua-users.org/wiki/CompressionAndArchiving
--   https://stackoverflow.com/questions/2829175/how-to-extract-files-from-a-zip-file-using-lua
--   
--require('strict')
class = require('middleclass')


local ZipRead = class('ZipRead')
local file_utils = require("utilities/file_utils")


local function size_of_file(f)
    return f:seek("end")
end

local function read_bytes_from_end(f, number_of_bytes)
    local size, err = size_of_file(f)
    if size == 0 then
        return nil, "Empty file"
    end
    if size == nil then
        --print("Problem getting file size", err)
        return nil, err
    end
    local pos, err
    if size > number_of_bytes then
        pos, err = f:seek("end", -number_of_bytes)
    else
        pos, err = f:seek("set")
    end
    if pos == nil then
        --print("Problem seeking", err)
        return nil, err
    end
    return f:read(number_of_bytes)
end

local function read_bytes(f, offset, number_of_bytes)
    local pos, err = f:seek("set", offset)
    if pos == nil then
        return nil, err
    end
    return f:read(number_of_bytes)
end

    

local function get_intel16(data, start)
    return data:byte(start) + data:byte(start+1) * 256
end

local function get_intel32(data, start)
    return data:byte(start) + data:byte(start+1) * 256 + 
    data:byte(start+2) * 65536 + data:byte(start+3) * 16777216
end

local CD_signature = 0x02014b50
local EoCD_signature = "\x50\x4b\x05\x06"

-- Offset 	Bytes 	Description[23]
-- 0 	4 	End of central directory signature = 0x06054b50
-- 4 	2 	Number of this disk
-- 6 	2 	Disk where central directory starts
-- 8 	2 	Number of central directory records on this disk
-- 10 	2 	Total number of central directory records
-- 12 	4 	Size of central directory (bytes)
-- 16 	4 	Offset of start of central directory, relative to start of archive
-- 20 	2 	Comment length (n)
-- 22 	n 	Comment
function ZipRead:validate_eocd(EoCDR, pos, file_offset)
    -- check fields above
    -- check start of central directory contains signature
    if EoCDR:sub(pos,pos+3) ~= EoCD_signature then
        return nil, "End of central directory signature missing"
    end
    local disk_num = get_intel16(EoCDR, pos+4)
    local central_dir_disk = get_intel16(EoCDR, pos+6)
    local num_central_dir_here = get_intel16(EoCDR, pos+8)
    local total_num_central_dir_records = get_intel16(EoCDR, pos+10)
    local central_dir_size = get_intel32(EoCDR, pos+12)
    local central_dir_offset = get_intel32(EoCDR, pos+16)
    local comment_length = get_intel16(EoCDR, pos+20)
    --[[
    print("Number of this disk", disk_num)
    print("Disk where central dir starts", central_dir_disk)
    print("Number of central dir on this disk", num_central_dir_here)
    print("Total number of central directory records", total_num_central_dir_records)
    print("Size of central directory (bytes)", central_dir_size)
    print("Offset of start of central directory, relative to start of archive", central_dir_offset)
    print("Comment length (n)", comment_length)
    --]]
    if disk_num ~= 0 and central_dir_disk ~= 0 then
        return nil, "More than one disk not supported"
    end
    if num_central_dir_here ~= total_num_central_dir_records then
        return nil, "Central directory record number mismatch"
    end
    --print(central_dir_size, central_dir_offset, (central_dir_size + central_dir_offset), pos)
    if (central_dir_size + central_dir_offset) > (pos + file_offset) then
        return nil, "Central directory overlaps End of central directory"
    end
    
    if (comment_length + pos + 21) > #EoCDR then
        --print(comment_length, pos, 22, #EoCDR)
        return nil, "Comments overlap end of file"
    end
    local data = read_bytes(self.f, central_dir_offset, 4)
    if not data or get_intel32(data, 1) ~= CD_signature then
        return nil, "Central directory signature missing"
    end
    
    self.CD_records = total_num_central_dir_records
    self.CD_pos = central_dir_offset
    self.CD_size = central_dir_size
    
    return true
end

--Central directory file header Offset 	Bytes 	Description[23]
--0 	4 	Central directory file header signature = 0x02014b50
--4 	2 	Version made by
--6 	2 	Version needed to extract (minimum)
--8 	2 	General purpose bit flag
--10 	2 	Compression method
--12 	2 	File last modification time
--14 	2 	File last modification date
--16 	4 	CRC-32
--20 	4 	Compressed size
--24 	4 	Uncompressed size
--28 	2 	File name length (n)
--30 	2 	Extra field length (m)
--32 	2 	File comment length (k)
--34 	2 	Disk number where file starts
--36 	2 	Internal file attributes
--38 	4 	External file attributes
--42 	4 	Relative offset of local file header. This is the number of bytes between the start of the first disk on which the file occurs, and the start of the local file header. This allows software reading the central directory to locate the position of the file inside the .ZIP file.
--46 	n 	File name
--46+n 	m 	Extra field
--46+n+m 	k 	File comment

function ZipRead:read_CD_entry(offset)
    local CD = {}
    CD.error = false

    local data, err = read_bytes(self.f, offset, 46)
    if data == nil then
        if self.debug then
            if self.debug then print("Read CD error:", err) end
        end
        CD.error = true
        return nil, err
    end
    CD.signature = get_intel32(data, 1)
    CD.version_made_by = get_intel16(data, 5)
    CD.minimum_version = get_intel16(data, 7)
    CD.flags = get_intel16(data, 9)
    CD.compression = get_intel16(data, 11)
    CD.last_mod_time = get_intel16(data, 13)
    CD.last_mod_date = get_intel16(data, 15)
    CD.CRC32 = get_intel32(data, 17)
    CD.compressed_size = get_intel32(data, 21)
    CD.uncompressed_size = get_intel32(data, 25)
    CD.filename_len = get_intel16(data, 29)
    CD.extra_field_len = get_intel16(data, 31)
    CD.file_comment_len = get_intel16(data, 33)
    CD.disk_num_file_starts = get_intel16(data, 35)
    CD.int_file_attributes = get_intel16(data, 37)
    CD.ext_file_attributes = get_intel32(data, 39)
    CD.relative_offset_of_local_file_header = get_intel32(data, 43)
    
    CD.total_length = 46 + CD.filename_len + CD.extra_field_len + CD.file_comment_len
    CD.filename, err = read_bytes(self.f, offset+46, CD.filename_len)
    
    if CD.filename == nil then
        if self.debug then print("Read filename error:", err) end
        CD.error = true
        return nil, err
    end
    -- ignore extra fields
    -- ignore file comment
    
    if self.debug then
        print(string.format("CD Signature 0x%x", CD.signature))
        print("Version made by", CD.version_made_by)
        print("Minimum version", CD.minimum_version)
        print(string.format("flags 0x%x", CD.flags))
        print(string.format("compression 0x%x", CD.compression))
        print("Last modification time", CD.last_mod_time)
        print("Last modification date", CD.last_mod_date)
        print(string.format("CD Signature 0x%x", CD.CRC32))
        print("Compressed size", CD.compressed_size)
        print("Uncompressed size", CD.uncompressed_size)
        print("Filename length", CD.filename_len)
        print("Extra field length", CD.extra_field_len)
        print("File comment length", CD.file_comment_len)
        print("Disk number file starts", CD.disk_num_file_starts)
        print(string.format("CD Signature 0x%x", CD.int_file_attributes))
        print(string.format("CD Signature 0x%x", CD.ext_file_attributes))
        print("Relative offset of local file header", CD.relative_offset_of_local_file_header)
        print("Filename:", CD.filename)
    end
    if CD.disk_num_file_starts ~= 0 then
        local err = "Multiple disks not supported"
        if self.debug then print(err) end
        CD.error = true
        return nil, err
    end
    if CD.signature ~= CD_signature then
        return nil, "CD Signature missing"
    end
    if CD.compression ~= 8 and CD.compression ~= 0 then
        return nil, "Unsupported compression"
    end
    
    return CD
end

function ZipRead:read_first_CD_entry()
    return self:read_CD_entry(central_dir_offset)
end

function ZipRead:_find_end_of_central_directory_record()
    -- there can be a comment of of to 65535 bytes plus the 22 bytes of EoCD header
    -- therefore we need to read 65555 bytes and search backwards from xx bytes 
    -- from the end looking for the signature.
    local EoCDR, err = read_bytes_from_end(self.f, 65535)
    if EoCDR == nil then
        return nil, err
    end
    local file_offset, err = self.f:seek("cur")
    if file_offset == nil then
        return nil, err
    end
    local start
    local f_end = 1
    local list_of_occurences = {}
    while true do
        start, f_end = EoCDR:find(EoCD_signature, f_end, true)
        if start == nil then break end
        table.insert(list_of_occurences, start)
    end
    
    -- let's assume that the comment section doesn't contain a valid EoCD record!
    local success = false
    local err = "No zip end record found"
    if self.debug then print("Zip file found end of central directory record signatures:", #list_of_occurences) end
    local pos = nil
    for i = #list_of_occurences, 1, -1 do
        pos = list_of_occurences[i]
        success, err = self:validate_eocd(EoCDR, pos, file_offset)
        if success then break end
    end
    if success then
        self.EoCDR_pos = pos
        err = ""
    end
    return success, err
end

local file_signature = 0x04034b50

--Local file header Offset 	Bytes 	Description[23]
--0 	4 	Local file header signature = 0x04034b50 (read as a little-endian number)
--4 	2 	Version needed to extract (minimum)
--6 	2 	General purpose bit flag
--8 	2 	Compression method
--10 	2 	File last modification time
--12 	2 	File last modification date
--14 	4 	CRC-32
--18 	4 	Compressed size
--22 	4 	Uncompressed size
--26 	2 	File name length (n)
--28 	2 	Extra field length (m)
--30 	n 	File name
--30+n 	m 	Extra field
--
-- If bit 3 (=8) is set, then CRC and file sizes are zero and a data descriptor follows the compressed data.
--
--Data descriptor Offset 	Bytes 	Description[23]
--0 	0/4 	Optional data descriptor signature = 0x08074b50
--0/4 	4 	CRC-32
--4/8 	4 	Compressed size
--8/12 	4 	Uncompressed size
function ZipRead:_read_file_header(offset, alt_crc, compressed_size_in, uncompressed_size_in)
    local filehead = {}
    filehead.error = false

    local data, err = read_bytes(self.f, offset, 30)
    if data == nil then
        if self.debug then
            if self.debug then print("Read file header error:", err) end
        end
        filehead.error = true
        return nil, err
    end
    filehead.signature = get_intel32(data, 1)
    filehead.minimum_version = get_intel16(data, 5)
    filehead.flags = get_intel16(data, 7)
    filehead.compression = get_intel16(data, 9)
    filehead.last_mod_time = get_intel16(data, 11)
    filehead.last_mod_date = get_intel16(data, 13)
    filehead.CRC32 = get_intel32(data, 15)
    filehead.compressed_size = get_intel32(data, 19)
    filehead.uncompressed_size = get_intel32(data, 23)
    filehead.filename_len = get_intel16(data, 27)
    filehead.extra_field_len = get_intel16(data, 29)

    filehead.file_data_offset = offset + 30 + filehead.filename_len + filehead.extra_field_len
    filehead.filename, err = read_bytes(self.f, offset + 30, filehead.filename_len)
    
    if filehead.filename == nil then
        if self.debug then print("Read filename error:", err) end
        filehead.error = true
        return nil, err
    end
    -- ignore extra fields
    
    if filehead.signature ~= file_signature then
        return nil, "File head Signature missing"
    end
    if filehead.compression ~= 8 and filehead.compression ~= 0 then
        return nil, "Unsupported compression in file head"
    end

    if bit32.btest(filehead.flags, 0x08) then
        -- Check for legality?
        -- There have been cases where these fields are filled in, e.g. uncompressed_size
        -- e.g. https://github.com/adamhathcock/sharpcompress/issues/88
        --if filehead.CRC32 ~=0 or filehead.compressed_size ~=0 or filehead.uncompressed_size ~0 then
        --    return nil, "Streaming data descriptor mode but CRC and/or sizes not zero in header!"
        --end

        
        -- read Data descriptor for CRC & sizes
        local dd_offset = compressed_size_in + filehead.file_data_offset
        local data2, err = read_bytes(self.f, dd_offset, 16)
        if data2 == nil then
            if self.debug then
                if self.debug then print("Read data descriptor error:", err) end
            end
            filehead.error = true
            return nil, err
        end

        local Optional_data_descriptor_signature = 0x08074b50
        local address0 = get_intel32(data2, 1)
        local address4 = get_intel32(data2, 5)
        if address0 == Optional_data_descriptor_signature then
            -- there is a potential signature
            -- but we need to check for pathological case
            if alt_crc == Optional_data_descriptor_signature then
                -- is this field at address 0 the really unlucky CRC or the real signature?
                if compressed_size_in == address4 then
                    -- there are three possibilities
                    -- 1. The compressed size is at address 4, and the CRC is at address 0
                    -- 2. The signature is at 0, the CRC is at 4 (and also the DD signature), and the compressed size is at 8 and also the signature value
                    -- 3. The data in the central directory is wrong and misleading us
                    -- We ignore the third one.
                    if filehead.compressed_size == Optional_data_descriptor_signature then
                        -- highly unlikely! But this means that the signature is likely at 0, the CRC at 4 and the compressed size is at 8
                        -- It is possible that we could check the uncompressed size, but this is getting silly, and verification will pick this up later.
                        filehead.CRC32 = address4
                        filehead.compressed_size = get_intel32(data2, 9)
                        filehead.uncompressed_size = get_intel32(data2, 13)
                    else
                        -- Yay! Something we can be confident in.
                        -- The CRC is at 0, and the compressed size is at 4
                        filehead.CRC32 = address0
                        filehead.compressed_size = address4
                        filehead.uncompressed_size = get_intel32(data2, 9)
                    end
                else
                    -- compressed size is not at address4, therefore it's likely at 8
                    -- this means sig at 0, crc at 4, compress at 8, uncompressed at 12
                    filehead.CRC32 = address4
                    filehead.compressed_size = get_intel32(data2, 9)
                    filehead.uncompressed_size = get_intel32(data2, 13)
                end
            else        -- signature present, and CRC is not some wierd value, pretty easy case
                filehead.CRC32 = address4
                filehead.compressed_size = get_intel32(data2, 9)
                filehead.uncompressed_size = get_intel32(data2, 13)
            end
        else    -- no signature ... the easy case
            filehead.CRC32 = address0
            filehead.compressed_size = address4
            filehead.uncompressed_size = get_intel32(data2, 9)
        end
        
    end
    

    return filehead
end

function ZipRead:check_file_header_against_central_directory(CD_file)
    local err = nil
    local filehead = CD_file.filehead
    
    if filehead.minimum_version ~= CD_file.minimum_version then
        err = "Minimum version in fileheader and central directory doesn't match"
    elseif filehead.flags ~= CD_file.flags then
        err = "flags in fileheader and central directory doesn't match"
    elseif filehead.compression ~= CD_file.compression then
        err = "compression in fileheader and central directory doesn't match"
    elseif filehead.last_mod_time ~= CD_file.last_mod_time then
        err = "last modification time in fileheader and central directory doesn't match"
    elseif filehead.last_mod_date ~= CD_file.last_mod_date then
        err = "last modification date in fileheader and central directory doesn't match"
    elseif filehead.CRC32 ~= CD_file.CRC32 then
        err = "CRC32 in fileheader and central directory doesn't match"
    elseif filehead.compressed_size ~= CD_file.compressed_size then
        err = "compressed size in fileheader and central directory doesn't match"
    elseif filehead.uncompressed_size ~= CD_file.uncompressed_size then
        err = "uncompressed size in fileheader and central directory doesn't match"
    elseif filehead.filename_len ~= CD_file.filename_len then
        err = "filename length in fileheader and central directory doesn't match"
    --elseif filehead.extra_field_len ~= CD_file.extra_field_len then
    --  err = "extra field length in fileheader and central directory doesn't match"
    end

    return err
end


function ZipRead:initialize(zip_filename, debug)
    -- Should we check the extension?
    -- No, let's trust it because it might be a .exe or .jar or somerthing
  
    self.debug = debug
    self.error = false
    local err
    self.f, err = io.open(zip_filename, "rb")
    if self.f == nil then
        self.error = true
        self.error_string = "Failed to open zip file"
        return
    end
    local success, err = self:_find_end_of_central_directory_record()
    if not success then
        self.error = true
        self.error_string = err
    else
        --print(self.CD_records)
        --print(self.CD_pos)
        --print(self.CD_size)
    
        local CD
        self.files = {}
        local offset = self.CD_pos
        for i = 1, self.CD_records do
            --CD, err = self:read_first_CD_entry()
            CD, err = self:read_CD_entry(offset)
            if not CD then
                self.error_string = err
                break
            end
            offset = offset + CD.total_length
            
            local filehead, err2 = self:_read_file_header(CD.relative_offset_of_local_file_header, CD.CRC32, CD.compressed_size, CD.uncompressed_size)
            if filehead == nil then
                self.error = true
                self.error_string = err2
                break
            else
                CD.filehead = filehead
                local err = self:check_file_header_against_central_directory(CD)
                if err then
                    self.error = true
                    self.error_string = err
                end
            end
            
            table.insert(self.files, CD)
            --print(CD.total_length)
        end
        
    end
end

--function ZipRead:open_file(zip_filename)
--end

--function ZipRead:save(target_filename)
--end

function ZipRead:list_files()
    if not self.error then
        for k,v in ipairs(self.files) do
            print(string.format("%d: %s", k, v.filename))
        end
    else
        print("Zip file error - can't list files")
    end
end



function ZipRead:uncompress(CD)
    local data_offset = CD.filehead.file_data_offset
    local data_compressed_size = CD.filehead.compressed_size
    local data_uncompressed_size = CD.filehead.uncompressed_size
    local result = nil
    local err, err_string
    
    -- read the data
    local data, err = read_bytes(self.f, data_offset, data_compressed_size)
    
    if data then
        if #data == data_compressed_size then
            -- inflate
            --print("Inflate", CD.filehead.filename, gulp_cpp.inflate(data))
            if CD.compression == 8 then
                result, err, err_string = gulp_cpp.inflate(data)
            elseif CD.compression == 0 then
                result = data
            else
                return nil, "Unknown compression method"
            end
            if #result == data_uncompressed_size then
                if gulp_cpp.calculate_crc32(result) == CD.filehead.CRC32 then
                    err = ""
                else
                    result = nil
                    err = "CRC failed on data"
                end
            else
                result = nil
                err = "Uncompressed data not correct size"
            end
        else
            err = "Compressed data not correct size"
        end
    end
    
    return result, err
end

function ZipRead:clean_up(destination)
    -- clean up the destination, we were using to unpack?
end


function ZipRead:_write_file_to_destination(destination, filepath_and_name, file_contents, exec)
    --if not check_folder_exists(destination) then
    --    return false, "Destination does not exist"
    --end
    if destination ~= "" and not file_utils.is_folder(destination) then
        return false, "Destination is not a folder"
    end
    local entire_path = file_utils.concat_path(destination, filepath_and_name)
    local filepath, filename = file_utils.split_path(filepath_and_name)
    local success, result = file_utils.create_paths_if_necessary(destination, filepath)
    if not success then 
        return success, result
    end
    if filename == "" then
        return true
    end
    --if check_file_exists(entire_path) then
    --    return false, "Destination exists"
    --    crc32 = read_crc32_from_disk(entire_path)
    --end
    
    local f, err = io.open(entire_path, "wb")
    if not f then
        return f, err
    end
    local f2, err_string = f:write(file_contents)
    f:close()
    if not f2 then
        return f2, err_string
    end
    if exec then
        if  gulp.platform == "Linux" or gulp.platform == "MacOSX" then
            local exec_success, exec_str, exec_res = os.execute('chmod +x "' .. entire_path ..'"')
            if not exec_success or exec_str ~= "exit" or exec_res ~= 0 then
                return false, "Failed to make " .. filepath_and_name .. " executable"
            end
        end
        --print(v.filename, "is executable")
    end
    return true
end



function ZipRead:uncompress_all(destination)
    if not self.error then
        for _,v in ipairs(self.files) do
            local data, err = self:uncompress(v)
            if not data then
                print("Problem with", v.filename)
                print("Error ", err)
                if destination then
                    self:clean_up(destination)
                end
                return false
            else
                if destination then
                    local attribs = bit32.band(0x1FF, bit32.rshift(v.ext_file_attributes, 16))
                    --print(string.format("%x", attribs))
                    local exec = bit32.band(0x40, attribs) ~= 0
                    local success, err = self:_write_file_to_destination(destination, v.filename, data, exec)
                    if not success then
                        return nil, err
                    end
                end
            end
        end
    else
        local err = "Zip file error - can't verify files"
        print(err)
        return false, err
    end
    
    return true
end

function ZipRead:verify_files()
    return self:uncompress_all()
end


function ZipRead:is_error()
    return self.error, self.error_string
end

function ZipRead:close()
    if self.f then
        self.f:close()
    end
    self.f = nil
end


--[[ Zip Tests

print("-------------------------------------------------------------")
local z = ZipRead("doesnt_exist.zip", true)
print("doesn't exist", z:is_error())
z:close()

-- test deleting no file
local success, errors = file_utils.remove_recursive("doesnt_exist")


os.execute("pwd")
os.execute("rm zip_test_files/test_delete_me.zip || true; cd zip_test_files/; zip --junk-paths test_delete_me.zip ../data/data/scripts/utilities/gulp_utf8.lua ../data/data/scripts/utilities/zip_read.lua")
gulp_cpp.SDL_Delay(200)

z = ZipRead("zip_test_files/test_delete_me.zip", false)
print("test.delete_me.zip - error?", z:is_error())
z:list_files()
print("Verify", z:verify_files())
print("Uncompress", z:uncompress_all("zip_test_files/test/"))
z:close()

-- test deleting a single file
local success, errors = file_utils.remove_recursive("zip_test_files/test_delete_me.zip")

os.execute("rm zip_test_files/SupportCode.zip || true ; zip -r SupportCode.zip SupportCode ; mv SupportCode.zip zip_test_files/")
gulp_cpp.SDL_Delay(200)

z = ZipRead("zip_test_files/SupportCode.zip", false)
print("SupportCode.zip - error?", z:is_error())
z:list_files()
print("Verify", z:verify_files())
print("Uncompress", z:uncompress_all("zip_test_files/test2/"))
z:close()

-- MacOS Finder actually uses ditto not zip.
os.execute("rm zip_test_files/SupportCode_ditto.zip || true ; ditto -c -k --sequesterRsrc --keepParent SupportCode SupportCode_ditto.zip ; mv SupportCode_ditto.zip zip_test_files/")
gulp_cpp.SDL_Delay(200)

z = ZipRead("zip_test_files/SupportCode_ditto.zip", false)
print("SupportCode_ditto.zip - error?", z:is_error())
z:list_files()
print("Verify", z:verify_files())
print("Uncompress", z:uncompress_all("zip_test_files/test3/"))
z:close()

------------------------------------------------------------------------------

-- retrieve the content of a URL
local http = require("socket.http")
local body, code = http.request("http://forlornfox.com/dl_files/ff/ForlornFox-0.06-DEV-Win32.zip")
if not body then error(code) end

-- save the content to a file
local f = assert(io.open('zip_test_files/win_downloaded.zip', 'wb')) -- open in "binary" mode
f:write(body)
f:close()

z = ZipRead("zip_test_files/win_downloaded.zip", false)
print("win_downloaded.zip - error?", z:is_error())
z:list_files()
print("Verify", z:verify_files())
print("Uncompress", z:uncompress_all("zip_test_files/test4/"))
z:close()

------------------------------------------------------------------------------

-- retrieve the content of a URL
http = require("socket.http")
body, code = http.request("http://forlornfox.com/dl_files/ff/ForlornFox-0.06-DEV-Mac.zip")
if not body then error(code) end

-- save the content to a file
f = assert(io.open('zip_test_files/mac_downloaded.zip', 'wb')) -- open in "binary" mode
f:write(body)
f:close()

z = ZipRead("zip_test_files/mac_downloaded.zip", false)
print("mac_downloaded.zip - error?", z:is_error())
z:list_files()
print("Verify", z:verify_files())
print("Uncompress", z:uncompress_all("zip_test_files/test5/"))
z:close()


print("-------------------------------------------------------------")

--]]

--[[

-- PKZIP crc32
-- http://reveng.sourceforge.net/crc-catalogue/17plus.htm#crc.cat.crc-32c
-- https://www.lammertbies.nl/comm/info/crc-calculation.html
-- width=32 poly=0x04c11db7 init=0xffffffff refin=true refout=true xorout=0xffffffff check=0xcbf43926 residue=0xdebb20e3 name="CRC-32"
-- 
-- CRC polynomial: 0xedb88320
-- init vector = 0xffffffffL
print("CRC1", gulp_cpp.calculate_crc32("123456789"), 0xe3069283)
print("CRC2", gulp_cpp.calculate_crc32("Rob Probin"), 0x397D31DE)
print("CRC3", gulp_cpp.calculate_crc32("The quick brown fox jumps over the lazy dog"), 0x414fa339)
print("CRC4", gulp_cpp.calculate_crc32("Test test"), 0x2475A17E)

print("Inflate", gulp_cpp.inflate("123456"))
--]]

return ZipRead