-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
require('strict')
local text_sizes = require("utilities/text_sizes")

function print_indexed_table(table_to_print, indent)
    local indent = indent or 0
    for k, v in ipairs(table_to_print) do
        local formatting = string.rep("  ", indent) .. "[" .. k .. "] = "
        if type(v) == "table" then
            debugMessage(formatting)
            print_indexed_table(v, indent+1)
        elseif type(v) == 'boolean' or type(v) == 'function' then
            debugMessage(formatting .. tostring(v))      
        else
            debugMessage(formatting .. v)
        end
    end
end

function print_table(table_to_print, indent)
    local indent = indent or 0
    for k, v in pairs(table_to_print) do
        
        local formatting
        if type(k) == 'table' then
            formatting = string.rep("  ", indent) .. "[table]" .. ": "
        else
            local key_ = k
            if type(k) == "boolean" then
                key_ = tostring(k)
            end
            formatting = string.rep("  ", indent) .. key_ .. ": "
        end
        
        if string.find(formatting, "__") == 1 then
            ;-- skip this internal item
        else
               
            if type(v) == "table" then
                debugMessage(formatting)
                if indent > 20 then
                    print("tables nested 20 deep?")
                else
                    print_table(v, indent+1)
                end
            elseif type(v) == 'boolean' or type(v) == 'function' or type (v) == 'userdata' then
                debugMessage(formatting .. tostring(v))      
            else
                debugMessage(formatting .. v)
            end
        end
        
    end
end

function round(float)
    if float > 0 then
        return math.floor(float+0.5)
    else
        return math.ceil(float-0.5)
    end
end


function show_versions()
    print(gulp.app_name)
    print("Version:", gulp.forlorn_fox_version)
    print("Engine:", string.format("%.2f", gulp.forlorn_fox_engine_version))
    local compiled = gulp_cpp.SDL_version()
    gulp_cpp.SDL_VERSION(compiled)
    print("SDL Compiled against:", compiled.major, compiled.minor, compiled.patch)
    local linked = gulp_cpp.SDL_version()
    gulp_cpp.SDL_GetVersion(linked)
    print("SDL Running against:", linked.major, linked.minor, linked.patch)
    print("SDL Revision", gulp_cpp.SDL_GetRevision())
    print("SDL Revision Num", gulp_cpp.SDL_GetRevisionNumber())
    print("Interpreter:", _VERSION)
    gulp_cpp.SDL_MIXER_VERSION(compiled)
    print("SDL Mixer Compiled against:", compiled.major, compiled.minor, compiled.patch)
    linked = gulp_cpp.Mix_Linked_Version()
    print("SDL Mixer Running against:", linked.major, linked.minor, linked.patch)
end



-- find things close to other things
-- 'me' is a table that must contain line, column and map_name
-- 'target' is a table of tables, which must each contain line, column and map_name
-- otherwise, you can check whether anything is close to anything else
-- 'iterator' controls how the list is stepped over, see PlayerList.lua for funky examples, 
--   or otherwise just leave it off the call for default operation
function find_nearby(me, targets, nearby_distance, iterator, mode)
    
    local nearby_keys = {}
    
    if not iterator then iterator = pairs end
    if not mode then mode = 'quick' end
    
    for key,target in iterator(targets) do
        
        local line_delta = math.abs(me.line - target.line)        
        local column_delta = math.abs(me.column - target.column)
        
        if  me.map_name == target.map_name then
            if mode == 'quick' then
                if line_delta < nearby_distance and column_delta < nearby_distance then
                    -- we are nearby!
                    table.insert(nearby_keys, key)
                end
                
            else 
                -- do proper distance calcs
                local d = math.sqrt((line_delta*line_delta) + (column_delta*column_delta))
            
                if d <= nearby_distance then
                    -- we are nearby
                    if mode == 'return_distances' then
                        table.insert(nearby_keys, {key, d})
                    else
                        table.insert(nearby_keys, key)
                    end
                end
    
            end
        end
    end
    
    return nearby_keys
end

function get_nearby_distances(me, targets, nearby_distance, iterator)
    if not iterator then iterator = pairs end
    return find_nearby(me, targets, nearby_distance, iterator, 'return_distances')
end


function random_entry(itable)
    return itable[math.random(#itable)]
end

function table_is_empty(table)
    return type(table)=='table' and next(table) == nil
end

function check_map_exists(map_name)
    return gulp.maps[map_name] ~= nil
end

function get_map(map_name)
    if gulp.maps[map_name] == nil then
        error("Map "..map_name.." not found")
    else
        return gulp.maps[map_name]
    end
end

function get_maze(map_name)
    local map = get_map(map_name)
    return map.auto.maze
end

function back_key(key)
    if (gulp.platform_type == "Desktop" and key == SDLK.ESCAPE)  or
       (gulp.platform_type == "Mobile"  and key == SDLK.AC_BACK) then
           return true
    end
    return false
end

    

-- basic deep copy
-- does not handle metatables 
-- does not handle recursive tables - class tables are recursive
function deep_copy(item)
    if type(item) ~= 'table' then 
        return item 
    end
    
    local copy = {}
    for k,v in pairs(item) do
        copy[k] = deep_copy(v)
    end
    
    return copy
end

function percent_chance(percentage)
    return math.random(100) < percentage
end

function time_function(message, fn, arg1, arg2, arg3, arg4, arg5)
    local start_t = gulp_cpp.SDL_GetTicks()
    local ret_val = fn(arg1, arg2, arg3, arg4, arg5)
    local time = gulp_cpp.SDL_GetTicks() - start_t
    
    if gulp.DebugLevelLoading then
        debugMessage(message, ":", time)
    end
    
    return ret_val
end

function gulp.debug_print(key, display_type, line, column, param1, param2, param3)
    if not gulp.server_debug_print_table then
        gulp.server_debug_print_table = {}
    end
    if display_type then
        gulp.server_debug_print_table[key] = { display_type, line, column, param1, param2, param3 }
    else
        gulp.server_debug_print_table[key] = nil
    end
end
    
function split(s, delimiter)
    local results = {}
    local max_width = 0
    
    for match in (s..delimiter):gmatch("(.-)"..delimiter) do
        local entry = { ['word'] = match, ['width'] = ANSI_printable_len(match) }
        table.insert(results, entry)
        if entry['width'] > max_width then max_width = entry['width'] end
    end
    return results, max_width
end    

function render_flag_decode(rend)

    local f = rend.flags

    local S = rend.software_flags_mask
    local A = rend.accleratated_flags_mask
    local V = rend.vsync_flags_mask
    local T = rend.target_texture_flags_mask

    local lookup = { S, A, V, T }
    local chars = { "S", "A", "V", "T" }
    local t = ""
    for i, k in ipairs(lookup) do
        if bit32.btest(f, k) then
            t = t .. chars[i]
            f = f - k
        end
    end
    if f ~= 0 then
        t = t .. "?"
    end
    
    return t
end

function debugMessage(...)
    local message = ""
    for i,v in ipairs({...}) do
        message = message..v.." "
    end
    
    gulp_cpp.debugMessage(message)
    
    --gulp_cpp.debugMessage(debug.traceback())
end

--[[
function update_all_offscreen_location_indicators()
    -- update the offscreen location indicator for all other players
    for _,player in all_remote_players() do
        player:update_offscreen_location_indicator()
    end
end    
--]]

function ToggleFullscreen()
    gulp_cpp.ToggleFullscreen()
    if gulp.localplayer then
        gulp.localplayer:calculate_offsets()
        --update_all_offscreen_location_indicators()
    end
end

function Zoom(zoom_in)
    if gulp.LuaStateCodeMode == 'CLIENT' then 
        if gulp.localplayer then
            local map = get_map(gulp.localplayer.map_name)
            local maze = map.auto.maze
            maze:zoom(zoom_in)
            gulp.localplayer:calculate_offsets()
            --update_all_offscreen_location_indicators()
            text_sizes.resize_all()
        end
    end
end    

function print_on_map(l, c, text, font, fg_colour, bg_colour)
    
    -- test with fonts / colours
    
    local update_colour = false
    if fg_colour or bg_colour then
        update_colour = true
    end
    
    c = c - 1       -- offset to account for index starting at 1
    
    for index = 1, text:len() do 
        local glyph = text:byte(index)
        if glyph >= 0x20 and glyph < 0x80 then
            
            local new_glyph = glyph
            if font then
                new_glyph = get_glyph(font, glyph)
            end
            
            local map = get_map(gulp.localplayer.map_name)
            local maze = map.auto.maze
            maze:set_glyph(l, c+index, new_glyph, 0, 1, 1)
            
            if update_colour then
                if fg_colour == nil then fg_colour = get_simple_colour_by_name(map.fg) end
                if bg_colour == nil then bg_colour = get_SDL_Color_by_name(map.bg) end
                maze:set_maze_colours(l, c+index, fg_colour, bg_colour)
            end
        end
    end
end

function demangle_singleplayer_map_name(mangled_map_name)
    local map_name = mangled_map_name
    local s,e = map_name:find("_^_")
    if s ~= nil then
        map_name = map_name:sub(1, s-1)
    end
    return map_name
end

function get_player_name(player_id)
    local player = get_player_by_id(player_id)
    if player then
        return player.name
    else
        return ""
    end
end
    

